/*
 * Headers
*/

#include <stdint.h>
#include <stddef.h>
#include <stdbool.h>


/*
 * Initialisation
*/

struct futhark_context_config ;
struct futhark_context_config *futhark_context_config_new();
void futhark_context_config_free(struct futhark_context_config *cfg);
void futhark_context_config_set_debugging(struct futhark_context_config *cfg,
                                          int flag);
void futhark_context_config_set_logging(struct futhark_context_config *cfg,
                                        int flag);
struct futhark_context ;
struct futhark_context *futhark_context_new(struct futhark_context_config *cfg);
void futhark_context_free(struct futhark_context *ctx);
int futhark_context_sync(struct futhark_context *ctx);
char *futhark_context_get_error(struct futhark_context *ctx);
void futhark_context_pause_profiling(struct futhark_context *ctx);
void futhark_context_unpause_profiling(struct futhark_context *ctx);

/*
 * Arrays
*/

struct futhark_f64_2d ;
struct futhark_f64_2d *futhark_new_f64_2d(struct futhark_context *ctx,
                                          double *data, int64_t dim0,
                                          int64_t dim1);
struct futhark_f64_2d *futhark_new_raw_f64_2d(struct futhark_context *ctx,
                                              char *data, int offset,
                                              int64_t dim0, int64_t dim1);
int futhark_free_f64_2d(struct futhark_context *ctx,
                        struct futhark_f64_2d *arr);
int futhark_values_f64_2d(struct futhark_context *ctx,
                          struct futhark_f64_2d *arr, double *data);
char *futhark_values_raw_f64_2d(struct futhark_context *ctx,
                                struct futhark_f64_2d *arr);
int64_t *futhark_shape_f64_2d(struct futhark_context *ctx,
                              struct futhark_f64_2d *arr);
struct futhark_f64_3d ;
struct futhark_f64_3d *futhark_new_f64_3d(struct futhark_context *ctx,
                                          double *data, int64_t dim0,
                                          int64_t dim1, int64_t dim2);
struct futhark_f64_3d *futhark_new_raw_f64_3d(struct futhark_context *ctx,
                                              char *data, int offset,
                                              int64_t dim0, int64_t dim1,
                                              int64_t dim2);
int futhark_free_f64_3d(struct futhark_context *ctx,
                        struct futhark_f64_3d *arr);
int futhark_values_f64_3d(struct futhark_context *ctx,
                          struct futhark_f64_3d *arr, double *data);
char *futhark_values_raw_f64_3d(struct futhark_context *ctx,
                                struct futhark_f64_3d *arr);
int64_t *futhark_shape_f64_3d(struct futhark_context *ctx,
                              struct futhark_f64_3d *arr);

/*
 * Opaque values
*/

struct futhark_opaque_arr_c_2d ;
int futhark_free_opaque_arr_c_2d(struct futhark_context *ctx,
                                 struct futhark_opaque_arr_c_2d *obj);

/*
 * Entry points
*/

int futhark_entry_fouriertest(struct futhark_context *ctx,
                              struct futhark_f64_3d **out0, const
                              struct futhark_f64_3d *in0);
int futhark_entry_fft2D(struct futhark_context *ctx,
                        struct futhark_opaque_arr_c_2d **out0, const
                        struct futhark_opaque_arr_c_2d *in0);

/*
 * Miscellaneous
*/

void futhark_debugging_report(struct futhark_context *ctx);
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <math.h>
#include <stdint.h>
#undef NDEBUG
#include <assert.h>
// Start of panic.h.

#include <stdarg.h>

static const char *fut_progname;

static void panic(int eval, const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
        fprintf(stderr, "%s: ", fut_progname);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
        exit(eval);
}

/* For generating arbitrary-sized error messages.  It is the callers
   responsibility to free the buffer at some point. */
static char* msgprintf(const char *s, ...) {
  va_list vl;
  va_start(vl, s);
  size_t needed = 1 + vsnprintf(NULL, 0, s, vl);
  char *buffer = (char*) malloc(needed);
  va_start(vl, s); /* Must re-init. */
  vsnprintf(buffer, needed, s, vl);
  return buffer;
}

// End of panic.h.

// Start of timing.h.

// The function get_wall_time() returns the wall time in microseconds
// (with an unspecified offset).

#ifdef _WIN32

#include <windows.h>

static int64_t get_wall_time(void) {
  LARGE_INTEGER time,freq;
  assert(QueryPerformanceFrequency(&freq));
  assert(QueryPerformanceCounter(&time));
  return ((double)time.QuadPart / freq.QuadPart) * 1000000;
}

#else
/* Assuming POSIX */

#include <time.h>
#include <sys/time.h>

static int64_t get_wall_time(void) {
  struct timeval time;
  assert(gettimeofday(&time,NULL) == 0);
  return time.tv_sec * 1000000 + time.tv_usec;
}

#endif

// End of timing.h.

#include <string.h>
#include <inttypes.h>
#include <errno.h>
#include <ctype.h>
#include <errno.h>
#include <getopt.h>
// Start of values.h.

//// Text I/O

typedef int (*writer)(FILE*, void*);
typedef int (*bin_reader)(void*);
typedef int (*str_reader)(const char *, void*);

struct array_reader {
  char* elems;
  int64_t n_elems_space;
  int64_t elem_size;
  int64_t n_elems_used;
  int64_t *shape;
  str_reader elem_reader;
};

static void skipspaces() {
  int c;
  do {
    c = getchar();
  } while (isspace(c));

  if (c != EOF) {
    ungetc(c, stdin);
  }
}

static int constituent(char c) {
  return isalnum(c) || c == '.' || c == '-' || c == '+' || c == '_';
}

// Produces an empty token only on EOF.
static void next_token(char *buf, int bufsize) {
 start:
  skipspaces();

  int i = 0;
  while (i < bufsize) {
    int c = getchar();
    buf[i] = c;

    if (c == EOF) {
      buf[i] = 0;
      return;
    } else if (c == '-' && i == 1 && buf[0] == '-') {
      // Line comment, so skip to end of line and start over.
      for (; c != '\n' && c != EOF; c = getchar());
      goto start;
    } else if (!constituent(c)) {
      if (i == 0) {
        // We permit single-character tokens that are not
        // constituents; this lets things like ']' and ',' be
        // tokens.
        buf[i+1] = 0;
        return;
      } else {
        ungetc(c, stdin);
        buf[i] = 0;
        return;
      }
    }

    i++;
  }

  buf[bufsize-1] = 0;
}

static int next_token_is(char *buf, int bufsize, const char* expected) {
  next_token(buf, bufsize);
  return strcmp(buf, expected) == 0;
}

static void remove_underscores(char *buf) {
  char *w = buf;

  for (char *r = buf; *r; r++) {
    if (*r != '_') {
      *w++ = *r;
    }
  }

  *w++ = 0;
}

static int read_str_elem(char *buf, struct array_reader *reader) {
  int ret;
  if (reader->n_elems_used == reader->n_elems_space) {
    reader->n_elems_space *= 2;
    reader->elems = (char*) realloc(reader->elems,
                                    reader->n_elems_space * reader->elem_size);
  }

  ret = reader->elem_reader(buf, reader->elems + reader->n_elems_used * reader->elem_size);

  if (ret == 0) {
    reader->n_elems_used++;
  }

  return ret;
}

static int read_str_array_elems(char *buf, int bufsize,
                                struct array_reader *reader, int dims) {
  int ret;
  int first = 1;
  char *knows_dimsize = (char*) calloc(dims,sizeof(char));
  int cur_dim = dims-1;
  int64_t *elems_read_in_dim = (int64_t*) calloc(dims,sizeof(int64_t));

  while (1) {
    next_token(buf, bufsize);

    if (strcmp(buf, "]") == 0) {
      if (knows_dimsize[cur_dim]) {
        if (reader->shape[cur_dim] != elems_read_in_dim[cur_dim]) {
          ret = 1;
          break;
        }
      } else {
        knows_dimsize[cur_dim] = 1;
        reader->shape[cur_dim] = elems_read_in_dim[cur_dim];
      }
      if (cur_dim == 0) {
        ret = 0;
        break;
      } else {
        cur_dim--;
        elems_read_in_dim[cur_dim]++;
      }
    } else if (strcmp(buf, ",") == 0) {
      next_token(buf, bufsize);
      if (strcmp(buf, "[") == 0) {
        if (cur_dim == dims - 1) {
          ret = 1;
          break;
        }
        first = 1;
        cur_dim++;
        elems_read_in_dim[cur_dim] = 0;
      } else if (cur_dim == dims - 1) {
        ret = read_str_elem(buf, reader);
        if (ret != 0) {
          break;
        }
        elems_read_in_dim[cur_dim]++;
      } else {
        ret = 1;
        break;
      }
    } else if (strlen(buf) == 0) {
      // EOF
      ret = 1;
      break;
    } else if (first) {
      if (strcmp(buf, "[") == 0) {
        if (cur_dim == dims - 1) {
          ret = 1;
          break;
        }
        cur_dim++;
        elems_read_in_dim[cur_dim] = 0;
      } else {
        ret = read_str_elem(buf, reader);
        if (ret != 0) {
          break;
        }
        elems_read_in_dim[cur_dim]++;
        first = 0;
      }
    } else {
      ret = 1;
      break;
    }
  }

  free(knows_dimsize);
  free(elems_read_in_dim);
  return ret;
}

static int read_str_empty_array(char *buf, int bufsize,
                                const char *type_name, int64_t *shape, int64_t dims) {
  if (strlen(buf) == 0) {
    // EOF
    return 1;
  }

  if (strcmp(buf, "empty") != 0) {
    return 1;
  }

  shape[0] = 0;

  if (!next_token_is(buf, bufsize, "(")) {
    return 1;
  }

  for (int i = 1; i < dims; i++) {
    if (!next_token_is(buf, bufsize, "[")) {
      return 1;
    }

    next_token(buf, bufsize);

    if (sscanf(buf, "%d", &shape[i]) != 1) {
      return 1;
    }

    if (!next_token_is(buf, bufsize, "]")) {
      return 1;
    }
  }

  if (!next_token_is(buf, bufsize, type_name)) {
    return 1;
  }


  if (!next_token_is(buf, bufsize, ")")) {
    return 1;
  }

  return 0;
}

static int read_str_array(int64_t elem_size, str_reader elem_reader,
                          const char *type_name,
                          void **data, int64_t *shape, int64_t dims) {
  int ret;
  struct array_reader reader;
  char buf[100];

  int dims_seen;
  for (dims_seen = 0; dims_seen < dims; dims_seen++) {
    if (!next_token_is(buf, sizeof(buf), "[")) {
      break;
    }
  }

  if (dims_seen == 0) {
    return read_str_empty_array(buf, sizeof(buf), type_name, shape, dims);
  }

  if (dims_seen != dims) {
    return 1;
  }

  reader.shape = shape;
  reader.n_elems_used = 0;
  reader.elem_size = elem_size;
  reader.n_elems_space = 16;
  reader.elems = (char*) realloc(*data, elem_size*reader.n_elems_space);
  reader.elem_reader = elem_reader;

  ret = read_str_array_elems(buf, sizeof(buf), &reader, dims);

  *data = reader.elems;

  return ret;
}

#define READ_STR(MACRO, PTR, SUFFIX)                                   \
  remove_underscores(buf);                                              \
  int j;                                                                \
  if (sscanf(buf, "%"MACRO"%n", (PTR*)dest, &j) == 1) {                 \
    return !(strcmp(buf+j, "") == 0 || strcmp(buf+j, SUFFIX) == 0);     \
  } else {                                                              \
    return 1;                                                           \
  }

static int read_str_i8(char *buf, void* dest) {
  /* Some platforms (WINDOWS) does not support scanf %hhd or its
     cousin, %SCNi8.  Read into int first to avoid corrupting
     memory.

     https://gcc.gnu.org/bugzilla/show_bug.cgi?id=63417  */
  remove_underscores(buf);
  int j, x;
  if (sscanf(buf, "%i%n", &x, &j) == 1) {
    *(int8_t*)dest = x;
    return !(strcmp(buf+j, "") == 0 || strcmp(buf+j, "i8") == 0);
  } else {
    return 1;
  }
}

static int read_str_u8(char *buf, void* dest) {
  /* Some platforms (WINDOWS) does not support scanf %hhd or its
     cousin, %SCNu8.  Read into int first to avoid corrupting
     memory.

     https://gcc.gnu.org/bugzilla/show_bug.cgi?id=63417  */
  remove_underscores(buf);
  int j, x;
  if (sscanf(buf, "%i%n", &x, &j) == 1) {
    *(uint8_t*)dest = x;
    return !(strcmp(buf+j, "") == 0 || strcmp(buf+j, "u8") == 0);
  } else {
    return 1;
  }
}

static int read_str_i16(char *buf, void* dest) {
  READ_STR(SCNi16, int16_t, "i16");
}

static int read_str_u16(char *buf, void* dest) {
  READ_STR(SCNi16, int16_t, "u16");
}

static int read_str_i32(char *buf, void* dest) {
  READ_STR(SCNi32, int32_t, "i32");
}

static int read_str_u32(char *buf, void* dest) {
  READ_STR(SCNi32, int32_t, "u32");
}

static int read_str_i64(char *buf, void* dest) {
  READ_STR(SCNi64, int64_t, "i64");
}

static int read_str_u64(char *buf, void* dest) {
  // FIXME: This is not correct, as SCNu64 only permits decimal
  // literals.  However, SCNi64 does not handle very large numbers
  // correctly (it's really for signed numbers, so that's fair).
  READ_STR(SCNu64, uint64_t, "u64");
}

static int read_str_f32(char *buf, void* dest) {
  remove_underscores(buf);
  if (strcmp(buf, "f32.nan") == 0) {
    *(float*)dest = NAN;
    return 0;
  } else if (strcmp(buf, "f32.inf") == 0) {
    *(float*)dest = INFINITY;
    return 0;
  } else if (strcmp(buf, "-f32.inf") == 0) {
    *(float*)dest = -INFINITY;
    return 0;
  } else {
    READ_STR("f", float, "f32");
  }
}

static int read_str_f64(char *buf, void* dest) {
  remove_underscores(buf);
  if (strcmp(buf, "f64.nan") == 0) {
    *(double*)dest = NAN;
    return 0;
  } else if (strcmp(buf, "f64.inf") == 0) {
    *(double*)dest = INFINITY;
    return 0;
  } else if (strcmp(buf, "-f64.inf") == 0) {
    *(double*)dest = -INFINITY;
    return 0;
  } else {
    READ_STR("lf", double, "f64");
  }
}

static int read_str_bool(char *buf, void* dest) {
  if (strcmp(buf, "true") == 0) {
    *(char*)dest = 1;
    return 0;
  } else if (strcmp(buf, "false") == 0) {
    *(char*)dest = 0;
    return 0;
  } else {
    return 1;
  }
}

static int write_str_i8(FILE *out, int8_t *src) {
  return fprintf(out, "%hhdi8", *src);
}

static int write_str_u8(FILE *out, uint8_t *src) {
  return fprintf(out, "%hhuu8", *src);
}

static int write_str_i16(FILE *out, int16_t *src) {
  return fprintf(out, "%hdi16", *src);
}

static int write_str_u16(FILE *out, uint16_t *src) {
  return fprintf(out, "%huu16", *src);
}

static int write_str_i32(FILE *out, int32_t *src) {
  return fprintf(out, "%di32", *src);
}

static int write_str_u32(FILE *out, uint32_t *src) {
  return fprintf(out, "%uu32", *src);
}

static int write_str_i64(FILE *out, int64_t *src) {
  return fprintf(out, "%"PRIi64"i64", *src);
}

static int write_str_u64(FILE *out, uint64_t *src) {
  return fprintf(out, "%"PRIu64"u64", *src);
}

static int write_str_f32(FILE *out, float *src) {
  float x = *src;
  if (isnan(x)) {
    return fprintf(out, "f32.nan");
  } else if (isinf(x) && x >= 0) {
    return fprintf(out, "f32.inf");
  } else if (isinf(x)) {
    return fprintf(out, "-f32.inf");
  } else {
    return fprintf(out, "%.6ff32", x);
  }
}

static int write_str_f64(FILE *out, double *src) {
  double x = *src;
  if (isnan(x)) {
    return fprintf(out, "f64.nan");
  } else if (isinf(x) && x >= 0) {
    return fprintf(out, "f64.inf");
  } else if (isinf(x)) {
    return fprintf(out, "-f64.inf");
  } else {
    return fprintf(out, "%.6ff64", *src);
  }
}

static int write_str_bool(FILE *out, void *src) {
  return fprintf(out, *(char*)src ? "true" : "false");
}

//// Binary I/O

#define BINARY_FORMAT_VERSION 2
#define IS_BIG_ENDIAN (!*(unsigned char *)&(uint16_t){1})

static void flip_bytes(int elem_size, unsigned char *elem) {
  for (int j=0; j<elem_size/2; j++) {
    unsigned char head = elem[j];
    int tail_index = elem_size-1-j;
    elem[j] = elem[tail_index];
    elem[tail_index] = head;
  }
}

// On Windows we need to explicitly set the file mode to not mangle
// newline characters.  On *nix there is no difference.
#ifdef _WIN32
#include <io.h>
#include <fcntl.h>
static void set_binary_mode(FILE *f) {
  setmode(fileno(f), O_BINARY);
}
#else
static void set_binary_mode(FILE *f) {
  (void)f;
}
#endif

static int read_byte(void* dest) {
  int num_elems_read = fread(dest, 1, 1, stdin);
  return num_elems_read == 1 ? 0 : 1;
}

//// Types

struct primtype_info_t {
  const char binname[4]; // Used for parsing binary data.
  const char* type_name; // Same name as in Futhark.
  const int size; // in bytes
  const writer write_str; // Write in text format.
  const str_reader read_str; // Read in text format.
};

static const struct primtype_info_t i8_info =
  {.binname = "  i8", .type_name = "i8",   .size = 1,
   .write_str = (writer)write_str_i8, .read_str = (str_reader)read_str_i8};
static const struct primtype_info_t i16_info =
  {.binname = " i16", .type_name = "i16",  .size = 2,
   .write_str = (writer)write_str_i16, .read_str = (str_reader)read_str_i16};
static const struct primtype_info_t i32_info =
  {.binname = " i32", .type_name = "i32",  .size = 4,
   .write_str = (writer)write_str_i32, .read_str = (str_reader)read_str_i32};
static const struct primtype_info_t i64_info =
  {.binname = " i64", .type_name = "i64",  .size = 8,
   .write_str = (writer)write_str_i64, .read_str = (str_reader)read_str_i64};
static const struct primtype_info_t u8_info =
  {.binname = "  u8", .type_name = "u8",   .size = 1,
   .write_str = (writer)write_str_u8, .read_str = (str_reader)read_str_u8};
static const struct primtype_info_t u16_info =
  {.binname = " u16", .type_name = "u16",  .size = 2,
   .write_str = (writer)write_str_u16, .read_str = (str_reader)read_str_u16};
static const struct primtype_info_t u32_info =
  {.binname = " u32", .type_name = "u32",  .size = 4,
   .write_str = (writer)write_str_u32, .read_str = (str_reader)read_str_u32};
static const struct primtype_info_t u64_info =
  {.binname = " u64", .type_name = "u64",  .size = 8,
   .write_str = (writer)write_str_u64, .read_str = (str_reader)read_str_u64};
static const struct primtype_info_t f32_info =
  {.binname = " f32", .type_name = "f32",  .size = 4,
   .write_str = (writer)write_str_f32, .read_str = (str_reader)read_str_f32};
static const struct primtype_info_t f64_info =
  {.binname = " f64", .type_name = "f64",  .size = 8,
   .write_str = (writer)write_str_f64, .read_str = (str_reader)read_str_f64};
static const struct primtype_info_t bool_info =
  {.binname = "bool", .type_name = "bool", .size = 1,
   .write_str = (writer)write_str_bool, .read_str = (str_reader)read_str_bool};

static const struct primtype_info_t* primtypes[] = {
  &i8_info, &i16_info, &i32_info, &i64_info,
  &u8_info, &u16_info, &u32_info, &u64_info,
  &f32_info, &f64_info,
  &bool_info,
  NULL // NULL-terminated
};

// General value interface.  All endian business taken care of at
// lower layers.

static int read_is_binary() {
  skipspaces();
  int c = getchar();
  if (c == 'b') {
    int8_t bin_version;
    int ret = read_byte(&bin_version);

    if (ret != 0) { panic(1, "binary-input: could not read version.\n"); }

    if (bin_version != BINARY_FORMAT_VERSION) {
      panic(1, "binary-input: File uses version %i, but I only understand version %i.\n",
            bin_version, BINARY_FORMAT_VERSION);
    }

    return 1;
  }
  ungetc(c, stdin);
  return 0;
}

static const struct primtype_info_t* read_bin_read_type_enum() {
  char read_binname[4];

  int num_matched = scanf("%4c", read_binname);
  if (num_matched != 1) { panic(1, "binary-input: Couldn't read element type.\n"); }

  const struct primtype_info_t **type = primtypes;

  for (; *type != NULL; type++) {
    // I compare the 4 characters manually instead of using strncmp because
    // this allows any value to be used, also NULL bytes
    if (memcmp(read_binname, (*type)->binname, 4) == 0) {
      return *type;
    }
  }
  panic(1, "binary-input: Did not recognize the type '%s'.\n", read_binname);
  return NULL;
}

static void read_bin_ensure_scalar(const struct primtype_info_t *expected_type) {
  int8_t bin_dims;
  int ret = read_byte(&bin_dims);
  if (ret != 0) { panic(1, "binary-input: Couldn't get dims.\n"); }

  if (bin_dims != 0) {
    panic(1, "binary-input: Expected scalar (0 dimensions), but got array with %i dimensions.\n",
          bin_dims);
  }

  const struct primtype_info_t *bin_type = read_bin_read_type_enum();
  if (bin_type != expected_type) {
    panic(1, "binary-input: Expected scalar of type %s but got scalar of type %s.\n",
          expected_type->type_name,
          bin_type->type_name);
  }
}

//// High-level interface

static int read_bin_array(const struct primtype_info_t *expected_type, void **data, int64_t *shape, int64_t dims) {
  int ret;

  int8_t bin_dims;
  ret = read_byte(&bin_dims);
  if (ret != 0) { panic(1, "binary-input: Couldn't get dims.\n"); }

  if (bin_dims != dims) {
    panic(1, "binary-input: Expected %i dimensions, but got array with %i dimensions.\n",
          dims, bin_dims);
  }

  const struct primtype_info_t *bin_primtype = read_bin_read_type_enum();
  if (expected_type != bin_primtype) {
    panic(1, "binary-input: Expected %iD-array with element type '%s' but got %iD-array with element type '%s'.\n",
          dims, expected_type->type_name, dims, bin_primtype->type_name);
  }

  uint64_t elem_count = 1;
  for (int i=0; i<dims; i++) {
    uint64_t bin_shape;
    ret = fread(&bin_shape, sizeof(bin_shape), 1, stdin);
    if (ret != 1) {
      panic(1, "binary-input: Couldn't read size for dimension %i of array.\n", i);
    }
    if (IS_BIG_ENDIAN) {
      flip_bytes(sizeof(bin_shape), (unsigned char*) &bin_shape);
    }
    elem_count *= bin_shape;
    shape[i] = (int64_t) bin_shape;
  }

  size_t elem_size = expected_type->size;
  void* tmp = realloc(*data, elem_count * elem_size);
  if (tmp == NULL) {
    panic(1, "binary-input: Failed to allocate array of size %i.\n",
          elem_count * elem_size);
  }
  *data = tmp;

  size_t num_elems_read = fread(*data, elem_size, elem_count, stdin);
  if (num_elems_read != elem_count) {
    panic(1, "binary-input: tried to read %i elements of an array, but only got %i elements.\n",
          elem_count, num_elems_read);
  }

  // If we're on big endian platform we must change all multibyte elements
  // from using little endian to big endian
  if (IS_BIG_ENDIAN && elem_size != 1) {
    flip_bytes(elem_size, (unsigned char*) *data);
  }

  return 0;
}

static int read_array(const struct primtype_info_t *expected_type, void **data, int64_t *shape, int64_t dims) {
  if (!read_is_binary()) {
    return read_str_array(expected_type->size, (str_reader)expected_type->read_str, expected_type->type_name, data, shape, dims);
  } else {
    return read_bin_array(expected_type, data, shape, dims);
  }
}

static int write_str_array(FILE *out, const struct primtype_info_t *elem_type, unsigned char *data, int64_t *shape, int8_t rank) {
  if (rank==0) {
    elem_type->write_str(out, (void*)data);
  } else {
    int64_t len = shape[0];
    int64_t slice_size = 1;

    int64_t elem_size = elem_type->size;
    for (int64_t i = 1; i < rank; i++) {
      slice_size *= shape[i];
    }

    if (len*slice_size == 0) {
      printf("empty(");
      for (int64_t i = 1; i < rank; i++) {
        printf("[%d]", shape[i]);
      }
      printf("%s", elem_type->type_name);
      printf(")");
    } else if (rank==1) {
      putchar('[');
      for (int64_t i = 0; i < len; i++) {
        elem_type->write_str(out, (void*) (data + i * elem_size));
        if (i != len-1) {
          printf(", ");
        }
      }
      putchar(']');
    } else {
      putchar('[');
      for (int64_t i = 0; i < len; i++) {
        write_str_array(out, elem_type, data + i * slice_size * elem_size, shape+1, rank-1);
        if (i != len-1) {
          printf(", ");
        }
      }
      putchar(']');
    }
  }
  return 0;
}

static int write_bin_array(FILE *out, const struct primtype_info_t *elem_type, unsigned char *data, int64_t *shape, int8_t rank) {
  int64_t num_elems = 1;
  for (int64_t i = 0; i < rank; i++) {
    num_elems *= shape[i];
  }

  fputc('b', out);
  fputc((char)BINARY_FORMAT_VERSION, out);
  fwrite(&rank, sizeof(int8_t), 1, out);
  fputs(elem_type->binname, out);
  if (shape != NULL) {
    fwrite(shape, sizeof(int64_t), rank, out);
  }

  if (IS_BIG_ENDIAN) {
    for (int64_t i = 0; i < num_elems; i++) {
      unsigned char *elem = data+i*elem_type->size;
      for (int64_t j = 0; j < elem_type->size; j++) {
        fwrite(&elem[elem_type->size-j], 1, 1, out);
      }
    }
  } else {
    fwrite(data, elem_type->size, num_elems, out);
  }

  return 0;
}

static int write_array(FILE *out, int write_binary,
                       const struct primtype_info_t *elem_type, void *data, int64_t *shape, int8_t rank) {
  if (write_binary) {
    return write_bin_array(out, elem_type, data, shape, rank);
  } else {
    return write_str_array(out, elem_type, data, shape, rank);
  }
}

static int read_scalar(const struct primtype_info_t *expected_type, void *dest) {
  if (!read_is_binary()) {
    char buf[100];
    next_token(buf, sizeof(buf));
    return expected_type->read_str(buf, dest);
  } else {
    read_bin_ensure_scalar(expected_type);
    size_t elem_size = expected_type->size;
    int num_elems_read = fread(dest, elem_size, 1, stdin);
    if (IS_BIG_ENDIAN) {
      flip_bytes(elem_size, (unsigned char*) dest);
    }
    return num_elems_read == 1 ? 0 : 1;
  }
}

static int write_scalar(FILE *out, int write_binary, const struct primtype_info_t *type, void *src) {
  if (write_binary) {
    return write_bin_array(out, type, src, NULL, 0);
  } else {
    return type->write_str(out, src);
  }
}

// End of values.h.

static int binary_output = 0;
static FILE *runtime_file;
static int perform_warmup = 0;
static int num_runs = 1;
static const char *entry_point = "main";
// Start of tuning.h.

static char* load_tuning_file(const char *fname,
                              void *cfg,
                              int (*set_size)(void*, const char*, size_t)) {
  const int max_line_len = 1024;
  char* line = (char*) malloc(max_line_len);

  FILE *f = fopen(fname, "r");

  if (f == NULL) {
    snprintf(line, max_line_len, "Cannot open file: %s", strerror(errno));
    return line;
  }

  int lineno = 0;
  while (fgets(line, max_line_len, f) != NULL) {
    lineno++;
    char *eql = strstr(line, "=");
    if (eql) {
      *eql = 0;
      int value = atoi(eql+1);
      if (set_size(cfg, line, value) != 0) {
        strncpy(eql+1, line, max_line_len-strlen(line)-1);
        snprintf(line, max_line_len, "Unknown name '%s' on line %d.", eql+1, lineno);
        return line;
      }
    } else {
      snprintf(line, max_line_len, "Invalid line %d (must be of form 'name=int').",
               lineno);
      return line;
    }
  }

  free(line);

  return NULL;
}

// End of tuning.h.

int parse_options(struct futhark_context_config *cfg, int argc,
                  char *const argv[])
{
    int ch;
    static struct option long_options[] = {{"write-runtime-to",
                                            required_argument, NULL, 1},
                                           {"runs", required_argument, NULL, 2},
                                           {"debugging", no_argument, NULL, 3},
                                           {"log", no_argument, NULL, 4},
                                           {"entry-point", required_argument,
                                            NULL, 5}, {"binary-output",
                                                       no_argument, NULL, 6},
                                           {0, 0, 0, 0}};
    
    while ((ch = getopt_long(argc, argv, ":t:r:DLe:b", long_options, NULL)) !=
           -1) {
        if (ch == 1 || ch == 't') {
            runtime_file = fopen(optarg, "w");
            if (runtime_file == NULL)
                panic(1, "Cannot open %s: %s\n", optarg, strerror(errno));
        }
        if (ch == 2 || ch == 'r') {
            num_runs = atoi(optarg);
            perform_warmup = 1;
            if (num_runs <= 0)
                panic(1, "Need a positive number of runs, not %s\n", optarg);
        }
        if (ch == 3 || ch == 'D')
            futhark_context_config_set_debugging(cfg, 1);
        if (ch == 4 || ch == 'L')
            futhark_context_config_set_logging(cfg, 1);
        if (ch == 5 || ch == 'e') {
            if (entry_point != NULL)
                entry_point = optarg;
        }
        if (ch == 6 || ch == 'b')
            binary_output = 1;
        if (ch == ':')
            panic(-1, "Missing argument for option %s\n", argv[optind - 1]);
        if (ch == '?') {
            fprintf(stderr, "Usage: %s: %s\n", fut_progname,
                    "[-t/--write-runtime-to FILE] [-r/--runs INT] [-D/--debugging] [-L/--log] [-e/--entry-point NAME] [-b/--binary-output]");
            panic(1, "Unknown option: %s\n", argv[optind - 1]);
        }
    }
    return optind;
}
static void futrts_cli_entry_fouriertest(struct futhark_context *ctx)
{
    int64_t t_start, t_end;
    int time_runs = 0, profile_run = 0;
    
    // We do not want to profile all the initialisation.
    futhark_context_pause_profiling(ctx);
    /* Declare and read input. */
    set_binary_mode(stdin);
    
    struct futhark_f64_3d *read_value_23018;
    int64_t read_shape_23019[3];
    double *read_arr_23020 = NULL;
    
    errno = 0;
    if (read_array(&f64_info, (void **) &read_arr_23020, read_shape_23019, 3) !=
        0)
        panic(1, "Cannot read input #%d of type %s%s (errno: %s).\n", 0,
              "[][][]", f64_info.type_name, strerror(errno));
    
    struct futhark_f64_3d *result_23021;
    
    if (perform_warmup) {
        int r;
        
        assert((read_value_23018 = futhark_new_f64_3d(ctx, read_arr_23020,
                                                      read_shape_23019[0],
                                                      read_shape_23019[1],
                                                      read_shape_23019[2])) !=
            0);
        assert(futhark_context_sync(ctx) == 0);
        // Only profile last run.
        if (profile_run)
            futhark_context_unpause_profiling(ctx);
        t_start = get_wall_time();
        r = futhark_entry_fouriertest(ctx, &result_23021, read_value_23018);
        if (r != 0)
            panic(1, "%s", futhark_context_get_error(ctx));
        assert(futhark_context_sync(ctx) == 0);
        if (profile_run)
            futhark_context_pause_profiling(ctx);
        t_end = get_wall_time();
        
        long elapsed_usec = t_end - t_start;
        
        if (time_runs && runtime_file != NULL)
            fprintf(runtime_file, "%lld\n", (long long) elapsed_usec);
        assert(futhark_free_f64_3d(ctx, read_value_23018) == 0);
        assert(futhark_free_f64_3d(ctx, result_23021) == 0);
    }
    time_runs = 1;
    /* Proper run. */
    for (int run = 0; run < num_runs; run++) {
        // Only profile last run.
        profile_run = run == num_runs - 1;
        
        int r;
        
        assert((read_value_23018 = futhark_new_f64_3d(ctx, read_arr_23020,
                                                      read_shape_23019[0],
                                                      read_shape_23019[1],
                                                      read_shape_23019[2])) !=
            0);
        assert(futhark_context_sync(ctx) == 0);
        // Only profile last run.
        if (profile_run)
            futhark_context_unpause_profiling(ctx);
        t_start = get_wall_time();
        r = futhark_entry_fouriertest(ctx, &result_23021, read_value_23018);
        if (r != 0)
            panic(1, "%s", futhark_context_get_error(ctx));
        assert(futhark_context_sync(ctx) == 0);
        if (profile_run)
            futhark_context_pause_profiling(ctx);
        t_end = get_wall_time();
        
        long elapsed_usec = t_end - t_start;
        
        if (time_runs && runtime_file != NULL)
            fprintf(runtime_file, "%lld\n", (long long) elapsed_usec);
        assert(futhark_free_f64_3d(ctx, read_value_23018) == 0);
        if (run < num_runs - 1) {
            assert(futhark_free_f64_3d(ctx, result_23021) == 0);
        }
    }
    free(read_arr_23020);
    if (binary_output)
        set_binary_mode(stdout);
    {
        double *arr = calloc(sizeof(double), futhark_shape_f64_3d(ctx,
                                                                  result_23021)[0] *
                             futhark_shape_f64_3d(ctx, result_23021)[1] *
                             futhark_shape_f64_3d(ctx, result_23021)[2]);
        
        assert(arr != NULL);
        assert(futhark_values_f64_3d(ctx, result_23021, arr) == 0);
        write_array(stdout, binary_output, &f64_info, arr,
                    futhark_shape_f64_3d(ctx, result_23021), 3);
        free(arr);
    }
    printf("\n");
    assert(futhark_free_f64_3d(ctx, result_23021) == 0);
}
static void futrts_cli_entry_fft2D(struct futhark_context *ctx)
{
    int64_t t_start, t_end;
    int time_runs = 0, profile_run = 0;
    
    // We do not want to profile all the initialisation.
    futhark_context_pause_profiling(ctx);
    /* Declare and read input. */
    set_binary_mode(stdin);
    panic(1, "Cannot read input #%d of type %s\n", 0, "arr_c_2d");
    
    struct futhark_opaque_arr_c_2d *result_23022;
    
    if (perform_warmup) {
        int r;
        
        ;
        assert(futhark_context_sync(ctx) == 0);
        // Only profile last run.
        if (profile_run)
            futhark_context_unpause_profiling(ctx);
        t_start = get_wall_time();
        r = futhark_entry_fft2D(ctx, &result_23022, NULL);
        if (r != 0)
            panic(1, "%s", futhark_context_get_error(ctx));
        assert(futhark_context_sync(ctx) == 0);
        if (profile_run)
            futhark_context_pause_profiling(ctx);
        t_end = get_wall_time();
        
        long elapsed_usec = t_end - t_start;
        
        if (time_runs && runtime_file != NULL)
            fprintf(runtime_file, "%lld\n", (long long) elapsed_usec);
        ;
        assert(futhark_free_opaque_arr_c_2d(ctx, result_23022) == 0);
    }
    time_runs = 1;
    /* Proper run. */
    for (int run = 0; run < num_runs; run++) {
        // Only profile last run.
        profile_run = run == num_runs - 1;
        
        int r;
        
        ;
        assert(futhark_context_sync(ctx) == 0);
        // Only profile last run.
        if (profile_run)
            futhark_context_unpause_profiling(ctx);
        t_start = get_wall_time();
        r = futhark_entry_fft2D(ctx, &result_23022, NULL);
        if (r != 0)
            panic(1, "%s", futhark_context_get_error(ctx));
        assert(futhark_context_sync(ctx) == 0);
        if (profile_run)
            futhark_context_pause_profiling(ctx);
        t_end = get_wall_time();
        
        long elapsed_usec = t_end - t_start;
        
        if (time_runs && runtime_file != NULL)
            fprintf(runtime_file, "%lld\n", (long long) elapsed_usec);
        ;
        if (run < num_runs - 1) {
            assert(futhark_free_opaque_arr_c_2d(ctx, result_23022) == 0);
        }
    }
    ;
    if (binary_output)
        set_binary_mode(stdout);
    printf("#<opaque %s>", "arr_c_2d");
    printf("\n");
    assert(futhark_free_opaque_arr_c_2d(ctx, result_23022) == 0);
}
typedef void entry_point_fun(struct futhark_context *);
struct entry_point_entry {
    const char *name;
    entry_point_fun *fun;
} ;
int main(int argc, char **argv)
{
    fut_progname = argv[0];
    
    struct entry_point_entry entry_points[] = {{.name ="fouriertest", .fun =
                                                futrts_cli_entry_fouriertest},
                                               {.name ="fft2D", .fun =
                                                futrts_cli_entry_fft2D}};
    struct futhark_context_config *cfg = futhark_context_config_new();
    
    assert(cfg != NULL);
    
    int parsed_options = parse_options(cfg, argc, argv);
    
    argc -= parsed_options;
    argv += parsed_options;
    if (argc != 0)
        panic(1, "Excess non-option: %s\n", argv[0]);
    
    struct futhark_context *ctx = futhark_context_new(cfg);
    
    assert(ctx != NULL);
    if (entry_point != NULL) {
        int num_entry_points = sizeof(entry_points) / sizeof(entry_points[0]);
        entry_point_fun *entry_point_fun = NULL;
        
        for (int i = 0; i < num_entry_points; i++) {
            if (strcmp(entry_points[i].name, entry_point) == 0) {
                entry_point_fun = entry_points[i].fun;
                break;
            }
        }
        if (entry_point_fun == NULL) {
            fprintf(stderr,
                    "No entry point '%s'.  Select another with --entry-point.  Options are:\n",
                    entry_point);
            for (int i = 0; i < num_entry_points; i++)
                fprintf(stderr, "%s\n", entry_points[i].name);
            return 1;
        }
        entry_point_fun(ctx);
        if (runtime_file != NULL)
            fclose(runtime_file);
        futhark_debugging_report(ctx);
    }
    futhark_context_free(ctx);
    futhark_context_config_free(cfg);
    return 0;
}
#ifdef _MSC_VER
#define inline __inline
#endif
#include <string.h>
#include <inttypes.h>
#include <ctype.h>
#include <errno.h>
#include <assert.h>
// Start of lock.h.

/* A very simple cross-platform implementation of locks.  Uses
   pthreads on Unix and some Windows thing there.  Futhark's
   host-level code is not multithreaded, but user code may be, so we
   need some mechanism for ensuring atomic access to API functions.
   This is that mechanism.  It is not exposed to user code at all, so
   we do not have to worry about name collisions. */

#ifdef _WIN32

typedef HANDLE lock_t;

static lock_t create_lock(lock_t *lock) {
  *lock = CreateMutex(NULL,  /* Default security attributes. */
                      FALSE, /* Initially unlocked. */
                      NULL); /* Unnamed. */
}

static void lock_lock(lock_t *lock) {
  assert(WaitForSingleObject(*lock, INFINITE) == WAIT_OBJECT_0);
}

static void lock_unlock(lock_t *lock) {
  assert(ReleaseMutex(*lock));
}

static void free_lock(lock_t *lock) {
  CloseHandle(*lock);
}

#else
/* Assuming POSIX */

#include <pthread.h>

typedef pthread_mutex_t lock_t;

static void create_lock(lock_t *lock) {
  int r = pthread_mutex_init(lock, NULL);
  assert(r == 0);
}

static void lock_lock(lock_t *lock) {
  int r = pthread_mutex_lock(lock);
  assert(r == 0);
}

static void lock_unlock(lock_t *lock) {
  int r = pthread_mutex_unlock(lock);
  assert(r == 0);
}

static void free_lock(lock_t *lock) {
  /* Nothing to do for pthreads. */
  (void)lock;
}

#endif

// End of lock.h.

struct memblock {
    int *references;
    char *mem;
    int64_t size;
    const char *desc;
} ;
struct futhark_context_config {
    int debugging;
} ;
struct futhark_context_config *futhark_context_config_new()
{
    struct futhark_context_config *cfg =
                                  (struct futhark_context_config *) malloc(sizeof(struct futhark_context_config));
    
    if (cfg == NULL)
        return NULL;
    cfg->debugging = 0;
    return cfg;
}
void futhark_context_config_free(struct futhark_context_config *cfg)
{
    free(cfg);
}
void futhark_context_config_set_debugging(struct futhark_context_config *cfg,
                                          int detail)
{
    cfg->debugging = detail;
}
void futhark_context_config_set_logging(struct futhark_context_config *cfg,
                                        int detail)
{
    /* Does nothing for this backend. */
    cfg = cfg;
    detail = detail;
}
struct futhark_context {
    int detail_memory;
    int debugging;
    int profiling;
    lock_t lock;
    char *error;
    int64_t peak_mem_usage_default;
    int64_t cur_mem_usage_default;
} ;
struct futhark_context *futhark_context_new(struct futhark_context_config *cfg)
{
    struct futhark_context *ctx =
                           (struct futhark_context *) malloc(sizeof(struct futhark_context));
    
    if (ctx == NULL)
        return NULL;
    ctx->detail_memory = cfg->debugging;
    ctx->debugging = cfg->debugging;
    ctx->error = NULL;
    create_lock(&ctx->lock);
    ctx->peak_mem_usage_default = 0;
    ctx->cur_mem_usage_default = 0;
    return ctx;
}
void futhark_context_free(struct futhark_context *ctx)
{
    free_lock(&ctx->lock);
    free(ctx);
}
int futhark_context_sync(struct futhark_context *ctx)
{
    ctx = ctx;
    return 0;
}
char *futhark_context_get_error(struct futhark_context *ctx)
{
    char *error = ctx->error;
    
    ctx->error = NULL;
    return error;
}
void futhark_context_pause_profiling(struct futhark_context *ctx)
{
    (void) ctx;
}
void futhark_context_unpause_profiling(struct futhark_context *ctx)
{
    (void) ctx;
}
static int memblock_unref(struct futhark_context *ctx, struct memblock *block,
                          const char *desc)
{
    if (block->references != NULL) {
        *block->references -= 1;
        if (ctx->detail_memory)
            fprintf(stderr,
                    "Unreferencing block %s (allocated as %s) in %s: %d references remaining.\n",
                    desc, block->desc, "default space", *block->references);
        if (*block->references == 0) {
            ctx->cur_mem_usage_default -= block->size;
            free(block->mem);
            free(block->references);
            if (ctx->detail_memory)
                fprintf(stderr,
                        "%lld bytes freed (now allocated: %lld bytes)\n",
                        (long long) block->size,
                        (long long) ctx->cur_mem_usage_default);
        }
        block->references = NULL;
    }
    return 0;
}
static int memblock_alloc(struct futhark_context *ctx, struct memblock *block,
                          int64_t size, const char *desc)
{
    if (size < 0)
        panic(1, "Negative allocation of %lld bytes attempted for %s in %s.\n",
              (long long) size, desc, "default space",
              ctx->cur_mem_usage_default);
    
    int ret = memblock_unref(ctx, block, desc);
    
    ctx->cur_mem_usage_default += size;
    if (ctx->detail_memory)
        fprintf(stderr,
                "Allocating %lld bytes for %s in %s (then allocated: %lld bytes)",
                (long long) size, desc, "default space",
                (long long) ctx->cur_mem_usage_default);
    if (ctx->cur_mem_usage_default > ctx->peak_mem_usage_default) {
        ctx->peak_mem_usage_default = ctx->cur_mem_usage_default;
        if (ctx->detail_memory)
            fprintf(stderr, " (new peak).\n");
    } else if (ctx->detail_memory)
        fprintf(stderr, ".\n");
    block->mem = (char *) malloc(size);
    block->references = (int *) malloc(sizeof(int));
    *block->references = 1;
    block->size = size;
    block->desc = desc;
    return ret;
}
static int memblock_set(struct futhark_context *ctx, struct memblock *lhs,
                        struct memblock *rhs, const char *lhs_desc)
{
    int ret = memblock_unref(ctx, lhs, lhs_desc);
    
    (*rhs->references)++;
    *lhs = *rhs;
    return ret;
}
void futhark_debugging_report(struct futhark_context *ctx)
{
    if (ctx->detail_memory || ctx->profiling) {
        fprintf(stderr, "Peak memory usage for default space: %lld bytes.\n",
                (long long) ctx->peak_mem_usage_default);
    }
    if (ctx->profiling) { }
}
static int futrts_fouriertest(struct futhark_context *ctx,
                              struct memblock *out_mem_p_23010,
                              int32_t *out_out_arrsizze_23011,
                              int32_t *out_out_arrsizze_23012,
                              int32_t *out_out_arrsizze_23013,
                              struct memblock input_mem_22678, int32_t n_22232,
                              int32_t m_22233, int32_t k_22234);
static int futrts_fft2D(struct futhark_context *ctx,
                        struct memblock *out_mem_p_23014,
                        int32_t *out_out_arrsizze_23015,
                        int32_t *out_out_arrsizze_23016,
                        struct memblock *out_mem_p_23017,
                        struct memblock arr_mem_22678,
                        struct memblock arr_mem_22679, int32_t n_21954,
                        int32_t m_21955);
static inline int8_t add8(int8_t x, int8_t y)
{
    return x + y;
}
static inline int16_t add16(int16_t x, int16_t y)
{
    return x + y;
}
static inline int32_t add32(int32_t x, int32_t y)
{
    return x + y;
}
static inline int64_t add64(int64_t x, int64_t y)
{
    return x + y;
}
static inline int8_t sub8(int8_t x, int8_t y)
{
    return x - y;
}
static inline int16_t sub16(int16_t x, int16_t y)
{
    return x - y;
}
static inline int32_t sub32(int32_t x, int32_t y)
{
    return x - y;
}
static inline int64_t sub64(int64_t x, int64_t y)
{
    return x - y;
}
static inline int8_t mul8(int8_t x, int8_t y)
{
    return x * y;
}
static inline int16_t mul16(int16_t x, int16_t y)
{
    return x * y;
}
static inline int32_t mul32(int32_t x, int32_t y)
{
    return x * y;
}
static inline int64_t mul64(int64_t x, int64_t y)
{
    return x * y;
}
static inline uint8_t udiv8(uint8_t x, uint8_t y)
{
    return x / y;
}
static inline uint16_t udiv16(uint16_t x, uint16_t y)
{
    return x / y;
}
static inline uint32_t udiv32(uint32_t x, uint32_t y)
{
    return x / y;
}
static inline uint64_t udiv64(uint64_t x, uint64_t y)
{
    return x / y;
}
static inline uint8_t umod8(uint8_t x, uint8_t y)
{
    return x % y;
}
static inline uint16_t umod16(uint16_t x, uint16_t y)
{
    return x % y;
}
static inline uint32_t umod32(uint32_t x, uint32_t y)
{
    return x % y;
}
static inline uint64_t umod64(uint64_t x, uint64_t y)
{
    return x % y;
}
static inline int8_t sdiv8(int8_t x, int8_t y)
{
    int8_t q = x / y;
    int8_t r = x % y;
    
    return q - ((r != 0 && r < 0 != y < 0) ? 1 : 0);
}
static inline int16_t sdiv16(int16_t x, int16_t y)
{
    int16_t q = x / y;
    int16_t r = x % y;
    
    return q - ((r != 0 && r < 0 != y < 0) ? 1 : 0);
}
static inline int32_t sdiv32(int32_t x, int32_t y)
{
    int32_t q = x / y;
    int32_t r = x % y;
    
    return q - ((r != 0 && r < 0 != y < 0) ? 1 : 0);
}
static inline int64_t sdiv64(int64_t x, int64_t y)
{
    int64_t q = x / y;
    int64_t r = x % y;
    
    return q - ((r != 0 && r < 0 != y < 0) ? 1 : 0);
}
static inline int8_t smod8(int8_t x, int8_t y)
{
    int8_t r = x % y;
    
    return r + (r == 0 || (x > 0 && y > 0) || (x < 0 && y < 0) ? 0 : y);
}
static inline int16_t smod16(int16_t x, int16_t y)
{
    int16_t r = x % y;
    
    return r + (r == 0 || (x > 0 && y > 0) || (x < 0 && y < 0) ? 0 : y);
}
static inline int32_t smod32(int32_t x, int32_t y)
{
    int32_t r = x % y;
    
    return r + (r == 0 || (x > 0 && y > 0) || (x < 0 && y < 0) ? 0 : y);
}
static inline int64_t smod64(int64_t x, int64_t y)
{
    int64_t r = x % y;
    
    return r + (r == 0 || (x > 0 && y > 0) || (x < 0 && y < 0) ? 0 : y);
}
static inline int8_t squot8(int8_t x, int8_t y)
{
    return x / y;
}
static inline int16_t squot16(int16_t x, int16_t y)
{
    return x / y;
}
static inline int32_t squot32(int32_t x, int32_t y)
{
    return x / y;
}
static inline int64_t squot64(int64_t x, int64_t y)
{
    return x / y;
}
static inline int8_t srem8(int8_t x, int8_t y)
{
    return x % y;
}
static inline int16_t srem16(int16_t x, int16_t y)
{
    return x % y;
}
static inline int32_t srem32(int32_t x, int32_t y)
{
    return x % y;
}
static inline int64_t srem64(int64_t x, int64_t y)
{
    return x % y;
}
static inline int8_t smin8(int8_t x, int8_t y)
{
    return x < y ? x : y;
}
static inline int16_t smin16(int16_t x, int16_t y)
{
    return x < y ? x : y;
}
static inline int32_t smin32(int32_t x, int32_t y)
{
    return x < y ? x : y;
}
static inline int64_t smin64(int64_t x, int64_t y)
{
    return x < y ? x : y;
}
static inline uint8_t umin8(uint8_t x, uint8_t y)
{
    return x < y ? x : y;
}
static inline uint16_t umin16(uint16_t x, uint16_t y)
{
    return x < y ? x : y;
}
static inline uint32_t umin32(uint32_t x, uint32_t y)
{
    return x < y ? x : y;
}
static inline uint64_t umin64(uint64_t x, uint64_t y)
{
    return x < y ? x : y;
}
static inline int8_t smax8(int8_t x, int8_t y)
{
    return x < y ? y : x;
}
static inline int16_t smax16(int16_t x, int16_t y)
{
    return x < y ? y : x;
}
static inline int32_t smax32(int32_t x, int32_t y)
{
    return x < y ? y : x;
}
static inline int64_t smax64(int64_t x, int64_t y)
{
    return x < y ? y : x;
}
static inline uint8_t umax8(uint8_t x, uint8_t y)
{
    return x < y ? y : x;
}
static inline uint16_t umax16(uint16_t x, uint16_t y)
{
    return x < y ? y : x;
}
static inline uint32_t umax32(uint32_t x, uint32_t y)
{
    return x < y ? y : x;
}
static inline uint64_t umax64(uint64_t x, uint64_t y)
{
    return x < y ? y : x;
}
static inline uint8_t shl8(uint8_t x, uint8_t y)
{
    return x << y;
}
static inline uint16_t shl16(uint16_t x, uint16_t y)
{
    return x << y;
}
static inline uint32_t shl32(uint32_t x, uint32_t y)
{
    return x << y;
}
static inline uint64_t shl64(uint64_t x, uint64_t y)
{
    return x << y;
}
static inline uint8_t lshr8(uint8_t x, uint8_t y)
{
    return x >> y;
}
static inline uint16_t lshr16(uint16_t x, uint16_t y)
{
    return x >> y;
}
static inline uint32_t lshr32(uint32_t x, uint32_t y)
{
    return x >> y;
}
static inline uint64_t lshr64(uint64_t x, uint64_t y)
{
    return x >> y;
}
static inline int8_t ashr8(int8_t x, int8_t y)
{
    return x >> y;
}
static inline int16_t ashr16(int16_t x, int16_t y)
{
    return x >> y;
}
static inline int32_t ashr32(int32_t x, int32_t y)
{
    return x >> y;
}
static inline int64_t ashr64(int64_t x, int64_t y)
{
    return x >> y;
}
static inline uint8_t and8(uint8_t x, uint8_t y)
{
    return x & y;
}
static inline uint16_t and16(uint16_t x, uint16_t y)
{
    return x & y;
}
static inline uint32_t and32(uint32_t x, uint32_t y)
{
    return x & y;
}
static inline uint64_t and64(uint64_t x, uint64_t y)
{
    return x & y;
}
static inline uint8_t or8(uint8_t x, uint8_t y)
{
    return x | y;
}
static inline uint16_t or16(uint16_t x, uint16_t y)
{
    return x | y;
}
static inline uint32_t or32(uint32_t x, uint32_t y)
{
    return x | y;
}
static inline uint64_t or64(uint64_t x, uint64_t y)
{
    return x | y;
}
static inline uint8_t xor8(uint8_t x, uint8_t y)
{
    return x ^ y;
}
static inline uint16_t xor16(uint16_t x, uint16_t y)
{
    return x ^ y;
}
static inline uint32_t xor32(uint32_t x, uint32_t y)
{
    return x ^ y;
}
static inline uint64_t xor64(uint64_t x, uint64_t y)
{
    return x ^ y;
}
static inline char ult8(uint8_t x, uint8_t y)
{
    return x < y;
}
static inline char ult16(uint16_t x, uint16_t y)
{
    return x < y;
}
static inline char ult32(uint32_t x, uint32_t y)
{
    return x < y;
}
static inline char ult64(uint64_t x, uint64_t y)
{
    return x < y;
}
static inline char ule8(uint8_t x, uint8_t y)
{
    return x <= y;
}
static inline char ule16(uint16_t x, uint16_t y)
{
    return x <= y;
}
static inline char ule32(uint32_t x, uint32_t y)
{
    return x <= y;
}
static inline char ule64(uint64_t x, uint64_t y)
{
    return x <= y;
}
static inline char slt8(int8_t x, int8_t y)
{
    return x < y;
}
static inline char slt16(int16_t x, int16_t y)
{
    return x < y;
}
static inline char slt32(int32_t x, int32_t y)
{
    return x < y;
}
static inline char slt64(int64_t x, int64_t y)
{
    return x < y;
}
static inline char sle8(int8_t x, int8_t y)
{
    return x <= y;
}
static inline char sle16(int16_t x, int16_t y)
{
    return x <= y;
}
static inline char sle32(int32_t x, int32_t y)
{
    return x <= y;
}
static inline char sle64(int64_t x, int64_t y)
{
    return x <= y;
}
static inline int8_t pow8(int8_t x, int8_t y)
{
    int8_t res = 1, rem = y;
    
    while (rem != 0) {
        if (rem & 1)
            res *= x;
        rem >>= 1;
        x *= x;
    }
    return res;
}
static inline int16_t pow16(int16_t x, int16_t y)
{
    int16_t res = 1, rem = y;
    
    while (rem != 0) {
        if (rem & 1)
            res *= x;
        rem >>= 1;
        x *= x;
    }
    return res;
}
static inline int32_t pow32(int32_t x, int32_t y)
{
    int32_t res = 1, rem = y;
    
    while (rem != 0) {
        if (rem & 1)
            res *= x;
        rem >>= 1;
        x *= x;
    }
    return res;
}
static inline int64_t pow64(int64_t x, int64_t y)
{
    int64_t res = 1, rem = y;
    
    while (rem != 0) {
        if (rem & 1)
            res *= x;
        rem >>= 1;
        x *= x;
    }
    return res;
}
static inline bool itob_i8_bool(int8_t x)
{
    return x;
}
static inline bool itob_i16_bool(int16_t x)
{
    return x;
}
static inline bool itob_i32_bool(int32_t x)
{
    return x;
}
static inline bool itob_i64_bool(int64_t x)
{
    return x;
}
static inline int8_t btoi_bool_i8(bool x)
{
    return x;
}
static inline int16_t btoi_bool_i16(bool x)
{
    return x;
}
static inline int32_t btoi_bool_i32(bool x)
{
    return x;
}
static inline int64_t btoi_bool_i64(bool x)
{
    return x;
}
#define sext_i8_i8(x) ((int8_t) (int8_t) x)
#define sext_i8_i16(x) ((int16_t) (int8_t) x)
#define sext_i8_i32(x) ((int32_t) (int8_t) x)
#define sext_i8_i64(x) ((int64_t) (int8_t) x)
#define sext_i16_i8(x) ((int8_t) (int16_t) x)
#define sext_i16_i16(x) ((int16_t) (int16_t) x)
#define sext_i16_i32(x) ((int32_t) (int16_t) x)
#define sext_i16_i64(x) ((int64_t) (int16_t) x)
#define sext_i32_i8(x) ((int8_t) (int32_t) x)
#define sext_i32_i16(x) ((int16_t) (int32_t) x)
#define sext_i32_i32(x) ((int32_t) (int32_t) x)
#define sext_i32_i64(x) ((int64_t) (int32_t) x)
#define sext_i64_i8(x) ((int8_t) (int64_t) x)
#define sext_i64_i16(x) ((int16_t) (int64_t) x)
#define sext_i64_i32(x) ((int32_t) (int64_t) x)
#define sext_i64_i64(x) ((int64_t) (int64_t) x)
#define zext_i8_i8(x) ((uint8_t) (uint8_t) x)
#define zext_i8_i16(x) ((uint16_t) (uint8_t) x)
#define zext_i8_i32(x) ((uint32_t) (uint8_t) x)
#define zext_i8_i64(x) ((uint64_t) (uint8_t) x)
#define zext_i16_i8(x) ((uint8_t) (uint16_t) x)
#define zext_i16_i16(x) ((uint16_t) (uint16_t) x)
#define zext_i16_i32(x) ((uint32_t) (uint16_t) x)
#define zext_i16_i64(x) ((uint64_t) (uint16_t) x)
#define zext_i32_i8(x) ((uint8_t) (uint32_t) x)
#define zext_i32_i16(x) ((uint16_t) (uint32_t) x)
#define zext_i32_i32(x) ((uint32_t) (uint32_t) x)
#define zext_i32_i64(x) ((uint64_t) (uint32_t) x)
#define zext_i64_i8(x) ((uint8_t) (uint64_t) x)
#define zext_i64_i16(x) ((uint16_t) (uint64_t) x)
#define zext_i64_i32(x) ((uint32_t) (uint64_t) x)
#define zext_i64_i64(x) ((uint64_t) (uint64_t) x)
#ifdef __OPENCL_VERSION__
int32_t futrts_popc8(int8_t x)
{
    return popcount(x);
}
int32_t futrts_popc16(int16_t x)
{
    return popcount(x);
}
int32_t futrts_popc32(int32_t x)
{
    return popcount(x);
}
int32_t futrts_popc64(int64_t x)
{
    return popcount(x);
}
#elif __CUDA_ARCH__
int32_t futrts_popc8(int8_t x)
{
    return __popc(zext_i8_i32(x));
}
int32_t futrts_popc16(int16_t x)
{
    return __popc(zext_i16_i32(x));
}
int32_t futrts_popc32(int32_t x)
{
    return __popc(x);
}
int32_t futrts_popc64(int64_t x)
{
    return __popcll(x);
}
#else
int32_t futrts_popc8(int8_t x)
{
    int c = 0;
    
    for (; x; ++c)
        x &= x - 1;
    return c;
}
int32_t futrts_popc16(int16_t x)
{
    int c = 0;
    
    for (; x; ++c)
        x &= x - 1;
    return c;
}
int32_t futrts_popc32(int32_t x)
{
    int c = 0;
    
    for (; x; ++c)
        x &= x - 1;
    return c;
}
int32_t futrts_popc64(int64_t x)
{
    int c = 0;
    
    for (; x; ++c)
        x &= x - 1;
    return c;
}
#endif
#ifdef __OPENCL_VERSION__
int32_t futrts_clzz8(int8_t x)
{
    return clz(x);
}
int32_t futrts_clzz16(int16_t x)
{
    return clz(x);
}
int32_t futrts_clzz32(int32_t x)
{
    return clz(x);
}
int32_t futrts_clzz64(int64_t x)
{
    return clz(x);
}
#elif __CUDA_ARCH__
int32_t futrts_clzz8(int8_t x)
{
    return __clz(zext_i8_i32(x)) - 24;
}
int32_t futrts_clzz16(int16_t x)
{
    return __clz(zext_i16_i32(x)) - 16;
}
int32_t futrts_clzz32(int32_t x)
{
    return __clz(x);
}
int32_t futrts_clzz64(int64_t x)
{
    return __clzll(x);
}
#else
int32_t futrts_clzz8(int8_t x)
{
    int n = 0;
    int bits = sizeof(x) * 8;
    
    for (int i = 0; i < bits; i++) {
        if (x < 0)
            break;
        n++;
        x <<= 1;
    }
    return n;
}
int32_t futrts_clzz16(int16_t x)
{
    int n = 0;
    int bits = sizeof(x) * 8;
    
    for (int i = 0; i < bits; i++) {
        if (x < 0)
            break;
        n++;
        x <<= 1;
    }
    return n;
}
int32_t futrts_clzz32(int32_t x)
{
    int n = 0;
    int bits = sizeof(x) * 8;
    
    for (int i = 0; i < bits; i++) {
        if (x < 0)
            break;
        n++;
        x <<= 1;
    }
    return n;
}
int32_t futrts_clzz64(int64_t x)
{
    int n = 0;
    int bits = sizeof(x) * 8;
    
    for (int i = 0; i < bits; i++) {
        if (x < 0)
            break;
        n++;
        x <<= 1;
    }
    return n;
}
#endif
static inline float fdiv32(float x, float y)
{
    return x / y;
}
static inline float fadd32(float x, float y)
{
    return x + y;
}
static inline float fsub32(float x, float y)
{
    return x - y;
}
static inline float fmul32(float x, float y)
{
    return x * y;
}
static inline float fmin32(float x, float y)
{
    return fmin(x, y);
}
static inline float fmax32(float x, float y)
{
    return fmax(x, y);
}
static inline float fpow32(float x, float y)
{
    return pow(x, y);
}
static inline char cmplt32(float x, float y)
{
    return x < y;
}
static inline char cmple32(float x, float y)
{
    return x <= y;
}
static inline float sitofp_i8_f32(int8_t x)
{
    return x;
}
static inline float sitofp_i16_f32(int16_t x)
{
    return x;
}
static inline float sitofp_i32_f32(int32_t x)
{
    return x;
}
static inline float sitofp_i64_f32(int64_t x)
{
    return x;
}
static inline float uitofp_i8_f32(uint8_t x)
{
    return x;
}
static inline float uitofp_i16_f32(uint16_t x)
{
    return x;
}
static inline float uitofp_i32_f32(uint32_t x)
{
    return x;
}
static inline float uitofp_i64_f32(uint64_t x)
{
    return x;
}
static inline int8_t fptosi_f32_i8(float x)
{
    return x;
}
static inline int16_t fptosi_f32_i16(float x)
{
    return x;
}
static inline int32_t fptosi_f32_i32(float x)
{
    return x;
}
static inline int64_t fptosi_f32_i64(float x)
{
    return x;
}
static inline uint8_t fptoui_f32_i8(float x)
{
    return x;
}
static inline uint16_t fptoui_f32_i16(float x)
{
    return x;
}
static inline uint32_t fptoui_f32_i32(float x)
{
    return x;
}
static inline uint64_t fptoui_f32_i64(float x)
{
    return x;
}
static inline double fdiv64(double x, double y)
{
    return x / y;
}
static inline double fadd64(double x, double y)
{
    return x + y;
}
static inline double fsub64(double x, double y)
{
    return x - y;
}
static inline double fmul64(double x, double y)
{
    return x * y;
}
static inline double fmin64(double x, double y)
{
    return fmin(x, y);
}
static inline double fmax64(double x, double y)
{
    return fmax(x, y);
}
static inline double fpow64(double x, double y)
{
    return pow(x, y);
}
static inline char cmplt64(double x, double y)
{
    return x < y;
}
static inline char cmple64(double x, double y)
{
    return x <= y;
}
static inline double sitofp_i8_f64(int8_t x)
{
    return x;
}
static inline double sitofp_i16_f64(int16_t x)
{
    return x;
}
static inline double sitofp_i32_f64(int32_t x)
{
    return x;
}
static inline double sitofp_i64_f64(int64_t x)
{
    return x;
}
static inline double uitofp_i8_f64(uint8_t x)
{
    return x;
}
static inline double uitofp_i16_f64(uint16_t x)
{
    return x;
}
static inline double uitofp_i32_f64(uint32_t x)
{
    return x;
}
static inline double uitofp_i64_f64(uint64_t x)
{
    return x;
}
static inline int8_t fptosi_f64_i8(double x)
{
    return x;
}
static inline int16_t fptosi_f64_i16(double x)
{
    return x;
}
static inline int32_t fptosi_f64_i32(double x)
{
    return x;
}
static inline int64_t fptosi_f64_i64(double x)
{
    return x;
}
static inline uint8_t fptoui_f64_i8(double x)
{
    return x;
}
static inline uint16_t fptoui_f64_i16(double x)
{
    return x;
}
static inline uint32_t fptoui_f64_i32(double x)
{
    return x;
}
static inline uint64_t fptoui_f64_i64(double x)
{
    return x;
}
static inline float fpconv_f32_f32(float x)
{
    return x;
}
static inline double fpconv_f32_f64(float x)
{
    return x;
}
static inline float fpconv_f64_f32(double x)
{
    return x;
}
static inline double fpconv_f64_f64(double x)
{
    return x;
}
static inline float futrts_log32(float x)
{
    return log(x);
}
static inline float futrts_log2_32(float x)
{
    return log2(x);
}
static inline float futrts_log10_32(float x)
{
    return log10(x);
}
static inline float futrts_sqrt32(float x)
{
    return sqrt(x);
}
static inline float futrts_exp32(float x)
{
    return exp(x);
}
static inline float futrts_cos32(float x)
{
    return cos(x);
}
static inline float futrts_sin32(float x)
{
    return sin(x);
}
static inline float futrts_tan32(float x)
{
    return tan(x);
}
static inline float futrts_acos32(float x)
{
    return acos(x);
}
static inline float futrts_asin32(float x)
{
    return asin(x);
}
static inline float futrts_atan32(float x)
{
    return atan(x);
}
static inline float futrts_atan2_32(float x, float y)
{
    return atan2(x, y);
}
static inline float futrts_gamma32(float x)
{
    return tgamma(x);
}
static inline float futrts_lgamma32(float x)
{
    return lgamma(x);
}
static inline char futrts_isnan32(float x)
{
    return isnan(x);
}
static inline char futrts_isinf32(float x)
{
    return isinf(x);
}
static inline int32_t futrts_to_bits32(float x)
{
    union {
        float f;
        int32_t t;
    } p;
    
    p.f = x;
    return p.t;
}
static inline float futrts_from_bits32(int32_t x)
{
    union {
        int32_t f;
        float t;
    } p;
    
    p.f = x;
    return p.t;
}
#ifdef __OPENCL_VERSION__
static inline float fmod32(float x, float y)
{
    return fmod(x, y);
}
static inline float futrts_round32(float x)
{
    return rint(x);
}
static inline float futrts_floor32(float x)
{
    return floor(x);
}
static inline float futrts_ceil32(float x)
{
    return ceil(x);
}
static inline float futrts_lerp32(float v0, float v1, float t)
{
    return mix(v0, v1, t);
}
#else
static inline float fmod32(float x, float y)
{
    return fmodf(x, y);
}
static inline float futrts_round32(float x)
{
    return rintf(x);
}
static inline float futrts_floor32(float x)
{
    return floorf(x);
}
static inline float futrts_ceil32(float x)
{
    return ceilf(x);
}
static inline float futrts_lerp32(float v0, float v1, float t)
{
    return v0 + (v1 - v0) * t;
}
#endif
static inline double futrts_log64(double x)
{
    return log(x);
}
static inline double futrts_log2_64(double x)
{
    return log2(x);
}
static inline double futrts_log10_64(double x)
{
    return log10(x);
}
static inline double futrts_sqrt64(double x)
{
    return sqrt(x);
}
static inline double futrts_exp64(double x)
{
    return exp(x);
}
static inline double futrts_cos64(double x)
{
    return cos(x);
}
static inline double futrts_sin64(double x)
{
    return sin(x);
}
static inline double futrts_tan64(double x)
{
    return tan(x);
}
static inline double futrts_acos64(double x)
{
    return acos(x);
}
static inline double futrts_asin64(double x)
{
    return asin(x);
}
static inline double futrts_atan64(double x)
{
    return atan(x);
}
static inline double futrts_atan2_64(double x, double y)
{
    return atan2(x, y);
}
static inline double futrts_gamma64(double x)
{
    return tgamma(x);
}
static inline double futrts_lgamma64(double x)
{
    return lgamma(x);
}
static inline double futrts_round64(double x)
{
    return rint(x);
}
static inline double futrts_ceil64(double x)
{
    return ceil(x);
}
static inline double futrts_floor64(double x)
{
    return floor(x);
}
static inline char futrts_isnan64(double x)
{
    return isnan(x);
}
static inline char futrts_isinf64(double x)
{
    return isinf(x);
}
static inline int64_t futrts_to_bits64(double x)
{
    union {
        double f;
        int64_t t;
    } p;
    
    p.f = x;
    return p.t;
}
static inline double futrts_from_bits64(int64_t x)
{
    union {
        int64_t f;
        double t;
    } p;
    
    p.f = x;
    return p.t;
}
static inline float fmod64(float x, float y)
{
    return fmod(x, y);
}
#ifdef __OPENCL_VERSION__
static inline double futrts_lerp64(double v0, double v1, double t)
{
    return mix(v0, v1, t);
}
#else
static inline double futrts_lerp64(double v0, double v1, double t)
{
    return v0 + (v1 - v0) * t;
}
#endif
static int futrts_fouriertest(struct futhark_context *ctx,
                              struct memblock *out_mem_p_23010,
                              int32_t *out_out_arrsizze_23011,
                              int32_t *out_out_arrsizze_23012,
                              int32_t *out_out_arrsizze_23013,
                              struct memblock input_mem_22678, int32_t n_22232,
                              int32_t m_22233, int32_t k_22234)
{
    struct memblock out_mem_22938;
    
    out_mem_22938.references = NULL;
    
    int32_t out_arrsizze_22939;
    int32_t out_arrsizze_22940;
    int32_t out_arrsizze_22941;
    bool loop_cond_22236 = slt32(1, m_22233);
    bool r_22237;
    int32_t r_22238;
    int32_t r_22239;
    bool loop_while_22240;
    int32_t r_22241;
    int32_t n_22242;
    
    loop_while_22240 = loop_cond_22236;
    r_22241 = 0;
    n_22242 = m_22233;
    while (loop_while_22240) {
        int32_t n_22243 = sdiv32(n_22242, 2);
        int32_t r_22244 = 1 + r_22241;
        bool loop_cond_22245 = slt32(1, n_22243);
        bool loop_while_tmp_22942 = loop_cond_22245;
        int32_t r_tmp_22943 = r_22244;
        int32_t n_tmp_22944;
        
        n_tmp_22944 = n_22243;
        loop_while_22240 = loop_while_tmp_22942;
        r_22241 = r_tmp_22943;
        n_22242 = n_tmp_22944;
    }
    r_22237 = loop_while_22240;
    r_22238 = r_22241;
    r_22239 = n_22242;
    
    bool loop_cond_22246 = slt32(1, k_22234);
    bool r_22247;
    int32_t r_22248;
    int32_t r_22249;
    bool loop_while_22250;
    int32_t r_22251;
    int32_t n_22252;
    
    loop_while_22250 = loop_cond_22246;
    r_22251 = 0;
    n_22252 = k_22234;
    while (loop_while_22250) {
        int32_t n_22253 = sdiv32(n_22252, 2);
        int32_t r_22254 = 1 + r_22251;
        bool loop_cond_22255 = slt32(1, n_22253);
        bool loop_while_tmp_22945 = loop_cond_22255;
        int32_t r_tmp_22946 = r_22254;
        int32_t n_tmp_22947;
        
        n_tmp_22947 = n_22253;
        loop_while_22250 = loop_while_tmp_22945;
        r_22251 = r_tmp_22946;
        n_22252 = n_tmp_22947;
    }
    r_22247 = loop_while_22250;
    r_22248 = r_22251;
    r_22249 = n_22252;
    
    int32_t iota_arg_22256 = sdiv32(k_22234, 2);
    int32_t iota_arg_22258 = sdiv32(m_22233, 2);
    bool r_22260;
    int32_t r_22261;
    int32_t r_22262;
    bool loop_while_22263;
    int32_t r_22264;
    int32_t n_22265;
    
    loop_while_22263 = loop_cond_22246;
    r_22264 = 0;
    n_22265 = k_22234;
    while (loop_while_22263) {
        int32_t n_22266 = sdiv32(n_22265, 2);
        int32_t r_22267 = 1 + r_22264;
        bool loop_cond_22268 = slt32(1, n_22266);
        bool loop_while_tmp_22948 = loop_cond_22268;
        int32_t r_tmp_22949 = r_22267;
        int32_t n_tmp_22950;
        
        n_tmp_22950 = n_22266;
        loop_while_22263 = loop_while_tmp_22948;
        r_22264 = r_tmp_22949;
        n_22265 = n_tmp_22950;
    }
    r_22260 = loop_while_22263;
    r_22261 = r_22264;
    r_22262 = n_22265;
    
    bool r_22269;
    int32_t r_22270;
    int32_t r_22271;
    bool loop_while_22272;
    int32_t r_22273;
    int32_t n_22274;
    
    loop_while_22272 = loop_cond_22236;
    r_22273 = 0;
    n_22274 = m_22233;
    while (loop_while_22272) {
        int32_t n_22275 = sdiv32(n_22274, 2);
        int32_t r_22276 = 1 + r_22273;
        bool loop_cond_22277 = slt32(1, n_22275);
        bool loop_while_tmp_22951 = loop_cond_22277;
        int32_t r_tmp_22952 = r_22276;
        int32_t n_tmp_22953;
        
        n_tmp_22953 = n_22275;
        loop_while_22272 = loop_while_tmp_22951;
        r_22273 = r_tmp_22952;
        n_22274 = n_tmp_22953;
    }
    r_22269 = loop_while_22272;
    r_22270 = r_22273;
    r_22271 = n_22274;
    
    int64_t binop_x_22680 = sext_i32_i64(n_22232);
    int64_t binop_y_22681 = sext_i32_i64(m_22233);
    int64_t binop_x_22682 = binop_x_22680 * binop_y_22681;
    int64_t binop_y_22683 = sext_i32_i64(k_22234);
    int64_t binop_x_22684 = binop_x_22682 * binop_y_22683;
    int64_t bytes_22679 = 8 * binop_x_22684;
    struct memblock mem_22685;
    
    mem_22685.references = NULL;
    if (memblock_alloc(ctx, &mem_22685, bytes_22679, "mem_22685"))
        return 1;
    
    int64_t binop_x_22690 = binop_y_22681 * binop_y_22683;
    int64_t bytes_22687 = 8 * binop_x_22690;
    struct memblock mem_22691;
    
    mem_22691.references = NULL;
    if (memblock_alloc(ctx, &mem_22691, bytes_22687, "mem_22691"))
        return 1;
    
    struct memblock mem_22696;
    
    mem_22696.references = NULL;
    if (memblock_alloc(ctx, &mem_22696, bytes_22687, "mem_22696"))
        return 1;
    
    int64_t bytes_22699 = 8 * binop_y_22683;
    struct memblock mem_22701;
    
    mem_22701.references = NULL;
    if (memblock_alloc(ctx, &mem_22701, bytes_22699, "mem_22701"))
        return 1;
    
    struct memblock mem_22704;
    
    mem_22704.references = NULL;
    if (memblock_alloc(ctx, &mem_22704, bytes_22699, "mem_22704"))
        return 1;
    
    struct memblock mem_22707;
    
    mem_22707.references = NULL;
    if (memblock_alloc(ctx, &mem_22707, bytes_22699, "mem_22707"))
        return 1;
    
    struct memblock mem_22710;
    
    mem_22710.references = NULL;
    if (memblock_alloc(ctx, &mem_22710, bytes_22699, "mem_22710"))
        return 1;
    
    int64_t binop_x_22728 = binop_y_22681 * binop_y_22683;
    int64_t bytes_22725 = 8 * binop_x_22728;
    struct memblock mem_22729;
    
    mem_22729.references = NULL;
    if (memblock_alloc(ctx, &mem_22729, bytes_22725, "mem_22729"))
        return 1;
    
    struct memblock mem_22734;
    
    mem_22734.references = NULL;
    if (memblock_alloc(ctx, &mem_22734, bytes_22725, "mem_22734"))
        return 1;
    
    int64_t bytes_22737 = 8 * binop_y_22681;
    struct memblock mem_22739;
    
    mem_22739.references = NULL;
    if (memblock_alloc(ctx, &mem_22739, bytes_22737, "mem_22739"))
        return 1;
    
    struct memblock mem_22742;
    
    mem_22742.references = NULL;
    if (memblock_alloc(ctx, &mem_22742, bytes_22737, "mem_22742"))
        return 1;
    
    struct memblock mem_22745;
    
    mem_22745.references = NULL;
    if (memblock_alloc(ctx, &mem_22745, bytes_22737, "mem_22745"))
        return 1;
    
    struct memblock mem_22748;
    
    mem_22748.references = NULL;
    if (memblock_alloc(ctx, &mem_22748, bytes_22737, "mem_22748"))
        return 1;
    
    struct memblock mem_22763;
    
    mem_22763.references = NULL;
    if (memblock_alloc(ctx, &mem_22763, bytes_22737, "mem_22763"))
        return 1;
    
    struct memblock mem_22766;
    
    mem_22766.references = NULL;
    if (memblock_alloc(ctx, &mem_22766, bytes_22737, "mem_22766"))
        return 1;
    
    struct memblock mem_22769;
    
    mem_22769.references = NULL;
    if (memblock_alloc(ctx, &mem_22769, bytes_22737, "mem_22769"))
        return 1;
    
    struct memblock mem_22772;
    
    mem_22772.references = NULL;
    if (memblock_alloc(ctx, &mem_22772, bytes_22737, "mem_22772"))
        return 1;
    
    int64_t bytes_22793 = 8 * binop_y_22683;
    struct memblock mem_22795;
    
    mem_22795.references = NULL;
    if (memblock_alloc(ctx, &mem_22795, bytes_22793, "mem_22795"))
        return 1;
    
    struct memblock mem_22798;
    
    mem_22798.references = NULL;
    if (memblock_alloc(ctx, &mem_22798, bytes_22793, "mem_22798"))
        return 1;
    
    struct memblock mem_22801;
    
    mem_22801.references = NULL;
    if (memblock_alloc(ctx, &mem_22801, bytes_22793, "mem_22801"))
        return 1;
    
    struct memblock mem_22804;
    
    mem_22804.references = NULL;
    if (memblock_alloc(ctx, &mem_22804, bytes_22793, "mem_22804"))
        return 1;
    for (int32_t i_22647 = 0; i_22647 < n_22232; i_22647++) {
        for (int32_t i_22541 = 0; i_22541 < m_22233; i_22541++) {
            for (int32_t i_22957 = 0; i_22957 < k_22234; i_22957++) {
                ((double *) mem_22701.mem)[i_22957] = 0.0;
            }
            memmove(mem_22704.mem + 0, input_mem_22678.mem + (i_22647 *
                                                              (k_22234 *
                                                               m_22233) +
                                                              i_22541 *
                                                              k_22234) * 8,
                    k_22234 * sizeof(double));
            memmove(mem_22707.mem + 0, input_mem_22678.mem + (i_22647 *
                                                              (k_22234 *
                                                               m_22233) +
                                                              i_22541 *
                                                              k_22234) * 8,
                    k_22234 * sizeof(double));
            for (int32_t i_22958 = 0; i_22958 < k_22234; i_22958++) {
                ((double *) mem_22710.mem)[i_22958] = 0.0;
            }
            
            struct memblock res_mem_22719;
            
            res_mem_22719.references = NULL;
            
            struct memblock res_mem_22720;
            
            res_mem_22720.references = NULL;
            
            struct memblock res_mem_22721;
            
            res_mem_22721.references = NULL;
            
            struct memblock res_mem_22722;
            
            res_mem_22722.references = NULL;
            
            struct memblock inputzq_mem_22711;
            
            inputzq_mem_22711.references = NULL;
            
            struct memblock inputzq_mem_22712;
            
            inputzq_mem_22712.references = NULL;
            
            struct memblock outputzq_mem_22713;
            
            outputzq_mem_22713.references = NULL;
            
            struct memblock outputzq_mem_22714;
            
            outputzq_mem_22714.references = NULL;
            if (memblock_set(ctx, &inputzq_mem_22711, &mem_22707,
                             "mem_22707") != 0)
                return 1;
            if (memblock_set(ctx, &inputzq_mem_22712, &mem_22710,
                             "mem_22710") != 0)
                return 1;
            if (memblock_set(ctx, &outputzq_mem_22713, &mem_22704,
                             "mem_22704") != 0)
                return 1;
            if (memblock_set(ctx, &outputzq_mem_22714, &mem_22701,
                             "mem_22701") != 0)
                return 1;
            for (int32_t i_22299 = 0; i_22299 < r_22248; i_22299++) {
                int32_t index_primexp_22300 = 1 << i_22299;
                int32_t i32_arg_22301 = 2 * index_primexp_22300;
                double res_22302 = sitofp_i32_f64(i32_arg_22301);
                
                for (int32_t write_iter_22507 = 0; write_iter_22507 <
                     iota_arg_22256; write_iter_22507++) {
                    double write_iv_22511 =
                           ((double *) inputzq_mem_22711.mem)[write_iter_22507];
                    double write_iv_22512 =
                           ((double *) inputzq_mem_22712.mem)[write_iter_22507];
                    int32_t i32_arg_22310 = smod32(write_iter_22507,
                                                   index_primexp_22300);
                    double res_22311 = sitofp_i32_f64(i32_arg_22310);
                    double res_22312 = -6.283185307179586 * res_22311;
                    double res_22313 = res_22312 / res_22302;
                    int32_t i_22314 = iota_arg_22256 + write_iter_22507;
                    double zt_arg_22315 =
                           ((double *) inputzq_mem_22711.mem)[i_22314];
                    double zt_arg_22316 =
                           ((double *) inputzq_mem_22712.mem)[i_22314];
                    double res_22317;
                    
                    res_22317 = futrts_cos64(res_22313);
                    
                    double res_22318;
                    
                    res_22318 = futrts_sin64(res_22313);
                    
                    double res_22319 = zt_arg_22315 * res_22317;
                    double res_22320 = zt_arg_22316 * res_22318;
                    double res_22321 = res_22319 - res_22320;
                    double res_22322 = zt_arg_22316 * res_22317;
                    double res_22323 = zt_arg_22315 * res_22318;
                    double res_22324 = res_22322 + res_22323;
                    double res_22325 = res_22321 + write_iv_22511;
                    double res_22326 = res_22324 + write_iv_22512;
                    double res_22327 = write_iv_22511 - res_22321;
                    double res_22328 = write_iv_22512 - res_22324;
                    int32_t x_22329 = sdiv32(write_iter_22507,
                                             index_primexp_22300);
                    int32_t x_22330 = index_primexp_22300 * x_22329;
                    int32_t x_22331 = 2 * x_22330;
                    int32_t idxD_22332 = i32_arg_22310 + x_22331;
                    int32_t res_22333 = index_primexp_22300 + idxD_22332;
                    bool less_than_zzero_22513 = slt32(idxD_22332, 0);
                    bool greater_than_sizze_22514 = sle32(k_22234, idxD_22332);
                    bool outside_bounds_dim_22515 = less_than_zzero_22513 ||
                         greater_than_sizze_22514;
                    
                    if (!outside_bounds_dim_22515) {
                        ((double *) outputzq_mem_22713.mem)[idxD_22332] =
                            res_22325;
                    }
                    
                    bool less_than_zzero_22519 = slt32(res_22333, 0);
                    bool greater_than_sizze_22520 = sle32(k_22234, res_22333);
                    bool outside_bounds_dim_22521 = less_than_zzero_22519 ||
                         greater_than_sizze_22520;
                    
                    if (!outside_bounds_dim_22521) {
                        ((double *) outputzq_mem_22713.mem)[res_22333] =
                            res_22327;
                    }
                    if (!outside_bounds_dim_22515) {
                        ((double *) outputzq_mem_22714.mem)[idxD_22332] =
                            res_22326;
                    }
                    if (!outside_bounds_dim_22521) {
                        ((double *) outputzq_mem_22714.mem)[res_22333] =
                            res_22328;
                    }
                }
                
                struct memblock inputzq_mem_tmp_22959;
                
                inputzq_mem_tmp_22959.references = NULL;
                if (memblock_set(ctx, &inputzq_mem_tmp_22959,
                                 &outputzq_mem_22713, "outputzq_mem_22713") !=
                    0)
                    return 1;
                
                struct memblock inputzq_mem_tmp_22960;
                
                inputzq_mem_tmp_22960.references = NULL;
                if (memblock_set(ctx, &inputzq_mem_tmp_22960,
                                 &outputzq_mem_22714, "outputzq_mem_22714") !=
                    0)
                    return 1;
                
                struct memblock outputzq_mem_tmp_22961;
                
                outputzq_mem_tmp_22961.references = NULL;
                if (memblock_set(ctx, &outputzq_mem_tmp_22961,
                                 &inputzq_mem_22711, "inputzq_mem_22711") != 0)
                    return 1;
                
                struct memblock outputzq_mem_tmp_22962;
                
                outputzq_mem_tmp_22962.references = NULL;
                if (memblock_set(ctx, &outputzq_mem_tmp_22962,
                                 &inputzq_mem_22712, "inputzq_mem_22712") != 0)
                    return 1;
                if (memblock_set(ctx, &inputzq_mem_22711,
                                 &inputzq_mem_tmp_22959,
                                 "inputzq_mem_tmp_22959") != 0)
                    return 1;
                if (memblock_set(ctx, &inputzq_mem_22712,
                                 &inputzq_mem_tmp_22960,
                                 "inputzq_mem_tmp_22960") != 0)
                    return 1;
                if (memblock_set(ctx, &outputzq_mem_22713,
                                 &outputzq_mem_tmp_22961,
                                 "outputzq_mem_tmp_22961") != 0)
                    return 1;
                if (memblock_set(ctx, &outputzq_mem_22714,
                                 &outputzq_mem_tmp_22962,
                                 "outputzq_mem_tmp_22962") != 0)
                    return 1;
                if (memblock_unref(ctx, &outputzq_mem_tmp_22962,
                                   "outputzq_mem_tmp_22962") != 0)
                    return 1;
                if (memblock_unref(ctx, &outputzq_mem_tmp_22961,
                                   "outputzq_mem_tmp_22961") != 0)
                    return 1;
                if (memblock_unref(ctx, &inputzq_mem_tmp_22960,
                                   "inputzq_mem_tmp_22960") != 0)
                    return 1;
                if (memblock_unref(ctx, &inputzq_mem_tmp_22959,
                                   "inputzq_mem_tmp_22959") != 0)
                    return 1;
            }
            if (memblock_set(ctx, &res_mem_22719, &inputzq_mem_22711,
                             "inputzq_mem_22711") != 0)
                return 1;
            if (memblock_set(ctx, &res_mem_22720, &inputzq_mem_22712,
                             "inputzq_mem_22712") != 0)
                return 1;
            if (memblock_set(ctx, &res_mem_22721, &outputzq_mem_22713,
                             "outputzq_mem_22713") != 0)
                return 1;
            if (memblock_set(ctx, &res_mem_22722, &outputzq_mem_22714,
                             "outputzq_mem_22714") != 0)
                return 1;
            memmove(mem_22696.mem + i_22541 * k_22234 * 8, res_mem_22720.mem +
                    0, k_22234 * sizeof(double));
            if (memblock_unref(ctx, &res_mem_22720, "res_mem_22720") != 0)
                return 1;
            memmove(mem_22691.mem + i_22541 * k_22234 * 8, res_mem_22719.mem +
                    0, k_22234 * sizeof(double));
            if (memblock_unref(ctx, &res_mem_22719, "res_mem_22719") != 0)
                return 1;
            if (memblock_unref(ctx, &outputzq_mem_22714,
                               "outputzq_mem_22714") != 0)
                return 1;
            if (memblock_unref(ctx, &outputzq_mem_22713,
                               "outputzq_mem_22713") != 0)
                return 1;
            if (memblock_unref(ctx, &inputzq_mem_22712, "inputzq_mem_22712") !=
                0)
                return 1;
            if (memblock_unref(ctx, &inputzq_mem_22711, "inputzq_mem_22711") !=
                0)
                return 1;
            if (memblock_unref(ctx, &res_mem_22722, "res_mem_22722") != 0)
                return 1;
            if (memblock_unref(ctx, &res_mem_22721, "res_mem_22721") != 0)
                return 1;
            if (memblock_unref(ctx, &res_mem_22720, "res_mem_22720") != 0)
                return 1;
            if (memblock_unref(ctx, &res_mem_22719, "res_mem_22719") != 0)
                return 1;
        }
        for (int32_t i_22608 = 0; i_22608 < k_22234; i_22608++) {
            for (int32_t i_22971 = 0; i_22971 < m_22233; i_22971++) {
                ((double *) mem_22739.mem)[i_22971] =
                    ((double *) mem_22691.mem)[i_22608 + i_22971 * k_22234];
            }
            for (int32_t i_22972 = 0; i_22972 < m_22233; i_22972++) {
                ((double *) mem_22742.mem)[i_22972] =
                    ((double *) mem_22696.mem)[i_22608 + i_22972 * k_22234];
            }
            for (int32_t i_22973 = 0; i_22973 < m_22233; i_22973++) {
                ((double *) mem_22745.mem)[i_22973] =
                    ((double *) mem_22691.mem)[i_22608 + i_22973 * k_22234];
            }
            for (int32_t i_22974 = 0; i_22974 < m_22233; i_22974++) {
                ((double *) mem_22748.mem)[i_22974] =
                    ((double *) mem_22696.mem)[i_22608 + i_22974 * k_22234];
            }
            
            struct memblock res_mem_22757;
            
            res_mem_22757.references = NULL;
            
            struct memblock res_mem_22758;
            
            res_mem_22758.references = NULL;
            
            struct memblock res_mem_22759;
            
            res_mem_22759.references = NULL;
            
            struct memblock res_mem_22760;
            
            res_mem_22760.references = NULL;
            
            struct memblock inputzq_mem_22749;
            
            inputzq_mem_22749.references = NULL;
            
            struct memblock inputzq_mem_22750;
            
            inputzq_mem_22750.references = NULL;
            
            struct memblock outputzq_mem_22751;
            
            outputzq_mem_22751.references = NULL;
            
            struct memblock outputzq_mem_22752;
            
            outputzq_mem_22752.references = NULL;
            if (memblock_set(ctx, &inputzq_mem_22749, &mem_22739,
                             "mem_22739") != 0)
                return 1;
            if (memblock_set(ctx, &inputzq_mem_22750, &mem_22742,
                             "mem_22742") != 0)
                return 1;
            if (memblock_set(ctx, &outputzq_mem_22751, &mem_22745,
                             "mem_22745") != 0)
                return 1;
            if (memblock_set(ctx, &outputzq_mem_22752, &mem_22748,
                             "mem_22748") != 0)
                return 1;
            for (int32_t i_22359 = 0; i_22359 < r_22238; i_22359++) {
                int32_t index_primexp_22360 = 1 << i_22359;
                int32_t i32_arg_22361 = 2 * index_primexp_22360;
                double res_22362 = sitofp_i32_f64(i32_arg_22361);
                
                for (int32_t write_iter_22544 = 0; write_iter_22544 <
                     iota_arg_22258; write_iter_22544++) {
                    double write_iv_22548 =
                           ((double *) inputzq_mem_22749.mem)[write_iter_22544];
                    double write_iv_22549 =
                           ((double *) inputzq_mem_22750.mem)[write_iter_22544];
                    int32_t i32_arg_22370 = smod32(write_iter_22544,
                                                   index_primexp_22360);
                    double res_22371 = sitofp_i32_f64(i32_arg_22370);
                    double res_22372 = -6.283185307179586 * res_22371;
                    double res_22373 = res_22372 / res_22362;
                    int32_t i_22374 = iota_arg_22258 + write_iter_22544;
                    double zt_arg_22375 =
                           ((double *) inputzq_mem_22749.mem)[i_22374];
                    double zt_arg_22376 =
                           ((double *) inputzq_mem_22750.mem)[i_22374];
                    double res_22377;
                    
                    res_22377 = futrts_cos64(res_22373);
                    
                    double res_22378;
                    
                    res_22378 = futrts_sin64(res_22373);
                    
                    double res_22379 = zt_arg_22375 * res_22377;
                    double res_22380 = zt_arg_22376 * res_22378;
                    double res_22381 = res_22379 - res_22380;
                    double res_22382 = zt_arg_22376 * res_22377;
                    double res_22383 = zt_arg_22375 * res_22378;
                    double res_22384 = res_22382 + res_22383;
                    double res_22385 = res_22381 + write_iv_22548;
                    double res_22386 = res_22384 + write_iv_22549;
                    double res_22387 = write_iv_22548 - res_22381;
                    double res_22388 = write_iv_22549 - res_22384;
                    int32_t x_22389 = sdiv32(write_iter_22544,
                                             index_primexp_22360);
                    int32_t x_22390 = index_primexp_22360 * x_22389;
                    int32_t x_22391 = 2 * x_22390;
                    int32_t idxD_22392 = i32_arg_22370 + x_22391;
                    int32_t res_22393 = index_primexp_22360 + idxD_22392;
                    bool less_than_zzero_22550 = slt32(idxD_22392, 0);
                    bool greater_than_sizze_22551 = sle32(m_22233, idxD_22392);
                    bool outside_bounds_dim_22552 = less_than_zzero_22550 ||
                         greater_than_sizze_22551;
                    
                    if (!outside_bounds_dim_22552) {
                        ((double *) outputzq_mem_22751.mem)[idxD_22392] =
                            res_22385;
                    }
                    
                    bool less_than_zzero_22556 = slt32(res_22393, 0);
                    bool greater_than_sizze_22557 = sle32(m_22233, res_22393);
                    bool outside_bounds_dim_22558 = less_than_zzero_22556 ||
                         greater_than_sizze_22557;
                    
                    if (!outside_bounds_dim_22558) {
                        ((double *) outputzq_mem_22751.mem)[res_22393] =
                            res_22387;
                    }
                    if (!outside_bounds_dim_22552) {
                        ((double *) outputzq_mem_22752.mem)[idxD_22392] =
                            res_22386;
                    }
                    if (!outside_bounds_dim_22558) {
                        ((double *) outputzq_mem_22752.mem)[res_22393] =
                            res_22388;
                    }
                }
                
                struct memblock inputzq_mem_tmp_22975;
                
                inputzq_mem_tmp_22975.references = NULL;
                if (memblock_set(ctx, &inputzq_mem_tmp_22975,
                                 &outputzq_mem_22751, "outputzq_mem_22751") !=
                    0)
                    return 1;
                
                struct memblock inputzq_mem_tmp_22976;
                
                inputzq_mem_tmp_22976.references = NULL;
                if (memblock_set(ctx, &inputzq_mem_tmp_22976,
                                 &outputzq_mem_22752, "outputzq_mem_22752") !=
                    0)
                    return 1;
                
                struct memblock outputzq_mem_tmp_22977;
                
                outputzq_mem_tmp_22977.references = NULL;
                if (memblock_set(ctx, &outputzq_mem_tmp_22977,
                                 &inputzq_mem_22749, "inputzq_mem_22749") != 0)
                    return 1;
                
                struct memblock outputzq_mem_tmp_22978;
                
                outputzq_mem_tmp_22978.references = NULL;
                if (memblock_set(ctx, &outputzq_mem_tmp_22978,
                                 &inputzq_mem_22750, "inputzq_mem_22750") != 0)
                    return 1;
                if (memblock_set(ctx, &inputzq_mem_22749,
                                 &inputzq_mem_tmp_22975,
                                 "inputzq_mem_tmp_22975") != 0)
                    return 1;
                if (memblock_set(ctx, &inputzq_mem_22750,
                                 &inputzq_mem_tmp_22976,
                                 "inputzq_mem_tmp_22976") != 0)
                    return 1;
                if (memblock_set(ctx, &outputzq_mem_22751,
                                 &outputzq_mem_tmp_22977,
                                 "outputzq_mem_tmp_22977") != 0)
                    return 1;
                if (memblock_set(ctx, &outputzq_mem_22752,
                                 &outputzq_mem_tmp_22978,
                                 "outputzq_mem_tmp_22978") != 0)
                    return 1;
                if (memblock_unref(ctx, &outputzq_mem_tmp_22978,
                                   "outputzq_mem_tmp_22978") != 0)
                    return 1;
                if (memblock_unref(ctx, &outputzq_mem_tmp_22977,
                                   "outputzq_mem_tmp_22977") != 0)
                    return 1;
                if (memblock_unref(ctx, &inputzq_mem_tmp_22976,
                                   "inputzq_mem_tmp_22976") != 0)
                    return 1;
                if (memblock_unref(ctx, &inputzq_mem_tmp_22975,
                                   "inputzq_mem_tmp_22975") != 0)
                    return 1;
            }
            if (memblock_set(ctx, &res_mem_22757, &inputzq_mem_22749,
                             "inputzq_mem_22749") != 0)
                return 1;
            if (memblock_set(ctx, &res_mem_22758, &inputzq_mem_22750,
                             "inputzq_mem_22750") != 0)
                return 1;
            if (memblock_set(ctx, &res_mem_22759, &outputzq_mem_22751,
                             "outputzq_mem_22751") != 0)
                return 1;
            if (memblock_set(ctx, &res_mem_22760, &outputzq_mem_22752,
                             "outputzq_mem_22752") != 0)
                return 1;
            memmove(mem_22763.mem + 0, res_mem_22757.mem + 0, m_22233 *
                    sizeof(double));
            memmove(mem_22766.mem + 0, res_mem_22758.mem + 0, m_22233 *
                    sizeof(double));
            memmove(mem_22769.mem + 0, res_mem_22757.mem + 0, m_22233 *
                    sizeof(double));
            if (memblock_unref(ctx, &res_mem_22757, "res_mem_22757") != 0)
                return 1;
            memmove(mem_22772.mem + 0, res_mem_22758.mem + 0, m_22233 *
                    sizeof(double));
            if (memblock_unref(ctx, &res_mem_22758, "res_mem_22758") != 0)
                return 1;
            
            struct memblock res_mem_22781;
            
            res_mem_22781.references = NULL;
            
            struct memblock res_mem_22782;
            
            res_mem_22782.references = NULL;
            
            struct memblock res_mem_22783;
            
            res_mem_22783.references = NULL;
            
            struct memblock res_mem_22784;
            
            res_mem_22784.references = NULL;
            
            struct memblock inputzq_mem_22773;
            
            inputzq_mem_22773.references = NULL;
            
            struct memblock inputzq_mem_22774;
            
            inputzq_mem_22774.references = NULL;
            
            struct memblock outputzq_mem_22775;
            
            outputzq_mem_22775.references = NULL;
            
            struct memblock outputzq_mem_22776;
            
            outputzq_mem_22776.references = NULL;
            if (memblock_set(ctx, &inputzq_mem_22773, &mem_22763,
                             "mem_22763") != 0)
                return 1;
            if (memblock_set(ctx, &inputzq_mem_22774, &mem_22766,
                             "mem_22766") != 0)
                return 1;
            if (memblock_set(ctx, &outputzq_mem_22775, &mem_22769,
                             "mem_22769") != 0)
                return 1;
            if (memblock_set(ctx, &outputzq_mem_22776, &mem_22772,
                             "mem_22772") != 0)
                return 1;
            for (int32_t i_22413 = 0; i_22413 < r_22270; i_22413++) {
                int32_t index_primexp_22414 = 1 << i_22413;
                int32_t i32_arg_22415 = 2 * index_primexp_22414;
                double res_22416 = sitofp_i32_f64(i32_arg_22415);
                
                for (int32_t write_iter_22574 = 0; write_iter_22574 <
                     iota_arg_22258; write_iter_22574++) {
                    double write_iv_22578 =
                           ((double *) inputzq_mem_22773.mem)[write_iter_22574];
                    double write_iv_22579 =
                           ((double *) inputzq_mem_22774.mem)[write_iter_22574];
                    int32_t i32_arg_22424 = smod32(write_iter_22574,
                                                   index_primexp_22414);
                    double res_22425 = sitofp_i32_f64(i32_arg_22424);
                    double res_22426 = -6.283185307179586 * res_22425;
                    double res_22427 = res_22426 / res_22416;
                    int32_t i_22428 = iota_arg_22258 + write_iter_22574;
                    double zt_arg_22429 =
                           ((double *) inputzq_mem_22773.mem)[i_22428];
                    double zt_arg_22430 =
                           ((double *) inputzq_mem_22774.mem)[i_22428];
                    double res_22431;
                    
                    res_22431 = futrts_cos64(res_22427);
                    
                    double res_22432;
                    
                    res_22432 = futrts_sin64(res_22427);
                    
                    double res_22433 = zt_arg_22429 * res_22431;
                    double res_22434 = zt_arg_22430 * res_22432;
                    double res_22435 = res_22433 - res_22434;
                    double res_22436 = zt_arg_22430 * res_22431;
                    double res_22437 = zt_arg_22429 * res_22432;
                    double res_22438 = res_22436 + res_22437;
                    double res_22439 = res_22435 + write_iv_22578;
                    double res_22440 = res_22438 + write_iv_22579;
                    double res_22441 = write_iv_22578 - res_22435;
                    double res_22442 = write_iv_22579 - res_22438;
                    int32_t x_22443 = sdiv32(write_iter_22574,
                                             index_primexp_22414);
                    int32_t x_22444 = index_primexp_22414 * x_22443;
                    int32_t x_22445 = 2 * x_22444;
                    int32_t idxD_22446 = i32_arg_22424 + x_22445;
                    int32_t res_22447 = index_primexp_22414 + idxD_22446;
                    bool less_than_zzero_22580 = slt32(idxD_22446, 0);
                    bool greater_than_sizze_22581 = sle32(m_22233, idxD_22446);
                    bool outside_bounds_dim_22582 = less_than_zzero_22580 ||
                         greater_than_sizze_22581;
                    
                    if (!outside_bounds_dim_22582) {
                        ((double *) outputzq_mem_22775.mem)[idxD_22446] =
                            res_22439;
                    }
                    
                    bool less_than_zzero_22586 = slt32(res_22447, 0);
                    bool greater_than_sizze_22587 = sle32(m_22233, res_22447);
                    bool outside_bounds_dim_22588 = less_than_zzero_22586 ||
                         greater_than_sizze_22587;
                    
                    if (!outside_bounds_dim_22588) {
                        ((double *) outputzq_mem_22775.mem)[res_22447] =
                            res_22441;
                    }
                    if (!outside_bounds_dim_22582) {
                        ((double *) outputzq_mem_22776.mem)[idxD_22446] =
                            res_22440;
                    }
                    if (!outside_bounds_dim_22588) {
                        ((double *) outputzq_mem_22776.mem)[res_22447] =
                            res_22442;
                    }
                }
                
                struct memblock inputzq_mem_tmp_22985;
                
                inputzq_mem_tmp_22985.references = NULL;
                if (memblock_set(ctx, &inputzq_mem_tmp_22985,
                                 &outputzq_mem_22775, "outputzq_mem_22775") !=
                    0)
                    return 1;
                
                struct memblock inputzq_mem_tmp_22986;
                
                inputzq_mem_tmp_22986.references = NULL;
                if (memblock_set(ctx, &inputzq_mem_tmp_22986,
                                 &outputzq_mem_22776, "outputzq_mem_22776") !=
                    0)
                    return 1;
                
                struct memblock outputzq_mem_tmp_22987;
                
                outputzq_mem_tmp_22987.references = NULL;
                if (memblock_set(ctx, &outputzq_mem_tmp_22987,
                                 &inputzq_mem_22773, "inputzq_mem_22773") != 0)
                    return 1;
                
                struct memblock outputzq_mem_tmp_22988;
                
                outputzq_mem_tmp_22988.references = NULL;
                if (memblock_set(ctx, &outputzq_mem_tmp_22988,
                                 &inputzq_mem_22774, "inputzq_mem_22774") != 0)
                    return 1;
                if (memblock_set(ctx, &inputzq_mem_22773,
                                 &inputzq_mem_tmp_22985,
                                 "inputzq_mem_tmp_22985") != 0)
                    return 1;
                if (memblock_set(ctx, &inputzq_mem_22774,
                                 &inputzq_mem_tmp_22986,
                                 "inputzq_mem_tmp_22986") != 0)
                    return 1;
                if (memblock_set(ctx, &outputzq_mem_22775,
                                 &outputzq_mem_tmp_22987,
                                 "outputzq_mem_tmp_22987") != 0)
                    return 1;
                if (memblock_set(ctx, &outputzq_mem_22776,
                                 &outputzq_mem_tmp_22988,
                                 "outputzq_mem_tmp_22988") != 0)
                    return 1;
                if (memblock_unref(ctx, &outputzq_mem_tmp_22988,
                                   "outputzq_mem_tmp_22988") != 0)
                    return 1;
                if (memblock_unref(ctx, &outputzq_mem_tmp_22987,
                                   "outputzq_mem_tmp_22987") != 0)
                    return 1;
                if (memblock_unref(ctx, &inputzq_mem_tmp_22986,
                                   "inputzq_mem_tmp_22986") != 0)
                    return 1;
                if (memblock_unref(ctx, &inputzq_mem_tmp_22985,
                                   "inputzq_mem_tmp_22985") != 0)
                    return 1;
            }
            if (memblock_set(ctx, &res_mem_22781, &inputzq_mem_22773,
                             "inputzq_mem_22773") != 0)
                return 1;
            if (memblock_set(ctx, &res_mem_22782, &inputzq_mem_22774,
                             "inputzq_mem_22774") != 0)
                return 1;
            if (memblock_set(ctx, &res_mem_22783, &outputzq_mem_22775,
                             "outputzq_mem_22775") != 0)
                return 1;
            if (memblock_set(ctx, &res_mem_22784, &outputzq_mem_22776,
                             "outputzq_mem_22776") != 0)
                return 1;
            memmove(mem_22734.mem + i_22608 * m_22233 * 8, res_mem_22782.mem +
                    0, m_22233 * sizeof(double));
            if (memblock_unref(ctx, &res_mem_22782, "res_mem_22782") != 0)
                return 1;
            memmove(mem_22729.mem + i_22608 * m_22233 * 8, res_mem_22781.mem +
                    0, m_22233 * sizeof(double));
            if (memblock_unref(ctx, &res_mem_22781, "res_mem_22781") != 0)
                return 1;
            if (memblock_unref(ctx, &outputzq_mem_22776,
                               "outputzq_mem_22776") != 0)
                return 1;
            if (memblock_unref(ctx, &outputzq_mem_22775,
                               "outputzq_mem_22775") != 0)
                return 1;
            if (memblock_unref(ctx, &inputzq_mem_22774, "inputzq_mem_22774") !=
                0)
                return 1;
            if (memblock_unref(ctx, &inputzq_mem_22773, "inputzq_mem_22773") !=
                0)
                return 1;
            if (memblock_unref(ctx, &res_mem_22784, "res_mem_22784") != 0)
                return 1;
            if (memblock_unref(ctx, &res_mem_22783, "res_mem_22783") != 0)
                return 1;
            if (memblock_unref(ctx, &res_mem_22782, "res_mem_22782") != 0)
                return 1;
            if (memblock_unref(ctx, &res_mem_22781, "res_mem_22781") != 0)
                return 1;
            if (memblock_unref(ctx, &outputzq_mem_22752,
                               "outputzq_mem_22752") != 0)
                return 1;
            if (memblock_unref(ctx, &outputzq_mem_22751,
                               "outputzq_mem_22751") != 0)
                return 1;
            if (memblock_unref(ctx, &inputzq_mem_22750, "inputzq_mem_22750") !=
                0)
                return 1;
            if (memblock_unref(ctx, &inputzq_mem_22749, "inputzq_mem_22749") !=
                0)
                return 1;
            if (memblock_unref(ctx, &res_mem_22760, "res_mem_22760") != 0)
                return 1;
            if (memblock_unref(ctx, &res_mem_22759, "res_mem_22759") != 0)
                return 1;
            if (memblock_unref(ctx, &res_mem_22758, "res_mem_22758") != 0)
                return 1;
            if (memblock_unref(ctx, &res_mem_22757, "res_mem_22757") != 0)
                return 1;
        }
        for (int32_t i_22643 = 0; i_22643 < m_22233; i_22643++) {
            for (int32_t i_22996 = 0; i_22996 < k_22234; i_22996++) {
                ((double *) mem_22795.mem)[i_22996] =
                    ((double *) mem_22729.mem)[i_22643 + i_22996 * m_22233];
            }
            for (int32_t i_22997 = 0; i_22997 < k_22234; i_22997++) {
                ((double *) mem_22798.mem)[i_22997] =
                    ((double *) mem_22734.mem)[i_22643 + i_22997 * m_22233];
            }
            for (int32_t i_22998 = 0; i_22998 < k_22234; i_22998++) {
                ((double *) mem_22801.mem)[i_22998] =
                    ((double *) mem_22729.mem)[i_22643 + i_22998 * m_22233];
            }
            for (int32_t i_22999 = 0; i_22999 < k_22234; i_22999++) {
                ((double *) mem_22804.mem)[i_22999] =
                    ((double *) mem_22734.mem)[i_22643 + i_22999 * m_22233];
            }
            
            struct memblock res_mem_22813;
            
            res_mem_22813.references = NULL;
            
            struct memblock res_mem_22814;
            
            res_mem_22814.references = NULL;
            
            struct memblock res_mem_22815;
            
            res_mem_22815.references = NULL;
            
            struct memblock res_mem_22816;
            
            res_mem_22816.references = NULL;
            
            struct memblock inputzq_mem_22805;
            
            inputzq_mem_22805.references = NULL;
            
            struct memblock inputzq_mem_22806;
            
            inputzq_mem_22806.references = NULL;
            
            struct memblock outputzq_mem_22807;
            
            outputzq_mem_22807.references = NULL;
            
            struct memblock outputzq_mem_22808;
            
            outputzq_mem_22808.references = NULL;
            if (memblock_set(ctx, &inputzq_mem_22805, &mem_22795,
                             "mem_22795") != 0)
                return 1;
            if (memblock_set(ctx, &inputzq_mem_22806, &mem_22798,
                             "mem_22798") != 0)
                return 1;
            if (memblock_set(ctx, &outputzq_mem_22807, &mem_22801,
                             "mem_22801") != 0)
                return 1;
            if (memblock_set(ctx, &outputzq_mem_22808, &mem_22804,
                             "mem_22804") != 0)
                return 1;
            for (int32_t i_22472 = 0; i_22472 < r_22261; i_22472++) {
                int32_t index_primexp_22473 = 1 << i_22472;
                int32_t i32_arg_22474 = 2 * index_primexp_22473;
                double res_22475 = sitofp_i32_f64(i32_arg_22474);
                
                for (int32_t write_iter_22611 = 0; write_iter_22611 <
                     iota_arg_22256; write_iter_22611++) {
                    double write_iv_22615 =
                           ((double *) inputzq_mem_22805.mem)[write_iter_22611];
                    double write_iv_22616 =
                           ((double *) inputzq_mem_22806.mem)[write_iter_22611];
                    int32_t i32_arg_22483 = smod32(write_iter_22611,
                                                   index_primexp_22473);
                    double res_22484 = sitofp_i32_f64(i32_arg_22483);
                    double res_22485 = -6.283185307179586 * res_22484;
                    double res_22486 = res_22485 / res_22475;
                    int32_t i_22487 = iota_arg_22256 + write_iter_22611;
                    double zt_arg_22488 =
                           ((double *) inputzq_mem_22805.mem)[i_22487];
                    double zt_arg_22489 =
                           ((double *) inputzq_mem_22806.mem)[i_22487];
                    double res_22490;
                    
                    res_22490 = futrts_cos64(res_22486);
                    
                    double res_22491;
                    
                    res_22491 = futrts_sin64(res_22486);
                    
                    double res_22492 = zt_arg_22488 * res_22490;
                    double res_22493 = zt_arg_22489 * res_22491;
                    double res_22494 = res_22492 - res_22493;
                    double res_22495 = zt_arg_22489 * res_22490;
                    double res_22496 = zt_arg_22488 * res_22491;
                    double res_22497 = res_22495 + res_22496;
                    double res_22498 = res_22494 + write_iv_22615;
                    double res_22499 = res_22497 + write_iv_22616;
                    double res_22500 = write_iv_22615 - res_22494;
                    double res_22501 = write_iv_22616 - res_22497;
                    int32_t x_22502 = sdiv32(write_iter_22611,
                                             index_primexp_22473);
                    int32_t x_22503 = index_primexp_22473 * x_22502;
                    int32_t x_22504 = 2 * x_22503;
                    int32_t idxD_22505 = i32_arg_22483 + x_22504;
                    int32_t res_22506 = index_primexp_22473 + idxD_22505;
                    bool less_than_zzero_22617 = slt32(idxD_22505, 0);
                    bool greater_than_sizze_22618 = sle32(k_22234, idxD_22505);
                    bool outside_bounds_dim_22619 = less_than_zzero_22617 ||
                         greater_than_sizze_22618;
                    
                    if (!outside_bounds_dim_22619) {
                        ((double *) outputzq_mem_22807.mem)[idxD_22505] =
                            res_22498;
                    }
                    
                    bool less_than_zzero_22623 = slt32(res_22506, 0);
                    bool greater_than_sizze_22624 = sle32(k_22234, res_22506);
                    bool outside_bounds_dim_22625 = less_than_zzero_22623 ||
                         greater_than_sizze_22624;
                    
                    if (!outside_bounds_dim_22625) {
                        ((double *) outputzq_mem_22807.mem)[res_22506] =
                            res_22500;
                    }
                    if (!outside_bounds_dim_22619) {
                        ((double *) outputzq_mem_22808.mem)[idxD_22505] =
                            res_22499;
                    }
                    if (!outside_bounds_dim_22625) {
                        ((double *) outputzq_mem_22808.mem)[res_22506] =
                            res_22501;
                    }
                }
                
                struct memblock inputzq_mem_tmp_23000;
                
                inputzq_mem_tmp_23000.references = NULL;
                if (memblock_set(ctx, &inputzq_mem_tmp_23000,
                                 &outputzq_mem_22807, "outputzq_mem_22807") !=
                    0)
                    return 1;
                
                struct memblock inputzq_mem_tmp_23001;
                
                inputzq_mem_tmp_23001.references = NULL;
                if (memblock_set(ctx, &inputzq_mem_tmp_23001,
                                 &outputzq_mem_22808, "outputzq_mem_22808") !=
                    0)
                    return 1;
                
                struct memblock outputzq_mem_tmp_23002;
                
                outputzq_mem_tmp_23002.references = NULL;
                if (memblock_set(ctx, &outputzq_mem_tmp_23002,
                                 &inputzq_mem_22805, "inputzq_mem_22805") != 0)
                    return 1;
                
                struct memblock outputzq_mem_tmp_23003;
                
                outputzq_mem_tmp_23003.references = NULL;
                if (memblock_set(ctx, &outputzq_mem_tmp_23003,
                                 &inputzq_mem_22806, "inputzq_mem_22806") != 0)
                    return 1;
                if (memblock_set(ctx, &inputzq_mem_22805,
                                 &inputzq_mem_tmp_23000,
                                 "inputzq_mem_tmp_23000") != 0)
                    return 1;
                if (memblock_set(ctx, &inputzq_mem_22806,
                                 &inputzq_mem_tmp_23001,
                                 "inputzq_mem_tmp_23001") != 0)
                    return 1;
                if (memblock_set(ctx, &outputzq_mem_22807,
                                 &outputzq_mem_tmp_23002,
                                 "outputzq_mem_tmp_23002") != 0)
                    return 1;
                if (memblock_set(ctx, &outputzq_mem_22808,
                                 &outputzq_mem_tmp_23003,
                                 "outputzq_mem_tmp_23003") != 0)
                    return 1;
                if (memblock_unref(ctx, &outputzq_mem_tmp_23003,
                                   "outputzq_mem_tmp_23003") != 0)
                    return 1;
                if (memblock_unref(ctx, &outputzq_mem_tmp_23002,
                                   "outputzq_mem_tmp_23002") != 0)
                    return 1;
                if (memblock_unref(ctx, &inputzq_mem_tmp_23001,
                                   "inputzq_mem_tmp_23001") != 0)
                    return 1;
                if (memblock_unref(ctx, &inputzq_mem_tmp_23000,
                                   "inputzq_mem_tmp_23000") != 0)
                    return 1;
            }
            if (memblock_set(ctx, &res_mem_22813, &inputzq_mem_22805,
                             "inputzq_mem_22805") != 0)
                return 1;
            if (memblock_set(ctx, &res_mem_22814, &inputzq_mem_22806,
                             "inputzq_mem_22806") != 0)
                return 1;
            if (memblock_set(ctx, &res_mem_22815, &outputzq_mem_22807,
                             "outputzq_mem_22807") != 0)
                return 1;
            if (memblock_set(ctx, &res_mem_22816, &outputzq_mem_22808,
                             "outputzq_mem_22808") != 0)
                return 1;
            memmove(mem_22685.mem + (i_22647 * (k_22234 * m_22233) + i_22643 *
                                     k_22234) * 8, res_mem_22813.mem + 0,
                    k_22234 * sizeof(double));
            if (memblock_unref(ctx, &res_mem_22813, "res_mem_22813") != 0)
                return 1;
            if (memblock_unref(ctx, &outputzq_mem_22808,
                               "outputzq_mem_22808") != 0)
                return 1;
            if (memblock_unref(ctx, &outputzq_mem_22807,
                               "outputzq_mem_22807") != 0)
                return 1;
            if (memblock_unref(ctx, &inputzq_mem_22806, "inputzq_mem_22806") !=
                0)
                return 1;
            if (memblock_unref(ctx, &inputzq_mem_22805, "inputzq_mem_22805") !=
                0)
                return 1;
            if (memblock_unref(ctx, &res_mem_22816, "res_mem_22816") != 0)
                return 1;
            if (memblock_unref(ctx, &res_mem_22815, "res_mem_22815") != 0)
                return 1;
            if (memblock_unref(ctx, &res_mem_22814, "res_mem_22814") != 0)
                return 1;
            if (memblock_unref(ctx, &res_mem_22813, "res_mem_22813") != 0)
                return 1;
        }
    }
    if (memblock_unref(ctx, &mem_22691, "mem_22691") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22696, "mem_22696") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22701, "mem_22701") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22704, "mem_22704") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22707, "mem_22707") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22710, "mem_22710") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22729, "mem_22729") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22734, "mem_22734") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22739, "mem_22739") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22742, "mem_22742") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22745, "mem_22745") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22748, "mem_22748") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22763, "mem_22763") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22766, "mem_22766") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22769, "mem_22769") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22772, "mem_22772") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22795, "mem_22795") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22798, "mem_22798") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22801, "mem_22801") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22804, "mem_22804") != 0)
        return 1;
    out_arrsizze_22939 = n_22232;
    out_arrsizze_22940 = m_22233;
    out_arrsizze_22941 = k_22234;
    if (memblock_set(ctx, &out_mem_22938, &mem_22685, "mem_22685") != 0)
        return 1;
    (*out_mem_p_23010).references = NULL;
    if (memblock_set(ctx, &*out_mem_p_23010, &out_mem_22938, "out_mem_22938") !=
        0)
        return 1;
    *out_out_arrsizze_23011 = out_arrsizze_22939;
    *out_out_arrsizze_23012 = out_arrsizze_22940;
    *out_out_arrsizze_23013 = out_arrsizze_22941;
    if (memblock_unref(ctx, &mem_22804, "mem_22804") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22801, "mem_22801") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22798, "mem_22798") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22795, "mem_22795") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22772, "mem_22772") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22769, "mem_22769") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22766, "mem_22766") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22763, "mem_22763") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22748, "mem_22748") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22745, "mem_22745") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22742, "mem_22742") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22739, "mem_22739") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22734, "mem_22734") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22729, "mem_22729") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22710, "mem_22710") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22707, "mem_22707") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22704, "mem_22704") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22701, "mem_22701") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22696, "mem_22696") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22691, "mem_22691") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22685, "mem_22685") != 0)
        return 1;
    if (memblock_unref(ctx, &out_mem_22938, "out_mem_22938") != 0)
        return 1;
    return 0;
}
static int futrts_fft2D(struct futhark_context *ctx,
                        struct memblock *out_mem_p_23014,
                        int32_t *out_out_arrsizze_23015,
                        int32_t *out_out_arrsizze_23016,
                        struct memblock *out_mem_p_23017,
                        struct memblock arr_mem_22678,
                        struct memblock arr_mem_22679, int32_t n_21954,
                        int32_t m_21955)
{
    struct memblock out_mem_22868;
    
    out_mem_22868.references = NULL;
    
    int32_t out_arrsizze_22869;
    int32_t out_arrsizze_22870;
    struct memblock out_mem_22871;
    
    out_mem_22871.references = NULL;
    
    bool loop_cond_21958 = slt32(1, n_21954);
    bool r_21959;
    int32_t r_21960;
    int32_t r_21961;
    bool loop_while_21962;
    int32_t r_21963;
    int32_t n_21964;
    
    loop_while_21962 = loop_cond_21958;
    r_21963 = 0;
    n_21964 = n_21954;
    while (loop_while_21962) {
        int32_t n_21965 = sdiv32(n_21964, 2);
        int32_t r_21966 = 1 + r_21963;
        bool loop_cond_21967 = slt32(1, n_21965);
        bool loop_while_tmp_22872 = loop_cond_21967;
        int32_t r_tmp_22873 = r_21966;
        int32_t n_tmp_22874;
        
        n_tmp_22874 = n_21965;
        loop_while_21962 = loop_while_tmp_22872;
        r_21963 = r_tmp_22873;
        n_21964 = n_tmp_22874;
    }
    r_21959 = loop_while_21962;
    r_21960 = r_21963;
    r_21961 = n_21964;
    
    bool loop_cond_21968 = slt32(1, m_21955);
    bool r_21969;
    int32_t r_21970;
    int32_t r_21971;
    bool loop_while_21972;
    int32_t r_21973;
    int32_t n_21974;
    
    loop_while_21972 = loop_cond_21968;
    r_21973 = 0;
    n_21974 = m_21955;
    while (loop_while_21972) {
        int32_t n_21975 = sdiv32(n_21974, 2);
        int32_t r_21976 = 1 + r_21973;
        bool loop_cond_21977 = slt32(1, n_21975);
        bool loop_while_tmp_22875 = loop_cond_21977;
        int32_t r_tmp_22876 = r_21976;
        int32_t n_tmp_22877;
        
        n_tmp_22877 = n_21975;
        loop_while_21972 = loop_while_tmp_22875;
        r_21973 = r_tmp_22876;
        n_21974 = n_tmp_22877;
    }
    r_21969 = loop_while_21972;
    r_21970 = r_21973;
    r_21971 = n_21974;
    
    int32_t iota_arg_21978 = sdiv32(m_21955, 2);
    int64_t binop_x_22681 = sext_i32_i64(n_21954);
    int64_t binop_y_22682 = sext_i32_i64(m_21955);
    int64_t binop_x_22683 = binop_x_22681 * binop_y_22682;
    int64_t bytes_22680 = 8 * binop_x_22683;
    struct memblock mem_22684;
    
    mem_22684.references = NULL;
    if (memblock_alloc(ctx, &mem_22684, bytes_22680, "mem_22684"))
        return 1;
    
    struct memblock mem_22689;
    
    mem_22689.references = NULL;
    if (memblock_alloc(ctx, &mem_22689, bytes_22680, "mem_22689"))
        return 1;
    
    int64_t bytes_22692 = 8 * binop_y_22682;
    struct memblock mem_22694;
    
    mem_22694.references = NULL;
    if (memblock_alloc(ctx, &mem_22694, bytes_22692, "mem_22694"))
        return 1;
    
    struct memblock mem_22697;
    
    mem_22697.references = NULL;
    if (memblock_alloc(ctx, &mem_22697, bytes_22692, "mem_22697"))
        return 1;
    
    struct memblock mem_22700;
    
    mem_22700.references = NULL;
    if (memblock_alloc(ctx, &mem_22700, bytes_22692, "mem_22700"))
        return 1;
    
    struct memblock mem_22703;
    
    mem_22703.references = NULL;
    if (memblock_alloc(ctx, &mem_22703, bytes_22692, "mem_22703"))
        return 1;
    for (int32_t i_22541 = 0; i_22541 < n_21954; i_22541++) {
        memmove(mem_22694.mem + 0, arr_mem_22678.mem + i_22541 * m_21955 * 8,
                m_21955 * sizeof(double));
        memmove(mem_22697.mem + 0, arr_mem_22679.mem + i_22541 * m_21955 * 8,
                m_21955 * sizeof(double));
        memmove(mem_22700.mem + 0, arr_mem_22678.mem + i_22541 * m_21955 * 8,
                m_21955 * sizeof(double));
        memmove(mem_22703.mem + 0, arr_mem_22679.mem + i_22541 * m_21955 * 8,
                m_21955 * sizeof(double));
        
        struct memblock res_mem_22712;
        
        res_mem_22712.references = NULL;
        
        struct memblock res_mem_22713;
        
        res_mem_22713.references = NULL;
        
        struct memblock res_mem_22714;
        
        res_mem_22714.references = NULL;
        
        struct memblock res_mem_22715;
        
        res_mem_22715.references = NULL;
        
        struct memblock inputzq_mem_22704;
        
        inputzq_mem_22704.references = NULL;
        
        struct memblock inputzq_mem_22705;
        
        inputzq_mem_22705.references = NULL;
        
        struct memblock outputzq_mem_22706;
        
        outputzq_mem_22706.references = NULL;
        
        struct memblock outputzq_mem_22707;
        
        outputzq_mem_22707.references = NULL;
        if (memblock_set(ctx, &inputzq_mem_22704, &mem_22694, "mem_22694") != 0)
            return 1;
        if (memblock_set(ctx, &inputzq_mem_22705, &mem_22697, "mem_22697") != 0)
            return 1;
        if (memblock_set(ctx, &outputzq_mem_22706, &mem_22700, "mem_22700") !=
            0)
            return 1;
        if (memblock_set(ctx, &outputzq_mem_22707, &mem_22703, "mem_22703") !=
            0)
            return 1;
        for (int32_t i_22003 = 0; i_22003 < r_21970; i_22003++) {
            int32_t index_primexp_22004 = 1 << i_22003;
            int32_t i32_arg_22005 = 2 * index_primexp_22004;
            double res_22006 = sitofp_i32_f64(i32_arg_22005);
            
            for (int32_t write_iter_22507 = 0; write_iter_22507 <
                 iota_arg_21978; write_iter_22507++) {
                double write_iv_22511 =
                       ((double *) inputzq_mem_22704.mem)[write_iter_22507];
                double write_iv_22512 =
                       ((double *) inputzq_mem_22705.mem)[write_iter_22507];
                int32_t i32_arg_22014 = smod32(write_iter_22507,
                                               index_primexp_22004);
                double res_22015 = sitofp_i32_f64(i32_arg_22014);
                double res_22016 = -6.283185307179586 * res_22015;
                double res_22017 = res_22016 / res_22006;
                int32_t i_22018 = iota_arg_21978 + write_iter_22507;
                double zt_arg_22019 =
                       ((double *) inputzq_mem_22704.mem)[i_22018];
                double zt_arg_22020 =
                       ((double *) inputzq_mem_22705.mem)[i_22018];
                double res_22021;
                
                res_22021 = futrts_cos64(res_22017);
                
                double res_22022;
                
                res_22022 = futrts_sin64(res_22017);
                
                double res_22023 = zt_arg_22019 * res_22021;
                double res_22024 = zt_arg_22020 * res_22022;
                double res_22025 = res_22023 - res_22024;
                double res_22026 = zt_arg_22020 * res_22021;
                double res_22027 = zt_arg_22019 * res_22022;
                double res_22028 = res_22026 + res_22027;
                double res_22029 = res_22025 + write_iv_22511;
                double res_22030 = res_22028 + write_iv_22512;
                double res_22031 = write_iv_22511 - res_22025;
                double res_22032 = write_iv_22512 - res_22028;
                int32_t x_22033 = sdiv32(write_iter_22507, index_primexp_22004);
                int32_t x_22034 = index_primexp_22004 * x_22033;
                int32_t x_22035 = 2 * x_22034;
                int32_t idxD_22036 = i32_arg_22014 + x_22035;
                int32_t res_22037 = index_primexp_22004 + idxD_22036;
                bool less_than_zzero_22513 = slt32(idxD_22036, 0);
                bool greater_than_sizze_22514 = sle32(m_21955, idxD_22036);
                bool outside_bounds_dim_22515 = less_than_zzero_22513 ||
                     greater_than_sizze_22514;
                
                if (!outside_bounds_dim_22515) {
                    ((double *) outputzq_mem_22706.mem)[idxD_22036] = res_22029;
                }
                
                bool less_than_zzero_22519 = slt32(res_22037, 0);
                bool greater_than_sizze_22520 = sle32(m_21955, res_22037);
                bool outside_bounds_dim_22521 = less_than_zzero_22519 ||
                     greater_than_sizze_22520;
                
                if (!outside_bounds_dim_22521) {
                    ((double *) outputzq_mem_22706.mem)[res_22037] = res_22031;
                }
                if (!outside_bounds_dim_22515) {
                    ((double *) outputzq_mem_22707.mem)[idxD_22036] = res_22030;
                }
                if (!outside_bounds_dim_22521) {
                    ((double *) outputzq_mem_22707.mem)[res_22037] = res_22032;
                }
            }
            
            struct memblock inputzq_mem_tmp_22880;
            
            inputzq_mem_tmp_22880.references = NULL;
            if (memblock_set(ctx, &inputzq_mem_tmp_22880, &outputzq_mem_22706,
                             "outputzq_mem_22706") != 0)
                return 1;
            
            struct memblock inputzq_mem_tmp_22881;
            
            inputzq_mem_tmp_22881.references = NULL;
            if (memblock_set(ctx, &inputzq_mem_tmp_22881, &outputzq_mem_22707,
                             "outputzq_mem_22707") != 0)
                return 1;
            
            struct memblock outputzq_mem_tmp_22882;
            
            outputzq_mem_tmp_22882.references = NULL;
            if (memblock_set(ctx, &outputzq_mem_tmp_22882, &inputzq_mem_22704,
                             "inputzq_mem_22704") != 0)
                return 1;
            
            struct memblock outputzq_mem_tmp_22883;
            
            outputzq_mem_tmp_22883.references = NULL;
            if (memblock_set(ctx, &outputzq_mem_tmp_22883, &inputzq_mem_22705,
                             "inputzq_mem_22705") != 0)
                return 1;
            if (memblock_set(ctx, &inputzq_mem_22704, &inputzq_mem_tmp_22880,
                             "inputzq_mem_tmp_22880") != 0)
                return 1;
            if (memblock_set(ctx, &inputzq_mem_22705, &inputzq_mem_tmp_22881,
                             "inputzq_mem_tmp_22881") != 0)
                return 1;
            if (memblock_set(ctx, &outputzq_mem_22706, &outputzq_mem_tmp_22882,
                             "outputzq_mem_tmp_22882") != 0)
                return 1;
            if (memblock_set(ctx, &outputzq_mem_22707, &outputzq_mem_tmp_22883,
                             "outputzq_mem_tmp_22883") != 0)
                return 1;
            if (memblock_unref(ctx, &outputzq_mem_tmp_22883,
                               "outputzq_mem_tmp_22883") != 0)
                return 1;
            if (memblock_unref(ctx, &outputzq_mem_tmp_22882,
                               "outputzq_mem_tmp_22882") != 0)
                return 1;
            if (memblock_unref(ctx, &inputzq_mem_tmp_22881,
                               "inputzq_mem_tmp_22881") != 0)
                return 1;
            if (memblock_unref(ctx, &inputzq_mem_tmp_22880,
                               "inputzq_mem_tmp_22880") != 0)
                return 1;
        }
        if (memblock_set(ctx, &res_mem_22712, &inputzq_mem_22704,
                         "inputzq_mem_22704") != 0)
            return 1;
        if (memblock_set(ctx, &res_mem_22713, &inputzq_mem_22705,
                         "inputzq_mem_22705") != 0)
            return 1;
        if (memblock_set(ctx, &res_mem_22714, &outputzq_mem_22706,
                         "outputzq_mem_22706") != 0)
            return 1;
        if (memblock_set(ctx, &res_mem_22715, &outputzq_mem_22707,
                         "outputzq_mem_22707") != 0)
            return 1;
        memmove(mem_22689.mem + i_22541 * m_21955 * 8, res_mem_22713.mem + 0,
                m_21955 * sizeof(double));
        if (memblock_unref(ctx, &res_mem_22713, "res_mem_22713") != 0)
            return 1;
        memmove(mem_22684.mem + i_22541 * m_21955 * 8, res_mem_22712.mem + 0,
                m_21955 * sizeof(double));
        if (memblock_unref(ctx, &res_mem_22712, "res_mem_22712") != 0)
            return 1;
        if (memblock_unref(ctx, &outputzq_mem_22707, "outputzq_mem_22707") != 0)
            return 1;
        if (memblock_unref(ctx, &outputzq_mem_22706, "outputzq_mem_22706") != 0)
            return 1;
        if (memblock_unref(ctx, &inputzq_mem_22705, "inputzq_mem_22705") != 0)
            return 1;
        if (memblock_unref(ctx, &inputzq_mem_22704, "inputzq_mem_22704") != 0)
            return 1;
        if (memblock_unref(ctx, &res_mem_22715, "res_mem_22715") != 0)
            return 1;
        if (memblock_unref(ctx, &res_mem_22714, "res_mem_22714") != 0)
            return 1;
        if (memblock_unref(ctx, &res_mem_22713, "res_mem_22713") != 0)
            return 1;
        if (memblock_unref(ctx, &res_mem_22712, "res_mem_22712") != 0)
            return 1;
    }
    if (memblock_unref(ctx, &mem_22694, "mem_22694") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22697, "mem_22697") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22700, "mem_22700") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22703, "mem_22703") != 0)
        return 1;
    
    int32_t iota_arg_22038 = sdiv32(n_21954, 2);
    bool r_22040;
    int32_t r_22041;
    int32_t r_22042;
    bool loop_while_22043;
    int32_t r_22044;
    int32_t n_22045;
    
    loop_while_22043 = loop_cond_21968;
    r_22044 = 0;
    n_22045 = m_21955;
    while (loop_while_22043) {
        int32_t n_22046 = sdiv32(n_22045, 2);
        int32_t r_22047 = 1 + r_22044;
        bool loop_cond_22048 = slt32(1, n_22046);
        bool loop_while_tmp_22890 = loop_cond_22048;
        int32_t r_tmp_22891 = r_22047;
        int32_t n_tmp_22892;
        
        n_tmp_22892 = n_22046;
        loop_while_22043 = loop_while_tmp_22890;
        r_22044 = r_tmp_22891;
        n_22045 = n_tmp_22892;
    }
    r_22040 = loop_while_22043;
    r_22041 = r_22044;
    r_22042 = n_22045;
    
    bool r_22049;
    int32_t r_22050;
    int32_t r_22051;
    bool loop_while_22052;
    int32_t r_22053;
    int32_t n_22054;
    
    loop_while_22052 = loop_cond_21958;
    r_22053 = 0;
    n_22054 = n_21954;
    while (loop_while_22052) {
        int32_t n_22055 = sdiv32(n_22054, 2);
        int32_t r_22056 = 1 + r_22053;
        bool loop_cond_22057 = slt32(1, n_22055);
        bool loop_while_tmp_22893 = loop_cond_22057;
        int32_t r_tmp_22894 = r_22056;
        int32_t n_tmp_22895;
        
        n_tmp_22895 = n_22055;
        loop_while_22052 = loop_while_tmp_22893;
        r_22053 = r_tmp_22894;
        n_22054 = n_tmp_22895;
    }
    r_22049 = loop_while_22052;
    r_22050 = r_22053;
    r_22051 = n_22054;
    
    int64_t binop_x_22721 = binop_x_22681 * binop_y_22682;
    int64_t bytes_22718 = 8 * binop_x_22721;
    struct memblock mem_22722;
    
    mem_22722.references = NULL;
    if (memblock_alloc(ctx, &mem_22722, bytes_22718, "mem_22722"))
        return 1;
    
    struct memblock mem_22727;
    
    mem_22727.references = NULL;
    if (memblock_alloc(ctx, &mem_22727, bytes_22718, "mem_22727"))
        return 1;
    
    int64_t bytes_22730 = 8 * binop_x_22681;
    struct memblock mem_22732;
    
    mem_22732.references = NULL;
    if (memblock_alloc(ctx, &mem_22732, bytes_22730, "mem_22732"))
        return 1;
    
    struct memblock mem_22735;
    
    mem_22735.references = NULL;
    if (memblock_alloc(ctx, &mem_22735, bytes_22730, "mem_22735"))
        return 1;
    
    struct memblock mem_22738;
    
    mem_22738.references = NULL;
    if (memblock_alloc(ctx, &mem_22738, bytes_22730, "mem_22738"))
        return 1;
    
    struct memblock mem_22741;
    
    mem_22741.references = NULL;
    if (memblock_alloc(ctx, &mem_22741, bytes_22730, "mem_22741"))
        return 1;
    
    struct memblock mem_22756;
    
    mem_22756.references = NULL;
    if (memblock_alloc(ctx, &mem_22756, bytes_22730, "mem_22756"))
        return 1;
    
    struct memblock mem_22759;
    
    mem_22759.references = NULL;
    if (memblock_alloc(ctx, &mem_22759, bytes_22730, "mem_22759"))
        return 1;
    
    struct memblock mem_22762;
    
    mem_22762.references = NULL;
    if (memblock_alloc(ctx, &mem_22762, bytes_22730, "mem_22762"))
        return 1;
    
    struct memblock mem_22765;
    
    mem_22765.references = NULL;
    if (memblock_alloc(ctx, &mem_22765, bytes_22730, "mem_22765"))
        return 1;
    for (int32_t i_22608 = 0; i_22608 < m_21955; i_22608++) {
        for (int32_t i_22898 = 0; i_22898 < n_21954; i_22898++) {
            ((double *) mem_22732.mem)[i_22898] =
                ((double *) mem_22684.mem)[i_22608 + i_22898 * m_21955];
        }
        for (int32_t i_22899 = 0; i_22899 < n_21954; i_22899++) {
            ((double *) mem_22735.mem)[i_22899] =
                ((double *) mem_22689.mem)[i_22608 + i_22899 * m_21955];
        }
        for (int32_t i_22900 = 0; i_22900 < n_21954; i_22900++) {
            ((double *) mem_22738.mem)[i_22900] =
                ((double *) mem_22684.mem)[i_22608 + i_22900 * m_21955];
        }
        for (int32_t i_22901 = 0; i_22901 < n_21954; i_22901++) {
            ((double *) mem_22741.mem)[i_22901] =
                ((double *) mem_22689.mem)[i_22608 + i_22901 * m_21955];
        }
        
        struct memblock res_mem_22750;
        
        res_mem_22750.references = NULL;
        
        struct memblock res_mem_22751;
        
        res_mem_22751.references = NULL;
        
        struct memblock res_mem_22752;
        
        res_mem_22752.references = NULL;
        
        struct memblock res_mem_22753;
        
        res_mem_22753.references = NULL;
        
        struct memblock inputzq_mem_22742;
        
        inputzq_mem_22742.references = NULL;
        
        struct memblock inputzq_mem_22743;
        
        inputzq_mem_22743.references = NULL;
        
        struct memblock outputzq_mem_22744;
        
        outputzq_mem_22744.references = NULL;
        
        struct memblock outputzq_mem_22745;
        
        outputzq_mem_22745.references = NULL;
        if (memblock_set(ctx, &inputzq_mem_22742, &mem_22732, "mem_22732") != 0)
            return 1;
        if (memblock_set(ctx, &inputzq_mem_22743, &mem_22735, "mem_22735") != 0)
            return 1;
        if (memblock_set(ctx, &outputzq_mem_22744, &mem_22738, "mem_22738") !=
            0)
            return 1;
        if (memblock_set(ctx, &outputzq_mem_22745, &mem_22741, "mem_22741") !=
            0)
            return 1;
        for (int32_t i_22083 = 0; i_22083 < r_21960; i_22083++) {
            int32_t index_primexp_22084 = 1 << i_22083;
            int32_t i32_arg_22085 = 2 * index_primexp_22084;
            double res_22086 = sitofp_i32_f64(i32_arg_22085);
            
            for (int32_t write_iter_22544 = 0; write_iter_22544 <
                 iota_arg_22038; write_iter_22544++) {
                double write_iv_22548 =
                       ((double *) inputzq_mem_22742.mem)[write_iter_22544];
                double write_iv_22549 =
                       ((double *) inputzq_mem_22743.mem)[write_iter_22544];
                int32_t i32_arg_22094 = smod32(write_iter_22544,
                                               index_primexp_22084);
                double res_22095 = sitofp_i32_f64(i32_arg_22094);
                double res_22096 = -6.283185307179586 * res_22095;
                double res_22097 = res_22096 / res_22086;
                int32_t i_22098 = iota_arg_22038 + write_iter_22544;
                double zt_arg_22099 =
                       ((double *) inputzq_mem_22742.mem)[i_22098];
                double zt_arg_22100 =
                       ((double *) inputzq_mem_22743.mem)[i_22098];
                double res_22101;
                
                res_22101 = futrts_cos64(res_22097);
                
                double res_22102;
                
                res_22102 = futrts_sin64(res_22097);
                
                double res_22103 = zt_arg_22099 * res_22101;
                double res_22104 = zt_arg_22100 * res_22102;
                double res_22105 = res_22103 - res_22104;
                double res_22106 = zt_arg_22100 * res_22101;
                double res_22107 = zt_arg_22099 * res_22102;
                double res_22108 = res_22106 + res_22107;
                double res_22109 = res_22105 + write_iv_22548;
                double res_22110 = res_22108 + write_iv_22549;
                double res_22111 = write_iv_22548 - res_22105;
                double res_22112 = write_iv_22549 - res_22108;
                int32_t x_22113 = sdiv32(write_iter_22544, index_primexp_22084);
                int32_t x_22114 = index_primexp_22084 * x_22113;
                int32_t x_22115 = 2 * x_22114;
                int32_t idxD_22116 = i32_arg_22094 + x_22115;
                int32_t res_22117 = index_primexp_22084 + idxD_22116;
                bool less_than_zzero_22550 = slt32(idxD_22116, 0);
                bool greater_than_sizze_22551 = sle32(n_21954, idxD_22116);
                bool outside_bounds_dim_22552 = less_than_zzero_22550 ||
                     greater_than_sizze_22551;
                
                if (!outside_bounds_dim_22552) {
                    ((double *) outputzq_mem_22744.mem)[idxD_22116] = res_22109;
                }
                
                bool less_than_zzero_22556 = slt32(res_22117, 0);
                bool greater_than_sizze_22557 = sle32(n_21954, res_22117);
                bool outside_bounds_dim_22558 = less_than_zzero_22556 ||
                     greater_than_sizze_22557;
                
                if (!outside_bounds_dim_22558) {
                    ((double *) outputzq_mem_22744.mem)[res_22117] = res_22111;
                }
                if (!outside_bounds_dim_22552) {
                    ((double *) outputzq_mem_22745.mem)[idxD_22116] = res_22110;
                }
                if (!outside_bounds_dim_22558) {
                    ((double *) outputzq_mem_22745.mem)[res_22117] = res_22112;
                }
            }
            
            struct memblock inputzq_mem_tmp_22902;
            
            inputzq_mem_tmp_22902.references = NULL;
            if (memblock_set(ctx, &inputzq_mem_tmp_22902, &outputzq_mem_22744,
                             "outputzq_mem_22744") != 0)
                return 1;
            
            struct memblock inputzq_mem_tmp_22903;
            
            inputzq_mem_tmp_22903.references = NULL;
            if (memblock_set(ctx, &inputzq_mem_tmp_22903, &outputzq_mem_22745,
                             "outputzq_mem_22745") != 0)
                return 1;
            
            struct memblock outputzq_mem_tmp_22904;
            
            outputzq_mem_tmp_22904.references = NULL;
            if (memblock_set(ctx, &outputzq_mem_tmp_22904, &inputzq_mem_22742,
                             "inputzq_mem_22742") != 0)
                return 1;
            
            struct memblock outputzq_mem_tmp_22905;
            
            outputzq_mem_tmp_22905.references = NULL;
            if (memblock_set(ctx, &outputzq_mem_tmp_22905, &inputzq_mem_22743,
                             "inputzq_mem_22743") != 0)
                return 1;
            if (memblock_set(ctx, &inputzq_mem_22742, &inputzq_mem_tmp_22902,
                             "inputzq_mem_tmp_22902") != 0)
                return 1;
            if (memblock_set(ctx, &inputzq_mem_22743, &inputzq_mem_tmp_22903,
                             "inputzq_mem_tmp_22903") != 0)
                return 1;
            if (memblock_set(ctx, &outputzq_mem_22744, &outputzq_mem_tmp_22904,
                             "outputzq_mem_tmp_22904") != 0)
                return 1;
            if (memblock_set(ctx, &outputzq_mem_22745, &outputzq_mem_tmp_22905,
                             "outputzq_mem_tmp_22905") != 0)
                return 1;
            if (memblock_unref(ctx, &outputzq_mem_tmp_22905,
                               "outputzq_mem_tmp_22905") != 0)
                return 1;
            if (memblock_unref(ctx, &outputzq_mem_tmp_22904,
                               "outputzq_mem_tmp_22904") != 0)
                return 1;
            if (memblock_unref(ctx, &inputzq_mem_tmp_22903,
                               "inputzq_mem_tmp_22903") != 0)
                return 1;
            if (memblock_unref(ctx, &inputzq_mem_tmp_22902,
                               "inputzq_mem_tmp_22902") != 0)
                return 1;
        }
        if (memblock_set(ctx, &res_mem_22750, &inputzq_mem_22742,
                         "inputzq_mem_22742") != 0)
            return 1;
        if (memblock_set(ctx, &res_mem_22751, &inputzq_mem_22743,
                         "inputzq_mem_22743") != 0)
            return 1;
        if (memblock_set(ctx, &res_mem_22752, &outputzq_mem_22744,
                         "outputzq_mem_22744") != 0)
            return 1;
        if (memblock_set(ctx, &res_mem_22753, &outputzq_mem_22745,
                         "outputzq_mem_22745") != 0)
            return 1;
        memmove(mem_22756.mem + 0, res_mem_22750.mem + 0, n_21954 *
                sizeof(double));
        memmove(mem_22759.mem + 0, res_mem_22751.mem + 0, n_21954 *
                sizeof(double));
        memmove(mem_22762.mem + 0, res_mem_22750.mem + 0, n_21954 *
                sizeof(double));
        if (memblock_unref(ctx, &res_mem_22750, "res_mem_22750") != 0)
            return 1;
        memmove(mem_22765.mem + 0, res_mem_22751.mem + 0, n_21954 *
                sizeof(double));
        if (memblock_unref(ctx, &res_mem_22751, "res_mem_22751") != 0)
            return 1;
        
        struct memblock res_mem_22774;
        
        res_mem_22774.references = NULL;
        
        struct memblock res_mem_22775;
        
        res_mem_22775.references = NULL;
        
        struct memblock res_mem_22776;
        
        res_mem_22776.references = NULL;
        
        struct memblock res_mem_22777;
        
        res_mem_22777.references = NULL;
        
        struct memblock inputzq_mem_22766;
        
        inputzq_mem_22766.references = NULL;
        
        struct memblock inputzq_mem_22767;
        
        inputzq_mem_22767.references = NULL;
        
        struct memblock outputzq_mem_22768;
        
        outputzq_mem_22768.references = NULL;
        
        struct memblock outputzq_mem_22769;
        
        outputzq_mem_22769.references = NULL;
        if (memblock_set(ctx, &inputzq_mem_22766, &mem_22756, "mem_22756") != 0)
            return 1;
        if (memblock_set(ctx, &inputzq_mem_22767, &mem_22759, "mem_22759") != 0)
            return 1;
        if (memblock_set(ctx, &outputzq_mem_22768, &mem_22762, "mem_22762") !=
            0)
            return 1;
        if (memblock_set(ctx, &outputzq_mem_22769, &mem_22765, "mem_22765") !=
            0)
            return 1;
        for (int32_t i_22137 = 0; i_22137 < r_22050; i_22137++) {
            int32_t index_primexp_22138 = 1 << i_22137;
            int32_t i32_arg_22139 = 2 * index_primexp_22138;
            double res_22140 = sitofp_i32_f64(i32_arg_22139);
            
            for (int32_t write_iter_22574 = 0; write_iter_22574 <
                 iota_arg_22038; write_iter_22574++) {
                double write_iv_22578 =
                       ((double *) inputzq_mem_22766.mem)[write_iter_22574];
                double write_iv_22579 =
                       ((double *) inputzq_mem_22767.mem)[write_iter_22574];
                int32_t i32_arg_22148 = smod32(write_iter_22574,
                                               index_primexp_22138);
                double res_22149 = sitofp_i32_f64(i32_arg_22148);
                double res_22150 = -6.283185307179586 * res_22149;
                double res_22151 = res_22150 / res_22140;
                int32_t i_22152 = iota_arg_22038 + write_iter_22574;
                double zt_arg_22153 =
                       ((double *) inputzq_mem_22766.mem)[i_22152];
                double zt_arg_22154 =
                       ((double *) inputzq_mem_22767.mem)[i_22152];
                double res_22155;
                
                res_22155 = futrts_cos64(res_22151);
                
                double res_22156;
                
                res_22156 = futrts_sin64(res_22151);
                
                double res_22157 = zt_arg_22153 * res_22155;
                double res_22158 = zt_arg_22154 * res_22156;
                double res_22159 = res_22157 - res_22158;
                double res_22160 = zt_arg_22154 * res_22155;
                double res_22161 = zt_arg_22153 * res_22156;
                double res_22162 = res_22160 + res_22161;
                double res_22163 = res_22159 + write_iv_22578;
                double res_22164 = res_22162 + write_iv_22579;
                double res_22165 = write_iv_22578 - res_22159;
                double res_22166 = write_iv_22579 - res_22162;
                int32_t x_22167 = sdiv32(write_iter_22574, index_primexp_22138);
                int32_t x_22168 = index_primexp_22138 * x_22167;
                int32_t x_22169 = 2 * x_22168;
                int32_t idxD_22170 = i32_arg_22148 + x_22169;
                int32_t res_22171 = index_primexp_22138 + idxD_22170;
                bool less_than_zzero_22580 = slt32(idxD_22170, 0);
                bool greater_than_sizze_22581 = sle32(n_21954, idxD_22170);
                bool outside_bounds_dim_22582 = less_than_zzero_22580 ||
                     greater_than_sizze_22581;
                
                if (!outside_bounds_dim_22582) {
                    ((double *) outputzq_mem_22768.mem)[idxD_22170] = res_22163;
                }
                
                bool less_than_zzero_22586 = slt32(res_22171, 0);
                bool greater_than_sizze_22587 = sle32(n_21954, res_22171);
                bool outside_bounds_dim_22588 = less_than_zzero_22586 ||
                     greater_than_sizze_22587;
                
                if (!outside_bounds_dim_22588) {
                    ((double *) outputzq_mem_22768.mem)[res_22171] = res_22165;
                }
                if (!outside_bounds_dim_22582) {
                    ((double *) outputzq_mem_22769.mem)[idxD_22170] = res_22164;
                }
                if (!outside_bounds_dim_22588) {
                    ((double *) outputzq_mem_22769.mem)[res_22171] = res_22166;
                }
            }
            
            struct memblock inputzq_mem_tmp_22912;
            
            inputzq_mem_tmp_22912.references = NULL;
            if (memblock_set(ctx, &inputzq_mem_tmp_22912, &outputzq_mem_22768,
                             "outputzq_mem_22768") != 0)
                return 1;
            
            struct memblock inputzq_mem_tmp_22913;
            
            inputzq_mem_tmp_22913.references = NULL;
            if (memblock_set(ctx, &inputzq_mem_tmp_22913, &outputzq_mem_22769,
                             "outputzq_mem_22769") != 0)
                return 1;
            
            struct memblock outputzq_mem_tmp_22914;
            
            outputzq_mem_tmp_22914.references = NULL;
            if (memblock_set(ctx, &outputzq_mem_tmp_22914, &inputzq_mem_22766,
                             "inputzq_mem_22766") != 0)
                return 1;
            
            struct memblock outputzq_mem_tmp_22915;
            
            outputzq_mem_tmp_22915.references = NULL;
            if (memblock_set(ctx, &outputzq_mem_tmp_22915, &inputzq_mem_22767,
                             "inputzq_mem_22767") != 0)
                return 1;
            if (memblock_set(ctx, &inputzq_mem_22766, &inputzq_mem_tmp_22912,
                             "inputzq_mem_tmp_22912") != 0)
                return 1;
            if (memblock_set(ctx, &inputzq_mem_22767, &inputzq_mem_tmp_22913,
                             "inputzq_mem_tmp_22913") != 0)
                return 1;
            if (memblock_set(ctx, &outputzq_mem_22768, &outputzq_mem_tmp_22914,
                             "outputzq_mem_tmp_22914") != 0)
                return 1;
            if (memblock_set(ctx, &outputzq_mem_22769, &outputzq_mem_tmp_22915,
                             "outputzq_mem_tmp_22915") != 0)
                return 1;
            if (memblock_unref(ctx, &outputzq_mem_tmp_22915,
                               "outputzq_mem_tmp_22915") != 0)
                return 1;
            if (memblock_unref(ctx, &outputzq_mem_tmp_22914,
                               "outputzq_mem_tmp_22914") != 0)
                return 1;
            if (memblock_unref(ctx, &inputzq_mem_tmp_22913,
                               "inputzq_mem_tmp_22913") != 0)
                return 1;
            if (memblock_unref(ctx, &inputzq_mem_tmp_22912,
                               "inputzq_mem_tmp_22912") != 0)
                return 1;
        }
        if (memblock_set(ctx, &res_mem_22774, &inputzq_mem_22766,
                         "inputzq_mem_22766") != 0)
            return 1;
        if (memblock_set(ctx, &res_mem_22775, &inputzq_mem_22767,
                         "inputzq_mem_22767") != 0)
            return 1;
        if (memblock_set(ctx, &res_mem_22776, &outputzq_mem_22768,
                         "outputzq_mem_22768") != 0)
            return 1;
        if (memblock_set(ctx, &res_mem_22777, &outputzq_mem_22769,
                         "outputzq_mem_22769") != 0)
            return 1;
        memmove(mem_22727.mem + i_22608 * n_21954 * 8, res_mem_22775.mem + 0,
                n_21954 * sizeof(double));
        if (memblock_unref(ctx, &res_mem_22775, "res_mem_22775") != 0)
            return 1;
        memmove(mem_22722.mem + i_22608 * n_21954 * 8, res_mem_22774.mem + 0,
                n_21954 * sizeof(double));
        if (memblock_unref(ctx, &res_mem_22774, "res_mem_22774") != 0)
            return 1;
        if (memblock_unref(ctx, &outputzq_mem_22769, "outputzq_mem_22769") != 0)
            return 1;
        if (memblock_unref(ctx, &outputzq_mem_22768, "outputzq_mem_22768") != 0)
            return 1;
        if (memblock_unref(ctx, &inputzq_mem_22767, "inputzq_mem_22767") != 0)
            return 1;
        if (memblock_unref(ctx, &inputzq_mem_22766, "inputzq_mem_22766") != 0)
            return 1;
        if (memblock_unref(ctx, &res_mem_22777, "res_mem_22777") != 0)
            return 1;
        if (memblock_unref(ctx, &res_mem_22776, "res_mem_22776") != 0)
            return 1;
        if (memblock_unref(ctx, &res_mem_22775, "res_mem_22775") != 0)
            return 1;
        if (memblock_unref(ctx, &res_mem_22774, "res_mem_22774") != 0)
            return 1;
        if (memblock_unref(ctx, &outputzq_mem_22745, "outputzq_mem_22745") != 0)
            return 1;
        if (memblock_unref(ctx, &outputzq_mem_22744, "outputzq_mem_22744") != 0)
            return 1;
        if (memblock_unref(ctx, &inputzq_mem_22743, "inputzq_mem_22743") != 0)
            return 1;
        if (memblock_unref(ctx, &inputzq_mem_22742, "inputzq_mem_22742") != 0)
            return 1;
        if (memblock_unref(ctx, &res_mem_22753, "res_mem_22753") != 0)
            return 1;
        if (memblock_unref(ctx, &res_mem_22752, "res_mem_22752") != 0)
            return 1;
        if (memblock_unref(ctx, &res_mem_22751, "res_mem_22751") != 0)
            return 1;
        if (memblock_unref(ctx, &res_mem_22750, "res_mem_22750") != 0)
            return 1;
    }
    if (memblock_unref(ctx, &mem_22684, "mem_22684") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22689, "mem_22689") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22732, "mem_22732") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22735, "mem_22735") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22738, "mem_22738") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22741, "mem_22741") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22756, "mem_22756") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22759, "mem_22759") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22762, "mem_22762") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22765, "mem_22765") != 0)
        return 1;
    
    struct memblock mem_22784;
    
    mem_22784.references = NULL;
    if (memblock_alloc(ctx, &mem_22784, bytes_22680, "mem_22784"))
        return 1;
    
    struct memblock mem_22789;
    
    mem_22789.references = NULL;
    if (memblock_alloc(ctx, &mem_22789, bytes_22680, "mem_22789"))
        return 1;
    
    int64_t bytes_22792 = 8 * binop_y_22682;
    struct memblock mem_22794;
    
    mem_22794.references = NULL;
    if (memblock_alloc(ctx, &mem_22794, bytes_22792, "mem_22794"))
        return 1;
    
    struct memblock mem_22797;
    
    mem_22797.references = NULL;
    if (memblock_alloc(ctx, &mem_22797, bytes_22792, "mem_22797"))
        return 1;
    
    struct memblock mem_22800;
    
    mem_22800.references = NULL;
    if (memblock_alloc(ctx, &mem_22800, bytes_22792, "mem_22800"))
        return 1;
    
    struct memblock mem_22803;
    
    mem_22803.references = NULL;
    if (memblock_alloc(ctx, &mem_22803, bytes_22792, "mem_22803"))
        return 1;
    for (int32_t i_22645 = 0; i_22645 < n_21954; i_22645++) {
        for (int32_t i_22924 = 0; i_22924 < m_21955; i_22924++) {
            ((double *) mem_22794.mem)[i_22924] =
                ((double *) mem_22722.mem)[i_22645 + i_22924 * n_21954];
        }
        for (int32_t i_22925 = 0; i_22925 < m_21955; i_22925++) {
            ((double *) mem_22797.mem)[i_22925] =
                ((double *) mem_22727.mem)[i_22645 + i_22925 * n_21954];
        }
        for (int32_t i_22926 = 0; i_22926 < m_21955; i_22926++) {
            ((double *) mem_22800.mem)[i_22926] =
                ((double *) mem_22722.mem)[i_22645 + i_22926 * n_21954];
        }
        for (int32_t i_22927 = 0; i_22927 < m_21955; i_22927++) {
            ((double *) mem_22803.mem)[i_22927] =
                ((double *) mem_22727.mem)[i_22645 + i_22927 * n_21954];
        }
        
        struct memblock res_mem_22812;
        
        res_mem_22812.references = NULL;
        
        struct memblock res_mem_22813;
        
        res_mem_22813.references = NULL;
        
        struct memblock res_mem_22814;
        
        res_mem_22814.references = NULL;
        
        struct memblock res_mem_22815;
        
        res_mem_22815.references = NULL;
        
        struct memblock inputzq_mem_22804;
        
        inputzq_mem_22804.references = NULL;
        
        struct memblock inputzq_mem_22805;
        
        inputzq_mem_22805.references = NULL;
        
        struct memblock outputzq_mem_22806;
        
        outputzq_mem_22806.references = NULL;
        
        struct memblock outputzq_mem_22807;
        
        outputzq_mem_22807.references = NULL;
        if (memblock_set(ctx, &inputzq_mem_22804, &mem_22794, "mem_22794") != 0)
            return 1;
        if (memblock_set(ctx, &inputzq_mem_22805, &mem_22797, "mem_22797") != 0)
            return 1;
        if (memblock_set(ctx, &outputzq_mem_22806, &mem_22800, "mem_22800") !=
            0)
            return 1;
        if (memblock_set(ctx, &outputzq_mem_22807, &mem_22803, "mem_22803") !=
            0)
            return 1;
        for (int32_t i_22197 = 0; i_22197 < r_22041; i_22197++) {
            int32_t index_primexp_22198 = 1 << i_22197;
            int32_t i32_arg_22199 = 2 * index_primexp_22198;
            double res_22200 = sitofp_i32_f64(i32_arg_22199);
            
            for (int32_t write_iter_22611 = 0; write_iter_22611 <
                 iota_arg_21978; write_iter_22611++) {
                double write_iv_22615 =
                       ((double *) inputzq_mem_22804.mem)[write_iter_22611];
                double write_iv_22616 =
                       ((double *) inputzq_mem_22805.mem)[write_iter_22611];
                int32_t i32_arg_22208 = smod32(write_iter_22611,
                                               index_primexp_22198);
                double res_22209 = sitofp_i32_f64(i32_arg_22208);
                double res_22210 = -6.283185307179586 * res_22209;
                double res_22211 = res_22210 / res_22200;
                int32_t i_22212 = iota_arg_21978 + write_iter_22611;
                double zt_arg_22213 =
                       ((double *) inputzq_mem_22804.mem)[i_22212];
                double zt_arg_22214 =
                       ((double *) inputzq_mem_22805.mem)[i_22212];
                double res_22215;
                
                res_22215 = futrts_cos64(res_22211);
                
                double res_22216;
                
                res_22216 = futrts_sin64(res_22211);
                
                double res_22217 = zt_arg_22213 * res_22215;
                double res_22218 = zt_arg_22214 * res_22216;
                double res_22219 = res_22217 - res_22218;
                double res_22220 = zt_arg_22214 * res_22215;
                double res_22221 = zt_arg_22213 * res_22216;
                double res_22222 = res_22220 + res_22221;
                double res_22223 = res_22219 + write_iv_22615;
                double res_22224 = res_22222 + write_iv_22616;
                double res_22225 = write_iv_22615 - res_22219;
                double res_22226 = write_iv_22616 - res_22222;
                int32_t x_22227 = sdiv32(write_iter_22611, index_primexp_22198);
                int32_t x_22228 = index_primexp_22198 * x_22227;
                int32_t x_22229 = 2 * x_22228;
                int32_t idxD_22230 = i32_arg_22208 + x_22229;
                int32_t res_22231 = index_primexp_22198 + idxD_22230;
                bool less_than_zzero_22617 = slt32(idxD_22230, 0);
                bool greater_than_sizze_22618 = sle32(m_21955, idxD_22230);
                bool outside_bounds_dim_22619 = less_than_zzero_22617 ||
                     greater_than_sizze_22618;
                
                if (!outside_bounds_dim_22619) {
                    ((double *) outputzq_mem_22806.mem)[idxD_22230] = res_22223;
                }
                
                bool less_than_zzero_22623 = slt32(res_22231, 0);
                bool greater_than_sizze_22624 = sle32(m_21955, res_22231);
                bool outside_bounds_dim_22625 = less_than_zzero_22623 ||
                     greater_than_sizze_22624;
                
                if (!outside_bounds_dim_22625) {
                    ((double *) outputzq_mem_22806.mem)[res_22231] = res_22225;
                }
                if (!outside_bounds_dim_22619) {
                    ((double *) outputzq_mem_22807.mem)[idxD_22230] = res_22224;
                }
                if (!outside_bounds_dim_22625) {
                    ((double *) outputzq_mem_22807.mem)[res_22231] = res_22226;
                }
            }
            
            struct memblock inputzq_mem_tmp_22928;
            
            inputzq_mem_tmp_22928.references = NULL;
            if (memblock_set(ctx, &inputzq_mem_tmp_22928, &outputzq_mem_22806,
                             "outputzq_mem_22806") != 0)
                return 1;
            
            struct memblock inputzq_mem_tmp_22929;
            
            inputzq_mem_tmp_22929.references = NULL;
            if (memblock_set(ctx, &inputzq_mem_tmp_22929, &outputzq_mem_22807,
                             "outputzq_mem_22807") != 0)
                return 1;
            
            struct memblock outputzq_mem_tmp_22930;
            
            outputzq_mem_tmp_22930.references = NULL;
            if (memblock_set(ctx, &outputzq_mem_tmp_22930, &inputzq_mem_22804,
                             "inputzq_mem_22804") != 0)
                return 1;
            
            struct memblock outputzq_mem_tmp_22931;
            
            outputzq_mem_tmp_22931.references = NULL;
            if (memblock_set(ctx, &outputzq_mem_tmp_22931, &inputzq_mem_22805,
                             "inputzq_mem_22805") != 0)
                return 1;
            if (memblock_set(ctx, &inputzq_mem_22804, &inputzq_mem_tmp_22928,
                             "inputzq_mem_tmp_22928") != 0)
                return 1;
            if (memblock_set(ctx, &inputzq_mem_22805, &inputzq_mem_tmp_22929,
                             "inputzq_mem_tmp_22929") != 0)
                return 1;
            if (memblock_set(ctx, &outputzq_mem_22806, &outputzq_mem_tmp_22930,
                             "outputzq_mem_tmp_22930") != 0)
                return 1;
            if (memblock_set(ctx, &outputzq_mem_22807, &outputzq_mem_tmp_22931,
                             "outputzq_mem_tmp_22931") != 0)
                return 1;
            if (memblock_unref(ctx, &outputzq_mem_tmp_22931,
                               "outputzq_mem_tmp_22931") != 0)
                return 1;
            if (memblock_unref(ctx, &outputzq_mem_tmp_22930,
                               "outputzq_mem_tmp_22930") != 0)
                return 1;
            if (memblock_unref(ctx, &inputzq_mem_tmp_22929,
                               "inputzq_mem_tmp_22929") != 0)
                return 1;
            if (memblock_unref(ctx, &inputzq_mem_tmp_22928,
                               "inputzq_mem_tmp_22928") != 0)
                return 1;
        }
        if (memblock_set(ctx, &res_mem_22812, &inputzq_mem_22804,
                         "inputzq_mem_22804") != 0)
            return 1;
        if (memblock_set(ctx, &res_mem_22813, &inputzq_mem_22805,
                         "inputzq_mem_22805") != 0)
            return 1;
        if (memblock_set(ctx, &res_mem_22814, &outputzq_mem_22806,
                         "outputzq_mem_22806") != 0)
            return 1;
        if (memblock_set(ctx, &res_mem_22815, &outputzq_mem_22807,
                         "outputzq_mem_22807") != 0)
            return 1;
        memmove(mem_22789.mem + i_22645 * m_21955 * 8, res_mem_22813.mem + 0,
                m_21955 * sizeof(double));
        if (memblock_unref(ctx, &res_mem_22813, "res_mem_22813") != 0)
            return 1;
        memmove(mem_22784.mem + i_22645 * m_21955 * 8, res_mem_22812.mem + 0,
                m_21955 * sizeof(double));
        if (memblock_unref(ctx, &res_mem_22812, "res_mem_22812") != 0)
            return 1;
        if (memblock_unref(ctx, &outputzq_mem_22807, "outputzq_mem_22807") != 0)
            return 1;
        if (memblock_unref(ctx, &outputzq_mem_22806, "outputzq_mem_22806") != 0)
            return 1;
        if (memblock_unref(ctx, &inputzq_mem_22805, "inputzq_mem_22805") != 0)
            return 1;
        if (memblock_unref(ctx, &inputzq_mem_22804, "inputzq_mem_22804") != 0)
            return 1;
        if (memblock_unref(ctx, &res_mem_22815, "res_mem_22815") != 0)
            return 1;
        if (memblock_unref(ctx, &res_mem_22814, "res_mem_22814") != 0)
            return 1;
        if (memblock_unref(ctx, &res_mem_22813, "res_mem_22813") != 0)
            return 1;
        if (memblock_unref(ctx, &res_mem_22812, "res_mem_22812") != 0)
            return 1;
    }
    if (memblock_unref(ctx, &mem_22722, "mem_22722") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22727, "mem_22727") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22794, "mem_22794") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22797, "mem_22797") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22800, "mem_22800") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22803, "mem_22803") != 0)
        return 1;
    out_arrsizze_22869 = n_21954;
    out_arrsizze_22870 = m_21955;
    if (memblock_set(ctx, &out_mem_22868, &mem_22784, "mem_22784") != 0)
        return 1;
    if (memblock_set(ctx, &out_mem_22871, &mem_22789, "mem_22789") != 0)
        return 1;
    (*out_mem_p_23014).references = NULL;
    if (memblock_set(ctx, &*out_mem_p_23014, &out_mem_22868, "out_mem_22868") !=
        0)
        return 1;
    *out_out_arrsizze_23015 = out_arrsizze_22869;
    *out_out_arrsizze_23016 = out_arrsizze_22870;
    (*out_mem_p_23017).references = NULL;
    if (memblock_set(ctx, &*out_mem_p_23017, &out_mem_22871, "out_mem_22871") !=
        0)
        return 1;
    if (memblock_unref(ctx, &mem_22803, "mem_22803") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22800, "mem_22800") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22797, "mem_22797") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22794, "mem_22794") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22789, "mem_22789") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22784, "mem_22784") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22765, "mem_22765") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22762, "mem_22762") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22759, "mem_22759") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22756, "mem_22756") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22741, "mem_22741") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22738, "mem_22738") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22735, "mem_22735") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22732, "mem_22732") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22727, "mem_22727") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22722, "mem_22722") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22703, "mem_22703") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22700, "mem_22700") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22697, "mem_22697") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22694, "mem_22694") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22689, "mem_22689") != 0)
        return 1;
    if (memblock_unref(ctx, &mem_22684, "mem_22684") != 0)
        return 1;
    if (memblock_unref(ctx, &out_mem_22871, "out_mem_22871") != 0)
        return 1;
    if (memblock_unref(ctx, &out_mem_22868, "out_mem_22868") != 0)
        return 1;
    return 0;
}
struct futhark_f64_2d {
    struct memblock mem;
    int64_t shape[2];
} ;
struct futhark_f64_2d *futhark_new_f64_2d(struct futhark_context *ctx,
                                          double *data, int64_t dim0,
                                          int64_t dim1)
{
    struct futhark_f64_2d *bad = NULL;
    struct futhark_f64_2d *arr =
                          (struct futhark_f64_2d *) malloc(sizeof(struct futhark_f64_2d));
    
    if (arr == NULL)
        return bad;
    lock_lock(&ctx->lock);
    arr->mem.references = NULL;
    if (memblock_alloc(ctx, &arr->mem, dim0 * dim1 * sizeof(double),
                       "arr->mem"))
        return NULL;
    arr->shape[0] = dim0;
    arr->shape[1] = dim1;
    memmove(arr->mem.mem + 0, data + 0, dim0 * dim1 * sizeof(double));
    lock_unlock(&ctx->lock);
    return arr;
}
struct futhark_f64_2d *futhark_new_raw_f64_2d(struct futhark_context *ctx,
                                              char *data, int offset,
                                              int64_t dim0, int64_t dim1)
{
    struct futhark_f64_2d *bad = NULL;
    struct futhark_f64_2d *arr =
                          (struct futhark_f64_2d *) malloc(sizeof(struct futhark_f64_2d));
    
    if (arr == NULL)
        return bad;
    lock_lock(&ctx->lock);
    arr->mem.references = NULL;
    if (memblock_alloc(ctx, &arr->mem, dim0 * dim1 * sizeof(double),
                       "arr->mem"))
        return NULL;
    arr->shape[0] = dim0;
    arr->shape[1] = dim1;
    memmove(arr->mem.mem + 0, data + offset, dim0 * dim1 * sizeof(double));
    lock_unlock(&ctx->lock);
    return arr;
}
int futhark_free_f64_2d(struct futhark_context *ctx, struct futhark_f64_2d *arr)
{
    lock_lock(&ctx->lock);
    if (memblock_unref(ctx, &arr->mem, "arr->mem") != 0)
        return 1;
    lock_unlock(&ctx->lock);
    free(arr);
    return 0;
}
int futhark_values_f64_2d(struct futhark_context *ctx,
                          struct futhark_f64_2d *arr, double *data)
{
    lock_lock(&ctx->lock);
    memmove(data + 0, arr->mem.mem + 0, arr->shape[0] * arr->shape[1] *
            sizeof(double));
    lock_unlock(&ctx->lock);
    return 0;
}
char *futhark_values_raw_f64_2d(struct futhark_context *ctx,
                                struct futhark_f64_2d *arr)
{
    (void) ctx;
    return arr->mem.mem;
}
int64_t *futhark_shape_f64_2d(struct futhark_context *ctx,
                              struct futhark_f64_2d *arr)
{
    (void) ctx;
    return arr->shape;
}
struct futhark_f64_3d {
    struct memblock mem;
    int64_t shape[3];
} ;
struct futhark_f64_3d *futhark_new_f64_3d(struct futhark_context *ctx,
                                          double *data, int64_t dim0,
                                          int64_t dim1, int64_t dim2)
{
    struct futhark_f64_3d *bad = NULL;
    struct futhark_f64_3d *arr =
                          (struct futhark_f64_3d *) malloc(sizeof(struct futhark_f64_3d));
    
    if (arr == NULL)
        return bad;
    lock_lock(&ctx->lock);
    arr->mem.references = NULL;
    if (memblock_alloc(ctx, &arr->mem, dim0 * dim1 * dim2 * sizeof(double),
                       "arr->mem"))
        return NULL;
    arr->shape[0] = dim0;
    arr->shape[1] = dim1;
    arr->shape[2] = dim2;
    memmove(arr->mem.mem + 0, data + 0, dim0 * dim1 * dim2 * sizeof(double));
    lock_unlock(&ctx->lock);
    return arr;
}
struct futhark_f64_3d *futhark_new_raw_f64_3d(struct futhark_context *ctx,
                                              char *data, int offset,
                                              int64_t dim0, int64_t dim1,
                                              int64_t dim2)
{
    struct futhark_f64_3d *bad = NULL;
    struct futhark_f64_3d *arr =
                          (struct futhark_f64_3d *) malloc(sizeof(struct futhark_f64_3d));
    
    if (arr == NULL)
        return bad;
    lock_lock(&ctx->lock);
    arr->mem.references = NULL;
    if (memblock_alloc(ctx, &arr->mem, dim0 * dim1 * dim2 * sizeof(double),
                       "arr->mem"))
        return NULL;
    arr->shape[0] = dim0;
    arr->shape[1] = dim1;
    arr->shape[2] = dim2;
    memmove(arr->mem.mem + 0, data + offset, dim0 * dim1 * dim2 *
            sizeof(double));
    lock_unlock(&ctx->lock);
    return arr;
}
int futhark_free_f64_3d(struct futhark_context *ctx, struct futhark_f64_3d *arr)
{
    lock_lock(&ctx->lock);
    if (memblock_unref(ctx, &arr->mem, "arr->mem") != 0)
        return 1;
    lock_unlock(&ctx->lock);
    free(arr);
    return 0;
}
int futhark_values_f64_3d(struct futhark_context *ctx,
                          struct futhark_f64_3d *arr, double *data)
{
    lock_lock(&ctx->lock);
    memmove(data + 0, arr->mem.mem + 0, arr->shape[0] * arr->shape[1] *
            arr->shape[2] * sizeof(double));
    lock_unlock(&ctx->lock);
    return 0;
}
char *futhark_values_raw_f64_3d(struct futhark_context *ctx,
                                struct futhark_f64_3d *arr)
{
    (void) ctx;
    return arr->mem.mem;
}
int64_t *futhark_shape_f64_3d(struct futhark_context *ctx,
                              struct futhark_f64_3d *arr)
{
    (void) ctx;
    return arr->shape;
}
struct futhark_opaque_arr_c_2d {
    struct futhark_f64_2d *v0;
    struct futhark_f64_2d *v1;
} ;
int futhark_free_opaque_arr_c_2d(struct futhark_context *ctx,
                                 struct futhark_opaque_arr_c_2d *obj)
{
    int ret = 0, tmp;
    
    if ((tmp = futhark_free_f64_2d(ctx, obj->v0)) != 0)
        ret = tmp;
    if ((tmp = futhark_free_f64_2d(ctx, obj->v1)) != 0)
        ret = tmp;
    free(obj);
    return ret;
}
int futhark_entry_fouriertest(struct futhark_context *ctx,
                              struct futhark_f64_3d **out0, const
                              struct futhark_f64_3d *in0)
{
    struct memblock input_mem_22678;
    
    input_mem_22678.references = NULL;
    
    int32_t n_22232;
    int32_t m_22233;
    int32_t k_22234;
    struct memblock out_mem_22938;
    
    out_mem_22938.references = NULL;
    
    int32_t out_arrsizze_22939;
    int32_t out_arrsizze_22940;
    int32_t out_arrsizze_22941;
    
    lock_lock(&ctx->lock);
    input_mem_22678 = in0->mem;
    n_22232 = in0->shape[0];
    m_22233 = in0->shape[1];
    k_22234 = in0->shape[2];
    
    int ret = futrts_fouriertest(ctx, &out_mem_22938, &out_arrsizze_22939,
                                 &out_arrsizze_22940, &out_arrsizze_22941,
                                 input_mem_22678, n_22232, m_22233, k_22234);
    
    if (ret == 0) {
        assert((*out0 =
                (struct futhark_f64_3d *) malloc(sizeof(struct futhark_f64_3d))) !=
            NULL);
        (*out0)->mem = out_mem_22938;
        (*out0)->shape[0] = out_arrsizze_22939;
        (*out0)->shape[1] = out_arrsizze_22940;
        (*out0)->shape[2] = out_arrsizze_22941;
    }
    lock_unlock(&ctx->lock);
    return ret;
}
int futhark_entry_fft2D(struct futhark_context *ctx,
                        struct futhark_opaque_arr_c_2d **out0, const
                        struct futhark_opaque_arr_c_2d *in0)
{
    struct memblock arr_mem_22678;
    
    arr_mem_22678.references = NULL;
    
    struct memblock arr_mem_22679;
    
    arr_mem_22679.references = NULL;
    
    int32_t n_21954;
    int32_t m_21955;
    struct memblock out_mem_22868;
    
    out_mem_22868.references = NULL;
    
    int32_t out_arrsizze_22869;
    int32_t out_arrsizze_22870;
    struct memblock out_mem_22871;
    
    out_mem_22871.references = NULL;
    lock_lock(&ctx->lock);
    arr_mem_22678 = in0->v0->mem;
    n_21954 = in0->v0->shape[0];
    m_21955 = in0->v0->shape[1];
    arr_mem_22679 = in0->v1->mem;
    n_21954 = in0->v1->shape[0];
    m_21955 = in0->v1->shape[1];
    
    int ret = futrts_fft2D(ctx, &out_mem_22868, &out_arrsizze_22869,
                           &out_arrsizze_22870, &out_mem_22871, arr_mem_22678,
                           arr_mem_22679, n_21954, m_21955);
    
    if (ret == 0) {
        assert((*out0 =
                (struct futhark_opaque_arr_c_2d *) malloc(sizeof(struct futhark_opaque_arr_c_2d))) !=
            NULL);
        assert(((*out0)->v0 =
                (struct futhark_f64_2d *) malloc(sizeof(struct futhark_f64_2d))) !=
            NULL);
        (*out0)->v0->mem = out_mem_22868;
        (*out0)->v0->shape[0] = out_arrsizze_22869;
        (*out0)->v0->shape[1] = out_arrsizze_22870;
        assert(((*out0)->v1 =
                (struct futhark_f64_2d *) malloc(sizeof(struct futhark_f64_2d))) !=
            NULL);
        (*out0)->v1->mem = out_mem_22871;
        (*out0)->v1->shape[0] = out_arrsizze_22869;
        (*out0)->v1->shape[1] = out_arrsizze_22870;
    }
    lock_unlock(&ctx->lock);
    return ret;
}
