/*
 * Headers
*/

#include <stdint.h>
#include <stddef.h>
#include <stdbool.h>
#include <cuda.h>
#include <nvrtc.h>


/*
 * Initialisation
*/

int futhark_get_num_sizes(void);
const char *futhark_get_size_name(int);
const char *futhark_get_size_class(int);
struct futhark_context_config ;
struct futhark_context_config *futhark_context_config_new(void);
void futhark_context_config_free(struct futhark_context_config *cfg);
void futhark_context_config_add_nvrtc_option(struct futhark_context_config *cfg,
                                             const char *opt);
void futhark_context_config_set_debugging(struct futhark_context_config *cfg,
                                          int flag);
void futhark_context_config_set_logging(struct futhark_context_config *cfg,
                                        int flag);
void futhark_context_config_set_device(struct futhark_context_config *cfg, const
                                       char *s);
void futhark_context_config_dump_program_to(struct futhark_context_config *cfg,
                                            const char *path);
void
futhark_context_config_load_program_from(struct futhark_context_config *cfg,
                                         const char *path);
void futhark_context_config_dump_ptx_to(struct futhark_context_config *cfg,
                                        const char *path);
void futhark_context_config_load_ptx_from(struct futhark_context_config *cfg,
                                          const char *path);
void
futhark_context_config_set_default_group_size(struct futhark_context_config *cfg,
                                              int size);
void
futhark_context_config_set_default_num_groups(struct futhark_context_config *cfg,
                                              int num);
void
futhark_context_config_set_default_tile_size(struct futhark_context_config *cfg,
                                             int num);
void
futhark_context_config_set_default_threshold(struct futhark_context_config *cfg,
                                             int num);
int futhark_context_config_set_size(struct futhark_context_config *cfg, const
                                    char *size_name, size_t size_value);
struct futhark_context ;
struct futhark_context *futhark_context_new(struct futhark_context_config *cfg);
void futhark_context_free(struct futhark_context *ctx);
int futhark_context_sync(struct futhark_context *ctx);
char *futhark_context_get_error(struct futhark_context *ctx);
void futhark_context_pause_profiling(struct futhark_context *ctx);
void futhark_context_unpause_profiling(struct futhark_context *ctx);

/*
 * Arrays
*/

struct futhark_i32_2d ;
struct futhark_i32_2d *futhark_new_i32_2d(struct futhark_context *ctx,
                                          int32_t *data, int64_t dim0,
                                          int64_t dim1);
struct futhark_i32_2d *futhark_new_raw_i32_2d(struct futhark_context *ctx,
                                              CUdeviceptr data, int offset,
                                              int64_t dim0, int64_t dim1);
int futhark_free_i32_2d(struct futhark_context *ctx,
                        struct futhark_i32_2d *arr);
int futhark_values_i32_2d(struct futhark_context *ctx,
                          struct futhark_i32_2d *arr, int32_t *data);
CUdeviceptr futhark_values_raw_i32_2d(struct futhark_context *ctx,
                                      struct futhark_i32_2d *arr);
int64_t *futhark_shape_i32_2d(struct futhark_context *ctx,
                              struct futhark_i32_2d *arr);

/*
 * Opaque values
*/


/*
 * Entry points
*/

int futhark_entry_main(struct futhark_context *ctx,
                       struct futhark_i32_2d **out0, const
                       struct futhark_i32_2d *in0);
int futhark_entry_quicksort1(struct futhark_context *ctx,
                             struct futhark_i32_2d **out0, const
                             struct futhark_i32_2d *in0);

/*
 * Miscellaneous
*/

void futhark_debugging_report(struct futhark_context *ctx);
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <math.h>
#include <stdint.h>
#undef NDEBUG
#include <assert.h>
// Start of panic.h.

#include <stdarg.h>

static const char *fut_progname;

static void panic(int eval, const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
        fprintf(stderr, "%s: ", fut_progname);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
        exit(eval);
}

/* For generating arbitrary-sized error messages.  It is the callers
   responsibility to free the buffer at some point. */
static char* msgprintf(const char *s, ...) {
  va_list vl;
  va_start(vl, s);
  size_t needed = 1 + vsnprintf(NULL, 0, s, vl);
  char *buffer = (char*) malloc(needed);
  va_start(vl, s); /* Must re-init. */
  vsnprintf(buffer, needed, s, vl);
  return buffer;
}

// End of panic.h.

// Start of timing.h.

// The function get_wall_time() returns the wall time in microseconds
// (with an unspecified offset).

#ifdef _WIN32

#include <windows.h>

static int64_t get_wall_time(void) {
  LARGE_INTEGER time,freq;
  assert(QueryPerformanceFrequency(&freq));
  assert(QueryPerformanceCounter(&time));
  return ((double)time.QuadPart / freq.QuadPart) * 1000000;
}

#else
/* Assuming POSIX */

#include <time.h>
#include <sys/time.h>

static int64_t get_wall_time(void) {
  struct timeval time;
  assert(gettimeofday(&time,NULL) == 0);
  return time.tv_sec * 1000000 + time.tv_usec;
}

#endif

// End of timing.h.

#include <string.h>
#include <inttypes.h>
#include <errno.h>
#include <ctype.h>
#include <errno.h>
#include <getopt.h>
// Start of values.h.

//// Text I/O

typedef int (*writer)(FILE*, void*);
typedef int (*bin_reader)(void*);
typedef int (*str_reader)(const char *, void*);

struct array_reader {
  char* elems;
  int64_t n_elems_space;
  int64_t elem_size;
  int64_t n_elems_used;
  int64_t *shape;
  str_reader elem_reader;
};

static void skipspaces() {
  int c;
  do {
    c = getchar();
  } while (isspace(c));

  if (c != EOF) {
    ungetc(c, stdin);
  }
}

static int constituent(char c) {
  return isalnum(c) || c == '.' || c == '-' || c == '+' || c == '_';
}

// Produces an empty token only on EOF.
static void next_token(char *buf, int bufsize) {
 start:
  skipspaces();

  int i = 0;
  while (i < bufsize) {
    int c = getchar();
    buf[i] = c;

    if (c == EOF) {
      buf[i] = 0;
      return;
    } else if (c == '-' && i == 1 && buf[0] == '-') {
      // Line comment, so skip to end of line and start over.
      for (; c != '\n' && c != EOF; c = getchar());
      goto start;
    } else if (!constituent(c)) {
      if (i == 0) {
        // We permit single-character tokens that are not
        // constituents; this lets things like ']' and ',' be
        // tokens.
        buf[i+1] = 0;
        return;
      } else {
        ungetc(c, stdin);
        buf[i] = 0;
        return;
      }
    }

    i++;
  }

  buf[bufsize-1] = 0;
}

static int next_token_is(char *buf, int bufsize, const char* expected) {
  next_token(buf, bufsize);
  return strcmp(buf, expected) == 0;
}

static void remove_underscores(char *buf) {
  char *w = buf;

  for (char *r = buf; *r; r++) {
    if (*r != '_') {
      *w++ = *r;
    }
  }

  *w++ = 0;
}

static int read_str_elem(char *buf, struct array_reader *reader) {
  int ret;
  if (reader->n_elems_used == reader->n_elems_space) {
    reader->n_elems_space *= 2;
    reader->elems = (char*) realloc(reader->elems,
                                    reader->n_elems_space * reader->elem_size);
  }

  ret = reader->elem_reader(buf, reader->elems + reader->n_elems_used * reader->elem_size);

  if (ret == 0) {
    reader->n_elems_used++;
  }

  return ret;
}

static int read_str_array_elems(char *buf, int bufsize,
                                struct array_reader *reader, int dims) {
  int ret;
  int first = 1;
  char *knows_dimsize = (char*) calloc(dims,sizeof(char));
  int cur_dim = dims-1;
  int64_t *elems_read_in_dim = (int64_t*) calloc(dims,sizeof(int64_t));

  while (1) {
    next_token(buf, bufsize);

    if (strcmp(buf, "]") == 0) {
      if (knows_dimsize[cur_dim]) {
        if (reader->shape[cur_dim] != elems_read_in_dim[cur_dim]) {
          ret = 1;
          break;
        }
      } else {
        knows_dimsize[cur_dim] = 1;
        reader->shape[cur_dim] = elems_read_in_dim[cur_dim];
      }
      if (cur_dim == 0) {
        ret = 0;
        break;
      } else {
        cur_dim--;
        elems_read_in_dim[cur_dim]++;
      }
    } else if (strcmp(buf, ",") == 0) {
      next_token(buf, bufsize);
      if (strcmp(buf, "[") == 0) {
        if (cur_dim == dims - 1) {
          ret = 1;
          break;
        }
        first = 1;
        cur_dim++;
        elems_read_in_dim[cur_dim] = 0;
      } else if (cur_dim == dims - 1) {
        ret = read_str_elem(buf, reader);
        if (ret != 0) {
          break;
        }
        elems_read_in_dim[cur_dim]++;
      } else {
        ret = 1;
        break;
      }
    } else if (strlen(buf) == 0) {
      // EOF
      ret = 1;
      break;
    } else if (first) {
      if (strcmp(buf, "[") == 0) {
        if (cur_dim == dims - 1) {
          ret = 1;
          break;
        }
        cur_dim++;
        elems_read_in_dim[cur_dim] = 0;
      } else {
        ret = read_str_elem(buf, reader);
        if (ret != 0) {
          break;
        }
        elems_read_in_dim[cur_dim]++;
        first = 0;
      }
    } else {
      ret = 1;
      break;
    }
  }

  free(knows_dimsize);
  free(elems_read_in_dim);
  return ret;
}

static int read_str_empty_array(char *buf, int bufsize,
                                const char *type_name, int64_t *shape, int64_t dims) {
  if (strlen(buf) == 0) {
    // EOF
    return 1;
  }

  if (strcmp(buf, "empty") != 0) {
    return 1;
  }

  shape[0] = 0;

  if (!next_token_is(buf, bufsize, "(")) {
    return 1;
  }

  for (int i = 1; i < dims; i++) {
    if (!next_token_is(buf, bufsize, "[")) {
      return 1;
    }

    next_token(buf, bufsize);

    if (sscanf(buf, "%d", &shape[i]) != 1) {
      return 1;
    }

    if (!next_token_is(buf, bufsize, "]")) {
      return 1;
    }
  }

  if (!next_token_is(buf, bufsize, type_name)) {
    return 1;
  }


  if (!next_token_is(buf, bufsize, ")")) {
    return 1;
  }

  return 0;
}

static int read_str_array(int64_t elem_size, str_reader elem_reader,
                          const char *type_name,
                          void **data, int64_t *shape, int64_t dims) {
  int ret;
  struct array_reader reader;
  char buf[100];

  int dims_seen;
  for (dims_seen = 0; dims_seen < dims; dims_seen++) {
    if (!next_token_is(buf, sizeof(buf), "[")) {
      break;
    }
  }

  if (dims_seen == 0) {
    return read_str_empty_array(buf, sizeof(buf), type_name, shape, dims);
  }

  if (dims_seen != dims) {
    return 1;
  }

  reader.shape = shape;
  reader.n_elems_used = 0;
  reader.elem_size = elem_size;
  reader.n_elems_space = 16;
  reader.elems = (char*) realloc(*data, elem_size*reader.n_elems_space);
  reader.elem_reader = elem_reader;

  ret = read_str_array_elems(buf, sizeof(buf), &reader, dims);

  *data = reader.elems;

  return ret;
}

#define READ_STR(MACRO, PTR, SUFFIX)                                   \
  remove_underscores(buf);                                              \
  int j;                                                                \
  if (sscanf(buf, "%"MACRO"%n", (PTR*)dest, &j) == 1) {                 \
    return !(strcmp(buf+j, "") == 0 || strcmp(buf+j, SUFFIX) == 0);     \
  } else {                                                              \
    return 1;                                                           \
  }

static int read_str_i8(char *buf, void* dest) {
  /* Some platforms (WINDOWS) does not support scanf %hhd or its
     cousin, %SCNi8.  Read into int first to avoid corrupting
     memory.

     https://gcc.gnu.org/bugzilla/show_bug.cgi?id=63417  */
  remove_underscores(buf);
  int j, x;
  if (sscanf(buf, "%i%n", &x, &j) == 1) {
    *(int8_t*)dest = x;
    return !(strcmp(buf+j, "") == 0 || strcmp(buf+j, "i8") == 0);
  } else {
    return 1;
  }
}

static int read_str_u8(char *buf, void* dest) {
  /* Some platforms (WINDOWS) does not support scanf %hhd or its
     cousin, %SCNu8.  Read into int first to avoid corrupting
     memory.

     https://gcc.gnu.org/bugzilla/show_bug.cgi?id=63417  */
  remove_underscores(buf);
  int j, x;
  if (sscanf(buf, "%i%n", &x, &j) == 1) {
    *(uint8_t*)dest = x;
    return !(strcmp(buf+j, "") == 0 || strcmp(buf+j, "u8") == 0);
  } else {
    return 1;
  }
}

static int read_str_i16(char *buf, void* dest) {
  READ_STR(SCNi16, int16_t, "i16");
}

static int read_str_u16(char *buf, void* dest) {
  READ_STR(SCNi16, int16_t, "u16");
}

static int read_str_i32(char *buf, void* dest) {
  READ_STR(SCNi32, int32_t, "i32");
}

static int read_str_u32(char *buf, void* dest) {
  READ_STR(SCNi32, int32_t, "u32");
}

static int read_str_i64(char *buf, void* dest) {
  READ_STR(SCNi64, int64_t, "i64");
}

static int read_str_u64(char *buf, void* dest) {
  // FIXME: This is not correct, as SCNu64 only permits decimal
  // literals.  However, SCNi64 does not handle very large numbers
  // correctly (it's really for signed numbers, so that's fair).
  READ_STR(SCNu64, uint64_t, "u64");
}

static int read_str_f32(char *buf, void* dest) {
  remove_underscores(buf);
  if (strcmp(buf, "f32.nan") == 0) {
    *(float*)dest = NAN;
    return 0;
  } else if (strcmp(buf, "f32.inf") == 0) {
    *(float*)dest = INFINITY;
    return 0;
  } else if (strcmp(buf, "-f32.inf") == 0) {
    *(float*)dest = -INFINITY;
    return 0;
  } else {
    READ_STR("f", float, "f32");
  }
}

static int read_str_f64(char *buf, void* dest) {
  remove_underscores(buf);
  if (strcmp(buf, "f64.nan") == 0) {
    *(double*)dest = NAN;
    return 0;
  } else if (strcmp(buf, "f64.inf") == 0) {
    *(double*)dest = INFINITY;
    return 0;
  } else if (strcmp(buf, "-f64.inf") == 0) {
    *(double*)dest = -INFINITY;
    return 0;
  } else {
    READ_STR("lf", double, "f64");
  }
}

static int read_str_bool(char *buf, void* dest) {
  if (strcmp(buf, "true") == 0) {
    *(char*)dest = 1;
    return 0;
  } else if (strcmp(buf, "false") == 0) {
    *(char*)dest = 0;
    return 0;
  } else {
    return 1;
  }
}

static int write_str_i8(FILE *out, int8_t *src) {
  return fprintf(out, "%hhdi8", *src);
}

static int write_str_u8(FILE *out, uint8_t *src) {
  return fprintf(out, "%hhuu8", *src);
}

static int write_str_i16(FILE *out, int16_t *src) {
  return fprintf(out, "%hdi16", *src);
}

static int write_str_u16(FILE *out, uint16_t *src) {
  return fprintf(out, "%huu16", *src);
}

static int write_str_i32(FILE *out, int32_t *src) {
  return fprintf(out, "%di32", *src);
}

static int write_str_u32(FILE *out, uint32_t *src) {
  return fprintf(out, "%uu32", *src);
}

static int write_str_i64(FILE *out, int64_t *src) {
  return fprintf(out, "%"PRIi64"i64", *src);
}

static int write_str_u64(FILE *out, uint64_t *src) {
  return fprintf(out, "%"PRIu64"u64", *src);
}

static int write_str_f32(FILE *out, float *src) {
  float x = *src;
  if (isnan(x)) {
    return fprintf(out, "f32.nan");
  } else if (isinf(x) && x >= 0) {
    return fprintf(out, "f32.inf");
  } else if (isinf(x)) {
    return fprintf(out, "-f32.inf");
  } else {
    return fprintf(out, "%.6ff32", x);
  }
}

static int write_str_f64(FILE *out, double *src) {
  double x = *src;
  if (isnan(x)) {
    return fprintf(out, "f64.nan");
  } else if (isinf(x) && x >= 0) {
    return fprintf(out, "f64.inf");
  } else if (isinf(x)) {
    return fprintf(out, "-f64.inf");
  } else {
    return fprintf(out, "%.6ff64", *src);
  }
}

static int write_str_bool(FILE *out, void *src) {
  return fprintf(out, *(char*)src ? "true" : "false");
}

//// Binary I/O

#define BINARY_FORMAT_VERSION 2
#define IS_BIG_ENDIAN (!*(unsigned char *)&(uint16_t){1})

static void flip_bytes(int elem_size, unsigned char *elem) {
  for (int j=0; j<elem_size/2; j++) {
    unsigned char head = elem[j];
    int tail_index = elem_size-1-j;
    elem[j] = elem[tail_index];
    elem[tail_index] = head;
  }
}

// On Windows we need to explicitly set the file mode to not mangle
// newline characters.  On *nix there is no difference.
#ifdef _WIN32
#include <io.h>
#include <fcntl.h>
static void set_binary_mode(FILE *f) {
  setmode(fileno(f), O_BINARY);
}
#else
static void set_binary_mode(FILE *f) {
  (void)f;
}
#endif

static int read_byte(void* dest) {
  int num_elems_read = fread(dest, 1, 1, stdin);
  return num_elems_read == 1 ? 0 : 1;
}

//// Types

struct primtype_info_t {
  const char binname[4]; // Used for parsing binary data.
  const char* type_name; // Same name as in Futhark.
  const int size; // in bytes
  const writer write_str; // Write in text format.
  const str_reader read_str; // Read in text format.
};

static const struct primtype_info_t i8_info =
  {.binname = "  i8", .type_name = "i8",   .size = 1,
   .write_str = (writer)write_str_i8, .read_str = (str_reader)read_str_i8};
static const struct primtype_info_t i16_info =
  {.binname = " i16", .type_name = "i16",  .size = 2,
   .write_str = (writer)write_str_i16, .read_str = (str_reader)read_str_i16};
static const struct primtype_info_t i32_info =
  {.binname = " i32", .type_name = "i32",  .size = 4,
   .write_str = (writer)write_str_i32, .read_str = (str_reader)read_str_i32};
static const struct primtype_info_t i64_info =
  {.binname = " i64", .type_name = "i64",  .size = 8,
   .write_str = (writer)write_str_i64, .read_str = (str_reader)read_str_i64};
static const struct primtype_info_t u8_info =
  {.binname = "  u8", .type_name = "u8",   .size = 1,
   .write_str = (writer)write_str_u8, .read_str = (str_reader)read_str_u8};
static const struct primtype_info_t u16_info =
  {.binname = " u16", .type_name = "u16",  .size = 2,
   .write_str = (writer)write_str_u16, .read_str = (str_reader)read_str_u16};
static const struct primtype_info_t u32_info =
  {.binname = " u32", .type_name = "u32",  .size = 4,
   .write_str = (writer)write_str_u32, .read_str = (str_reader)read_str_u32};
static const struct primtype_info_t u64_info =
  {.binname = " u64", .type_name = "u64",  .size = 8,
   .write_str = (writer)write_str_u64, .read_str = (str_reader)read_str_u64};
static const struct primtype_info_t f32_info =
  {.binname = " f32", .type_name = "f32",  .size = 4,
   .write_str = (writer)write_str_f32, .read_str = (str_reader)read_str_f32};
static const struct primtype_info_t f64_info =
  {.binname = " f64", .type_name = "f64",  .size = 8,
   .write_str = (writer)write_str_f64, .read_str = (str_reader)read_str_f64};
static const struct primtype_info_t bool_info =
  {.binname = "bool", .type_name = "bool", .size = 1,
   .write_str = (writer)write_str_bool, .read_str = (str_reader)read_str_bool};

static const struct primtype_info_t* primtypes[] = {
  &i8_info, &i16_info, &i32_info, &i64_info,
  &u8_info, &u16_info, &u32_info, &u64_info,
  &f32_info, &f64_info,
  &bool_info,
  NULL // NULL-terminated
};

// General value interface.  All endian business taken care of at
// lower layers.

static int read_is_binary() {
  skipspaces();
  int c = getchar();
  if (c == 'b') {
    int8_t bin_version;
    int ret = read_byte(&bin_version);

    if (ret != 0) { panic(1, "binary-input: could not read version.\n"); }

    if (bin_version != BINARY_FORMAT_VERSION) {
      panic(1, "binary-input: File uses version %i, but I only understand version %i.\n",
            bin_version, BINARY_FORMAT_VERSION);
    }

    return 1;
  }
  ungetc(c, stdin);
  return 0;
}

static const struct primtype_info_t* read_bin_read_type_enum() {
  char read_binname[4];

  int num_matched = scanf("%4c", read_binname);
  if (num_matched != 1) { panic(1, "binary-input: Couldn't read element type.\n"); }

  const struct primtype_info_t **type = primtypes;

  for (; *type != NULL; type++) {
    // I compare the 4 characters manually instead of using strncmp because
    // this allows any value to be used, also NULL bytes
    if (memcmp(read_binname, (*type)->binname, 4) == 0) {
      return *type;
    }
  }
  panic(1, "binary-input: Did not recognize the type '%s'.\n", read_binname);
  return NULL;
}

static void read_bin_ensure_scalar(const struct primtype_info_t *expected_type) {
  int8_t bin_dims;
  int ret = read_byte(&bin_dims);
  if (ret != 0) { panic(1, "binary-input: Couldn't get dims.\n"); }

  if (bin_dims != 0) {
    panic(1, "binary-input: Expected scalar (0 dimensions), but got array with %i dimensions.\n",
          bin_dims);
  }

  const struct primtype_info_t *bin_type = read_bin_read_type_enum();
  if (bin_type != expected_type) {
    panic(1, "binary-input: Expected scalar of type %s but got scalar of type %s.\n",
          expected_type->type_name,
          bin_type->type_name);
  }
}

//// High-level interface

static int read_bin_array(const struct primtype_info_t *expected_type, void **data, int64_t *shape, int64_t dims) {
  int ret;

  int8_t bin_dims;
  ret = read_byte(&bin_dims);
  if (ret != 0) { panic(1, "binary-input: Couldn't get dims.\n"); }

  if (bin_dims != dims) {
    panic(1, "binary-input: Expected %i dimensions, but got array with %i dimensions.\n",
          dims, bin_dims);
  }

  const struct primtype_info_t *bin_primtype = read_bin_read_type_enum();
  if (expected_type != bin_primtype) {
    panic(1, "binary-input: Expected %iD-array with element type '%s' but got %iD-array with element type '%s'.\n",
          dims, expected_type->type_name, dims, bin_primtype->type_name);
  }

  uint64_t elem_count = 1;
  for (int i=0; i<dims; i++) {
    uint64_t bin_shape;
    ret = fread(&bin_shape, sizeof(bin_shape), 1, stdin);
    if (ret != 1) {
      panic(1, "binary-input: Couldn't read size for dimension %i of array.\n", i);
    }
    if (IS_BIG_ENDIAN) {
      flip_bytes(sizeof(bin_shape), (unsigned char*) &bin_shape);
    }
    elem_count *= bin_shape;
    shape[i] = (int64_t) bin_shape;
  }

  size_t elem_size = expected_type->size;
  void* tmp = realloc(*data, elem_count * elem_size);
  if (tmp == NULL) {
    panic(1, "binary-input: Failed to allocate array of size %i.\n",
          elem_count * elem_size);
  }
  *data = tmp;

  size_t num_elems_read = fread(*data, elem_size, elem_count, stdin);
  if (num_elems_read != elem_count) {
    panic(1, "binary-input: tried to read %i elements of an array, but only got %i elements.\n",
          elem_count, num_elems_read);
  }

  // If we're on big endian platform we must change all multibyte elements
  // from using little endian to big endian
  if (IS_BIG_ENDIAN && elem_size != 1) {
    flip_bytes(elem_size, (unsigned char*) *data);
  }

  return 0;
}

static int read_array(const struct primtype_info_t *expected_type, void **data, int64_t *shape, int64_t dims) {
  if (!read_is_binary()) {
    return read_str_array(expected_type->size, (str_reader)expected_type->read_str, expected_type->type_name, data, shape, dims);
  } else {
    return read_bin_array(expected_type, data, shape, dims);
  }
}

static int write_str_array(FILE *out, const struct primtype_info_t *elem_type, unsigned char *data, int64_t *shape, int8_t rank) {
  if (rank==0) {
    elem_type->write_str(out, (void*)data);
  } else {
    int64_t len = shape[0];
    int64_t slice_size = 1;

    int64_t elem_size = elem_type->size;
    for (int64_t i = 1; i < rank; i++) {
      slice_size *= shape[i];
    }

    if (len*slice_size == 0) {
      printf("empty(");
      for (int64_t i = 1; i < rank; i++) {
        printf("[%d]", shape[i]);
      }
      printf("%s", elem_type->type_name);
      printf(")");
    } else if (rank==1) {
      putchar('[');
      for (int64_t i = 0; i < len; i++) {
        elem_type->write_str(out, (void*) (data + i * elem_size));
        if (i != len-1) {
          printf(", ");
        }
      }
      putchar(']');
    } else {
      putchar('[');
      for (int64_t i = 0; i < len; i++) {
        write_str_array(out, elem_type, data + i * slice_size * elem_size, shape+1, rank-1);
        if (i != len-1) {
          printf(", ");
        }
      }
      putchar(']');
    }
  }
  return 0;
}

static int write_bin_array(FILE *out, const struct primtype_info_t *elem_type, unsigned char *data, int64_t *shape, int8_t rank) {
  int64_t num_elems = 1;
  for (int64_t i = 0; i < rank; i++) {
    num_elems *= shape[i];
  }

  fputc('b', out);
  fputc((char)BINARY_FORMAT_VERSION, out);
  fwrite(&rank, sizeof(int8_t), 1, out);
  fputs(elem_type->binname, out);
  if (shape != NULL) {
    fwrite(shape, sizeof(int64_t), rank, out);
  }

  if (IS_BIG_ENDIAN) {
    for (int64_t i = 0; i < num_elems; i++) {
      unsigned char *elem = data+i*elem_type->size;
      for (int64_t j = 0; j < elem_type->size; j++) {
        fwrite(&elem[elem_type->size-j], 1, 1, out);
      }
    }
  } else {
    fwrite(data, elem_type->size, num_elems, out);
  }

  return 0;
}

static int write_array(FILE *out, int write_binary,
                       const struct primtype_info_t *elem_type, void *data, int64_t *shape, int8_t rank) {
  if (write_binary) {
    return write_bin_array(out, elem_type, data, shape, rank);
  } else {
    return write_str_array(out, elem_type, data, shape, rank);
  }
}

static int read_scalar(const struct primtype_info_t *expected_type, void *dest) {
  if (!read_is_binary()) {
    char buf[100];
    next_token(buf, sizeof(buf));
    return expected_type->read_str(buf, dest);
  } else {
    read_bin_ensure_scalar(expected_type);
    size_t elem_size = expected_type->size;
    int num_elems_read = fread(dest, elem_size, 1, stdin);
    if (IS_BIG_ENDIAN) {
      flip_bytes(elem_size, (unsigned char*) dest);
    }
    return num_elems_read == 1 ? 0 : 1;
  }
}

static int write_scalar(FILE *out, int write_binary, const struct primtype_info_t *type, void *src) {
  if (write_binary) {
    return write_bin_array(out, type, src, NULL, 0);
  } else {
    return type->write_str(out, src);
  }
}

// End of values.h.

static int binary_output = 0;
static FILE *runtime_file;
static int perform_warmup = 0;
static int num_runs = 1;
static const char *entry_point = "main";
// Start of tuning.h.

static char* load_tuning_file(const char *fname,
                              void *cfg,
                              int (*set_size)(void*, const char*, size_t)) {
  const int max_line_len = 1024;
  char* line = (char*) malloc(max_line_len);

  FILE *f = fopen(fname, "r");

  if (f == NULL) {
    snprintf(line, max_line_len, "Cannot open file: %s", strerror(errno));
    return line;
  }

  int lineno = 0;
  while (fgets(line, max_line_len, f) != NULL) {
    lineno++;
    char *eql = strstr(line, "=");
    if (eql) {
      *eql = 0;
      int value = atoi(eql+1);
      if (set_size(cfg, line, value) != 0) {
        strncpy(eql+1, line, max_line_len-strlen(line)-1);
        snprintf(line, max_line_len, "Unknown name '%s' on line %d.", eql+1, lineno);
        return line;
      }
    } else {
      snprintf(line, max_line_len, "Invalid line %d (must be of form 'name=int').",
               lineno);
      return line;
    }
  }

  free(line);

  return NULL;
}

// End of tuning.h.

int parse_options(struct futhark_context_config *cfg, int argc,
                  char *const argv[])
{
    int ch;
    static struct option long_options[] = {{"write-runtime-to",
                                            required_argument, NULL, 1},
                                           {"runs", required_argument, NULL, 2},
                                           {"debugging", no_argument, NULL, 3},
                                           {"log", no_argument, NULL, 4},
                                           {"entry-point", required_argument,
                                            NULL, 5}, {"binary-output",
                                                       no_argument, NULL, 6},
                                           {"dump-cuda", required_argument,
                                            NULL, 7}, {"load-cuda",
                                                       required_argument, NULL,
                                                       8}, {"dump-ptx",
                                                            required_argument,
                                                            NULL, 9},
                                           {"load-ptx", required_argument, NULL,
                                            10}, {"nvrtc-option",
                                                  required_argument, NULL, 11},
                                           {"print-sizes", no_argument, NULL,
                                            12}, {"default-group-size",
                                                  required_argument, NULL, 13},
                                           {"default-num-groups",
                                            required_argument, NULL, 14},
                                           {"default-tile-size",
                                            required_argument, NULL, 15},
                                           {"default-threshold",
                                            required_argument, NULL, 16},
                                           {"tuning", required_argument, NULL,
                                            17}, {0, 0, 0, 0}};
    
    while ((ch = getopt_long(argc, argv, ":t:r:DLe:b", long_options, NULL)) !=
           -1) {
        if (ch == 1 || ch == 't') {
            runtime_file = fopen(optarg, "w");
            if (runtime_file == NULL)
                panic(1, "Cannot open %s: %s\n", optarg, strerror(errno));
        }
        if (ch == 2 || ch == 'r') {
            num_runs = atoi(optarg);
            perform_warmup = 1;
            if (num_runs <= 0)
                panic(1, "Need a positive number of runs, not %s\n", optarg);
        }
        if (ch == 3 || ch == 'D')
            futhark_context_config_set_debugging(cfg, 1);
        if (ch == 4 || ch == 'L')
            futhark_context_config_set_logging(cfg, 1);
        if (ch == 5 || ch == 'e') {
            if (entry_point != NULL)
                entry_point = optarg;
        }
        if (ch == 6 || ch == 'b')
            binary_output = 1;
        if (ch == 7) {
            futhark_context_config_dump_program_to(cfg, optarg);
            entry_point = NULL;
        }
        if (ch == 8)
            futhark_context_config_load_program_from(cfg, optarg);
        if (ch == 9) {
            futhark_context_config_dump_ptx_to(cfg, optarg);
            entry_point = NULL;
        }
        if (ch == 10)
            futhark_context_config_load_ptx_from(cfg, optarg);
        if (ch == 11)
            futhark_context_config_add_nvrtc_option(cfg, optarg);
        if (ch == 12) {
            int n = futhark_get_num_sizes();
            
            for (int i = 0; i < n; i++)
                printf("%s (%s)\n", futhark_get_size_name(i),
                       futhark_get_size_class(i));
            exit(0);
        }
        if (ch == 13)
            futhark_context_config_set_default_group_size(cfg, atoi(optarg));
        if (ch == 14)
            futhark_context_config_set_default_num_groups(cfg, atoi(optarg));
        if (ch == 15)
            futhark_context_config_set_default_tile_size(cfg, atoi(optarg));
        if (ch == 16)
            futhark_context_config_set_default_threshold(cfg, atoi(optarg));
        if (ch == 17) {
            char *fname = optarg;
            char *ret = load_tuning_file(optarg, cfg, (int (*)(void *, const
                                                               char *,
                                                               size_t)) futhark_context_config_set_size);
            
            if (ret != NULL)
                panic(1, "When loading tuning from '%s': %s\n", optarg, ret);
        }
        if (ch == ':')
            panic(-1, "Missing argument for option %s\n", argv[optind - 1]);
        if (ch == '?') {
            fprintf(stderr, "Usage: %s: %s\n", fut_progname,
                    "[-t/--write-runtime-to FILE] [-r/--runs INT] [-D/--debugging] [-L/--log] [-e/--entry-point NAME] [-b/--binary-output] [--dump-cuda FILE] [--load-cuda FILE] [--dump-ptx FILE] [--load-ptx FILE] [--nvrtc-option OPT] [--print-sizes] [--default-group-size INT] [--default-num-groups INT] [--default-tile-size INT] [--default-threshold INT] [--tuning FILE]");
            panic(1, "Unknown option: %s\n", argv[optind - 1]);
        }
    }
    return optind;
}
static void futrts_cli_entry_main(struct futhark_context *ctx)
{
    int64_t t_start, t_end;
    int time_runs = 0, profile_run = 0;
    
    // We do not want to profile all the initialisation.
    futhark_context_pause_profiling(ctx);
    /* Declare and read input. */
    set_binary_mode(stdin);
    
    struct futhark_i32_2d *read_value_33861;
    int64_t read_shape_33862[2];
    int32_t *read_arr_33863 = NULL;
    
    errno = 0;
    if (read_array(&i32_info, (void **) &read_arr_33863, read_shape_33862, 2) !=
        0)
        panic(1, "Cannot read input #%d of type %s%s (errno: %s).\n", 0, "[][]",
              i32_info.type_name, strerror(errno));
    
    struct futhark_i32_2d *result_33864;
    
    if (perform_warmup) {
        int r;
        
        assert((read_value_33861 = futhark_new_i32_2d(ctx, read_arr_33863,
                                                      read_shape_33862[0],
                                                      read_shape_33862[1])) !=
            0);
        assert(futhark_context_sync(ctx) == 0);
        // Only profile last run.
        if (profile_run)
            futhark_context_unpause_profiling(ctx);
        t_start = get_wall_time();
        r = futhark_entry_main(ctx, &result_33864, read_value_33861);
        if (r != 0)
            panic(1, "%s", futhark_context_get_error(ctx));
        assert(futhark_context_sync(ctx) == 0);
        if (profile_run)
            futhark_context_pause_profiling(ctx);
        t_end = get_wall_time();
        
        long elapsed_usec = t_end - t_start;
        
        if (time_runs && runtime_file != NULL)
            fprintf(runtime_file, "%lld\n", (long long) elapsed_usec);
        assert(futhark_free_i32_2d(ctx, read_value_33861) == 0);
        assert(futhark_free_i32_2d(ctx, result_33864) == 0);
    }
    time_runs = 1;
    /* Proper run. */
    for (int run = 0; run < num_runs; run++) {
        // Only profile last run.
        profile_run = run == num_runs - 1;
        
        int r;
        
        assert((read_value_33861 = futhark_new_i32_2d(ctx, read_arr_33863,
                                                      read_shape_33862[0],
                                                      read_shape_33862[1])) !=
            0);
        assert(futhark_context_sync(ctx) == 0);
        // Only profile last run.
        if (profile_run)
            futhark_context_unpause_profiling(ctx);
        t_start = get_wall_time();
        r = futhark_entry_main(ctx, &result_33864, read_value_33861);
        if (r != 0)
            panic(1, "%s", futhark_context_get_error(ctx));
        assert(futhark_context_sync(ctx) == 0);
        if (profile_run)
            futhark_context_pause_profiling(ctx);
        t_end = get_wall_time();
        
        long elapsed_usec = t_end - t_start;
        
        if (time_runs && runtime_file != NULL)
            fprintf(runtime_file, "%lld\n", (long long) elapsed_usec);
        assert(futhark_free_i32_2d(ctx, read_value_33861) == 0);
        if (run < num_runs - 1) {
            assert(futhark_free_i32_2d(ctx, result_33864) == 0);
        }
    }
    free(read_arr_33863);
    if (binary_output)
        set_binary_mode(stdout);
    {
        int32_t *arr = calloc(sizeof(int32_t), futhark_shape_i32_2d(ctx,
                                                                    result_33864)[0] *
                              futhark_shape_i32_2d(ctx, result_33864)[1]);
        
        assert(arr != NULL);
        assert(futhark_values_i32_2d(ctx, result_33864, arr) == 0);
        write_array(stdout, binary_output, &i32_info, arr,
                    futhark_shape_i32_2d(ctx, result_33864), 2);
        free(arr);
    }
    printf("\n");
    assert(futhark_free_i32_2d(ctx, result_33864) == 0);
}
static void futrts_cli_entry_quicksort1(struct futhark_context *ctx)
{
    int64_t t_start, t_end;
    int time_runs = 0, profile_run = 0;
    
    // We do not want to profile all the initialisation.
    futhark_context_pause_profiling(ctx);
    /* Declare and read input. */
    set_binary_mode(stdin);
    
    struct futhark_i32_2d *read_value_33865;
    int64_t read_shape_33866[2];
    int32_t *read_arr_33867 = NULL;
    
    errno = 0;
    if (read_array(&i32_info, (void **) &read_arr_33867, read_shape_33866, 2) !=
        0)
        panic(1, "Cannot read input #%d of type %s%s (errno: %s).\n", 0, "[][]",
              i32_info.type_name, strerror(errno));
    
    struct futhark_i32_2d *result_33868;
    
    if (perform_warmup) {
        int r;
        
        assert((read_value_33865 = futhark_new_i32_2d(ctx, read_arr_33867,
                                                      read_shape_33866[0],
                                                      read_shape_33866[1])) !=
            0);
        assert(futhark_context_sync(ctx) == 0);
        // Only profile last run.
        if (profile_run)
            futhark_context_unpause_profiling(ctx);
        t_start = get_wall_time();
        r = futhark_entry_quicksort1(ctx, &result_33868, read_value_33865);
        if (r != 0)
            panic(1, "%s", futhark_context_get_error(ctx));
        assert(futhark_context_sync(ctx) == 0);
        if (profile_run)
            futhark_context_pause_profiling(ctx);
        t_end = get_wall_time();
        
        long elapsed_usec = t_end - t_start;
        
        if (time_runs && runtime_file != NULL)
            fprintf(runtime_file, "%lld\n", (long long) elapsed_usec);
        assert(futhark_free_i32_2d(ctx, read_value_33865) == 0);
        assert(futhark_free_i32_2d(ctx, result_33868) == 0);
    }
    time_runs = 1;
    /* Proper run. */
    for (int run = 0; run < num_runs; run++) {
        // Only profile last run.
        profile_run = run == num_runs - 1;
        
        int r;
        
        assert((read_value_33865 = futhark_new_i32_2d(ctx, read_arr_33867,
                                                      read_shape_33866[0],
                                                      read_shape_33866[1])) !=
            0);
        assert(futhark_context_sync(ctx) == 0);
        // Only profile last run.
        if (profile_run)
            futhark_context_unpause_profiling(ctx);
        t_start = get_wall_time();
        r = futhark_entry_quicksort1(ctx, &result_33868, read_value_33865);
        if (r != 0)
            panic(1, "%s", futhark_context_get_error(ctx));
        assert(futhark_context_sync(ctx) == 0);
        if (profile_run)
            futhark_context_pause_profiling(ctx);
        t_end = get_wall_time();
        
        long elapsed_usec = t_end - t_start;
        
        if (time_runs && runtime_file != NULL)
            fprintf(runtime_file, "%lld\n", (long long) elapsed_usec);
        assert(futhark_free_i32_2d(ctx, read_value_33865) == 0);
        if (run < num_runs - 1) {
            assert(futhark_free_i32_2d(ctx, result_33868) == 0);
        }
    }
    free(read_arr_33867);
    if (binary_output)
        set_binary_mode(stdout);
    {
        int32_t *arr = calloc(sizeof(int32_t), futhark_shape_i32_2d(ctx,
                                                                    result_33868)[0] *
                              futhark_shape_i32_2d(ctx, result_33868)[1]);
        
        assert(arr != NULL);
        assert(futhark_values_i32_2d(ctx, result_33868, arr) == 0);
        write_array(stdout, binary_output, &i32_info, arr,
                    futhark_shape_i32_2d(ctx, result_33868), 2);
        free(arr);
    }
    printf("\n");
    assert(futhark_free_i32_2d(ctx, result_33868) == 0);
}
typedef void entry_point_fun(struct futhark_context *);
struct entry_point_entry {
    const char *name;
    entry_point_fun *fun;
} ;
int main(int argc, char **argv)
{
    fut_progname = argv[0];
    
    struct entry_point_entry entry_points[] = {{.name ="main", .fun =
                                                futrts_cli_entry_main}, {.name =
                                                                         "quicksort1",
                                                                         .fun =
                                                                         futrts_cli_entry_quicksort1}};
    struct futhark_context_config *cfg = futhark_context_config_new();
    
    assert(cfg != NULL);
    
    int parsed_options = parse_options(cfg, argc, argv);
    
    argc -= parsed_options;
    argv += parsed_options;
    if (argc != 0)
        panic(1, "Excess non-option: %s\n", argv[0]);
    
    struct futhark_context *ctx = futhark_context_new(cfg);
    
    assert(ctx != NULL);
    if (entry_point != NULL) {
        int num_entry_points = sizeof(entry_points) / sizeof(entry_points[0]);
        entry_point_fun *entry_point_fun = NULL;
        
        for (int i = 0; i < num_entry_points; i++) {
            if (strcmp(entry_points[i].name, entry_point) == 0) {
                entry_point_fun = entry_points[i].fun;
                break;
            }
        }
        if (entry_point_fun == NULL) {
            fprintf(stderr,
                    "No entry point '%s'.  Select another with --entry-point.  Options are:\n",
                    entry_point);
            for (int i = 0; i < num_entry_points; i++)
                fprintf(stderr, "%s\n", entry_points[i].name);
            return 1;
        }
        entry_point_fun(ctx);
        if (runtime_file != NULL)
            fclose(runtime_file);
        futhark_debugging_report(ctx);
    }
    futhark_context_free(ctx);
    futhark_context_config_free(cfg);
    return 0;
}
#ifdef _MSC_VER
#define inline __inline
#endif
#include <string.h>
#include <inttypes.h>
#include <ctype.h>
#include <errno.h>
#include <assert.h>
// Start of lock.h.

/* A very simple cross-platform implementation of locks.  Uses
   pthreads on Unix and some Windows thing there.  Futhark's
   host-level code is not multithreaded, but user code may be, so we
   need some mechanism for ensuring atomic access to API functions.
   This is that mechanism.  It is not exposed to user code at all, so
   we do not have to worry about name collisions. */

#ifdef _WIN32

typedef HANDLE lock_t;

static lock_t create_lock(lock_t *lock) {
  *lock = CreateMutex(NULL,  /* Default security attributes. */
                      FALSE, /* Initially unlocked. */
                      NULL); /* Unnamed. */
}

static void lock_lock(lock_t *lock) {
  assert(WaitForSingleObject(*lock, INFINITE) == WAIT_OBJECT_0);
}

static void lock_unlock(lock_t *lock) {
  assert(ReleaseMutex(*lock));
}

static void free_lock(lock_t *lock) {
  CloseHandle(*lock);
}

#else
/* Assuming POSIX */

#include <pthread.h>

typedef pthread_mutex_t lock_t;

static void create_lock(lock_t *lock) {
  int r = pthread_mutex_init(lock, NULL);
  assert(r == 0);
}

static void lock_lock(lock_t *lock) {
  int r = pthread_mutex_lock(lock);
  assert(r == 0);
}

static void lock_unlock(lock_t *lock) {
  int r = pthread_mutex_unlock(lock);
  assert(r == 0);
}

static void free_lock(lock_t *lock) {
  /* Nothing to do for pthreads. */
  (void)lock;
}

#endif

// End of lock.h.

#include <cuda.h>
#include <nvrtc.h>
typedef CUdeviceptr fl_mem_t;
// Start of free_list.h.

/* An entry in the free list.  May be invalid, to avoid having to
   deallocate entries as soon as they are removed.  There is also a
   tag, to help with memory reuse. */
struct free_list_entry {
  size_t size;
  fl_mem_t mem;
  const char *tag;
  unsigned char valid;
};

struct free_list {
  struct free_list_entry *entries;        // Pointer to entries.
  int capacity;                           // Number of entries.
  int used;                               // Number of valid entries.
};

static void free_list_init(struct free_list *l) {
  l->capacity = 30; // Picked arbitrarily.
  l->used = 0;
  l->entries = (struct free_list_entry*) malloc(sizeof(struct free_list_entry) * l->capacity);
  for (int i = 0; i < l->capacity; i++) {
    l->entries[i].valid = 0;
  }
}

/* Remove invalid entries from the free list. */
static void free_list_pack(struct free_list *l) {
  int p = 0;
  for (int i = 0; i < l->capacity; i++) {
    if (l->entries[i].valid) {
      l->entries[p] = l->entries[i];
      p++;
    }
  }
  // Now p == l->used.
  l->entries = realloc(l->entries, l->used * sizeof(struct free_list_entry));
  l->capacity = l->used;
}

static void free_list_destroy(struct free_list *l) {
  assert(l->used == 0);
  free(l->entries);
}

static int free_list_find_invalid(struct free_list *l) {
  int i;
  for (i = 0; i < l->capacity; i++) {
    if (!l->entries[i].valid) {
      break;
    }
  }
  return i;
}

static void free_list_insert(struct free_list *l, size_t size, fl_mem_t mem, const char *tag) {
  int i = free_list_find_invalid(l);

  if (i == l->capacity) {
    // List is full; so we have to grow it.
    int new_capacity = l->capacity * 2 * sizeof(struct free_list_entry);
    l->entries = realloc(l->entries, new_capacity);
    for (int j = 0; j < l->capacity; j++) {
      l->entries[j+l->capacity].valid = 0;
    }
    l->capacity *= 2;
  }

  // Now 'i' points to the first invalid entry.
  l->entries[i].valid = 1;
  l->entries[i].size = size;
  l->entries[i].mem = mem;
  l->entries[i].tag = tag;

  l->used++;
}

/* Find and remove a memory block of at least the desired size and
   tag.  Returns 0 on success.  */
static int free_list_find(struct free_list *l, const char *tag, size_t *size_out, fl_mem_t *mem_out) {
  int i;
  for (i = 0; i < l->capacity; i++) {
    if (l->entries[i].valid && l->entries[i].tag == tag) {
      l->entries[i].valid = 0;
      *size_out = l->entries[i].size;
      *mem_out = l->entries[i].mem;
      l->used--;
      return 0;
    }
  }

  return 1;
}

/* Remove the first block in the free list.  Returns 0 if a block was
   removed, and nonzero if the free list was already empty. */
static int free_list_first(struct free_list *l, fl_mem_t *mem_out) {
  for (int i = 0; i < l->capacity; i++) {
    if (l->entries[i].valid) {
      l->entries[i].valid = 0;
      *mem_out = l->entries[i].mem;
      l->used--;
      return 0;
    }
  }

  return 1;
}

// End of free_list.h.

// Start of cuda.h.

#define CUDA_SUCCEED(x) cuda_api_succeed(x, #x, __FILE__, __LINE__)
#define NVRTC_SUCCEED(x) nvrtc_api_succeed(x, #x, __FILE__, __LINE__)

static inline void cuda_api_succeed(CUresult res, const char *call,
    const char *file, int line) {
  if (res != CUDA_SUCCESS) {
    const char *err_str;
    cuGetErrorString(res, &err_str);
    if (err_str == NULL) { err_str = "Unknown"; }
    panic(-1, "%s:%d: CUDA call\n  %s\nfailed with error code %d (%s)\n",
        file, line, call, res, err_str);
  }
}

static inline void nvrtc_api_succeed(nvrtcResult res, const char *call,
                                     const char *file, int line) {
  if (res != NVRTC_SUCCESS) {
    const char *err_str = nvrtcGetErrorString(res);
    panic(-1, "%s:%d: NVRTC call\n  %s\nfailed with error code %d (%s)\n",
        file, line, call, res, err_str);
  }
}

struct cuda_config {
  int debugging;
  int logging;
  const char *preferred_device;

  const char *dump_program_to;
  const char *load_program_from;

  const char *dump_ptx_to;
  const char *load_ptx_from;

  size_t default_block_size;
  size_t default_grid_size;
  size_t default_tile_size;
  size_t default_threshold;

  int default_block_size_changed;
  int default_grid_size_changed;
  int default_tile_size_changed;

  int num_sizes;
  const char **size_names;
  const char **size_vars;
  size_t *size_values;
  const char **size_classes;
};

static void cuda_config_init(struct cuda_config *cfg,
                             int num_sizes,
                             const char *size_names[],
                             const char *size_vars[],
                             size_t *size_values,
                             const char *size_classes[]) {
  cfg->debugging = 0;
  cfg->logging = 0;
  cfg->preferred_device = "";

  cfg->dump_program_to = NULL;
  cfg->load_program_from = NULL;

  cfg->dump_ptx_to = NULL;
  cfg->load_ptx_from = NULL;

  cfg->default_block_size = 256;
  cfg->default_grid_size = 128;
  cfg->default_tile_size = 32;
  cfg->default_threshold = 32*1024;

  cfg->default_block_size_changed = 0;
  cfg->default_grid_size_changed = 0;
  cfg->default_tile_size_changed = 0;

  cfg->num_sizes = num_sizes;
  cfg->size_names = size_names;
  cfg->size_vars = size_vars;
  cfg->size_values = size_values;
  cfg->size_classes = size_classes;
}

struct cuda_context {
  CUdevice dev;
  CUcontext cu_ctx;
  CUmodule module;

  struct cuda_config cfg;

  struct free_list free_list;

  size_t max_block_size;
  size_t max_grid_size;
  size_t max_tile_size;
  size_t max_threshold;
  size_t max_shared_memory;

  size_t lockstep_width;
};

#define CU_DEV_ATTR(x) (CU_DEVICE_ATTRIBUTE_##x)
#define device_query(dev,attrib) _device_query(dev, CU_DEV_ATTR(attrib))
static int _device_query(CUdevice dev, CUdevice_attribute attrib) {
  int val;
  CUDA_SUCCEED(cuDeviceGetAttribute(&val, attrib, dev));
  return val;
}

#define CU_FUN_ATTR(x) (CU_FUNC_ATTRIBUTE_##x)
#define function_query(fn,attrib) _function_query(dev, CU_FUN_ATTR(attrib))
static int _function_query(CUfunction dev, CUfunction_attribute attrib) {
  int val;
  CUDA_SUCCEED(cuFuncGetAttribute(&val, attrib, dev));
  return val;
}

static void set_preferred_device(struct cuda_config *cfg, const char *s) {
  cfg->preferred_device = s;
}

static int cuda_device_setup(struct cuda_context *ctx) {
  char name[256];
  int count, chosen = -1, best_cc = -1;
  int cc_major_best, cc_minor_best;
  int cc_major, cc_minor;
  CUdevice dev;

  CUDA_SUCCEED(cuDeviceGetCount(&count));
  if (count == 0) { return 1; }

  // XXX: Current device selection policy is to choose the device with the
  // highest compute capability (if no preferred device is set).
  // This should maybe be changed, since greater compute capability is not
  // necessarily an indicator of better performance.
  for (int i = 0; i < count; i++) {
    CUDA_SUCCEED(cuDeviceGet(&dev, i));

    cc_major = device_query(dev, COMPUTE_CAPABILITY_MAJOR);
    cc_minor = device_query(dev, COMPUTE_CAPABILITY_MINOR);

    CUDA_SUCCEED(cuDeviceGetName(name, sizeof(name)/sizeof(name[0]) - 1, dev));
    name[sizeof(name)/sizeof(name[0])] = 0;

    if (ctx->cfg.debugging) {
      fprintf(stderr, "Device #%d: name=\"%s\", compute capability=%d.%d\n",
          i, name, cc_major, cc_minor);
    }

    if (device_query(dev, COMPUTE_MODE) == CU_COMPUTEMODE_PROHIBITED) {
      if (ctx->cfg.debugging) {
        fprintf(stderr, "Device #%d is compute-prohibited, ignoring\n", i);
      }
      continue;
    }

    if (best_cc == -1 || cc_major > cc_major_best ||
        (cc_major == cc_major_best && cc_minor > cc_minor_best)) {
      best_cc = i;
      cc_major_best = cc_major;
      cc_minor_best = cc_minor;
    }

    if (chosen == -1 && strstr(name, ctx->cfg.preferred_device) == name) {
      chosen = i;
    }
  }

  if (chosen == -1) { chosen = best_cc; }
  if (chosen == -1) { return 1; }

  if (ctx->cfg.debugging) {
    fprintf(stderr, "Using device #%d\n", chosen);
  }

  CUDA_SUCCEED(cuDeviceGet(&ctx->dev, chosen));
  return 0;
}

static char *concat_fragments(const char *src_fragments[]) {
  size_t src_len = 0;
  const char **p;

  for (p = src_fragments; *p; p++) {
    src_len += strlen(*p);
  }

  char *src = (char*) malloc(src_len + 1);
  size_t n = 0;
  for (p = src_fragments; *p; p++) {
    strcpy(src + n, *p);
    n += strlen(*p);
  }

  return src;
}

static const char *cuda_nvrtc_get_arch(CUdevice dev) {
  struct {
    int major;
    int minor;
    const char *arch_str;
  } static const x[] = {
    { 3, 0, "compute_30" },
    { 3, 2, "compute_32" },
    { 3, 5, "compute_35" },
    { 3, 7, "compute_37" },
    { 5, 0, "compute_50" },
    { 5, 2, "compute_52" },
    { 5, 3, "compute_53" },
    { 6, 0, "compute_60" },
    { 6, 1, "compute_61" },
    { 6, 2, "compute_62" },
    { 7, 0, "compute_70" },
    { 7, 2, "compute_72" }
  };

  int major = device_query(dev, COMPUTE_CAPABILITY_MAJOR);
  int minor = device_query(dev, COMPUTE_CAPABILITY_MINOR);

  int chosen = -1;
  for (int i = 0; i < sizeof(x)/sizeof(x[0]); i++) {
    if (x[i].major < major || (x[i].major == major && x[i].minor <= minor)) {
      chosen = i;
    } else {
      break;
    }
  }

  if (chosen == -1) {
    panic(-1, "Unsupported compute capability %d.%d\n", major, minor);
  }
  return x[chosen].arch_str;
}

static char *cuda_nvrtc_build(struct cuda_context *ctx, const char *src,
                              const char *extra_opts[]) {
  nvrtcProgram prog;
  NVRTC_SUCCEED(nvrtcCreateProgram(&prog, src, "futhark-cuda", 0, NULL, NULL));
  int arch_set = 0, num_extra_opts;

  // nvrtc cannot handle multiple -arch options.  Hence, if one of the
  // extra_opts is -arch, we have to be careful not to do our usual
  // automatic generation.
  for (num_extra_opts = 0; extra_opts[num_extra_opts] != NULL; num_extra_opts++) {
    if (strstr(extra_opts[num_extra_opts], "-arch")
        == extra_opts[num_extra_opts] ||
        strstr(extra_opts[num_extra_opts], "--gpu-architecture")
        == extra_opts[num_extra_opts]) {
      arch_set = 1;
    }
  }

  size_t n_opts, i = 0, i_dyn, n_opts_alloc = 20 + num_extra_opts + ctx->cfg.num_sizes;
  const char **opts = (const char**) malloc(n_opts_alloc * sizeof(const char *));
  if (!arch_set) {
    opts[i++] = "-arch";
    opts[i++] = cuda_nvrtc_get_arch(ctx->dev);
  }
  opts[i++] = "-default-device";
  if (ctx->cfg.debugging) {
    opts[i++] = "-G";
    opts[i++] = "-lineinfo";
  } else {
    opts[i++] = "--disable-warnings";
  }
  i_dyn = i;
  for (size_t j = 0; j < ctx->cfg.num_sizes; j++) {
    opts[i++] = msgprintf("-D%s=%zu", ctx->cfg.size_vars[j],
        ctx->cfg.size_values[j]);
  }
  opts[i++] = msgprintf("-DLOCKSTEP_WIDTH=%zu", ctx->lockstep_width);
  opts[i++] = msgprintf("-DMAX_THREADS_PER_BLOCK=%zu", ctx->max_block_size);

  // It is crucial that the extra_opts are last, so that the free()
  // logic below does not cause problems.
  for (int j = 0; extra_opts[j] != NULL; j++) {
    opts[i++] = extra_opts[j];
  }

  n_opts = i;

  if (ctx->cfg.debugging) {
    fprintf(stderr, "NVRTC compile options:\n");
    for (size_t j = 0; j < n_opts; j++) {
      fprintf(stderr, "\t%s\n", opts[j]);
    }
    fprintf(stderr, "\n");
  }

  nvrtcResult res = nvrtcCompileProgram(prog, n_opts, opts);
  if (res != NVRTC_SUCCESS) {
    size_t log_size;
    if (nvrtcGetProgramLogSize(prog, &log_size) == NVRTC_SUCCESS) {
      char *log = (char*) malloc(log_size);
      if (nvrtcGetProgramLog(prog, log) == NVRTC_SUCCESS) {
        fprintf(stderr,"Compilation log:\n%s\n", log);
      }
      free(log);
    }
    NVRTC_SUCCEED(res);
  }

  for (i = i_dyn; i < n_opts-num_extra_opts; i++) { free((char *)opts[i]); }
  free(opts);

  char *ptx;
  size_t ptx_size;
  NVRTC_SUCCEED(nvrtcGetPTXSize(prog, &ptx_size));
  ptx = (char*) malloc(ptx_size);
  NVRTC_SUCCEED(nvrtcGetPTX(prog, ptx));

  NVRTC_SUCCEED(nvrtcDestroyProgram(&prog));

  return ptx;
}

static void cuda_size_setup(struct cuda_context *ctx)
{
  if (ctx->cfg.default_block_size > ctx->max_block_size) {
    if (ctx->cfg.default_block_size_changed) {
      fprintf(stderr,
          "Note: Device limits default block size to %zu (down from %zu).\n",
          ctx->max_block_size, ctx->cfg.default_block_size);
    }
    ctx->cfg.default_block_size = ctx->max_block_size;
  }
  if (ctx->cfg.default_grid_size > ctx->max_grid_size) {
    if (ctx->cfg.default_grid_size_changed) {
      fprintf(stderr,
          "Note: Device limits default grid size to %zu (down from %zu).\n",
          ctx->max_grid_size, ctx->cfg.default_grid_size);
    }
    ctx->cfg.default_grid_size = ctx->max_grid_size;
  }
  if (ctx->cfg.default_tile_size > ctx->max_tile_size) {
    if (ctx->cfg.default_tile_size_changed) {
      fprintf(stderr,
          "Note: Device limits default tile size to %zu (down from %zu).\n",
          ctx->max_tile_size, ctx->cfg.default_tile_size);
    }
    ctx->cfg.default_tile_size = ctx->max_tile_size;
  }

  for (int i = 0; i < ctx->cfg.num_sizes; i++) {
    const char *size_class, *size_name;
    size_t *size_value, max_value, default_value;

    size_class = ctx->cfg.size_classes[i];
    size_value = &ctx->cfg.size_values[i];
    size_name = ctx->cfg.size_names[i];

    if (strstr(size_class, "group_size") == size_class) {
      max_value = ctx->max_block_size;
      default_value = ctx->cfg.default_block_size;
    } else if (strstr(size_class, "num_groups") == size_class) {
      max_value = ctx->max_grid_size;
      default_value = ctx->cfg.default_grid_size;
    } else if (strstr(size_class, "tile_size") == size_class) {
      max_value = ctx->max_tile_size;
      default_value = ctx->cfg.default_tile_size;
    } else if (strstr(size_class, "threshold") == size_class) {
      max_value = ctx->max_threshold;
      default_value = ctx->cfg.default_threshold;
    } else {
      panic(1, "Unknown size class for size '%s': %s\n", size_name, size_class);
    }

    if (*size_value == 0) {
      *size_value = default_value;
    } else if (max_value > 0 && *size_value > max_value) {
      fprintf(stderr, "Note: Device limits %s to %zu (down from %zu)\n",
              size_name, max_value, *size_value);
      *size_value = max_value;
    }
  }
}

static void dump_string_to_file(const char *file, const char *buf) {
  FILE *f = fopen(file, "w");
  assert(f != NULL);
  assert(fputs(buf, f) != EOF);
  assert(fclose(f) == 0);
}

static void load_string_from_file(const char *file, char **obuf, size_t *olen) {
  char *buf;
  size_t len;
  FILE *f = fopen(file, "r");

  assert(f != NULL);
  assert(fseek(f, 0, SEEK_END) == 0);
  len = ftell(f);
  assert(fseek(f, 0, SEEK_SET) == 0);

  buf = (char*) malloc(len + 1);
  assert(fread(buf, 1, len, f) == len);
  buf[len] = 0;
  *obuf = buf;
  if (olen != NULL) {
    *olen = len;
  }

  assert(fclose(f) == 0);
}

static void cuda_module_setup(struct cuda_context *ctx,
                              const char *src_fragments[],
                              const char *extra_opts[]) {
  char *ptx = NULL, *src = NULL;

  if (ctx->cfg.load_ptx_from == NULL && ctx->cfg.load_program_from == NULL) {
    src = concat_fragments(src_fragments);
    ptx = cuda_nvrtc_build(ctx, src, extra_opts);
  } else if (ctx->cfg.load_ptx_from == NULL) {
    load_string_from_file(ctx->cfg.load_program_from, &src, NULL);
    ptx = cuda_nvrtc_build(ctx, src, extra_opts);
  } else {
    if (ctx->cfg.load_program_from != NULL) {
      fprintf(stderr,
              "WARNING: Loading PTX from %s instead of C code from %s\n",
              ctx->cfg.load_ptx_from, ctx->cfg.load_program_from);
    }

    load_string_from_file(ctx->cfg.load_ptx_from, &ptx, NULL);
  }

  if (ctx->cfg.dump_program_to != NULL) {
    if (src == NULL) {
      src = concat_fragments(src_fragments);
    }
    dump_string_to_file(ctx->cfg.dump_program_to, src);
  }
  if (ctx->cfg.dump_ptx_to != NULL) {
    dump_string_to_file(ctx->cfg.dump_ptx_to, ptx);
  }

  CUDA_SUCCEED(cuModuleLoadData(&ctx->module, ptx));

  free(ptx);
  if (src != NULL) {
    free(src);
  }
}

static void cuda_setup(struct cuda_context *ctx, const char *src_fragments[], const char *extra_opts[]) {
  CUDA_SUCCEED(cuInit(0));

  if (cuda_device_setup(ctx) != 0) {
    panic(-1, "No suitable CUDA device found.\n");
  }
  CUDA_SUCCEED(cuCtxCreate(&ctx->cu_ctx, 0, ctx->dev));

  free_list_init(&ctx->free_list);

  ctx->max_shared_memory = device_query(ctx->dev, MAX_SHARED_MEMORY_PER_BLOCK);
  ctx->max_block_size = device_query(ctx->dev, MAX_THREADS_PER_BLOCK);
  ctx->max_grid_size = device_query(ctx->dev, MAX_GRID_DIM_X);
  ctx->max_tile_size = sqrt(ctx->max_block_size);
  ctx->max_threshold = 0;
  ctx->lockstep_width = device_query(ctx->dev, WARP_SIZE);

  cuda_size_setup(ctx);
  cuda_module_setup(ctx, src_fragments, extra_opts);
}

static CUresult cuda_free_all(struct cuda_context *ctx);

static void cuda_cleanup(struct cuda_context *ctx) {
  CUDA_SUCCEED(cuda_free_all(ctx));
  CUDA_SUCCEED(cuModuleUnload(ctx->module));
  CUDA_SUCCEED(cuCtxDestroy(ctx->cu_ctx));
}

static CUresult cuda_alloc(struct cuda_context *ctx, size_t min_size,
                           const char *tag, CUdeviceptr *mem_out) {
  if (min_size < sizeof(int)) {
    min_size = sizeof(int);
  }

  size_t size;
  if (free_list_find(&ctx->free_list, tag, &size, mem_out) == 0) {
    if (size >= min_size) {
      return CUDA_SUCCESS;
    } else {
      CUresult res = cuMemFree(*mem_out);
      if (res != CUDA_SUCCESS) {
        return res;
      }
    }
  }

  CUresult res = cuMemAlloc(mem_out, min_size);
  while (res == CUDA_ERROR_OUT_OF_MEMORY) {
    CUdeviceptr mem;
    if (free_list_first(&ctx->free_list, &mem) == 0) {
      res = cuMemFree(mem);
      if (res != CUDA_SUCCESS) {
        return res;
      }
    } else {
      break;
    }
    res = cuMemAlloc(mem_out, min_size);
  }

  return res;
}

static CUresult cuda_free(struct cuda_context *ctx, CUdeviceptr mem,
                          const char *tag) {
  size_t size;
  CUdeviceptr existing_mem;

  // If there is already a block with this tag, then remove it.
  if (free_list_find(&ctx->free_list, tag, &size, &existing_mem) == 0) {
    CUresult res = cuMemFree(existing_mem);
    if (res != CUDA_SUCCESS) {
      return res;
    }
  }

  CUresult res = cuMemGetAddressRange(NULL, &size, mem);
  if (res == CUDA_SUCCESS) {
    free_list_insert(&ctx->free_list, size, mem, tag);
  }

  return res;
}

static CUresult cuda_free_all(struct cuda_context *ctx) {
  CUdeviceptr mem;
  free_list_pack(&ctx->free_list);
  while (free_list_first(&ctx->free_list, &mem) == 0) {
    CUresult res = cuMemFree(mem);
    if (res != CUDA_SUCCESS) {
      return res;
    }
  }

  return CUDA_SUCCESS;
}

// End of cuda.h.

const char *cuda_program[] =
           {"typedef char int8_t;\ntypedef short int16_t;\ntypedef int int32_t;\ntypedef long long int64_t;\ntypedef unsigned char uint8_t;\ntypedef unsigned short uint16_t;\ntypedef unsigned int uint32_t;\ntypedef unsigned long long uint64_t;\ntypedef uint8_t uchar;\ntypedef uint16_t ushort;\ntypedef uint32_t uint;\ntypedef uint64_t ulong;\n#define __kernel extern \"C\" __global__ __launch_bounds__(MAX_THREADS_PER_BLOCK)\n#define __global\n#define __local\n#define __private\n#define __constant\n#define __write_only\n#define __read_only\nstatic inline int get_group_id_fn(int block_dim0, int block_dim1,\n                                  int block_dim2, int d)\n{\n    switch (d) {\n        \n      case 0:\n        d = block_dim0;\n        break;\n        \n      case 1:\n        d = block_dim1;\n        break;\n        \n      case 2:\n        d = block_dim2;\n        break;\n    }\n    switch (d) {\n        \n      case 0:\n        return blockIdx.x;\n        \n      case 1:\n        return blockIdx.y;\n        \n      case 2:\n        return blockIdx.z;\n        \n      default:\n        return 0;\n    }\n}\n#define get_group_id(d) get_group_id_fn(block_dim0, block_dim1, block_dim2, d)\nstatic inline int get_num_groups_fn(int block_dim0, int block_dim1,\n                                    int block_dim2, int d)\n{\n    switch (d) {\n        \n      case 0:\n        d = block_dim0;\n        break;\n        \n      case 1:\n        d = block_dim1;\n        break;\n        \n      case 2:\n        d = block_dim2;\n        break;\n    }\n    switch (d) {\n        \n      case 0:\n        return gridDim.x;\n        \n      case 1:\n        return gridDim.y;\n        \n      case 2:\n        return gridDim.z;\n        \n      default:\n        return 0;\n    }\n}\n#define get_num_groups(d) get_num_groups_fn(block_dim0, block_dim1, block_dim2, d)\nstatic inline int get_local_id(int d)\n{\n    switch (d) {\n        \n      case 0:\n        return threadIdx.x;\n        \n      case 1:\n        return threadIdx.y;\n        \n      case 2:\n        return threadIdx.z;\n        \n      ",
            "default:\n        return 0;\n    }\n}\nstatic inline int get_local_size(int d)\n{\n    switch (d) {\n        \n      case 0:\n        return blockDim.x;\n        \n      case 1:\n        return blockDim.y;\n        \n      case 2:\n        return blockDim.z;\n        \n      default:\n        return 0;\n    }\n}\nstatic inline int get_global_id_fn(int block_dim0, int block_dim1,\n                                   int block_dim2, int d)\n{\n    return get_group_id(d) * get_local_size(d) + get_local_id(d);\n}\n#define get_global_id(d) get_global_id_fn(block_dim0, block_dim1, block_dim2, d)\nstatic inline int get_global_size(int block_dim0, int block_dim1,\n                                  int block_dim2, int d)\n{\n    return get_num_groups(d) * get_local_size(d);\n}\n#define CLK_LOCAL_MEM_FENCE 1\n#define CLK_GLOBAL_MEM_FENCE 2\nstatic inline void barrier(int x)\n{\n    __syncthreads();\n}\nstatic inline void mem_fence_local()\n{\n    __threadfence_block();\n}\nstatic inline void mem_fence_global()\n{\n    __threadfence();\n}\n#define NAN (0.0/0.0)\n#define INFINITY (1.0/0.0)\nextern volatile __shared__ char shared_mem[];\nstatic inline int atomic_add(volatile int *p, int val)\n{\n    return atomicAdd((int *) p, val);\n}\nstatic inline unsigned int atomic_add(volatile unsigned int *p,\n                                      unsigned int val)\n{\n    return atomicAdd((unsigned int *) p, val);\n}\nstatic inline unsigned long long atomic_add(volatile unsigned long long *p,\n                                            unsigned long long val)\n{\n    return atomicAdd((unsigned long long *) p, val);\n}\nstatic inline int atomic_max(volatile int *p, int val)\n{\n    return atomicMax((int *) p, val);\n}\nstatic inline unsigned int atomic_max(volatile unsigned int *p,\n                                      unsigned int val)\n{\n    return atomicMax((unsigned int *) p, val);\n}\nstatic inline unsigned long long atomic_max(volatile unsigned long long *p,\n                                            unsigned long long val)\n{\n    return atomicMax((un",
            "signed long long *) p, val);\n}\nstatic inline int atomic_min(volatile int *p, int val)\n{\n    return atomicMin((int *) p, val);\n}\nstatic inline unsigned int atomic_min(volatile unsigned int *p,\n                                      unsigned int val)\n{\n    return atomicMin((unsigned int *) p, val);\n}\nstatic inline unsigned long long atomic_min(volatile unsigned long long *p,\n                                            unsigned long long val)\n{\n    return atomicMin((unsigned long long *) p, val);\n}\nstatic inline int atomic_and(volatile int *p, int val)\n{\n    return atomicAnd((int *) p, val);\n}\nstatic inline unsigned int atomic_and(volatile unsigned int *p,\n                                      unsigned int val)\n{\n    return atomicAnd((unsigned int *) p, val);\n}\nstatic inline unsigned long long atomic_and(volatile unsigned long long *p,\n                                            unsigned long long val)\n{\n    return atomicAnd((unsigned long long *) p, val);\n}\nstatic inline int atomic_or(volatile int *p, int val)\n{\n    return atomicOr((int *) p, val);\n}\nstatic inline unsigned int atomic_or(volatile unsigned int *p, unsigned int val)\n{\n    return atomicOr((unsigned int *) p, val);\n}\nstatic inline unsigned long long atomic_or(volatile unsigned long long *p,\n                                           unsigned long long val)\n{\n    return atomicOr((unsigned long long *) p, val);\n}\nstatic inline int atomic_xor(volatile int *p, int val)\n{\n    return atomicXor((int *) p, val);\n}\nstatic inline unsigned int atomic_xor(volatile unsigned int *p,\n                                      unsigned int val)\n{\n    return atomicXor((unsigned int *) p, val);\n}\nstatic inline unsigned long long atomic_xor(volatile unsigned long long *p,\n                                            unsigned long long val)\n{\n    return atomicXor((unsigned long long *) p, val);\n}\nstatic inline int atomic_xchg(volatile int *p, int val)\n{\n    return atomicExch((int *) p, val);\n}\nstatic inline unsigned int atomic_xchg(",
            "volatile unsigned int *p,\n                                       unsigned int val)\n{\n    return atomicExch((unsigned int *) p, val);\n}\nstatic inline unsigned long long atomic_xchg(volatile unsigned long long *p,\n                                             unsigned long long val)\n{\n    return atomicExch((unsigned long long *) p, val);\n}\nstatic inline int atomic_cmpxchg(volatile int *p, int cmp, int val)\n{\n    return atomicCAS((int *) p, cmp, val);\n}\nstatic inline unsigned int atomic_cmpxchg(volatile unsigned int *p,\n                                          unsigned int cmp, unsigned int val)\n{\n    return atomicCAS((unsigned int *) p, cmp, val);\n}\nstatic inline unsigned long long atomic_cmpxchg(volatile unsigned long long *p,\n                                                unsigned long long cmp,\n                                                unsigned long long val)\n{\n    return atomicCAS((unsigned long long *) p, cmp, val);\n}\nstatic inline int8_t add8(int8_t x, int8_t y)\n{\n    return x + y;\n}\nstatic inline int16_t add16(int16_t x, int16_t y)\n{\n    return x + y;\n}\nstatic inline int32_t add32(int32_t x, int32_t y)\n{\n    return x + y;\n}\nstatic inline int64_t add64(int64_t x, int64_t y)\n{\n    return x + y;\n}\nstatic inline int8_t sub8(int8_t x, int8_t y)\n{\n    return x - y;\n}\nstatic inline int16_t sub16(int16_t x, int16_t y)\n{\n    return x - y;\n}\nstatic inline int32_t sub32(int32_t x, int32_t y)\n{\n    return x - y;\n}\nstatic inline int64_t sub64(int64_t x, int64_t y)\n{\n    return x - y;\n}\nstatic inline int8_t mul8(int8_t x, int8_t y)\n{\n    return x * y;\n}\nstatic inline int16_t mul16(int16_t x, int16_t y)\n{\n    return x * y;\n}\nstatic inline int32_t mul32(int32_t x, int32_t y)\n{\n    return x * y;\n}\nstatic inline int64_t mul64(int64_t x, int64_t y)\n{\n    return x * y;\n}\nstatic inline uint8_t udiv8(uint8_t x, uint8_t y)\n{\n    return x / y;\n}\nstatic inline uint16_t udiv16(uint16_t x, uint16_t y)\n{\n    return x / y;\n}\nstatic inline uint32_t udiv32(uint32_t x, uint32_t y)\n{\n  ",
            "  return x / y;\n}\nstatic inline uint64_t udiv64(uint64_t x, uint64_t y)\n{\n    return x / y;\n}\nstatic inline uint8_t umod8(uint8_t x, uint8_t y)\n{\n    return x % y;\n}\nstatic inline uint16_t umod16(uint16_t x, uint16_t y)\n{\n    return x % y;\n}\nstatic inline uint32_t umod32(uint32_t x, uint32_t y)\n{\n    return x % y;\n}\nstatic inline uint64_t umod64(uint64_t x, uint64_t y)\n{\n    return x % y;\n}\nstatic inline int8_t sdiv8(int8_t x, int8_t y)\n{\n    int8_t q = x / y;\n    int8_t r = x % y;\n    \n    return q - ((r != 0 && r < 0 != y < 0) ? 1 : 0);\n}\nstatic inline int16_t sdiv16(int16_t x, int16_t y)\n{\n    int16_t q = x / y;\n    int16_t r = x % y;\n    \n    return q - ((r != 0 && r < 0 != y < 0) ? 1 : 0);\n}\nstatic inline int32_t sdiv32(int32_t x, int32_t y)\n{\n    int32_t q = x / y;\n    int32_t r = x % y;\n    \n    return q - ((r != 0 && r < 0 != y < 0) ? 1 : 0);\n}\nstatic inline int64_t sdiv64(int64_t x, int64_t y)\n{\n    int64_t q = x / y;\n    int64_t r = x % y;\n    \n    return q - ((r != 0 && r < 0 != y < 0) ? 1 : 0);\n}\nstatic inline int8_t smod8(int8_t x, int8_t y)\n{\n    int8_t r = x % y;\n    \n    return r + (r == 0 || (x > 0 && y > 0) || (x < 0 && y < 0) ? 0 : y);\n}\nstatic inline int16_t smod16(int16_t x, int16_t y)\n{\n    int16_t r = x % y;\n    \n    return r + (r == 0 || (x > 0 && y > 0) || (x < 0 && y < 0) ? 0 : y);\n}\nstatic inline int32_t smod32(int32_t x, int32_t y)\n{\n    int32_t r = x % y;\n    \n    return r + (r == 0 || (x > 0 && y > 0) || (x < 0 && y < 0) ? 0 : y);\n}\nstatic inline int64_t smod64(int64_t x, int64_t y)\n{\n    int64_t r = x % y;\n    \n    return r + (r == 0 || (x > 0 && y > 0) || (x < 0 && y < 0) ? 0 : y);\n}\nstatic inline int8_t squot8(int8_t x, int8_t y)\n{\n    return x / y;\n}\nstatic inline int16_t squot16(int16_t x, int16_t y)\n{\n    return x / y;\n}\nstatic inline int32_t squot32(int32_t x, int32_t y)\n{\n    return x / y;\n}\nstatic inline int64_t squot64(int64_t x, int64_t y)\n{\n    return x / y;\n}\nstatic inline int8_t srem8(int8_t x, int8_t y)\n{\n    return x % y",
            ";\n}\nstatic inline int16_t srem16(int16_t x, int16_t y)\n{\n    return x % y;\n}\nstatic inline int32_t srem32(int32_t x, int32_t y)\n{\n    return x % y;\n}\nstatic inline int64_t srem64(int64_t x, int64_t y)\n{\n    return x % y;\n}\nstatic inline int8_t smin8(int8_t x, int8_t y)\n{\n    return x < y ? x : y;\n}\nstatic inline int16_t smin16(int16_t x, int16_t y)\n{\n    return x < y ? x : y;\n}\nstatic inline int32_t smin32(int32_t x, int32_t y)\n{\n    return x < y ? x : y;\n}\nstatic inline int64_t smin64(int64_t x, int64_t y)\n{\n    return x < y ? x : y;\n}\nstatic inline uint8_t umin8(uint8_t x, uint8_t y)\n{\n    return x < y ? x : y;\n}\nstatic inline uint16_t umin16(uint16_t x, uint16_t y)\n{\n    return x < y ? x : y;\n}\nstatic inline uint32_t umin32(uint32_t x, uint32_t y)\n{\n    return x < y ? x : y;\n}\nstatic inline uint64_t umin64(uint64_t x, uint64_t y)\n{\n    return x < y ? x : y;\n}\nstatic inline int8_t smax8(int8_t x, int8_t y)\n{\n    return x < y ? y : x;\n}\nstatic inline int16_t smax16(int16_t x, int16_t y)\n{\n    return x < y ? y : x;\n}\nstatic inline int32_t smax32(int32_t x, int32_t y)\n{\n    return x < y ? y : x;\n}\nstatic inline int64_t smax64(int64_t x, int64_t y)\n{\n    return x < y ? y : x;\n}\nstatic inline uint8_t umax8(uint8_t x, uint8_t y)\n{\n    return x < y ? y : x;\n}\nstatic inline uint16_t umax16(uint16_t x, uint16_t y)\n{\n    return x < y ? y : x;\n}\nstatic inline uint32_t umax32(uint32_t x, uint32_t y)\n{\n    return x < y ? y : x;\n}\nstatic inline uint64_t umax64(uint64_t x, uint64_t y)\n{\n    return x < y ? y : x;\n}\nstatic inline uint8_t shl8(uint8_t x, uint8_t y)\n{\n    return x << y;\n}\nstatic inline uint16_t shl16(uint16_t x, uint16_t y)\n{\n    return x << y;\n}\nstatic inline uint32_t shl32(uint32_t x, uint32_t y)\n{\n    return x << y;\n}\nstatic inline uint64_t shl64(uint64_t x, uint64_t y)\n{\n    return x << y;\n}\nstatic inline uint8_t lshr8(uint8_t x, uint8_t y)\n{\n    return x >> y;\n}\nstatic inline uint16_t lshr16(uint16_t x, uint16_t y)\n{\n    return x >> y;\n}\nstatic inline uint32_t ",
            "lshr32(uint32_t x, uint32_t y)\n{\n    return x >> y;\n}\nstatic inline uint64_t lshr64(uint64_t x, uint64_t y)\n{\n    return x >> y;\n}\nstatic inline int8_t ashr8(int8_t x, int8_t y)\n{\n    return x >> y;\n}\nstatic inline int16_t ashr16(int16_t x, int16_t y)\n{\n    return x >> y;\n}\nstatic inline int32_t ashr32(int32_t x, int32_t y)\n{\n    return x >> y;\n}\nstatic inline int64_t ashr64(int64_t x, int64_t y)\n{\n    return x >> y;\n}\nstatic inline uint8_t and8(uint8_t x, uint8_t y)\n{\n    return x & y;\n}\nstatic inline uint16_t and16(uint16_t x, uint16_t y)\n{\n    return x & y;\n}\nstatic inline uint32_t and32(uint32_t x, uint32_t y)\n{\n    return x & y;\n}\nstatic inline uint64_t and64(uint64_t x, uint64_t y)\n{\n    return x & y;\n}\nstatic inline uint8_t or8(uint8_t x, uint8_t y)\n{\n    return x | y;\n}\nstatic inline uint16_t or16(uint16_t x, uint16_t y)\n{\n    return x | y;\n}\nstatic inline uint32_t or32(uint32_t x, uint32_t y)\n{\n    return x | y;\n}\nstatic inline uint64_t or64(uint64_t x, uint64_t y)\n{\n    return x | y;\n}\nstatic inline uint8_t xor8(uint8_t x, uint8_t y)\n{\n    return x ^ y;\n}\nstatic inline uint16_t xor16(uint16_t x, uint16_t y)\n{\n    return x ^ y;\n}\nstatic inline uint32_t xor32(uint32_t x, uint32_t y)\n{\n    return x ^ y;\n}\nstatic inline uint64_t xor64(uint64_t x, uint64_t y)\n{\n    return x ^ y;\n}\nstatic inline char ult8(uint8_t x, uint8_t y)\n{\n    return x < y;\n}\nstatic inline char ult16(uint16_t x, uint16_t y)\n{\n    return x < y;\n}\nstatic inline char ult32(uint32_t x, uint32_t y)\n{\n    return x < y;\n}\nstatic inline char ult64(uint64_t x, uint64_t y)\n{\n    return x < y;\n}\nstatic inline char ule8(uint8_t x, uint8_t y)\n{\n    return x <= y;\n}\nstatic inline char ule16(uint16_t x, uint16_t y)\n{\n    return x <= y;\n}\nstatic inline char ule32(uint32_t x, uint32_t y)\n{\n    return x <= y;\n}\nstatic inline char ule64(uint64_t x, uint64_t y)\n{\n    return x <= y;\n}\nstatic inline char slt8(int8_t x, int8_t y)\n{\n    return x < y;\n}\nstatic inline char slt16(int16_t x, int16_t y)\n{\n    return x",
            " < y;\n}\nstatic inline char slt32(int32_t x, int32_t y)\n{\n    return x < y;\n}\nstatic inline char slt64(int64_t x, int64_t y)\n{\n    return x < y;\n}\nstatic inline char sle8(int8_t x, int8_t y)\n{\n    return x <= y;\n}\nstatic inline char sle16(int16_t x, int16_t y)\n{\n    return x <= y;\n}\nstatic inline char sle32(int32_t x, int32_t y)\n{\n    return x <= y;\n}\nstatic inline char sle64(int64_t x, int64_t y)\n{\n    return x <= y;\n}\nstatic inline int8_t pow8(int8_t x, int8_t y)\n{\n    int8_t res = 1, rem = y;\n    \n    while (rem != 0) {\n        if (rem & 1)\n            res *= x;\n        rem >>= 1;\n        x *= x;\n    }\n    return res;\n}\nstatic inline int16_t pow16(int16_t x, int16_t y)\n{\n    int16_t res = 1, rem = y;\n    \n    while (rem != 0) {\n        if (rem & 1)\n            res *= x;\n        rem >>= 1;\n        x *= x;\n    }\n    return res;\n}\nstatic inline int32_t pow32(int32_t x, int32_t y)\n{\n    int32_t res = 1, rem = y;\n    \n    while (rem != 0) {\n        if (rem & 1)\n            res *= x;\n        rem >>= 1;\n        x *= x;\n    }\n    return res;\n}\nstatic inline int64_t pow64(int64_t x, int64_t y)\n{\n    int64_t res = 1, rem = y;\n    \n    while (rem != 0) {\n        if (rem & 1)\n            res *= x;\n        rem >>= 1;\n        x *= x;\n    }\n    return res;\n}\nstatic inline bool itob_i8_bool(int8_t x)\n{\n    return x;\n}\nstatic inline bool itob_i16_bool(int16_t x)\n{\n    return x;\n}\nstatic inline bool itob_i32_bool(int32_t x)\n{\n    return x;\n}\nstatic inline bool itob_i64_bool(int64_t x)\n{\n    return x;\n}\nstatic inline int8_t btoi_bool_i8(bool x)\n{\n    return x;\n}\nstatic inline int16_t btoi_bool_i16(bool x)\n{\n    return x;\n}\nstatic inline int32_t btoi_bool_i32(bool x)\n{\n    return x;\n}\nstatic inline int64_t btoi_bool_i64(bool x)\n{\n    return x;\n}\n#define sext_i8_i8(x) ((int8_t) (int8_t) x)\n#define sext_i8_i16(x) ((int16_t) (int8_t) x)\n#define sext_i8_i32(x) ((int32_t) (int8_t) x)\n#define sext_i8_i64(x) ((int64_t) (int8_t) x)\n#define sext_i16_i8(x) ((int8_t) (int16_t) x)\n#define sext_i",
            "16_i16(x) ((int16_t) (int16_t) x)\n#define sext_i16_i32(x) ((int32_t) (int16_t) x)\n#define sext_i16_i64(x) ((int64_t) (int16_t) x)\n#define sext_i32_i8(x) ((int8_t) (int32_t) x)\n#define sext_i32_i16(x) ((int16_t) (int32_t) x)\n#define sext_i32_i32(x) ((int32_t) (int32_t) x)\n#define sext_i32_i64(x) ((int64_t) (int32_t) x)\n#define sext_i64_i8(x) ((int8_t) (int64_t) x)\n#define sext_i64_i16(x) ((int16_t) (int64_t) x)\n#define sext_i64_i32(x) ((int32_t) (int64_t) x)\n#define sext_i64_i64(x) ((int64_t) (int64_t) x)\n#define zext_i8_i8(x) ((uint8_t) (uint8_t) x)\n#define zext_i8_i16(x) ((uint16_t) (uint8_t) x)\n#define zext_i8_i32(x) ((uint32_t) (uint8_t) x)\n#define zext_i8_i64(x) ((uint64_t) (uint8_t) x)\n#define zext_i16_i8(x) ((uint8_t) (uint16_t) x)\n#define zext_i16_i16(x) ((uint16_t) (uint16_t) x)\n#define zext_i16_i32(x) ((uint32_t) (uint16_t) x)\n#define zext_i16_i64(x) ((uint64_t) (uint16_t) x)\n#define zext_i32_i8(x) ((uint8_t) (uint32_t) x)\n#define zext_i32_i16(x) ((uint16_t) (uint32_t) x)\n#define zext_i32_i32(x) ((uint32_t) (uint32_t) x)\n#define zext_i32_i64(x) ((uint64_t) (uint32_t) x)\n#define zext_i64_i8(x) ((uint8_t) (uint64_t) x)\n#define zext_i64_i16(x) ((uint16_t) (uint64_t) x)\n#define zext_i64_i32(x) ((uint32_t) (uint64_t) x)\n#define zext_i64_i64(x) ((uint64_t) (uint64_t) x)\n#ifdef __OPENCL_VERSION__\nint32_t futrts_popc8(int8_t x)\n{\n    return popcount(x);\n}\nint32_t futrts_popc16(int16_t x)\n{\n    return popcount(x);\n}\nint32_t futrts_popc32(int32_t x)\n{\n    return popcount(x);\n}\nint32_t futrts_popc64(int64_t x)\n{\n    return popcount(x);\n}\n#elif __CUDA_ARCH__\nint32_t futrts_popc8(int8_t x)\n{\n    return __popc(zext_i8_i32(x));\n}\nint32_t futrts_popc16(int16_t x)\n{\n    return __popc(zext_i16_i32(x));\n}\nint32_t futrts_popc32(int32_t x)\n{\n    return __popc(x);\n}\nint32_t futrts_popc64(int64_t x)\n{\n    return __popcll(x);\n}\n#else\nint32_t futrts_popc8(int8_t x)\n{\n    int c = 0;\n    \n    for (; x; ++c)\n        x &= x - 1;\n    return c;\n}\nint32_t futrts_popc16(int16_t x)\n{\n    in",
            "t c = 0;\n    \n    for (; x; ++c)\n        x &= x - 1;\n    return c;\n}\nint32_t futrts_popc32(int32_t x)\n{\n    int c = 0;\n    \n    for (; x; ++c)\n        x &= x - 1;\n    return c;\n}\nint32_t futrts_popc64(int64_t x)\n{\n    int c = 0;\n    \n    for (; x; ++c)\n        x &= x - 1;\n    return c;\n}\n#endif\n#ifdef __OPENCL_VERSION__\nint32_t futrts_clzz8(int8_t x)\n{\n    return clz(x);\n}\nint32_t futrts_clzz16(int16_t x)\n{\n    return clz(x);\n}\nint32_t futrts_clzz32(int32_t x)\n{\n    return clz(x);\n}\nint32_t futrts_clzz64(int64_t x)\n{\n    return clz(x);\n}\n#elif __CUDA_ARCH__\nint32_t futrts_clzz8(int8_t x)\n{\n    return __clz(zext_i8_i32(x)) - 24;\n}\nint32_t futrts_clzz16(int16_t x)\n{\n    return __clz(zext_i16_i32(x)) - 16;\n}\nint32_t futrts_clzz32(int32_t x)\n{\n    return __clz(x);\n}\nint32_t futrts_clzz64(int64_t x)\n{\n    return __clzll(x);\n}\n#else\nint32_t futrts_clzz8(int8_t x)\n{\n    int n = 0;\n    int bits = sizeof(x) * 8;\n    \n    for (int i = 0; i < bits; i++) {\n        if (x < 0)\n            break;\n        n++;\n        x <<= 1;\n    }\n    return n;\n}\nint32_t futrts_clzz16(int16_t x)\n{\n    int n = 0;\n    int bits = sizeof(x) * 8;\n    \n    for (int i = 0; i < bits; i++) {\n        if (x < 0)\n            break;\n        n++;\n        x <<= 1;\n    }\n    return n;\n}\nint32_t futrts_clzz32(int32_t x)\n{\n    int n = 0;\n    int bits = sizeof(x) * 8;\n    \n    for (int i = 0; i < bits; i++) {\n        if (x < 0)\n            break;\n        n++;\n        x <<= 1;\n    }\n    return n;\n}\nint32_t futrts_clzz64(int64_t x)\n{\n    int n = 0;\n    int bits = sizeof(x) * 8;\n    \n    for (int i = 0; i < bits; i++) {\n        if (x < 0)\n            break;\n        n++;\n        x <<= 1;\n    }\n    return n;\n}\n#endif\nstatic inline float fdiv32(float x, float y)\n{\n    return x / y;\n}\nstatic inline float fadd32(float x, float y)\n{\n    return x + y;\n}\nstatic inline float fsub32(float x, float y)\n{\n    return x - y;\n}\nstatic inline float fmul32(float x, float y)\n{\n    return x * y;\n}\nstatic inline float fmin32(float x, floa",
            "t y)\n{\n    return fmin(x, y);\n}\nstatic inline float fmax32(float x, float y)\n{\n    return fmax(x, y);\n}\nstatic inline float fpow32(float x, float y)\n{\n    return pow(x, y);\n}\nstatic inline char cmplt32(float x, float y)\n{\n    return x < y;\n}\nstatic inline char cmple32(float x, float y)\n{\n    return x <= y;\n}\nstatic inline float sitofp_i8_f32(int8_t x)\n{\n    return x;\n}\nstatic inline float sitofp_i16_f32(int16_t x)\n{\n    return x;\n}\nstatic inline float sitofp_i32_f32(int32_t x)\n{\n    return x;\n}\nstatic inline float sitofp_i64_f32(int64_t x)\n{\n    return x;\n}\nstatic inline float uitofp_i8_f32(uint8_t x)\n{\n    return x;\n}\nstatic inline float uitofp_i16_f32(uint16_t x)\n{\n    return x;\n}\nstatic inline float uitofp_i32_f32(uint32_t x)\n{\n    return x;\n}\nstatic inline float uitofp_i64_f32(uint64_t x)\n{\n    return x;\n}\nstatic inline int8_t fptosi_f32_i8(float x)\n{\n    return x;\n}\nstatic inline int16_t fptosi_f32_i16(float x)\n{\n    return x;\n}\nstatic inline int32_t fptosi_f32_i32(float x)\n{\n    return x;\n}\nstatic inline int64_t fptosi_f32_i64(float x)\n{\n    return x;\n}\nstatic inline uint8_t fptoui_f32_i8(float x)\n{\n    return x;\n}\nstatic inline uint16_t fptoui_f32_i16(float x)\n{\n    return x;\n}\nstatic inline uint32_t fptoui_f32_i32(float x)\n{\n    return x;\n}\nstatic inline uint64_t fptoui_f32_i64(float x)\n{\n    return x;\n}\nstatic inline float futrts_log32(float x)\n{\n    return log(x);\n}\nstatic inline float futrts_log2_32(float x)\n{\n    return log2(x);\n}\nstatic inline float futrts_log10_32(float x)\n{\n    return log10(x);\n}\nstatic inline float futrts_sqrt32(float x)\n{\n    return sqrt(x);\n}\nstatic inline float futrts_exp32(float x)\n{\n    return exp(x);\n}\nstatic inline float futrts_cos32(float x)\n{\n    return cos(x);\n}\nstatic inline float futrts_sin32(float x)\n{\n    return sin(x);\n}\nstatic inline float futrts_tan32(float x)\n{\n    return tan(x);\n}\nstatic inline float futrts_acos32(float x)\n{\n    return acos(x);\n}\nstatic inline float futrts_asin32(float x)\n{\n    return asin(x);\n}\nst",
            "atic inline float futrts_atan32(float x)\n{\n    return atan(x);\n}\nstatic inline float futrts_atan2_32(float x, float y)\n{\n    return atan2(x, y);\n}\nstatic inline float futrts_gamma32(float x)\n{\n    return tgamma(x);\n}\nstatic inline float futrts_lgamma32(float x)\n{\n    return lgamma(x);\n}\nstatic inline char futrts_isnan32(float x)\n{\n    return isnan(x);\n}\nstatic inline char futrts_isinf32(float x)\n{\n    return isinf(x);\n}\nstatic inline int32_t futrts_to_bits32(float x)\n{\n    union {\n        float f;\n        int32_t t;\n    } p;\n    \n    p.f = x;\n    return p.t;\n}\nstatic inline float futrts_from_bits32(int32_t x)\n{\n    union {\n        int32_t f;\n        float t;\n    } p;\n    \n    p.f = x;\n    return p.t;\n}\n#ifdef __OPENCL_VERSION__\nstatic inline float fmod32(float x, float y)\n{\n    return fmod(x, y);\n}\nstatic inline float futrts_round32(float x)\n{\n    return rint(x);\n}\nstatic inline float futrts_floor32(float x)\n{\n    return floor(x);\n}\nstatic inline float futrts_ceil32(float x)\n{\n    return ceil(x);\n}\nstatic inline float futrts_lerp32(float v0, float v1, float t)\n{\n    return mix(v0, v1, t);\n}\n#else\nstatic inline float fmod32(float x, float y)\n{\n    return fmodf(x, y);\n}\nstatic inline float futrts_round32(float x)\n{\n    return rintf(x);\n}\nstatic inline float futrts_floor32(float x)\n{\n    return floorf(x);\n}\nstatic inline float futrts_ceil32(float x)\n{\n    return ceilf(x);\n}\nstatic inline float futrts_lerp32(float v0, float v1, float t)\n{\n    return v0 + (v1 - v0) * t;\n}\n#endif\nstatic inline double fdiv64(double x, double y)\n{\n    return x / y;\n}\nstatic inline double fadd64(double x, double y)\n{\n    return x + y;\n}\nstatic inline double fsub64(double x, double y)\n{\n    return x - y;\n}\nstatic inline double fmul64(double x, double y)\n{\n    return x * y;\n}\nstatic inline double fmin64(double x, double y)\n{\n    return fmin(x, y);\n}\nstatic inline double fmax64(double x, double y)\n{\n    return fmax(x, y);\n}\nstatic inline double fpow64(double x, double y)\n{\n    return pow(x, y);",
            "\n}\nstatic inline char cmplt64(double x, double y)\n{\n    return x < y;\n}\nstatic inline char cmple64(double x, double y)\n{\n    return x <= y;\n}\nstatic inline double sitofp_i8_f64(int8_t x)\n{\n    return x;\n}\nstatic inline double sitofp_i16_f64(int16_t x)\n{\n    return x;\n}\nstatic inline double sitofp_i32_f64(int32_t x)\n{\n    return x;\n}\nstatic inline double sitofp_i64_f64(int64_t x)\n{\n    return x;\n}\nstatic inline double uitofp_i8_f64(uint8_t x)\n{\n    return x;\n}\nstatic inline double uitofp_i16_f64(uint16_t x)\n{\n    return x;\n}\nstatic inline double uitofp_i32_f64(uint32_t x)\n{\n    return x;\n}\nstatic inline double uitofp_i64_f64(uint64_t x)\n{\n    return x;\n}\nstatic inline int8_t fptosi_f64_i8(double x)\n{\n    return x;\n}\nstatic inline int16_t fptosi_f64_i16(double x)\n{\n    return x;\n}\nstatic inline int32_t fptosi_f64_i32(double x)\n{\n    return x;\n}\nstatic inline int64_t fptosi_f64_i64(double x)\n{\n    return x;\n}\nstatic inline uint8_t fptoui_f64_i8(double x)\n{\n    return x;\n}\nstatic inline uint16_t fptoui_f64_i16(double x)\n{\n    return x;\n}\nstatic inline uint32_t fptoui_f64_i32(double x)\n{\n    return x;\n}\nstatic inline uint64_t fptoui_f64_i64(double x)\n{\n    return x;\n}\nstatic inline double futrts_log64(double x)\n{\n    return log(x);\n}\nstatic inline double futrts_log2_64(double x)\n{\n    return log2(x);\n}\nstatic inline double futrts_log10_64(double x)\n{\n    return log10(x);\n}\nstatic inline double futrts_sqrt64(double x)\n{\n    return sqrt(x);\n}\nstatic inline double futrts_exp64(double x)\n{\n    return exp(x);\n}\nstatic inline double futrts_cos64(double x)\n{\n    return cos(x);\n}\nstatic inline double futrts_sin64(double x)\n{\n    return sin(x);\n}\nstatic inline double futrts_tan64(double x)\n{\n    return tan(x);\n}\nstatic inline double futrts_acos64(double x)\n{\n    return acos(x);\n}\nstatic inline double futrts_asin64(double x)\n{\n    return asin(x);\n}\nstatic inline double futrts_atan64(double x)\n{\n    return atan(x);\n}\nstatic inline double futrts_atan2_64(double x, double y)\n{\n    re",
            "turn atan2(x, y);\n}\nstatic inline double futrts_gamma64(double x)\n{\n    return tgamma(x);\n}\nstatic inline double futrts_lgamma64(double x)\n{\n    return lgamma(x);\n}\nstatic inline double futrts_round64(double x)\n{\n    return rint(x);\n}\nstatic inline double futrts_ceil64(double x)\n{\n    return ceil(x);\n}\nstatic inline double futrts_floor64(double x)\n{\n    return floor(x);\n}\nstatic inline char futrts_isnan64(double x)\n{\n    return isnan(x);\n}\nstatic inline char futrts_isinf64(double x)\n{\n    return isinf(x);\n}\nstatic inline int64_t futrts_to_bits64(double x)\n{\n    union {\n        double f;\n        int64_t t;\n    } p;\n    \n    p.f = x;\n    return p.t;\n}\nstatic inline double futrts_from_bits64(int64_t x)\n{\n    union {\n        int64_t f;\n        double t;\n    } p;\n    \n    p.f = x;\n    return p.t;\n}\nstatic inline float fmod64(float x, float y)\n{\n    return fmod(x, y);\n}\n#ifdef __OPENCL_VERSION__\nstatic inline double futrts_lerp64(double v0, double v1, double t)\n{\n    return mix(v0, v1, t);\n}\n#else\nstatic inline double futrts_lerp64(double v0, double v1, double t)\n{\n    return v0 + (v1 - v0) * t;\n}\n#endif\nstatic inline float fpconv_f32_f32(float x)\n{\n    return x;\n}\nstatic inline double fpconv_f32_f64(float x)\n{\n    return x;\n}\nstatic inline float fpconv_f64_f32(double x)\n{\n    return x;\n}\nstatic inline double fpconv_f64_f64(double x)\n{\n    return x;\n}\n__kernel void map_transpose_bool(const int block_dim0, const int block_dim1,\n                                 const int block_dim2,\n                                 uint block_11_backing_offset_0,\n                                 int32_t destoffset_1, int32_t srcoffset_3,\n                                 int32_t num_arrays_4, int32_t x_elems_5,\n                                 int32_t y_elems_6, int32_t in_elems_7,\n                                 int32_t out_elems_8, int32_t mulx_9,\n                                 int32_t muly_10, __global\n                                 unsigned char *destmem_0, __global\n                ",
            "                 unsigned char *srcmem_2)\n{\n    volatile char *block_11_backing_0 = &shared_mem[block_11_backing_offset_0];\n    __local char *block_11;\n    \n    block_11 = (__local char *) block_11_backing_0;\n    \n    int32_t get_global_id_0_37;\n    \n    get_global_id_0_37 = get_global_id(0);\n    \n    int32_t get_local_id_0_38;\n    \n    get_local_id_0_38 = get_local_id(0);\n    \n    int32_t get_local_id_1_39;\n    \n    get_local_id_1_39 = get_local_id(1);\n    \n    int32_t get_group_id_0_40;\n    \n    get_group_id_0_40 = get_group_id(0);\n    \n    int32_t get_group_id_1_41;\n    \n    get_group_id_1_41 = get_group_id(1);\n    \n    int32_t get_group_id_2_42;\n    \n    get_group_id_2_42 = get_group_id(2);\n    \n    int32_t our_array_offset_30 = get_group_id_2_42 * x_elems_5 * y_elems_6;\n    int32_t odata_offset_33 = destoffset_1 + our_array_offset_30;\n    int32_t idata_offset_34 = srcoffset_3 + our_array_offset_30;\n    int32_t x_index_31 = get_global_id_0_37;\n    int32_t y_index_32 = get_group_id_1_41 * 32 + get_local_id_1_39;\n    \n    if (slt32(x_index_31, x_elems_5)) {\n        for (int32_t j_43 = 0; j_43 < 4; j_43++) {\n            int32_t index_in_35 = (y_index_32 + j_43 * 8) * x_elems_5 +\n                    x_index_31;\n            \n            if (slt32(y_index_32 + j_43 * 8, y_elems_6) && slt32(index_in_35,\n                                                                 in_elems_7)) {\n                ((__local bool *) block_11)[(get_local_id_1_39 + j_43 * 8) *\n                                            33 + get_local_id_0_38] = ((__global\n                                                                        bool *) srcmem_2)[idata_offset_34 +\n                                                                                          index_in_35];\n            }\n        }\n    }\n    barrier(CLK_LOCAL_MEM_FENCE);\n    x_index_31 = get_group_id_1_41 * 32 + get_local_id_0_38;\n    y_index_32 = get_group_id_0_40 * 32 + get_local_id_1_39;\n    if (slt32(x_index_31, y_elems_6)) {\n  ",
            "      for (int32_t j_43 = 0; j_43 < 4; j_43++) {\n            int32_t index_out_36 = (y_index_32 + j_43 * 8) * y_elems_6 +\n                    x_index_31;\n            \n            if (slt32(y_index_32 + j_43 * 8, x_elems_5) && slt32(index_out_36,\n                                                                 out_elems_8)) {\n                ((__global bool *) destmem_0)[odata_offset_33 + index_out_36] =\n                    ((__local bool *) block_11)[get_local_id_0_38 * 33 +\n                                                get_local_id_1_39 + j_43 * 8];\n            }\n        }\n    }\n}\n__kernel void map_transpose_bool_low_height(const int block_dim0, const\n                                            int block_dim1, const\n                                            int block_dim2,\n                                            uint block_11_backing_offset_0,\n                                            int32_t destoffset_1,\n                                            int32_t srcoffset_3,\n                                            int32_t num_arrays_4,\n                                            int32_t x_elems_5,\n                                            int32_t y_elems_6,\n                                            int32_t in_elems_7,\n                                            int32_t out_elems_8, int32_t mulx_9,\n                                            int32_t muly_10, __global\n                                            unsigned char *destmem_0, __global\n                                            unsigned char *srcmem_2)\n{\n    volatile char *block_11_backing_0 = &shared_mem[block_11_backing_offset_0];\n    __local char *block_11;\n    \n    block_11 = (__local char *) block_11_backing_0;\n    \n    int32_t get_global_id_0_37;\n    \n    get_global_id_0_37 = get_global_id(0);\n    \n    int32_t get_local_id_0_38;\n    \n    get_local_id_0_38 = get_local_id(0);\n    \n    int32_t get_local_id_1_39;\n    \n    get_local_id_1_39 = get_local_id(1);\n    \n    int32_t get_group_id_0_4",
            "0;\n    \n    get_group_id_0_40 = get_group_id(0);\n    \n    int32_t get_group_id_1_41;\n    \n    get_group_id_1_41 = get_group_id(1);\n    \n    int32_t get_group_id_2_42;\n    \n    get_group_id_2_42 = get_group_id(2);\n    \n    int32_t our_array_offset_30 = get_group_id_2_42 * x_elems_5 * y_elems_6;\n    int32_t odata_offset_33 = destoffset_1 + our_array_offset_30;\n    int32_t idata_offset_34 = srcoffset_3 + our_array_offset_30;\n    int32_t x_index_31 = get_group_id_0_40 * 16 * mulx_9 + get_local_id_0_38 +\n            srem32(get_local_id_1_39, mulx_9) * 16;\n    int32_t y_index_32 = get_group_id_1_41 * 16 + squot32(get_local_id_1_39,\n                                                          mulx_9);\n    int32_t index_in_35 = y_index_32 * x_elems_5 + x_index_31;\n    \n    if (slt32(x_index_31, x_elems_5) && (slt32(y_index_32, y_elems_6) &&\n                                         slt32(index_in_35, in_elems_7))) {\n        ((__local bool *) block_11)[get_local_id_1_39 * 17 +\n                                    get_local_id_0_38] = ((__global\n                                                           bool *) srcmem_2)[idata_offset_34 +\n                                                                             index_in_35];\n    }\n    barrier(CLK_LOCAL_MEM_FENCE);\n    x_index_31 = get_group_id_1_41 * 16 + squot32(get_local_id_0_38, mulx_9);\n    y_index_32 = get_group_id_0_40 * 16 * mulx_9 + get_local_id_1_39 +\n        srem32(get_local_id_0_38, mulx_9) * 16;\n    \n    int32_t index_out_36 = y_index_32 * y_elems_6 + x_index_31;\n    \n    if (slt32(x_index_31, y_elems_6) && (slt32(y_index_32, x_elems_5) &&\n                                         slt32(index_out_36, out_elems_8))) {\n        ((__global bool *) destmem_0)[odata_offset_33 + index_out_36] =\n            ((__local bool *) block_11)[get_local_id_0_38 * 17 +\n                                        get_local_id_1_39];\n    }\n}\n__kernel void map_transpose_bool_low_width(const int block_dim0, const\n                             ",
            "              int block_dim1, const int block_dim2,\n                                           uint block_11_backing_offset_0,\n                                           int32_t destoffset_1,\n                                           int32_t srcoffset_3,\n                                           int32_t num_arrays_4,\n                                           int32_t x_elems_5, int32_t y_elems_6,\n                                           int32_t in_elems_7,\n                                           int32_t out_elems_8, int32_t mulx_9,\n                                           int32_t muly_10, __global\n                                           unsigned char *destmem_0, __global\n                                           unsigned char *srcmem_2)\n{\n    volatile char *block_11_backing_0 = &shared_mem[block_11_backing_offset_0];\n    __local char *block_11;\n    \n    block_11 = (__local char *) block_11_backing_0;\n    \n    int32_t get_global_id_0_37;\n    \n    get_global_id_0_37 = get_global_id(0);\n    \n    int32_t get_local_id_0_38;\n    \n    get_local_id_0_38 = get_local_id(0);\n    \n    int32_t get_local_id_1_39;\n    \n    get_local_id_1_39 = get_local_id(1);\n    \n    int32_t get_group_id_0_40;\n    \n    get_group_id_0_40 = get_group_id(0);\n    \n    int32_t get_group_id_1_41;\n    \n    get_group_id_1_41 = get_group_id(1);\n    \n    int32_t get_group_id_2_42;\n    \n    get_group_id_2_42 = get_group_id(2);\n    \n    int32_t our_array_offset_30 = get_group_id_2_42 * x_elems_5 * y_elems_6;\n    int32_t odata_offset_33 = destoffset_1 + our_array_offset_30;\n    int32_t idata_offset_34 = srcoffset_3 + our_array_offset_30;\n    int32_t x_index_31 = get_group_id_0_40 * 16 + squot32(get_local_id_0_38,\n                                                          muly_10);\n    int32_t y_index_32 = get_group_id_1_41 * 16 * muly_10 + get_local_id_1_39 +\n            srem32(get_local_id_0_38, muly_10) * 16;\n    int32_t index_in_35 = y_index_32 * x_elems_5 + x_index_31;\n    \n    if (slt32(x_ind",
            "ex_31, x_elems_5) && (slt32(y_index_32, y_elems_6) &&\n                                         slt32(index_in_35, in_elems_7))) {\n        ((__local bool *) block_11)[get_local_id_1_39 * 17 +\n                                    get_local_id_0_38] = ((__global\n                                                           bool *) srcmem_2)[idata_offset_34 +\n                                                                             index_in_35];\n    }\n    barrier(CLK_LOCAL_MEM_FENCE);\n    x_index_31 = get_group_id_1_41 * 16 * muly_10 + get_local_id_0_38 +\n        srem32(get_local_id_1_39, muly_10) * 16;\n    y_index_32 = get_group_id_0_40 * 16 + squot32(get_local_id_1_39, muly_10);\n    \n    int32_t index_out_36 = y_index_32 * y_elems_6 + x_index_31;\n    \n    if (slt32(x_index_31, y_elems_6) && (slt32(y_index_32, x_elems_5) &&\n                                         slt32(index_out_36, out_elems_8))) {\n        ((__global bool *) destmem_0)[odata_offset_33 + index_out_36] =\n            ((__local bool *) block_11)[get_local_id_0_38 * 17 +\n                                        get_local_id_1_39];\n    }\n}\n__kernel void map_transpose_bool_small(uint block_11_backing_offset_0,\n                                       int32_t destoffset_1,\n                                       int32_t srcoffset_3,\n                                       int32_t num_arrays_4, int32_t x_elems_5,\n                                       int32_t y_elems_6, int32_t in_elems_7,\n                                       int32_t out_elems_8, int32_t mulx_9,\n                                       int32_t muly_10, __global\n                                       unsigned char *destmem_0, __global\n                                       unsigned char *srcmem_2)\n{\n    const int block_dim0 = 0;\n    const int block_dim1 = 1;\n    const int block_dim2 = 2;\n    volatile char *block_11_backing_0 = &shared_mem[block_11_backing_offset_0];\n    __local char *block_11;\n    \n    block_11 = (__local char *) block_11_backing_0;",
            "\n    \n    int32_t get_global_id_0_37;\n    \n    get_global_id_0_37 = get_global_id(0);\n    \n    int32_t get_local_id_0_38;\n    \n    get_local_id_0_38 = get_local_id(0);\n    \n    int32_t get_local_id_1_39;\n    \n    get_local_id_1_39 = get_local_id(1);\n    \n    int32_t get_group_id_0_40;\n    \n    get_group_id_0_40 = get_group_id(0);\n    \n    int32_t get_group_id_1_41;\n    \n    get_group_id_1_41 = get_group_id(1);\n    \n    int32_t get_group_id_2_42;\n    \n    get_group_id_2_42 = get_group_id(2);\n    \n    int32_t our_array_offset_30 = squot32(get_global_id_0_37, y_elems_6 *\n                                          x_elems_5) * (y_elems_6 * x_elems_5);\n    int32_t x_index_31 = squot32(srem32(get_global_id_0_37, y_elems_6 *\n                                        x_elems_5), y_elems_6);\n    int32_t y_index_32 = srem32(get_global_id_0_37, y_elems_6);\n    int32_t odata_offset_33 = destoffset_1 + our_array_offset_30;\n    int32_t idata_offset_34 = srcoffset_3 + our_array_offset_30;\n    int32_t index_in_35 = y_index_32 * x_elems_5 + x_index_31;\n    int32_t index_out_36 = x_index_31 * y_elems_6 + y_index_32;\n    \n    if (slt32(get_global_id_0_37, in_elems_7)) {\n        ((__global bool *) destmem_0)[odata_offset_33 + index_out_36] =\n            ((__global bool *) srcmem_2)[idata_offset_34 + index_in_35];\n    }\n}\n__kernel void map_transpose_i32(const int block_dim0, const int block_dim1,\n                                const int block_dim2,\n                                uint block_11_backing_offset_0,\n                                int32_t destoffset_1, int32_t srcoffset_3,\n                                int32_t num_arrays_4, int32_t x_elems_5,\n                                int32_t y_elems_6, int32_t in_elems_7,\n                                int32_t out_elems_8, int32_t mulx_9,\n                                int32_t muly_10, __global\n                                unsigned char *destmem_0, __global\n                                unsigned char *srcmem_2)\n{\n    volatile c",
            "har *block_11_backing_0 = &shared_mem[block_11_backing_offset_0];\n    __local char *block_11;\n    \n    block_11 = (__local char *) block_11_backing_0;\n    \n    int32_t get_global_id_0_37;\n    \n    get_global_id_0_37 = get_global_id(0);\n    \n    int32_t get_local_id_0_38;\n    \n    get_local_id_0_38 = get_local_id(0);\n    \n    int32_t get_local_id_1_39;\n    \n    get_local_id_1_39 = get_local_id(1);\n    \n    int32_t get_group_id_0_40;\n    \n    get_group_id_0_40 = get_group_id(0);\n    \n    int32_t get_group_id_1_41;\n    \n    get_group_id_1_41 = get_group_id(1);\n    \n    int32_t get_group_id_2_42;\n    \n    get_group_id_2_42 = get_group_id(2);\n    \n    int32_t our_array_offset_30 = get_group_id_2_42 * x_elems_5 * y_elems_6;\n    int32_t odata_offset_33 = squot32(destoffset_1, 4) + our_array_offset_30;\n    int32_t idata_offset_34 = squot32(srcoffset_3, 4) + our_array_offset_30;\n    int32_t x_index_31 = get_global_id_0_37;\n    int32_t y_index_32 = get_group_id_1_41 * 32 + get_local_id_1_39;\n    \n    if (slt32(x_index_31, x_elems_5)) {\n        for (int32_t j_43 = 0; j_43 < 4; j_43++) {\n            int32_t index_in_35 = (y_index_32 + j_43 * 8) * x_elems_5 +\n                    x_index_31;\n            \n            if (slt32(y_index_32 + j_43 * 8, y_elems_6) && slt32(index_in_35,\n                                                                 in_elems_7)) {\n                ((__local int32_t *) block_11)[(get_local_id_1_39 + j_43 * 8) *\n                                               33 + get_local_id_0_38] =\n                    ((__global int32_t *) srcmem_2)[idata_offset_34 +\n                                                    index_in_35];\n            }\n        }\n    }\n    barrier(CLK_LOCAL_MEM_FENCE);\n    x_index_31 = get_group_id_1_41 * 32 + get_local_id_0_38;\n    y_index_32 = get_group_id_0_40 * 32 + get_local_id_1_39;\n    if (slt32(x_index_31, y_elems_6)) {\n        for (int32_t j_43 = 0; j_43 < 4; j_43++) {\n            int32_t index_out_36 = (y_index_32 + j_43 * 8) * y_ele",
            "ms_6 +\n                    x_index_31;\n            \n            if (slt32(y_index_32 + j_43 * 8, x_elems_5) && slt32(index_out_36,\n                                                                 out_elems_8)) {\n                ((__global int32_t *) destmem_0)[odata_offset_33 +\n                                                 index_out_36] = ((__local\n                                                                   int32_t *) block_11)[get_local_id_0_38 *\n                                                                                        33 +\n                                                                                        get_local_id_1_39 +\n                                                                                        j_43 *\n                                                                                        8];\n            }\n        }\n    }\n}\n__kernel void map_transpose_i32_low_height(const int block_dim0, const\n                                           int block_dim1, const int block_dim2,\n                                           uint block_11_backing_offset_0,\n                                           int32_t destoffset_1,\n                                           int32_t srcoffset_3,\n                                           int32_t num_arrays_4,\n                                           int32_t x_elems_5, int32_t y_elems_6,\n                                           int32_t in_elems_7,\n                                           int32_t out_elems_8, int32_t mulx_9,\n                                           int32_t muly_10, __global\n                                           unsigned char *destmem_0, __global\n                                           unsigned char *srcmem_2)\n{\n    volatile char *block_11_backing_0 = &shared_mem[block_11_backing_offset_0];\n    __local char *block_11;\n    \n    block_11 = (__local char *) block_11_backing_0;\n    \n    int32_t get_global_id_0_37;\n    \n    get_global_id_0_37 = get_global_id(0);\n    \n ",
            "   int32_t get_local_id_0_38;\n    \n    get_local_id_0_38 = get_local_id(0);\n    \n    int32_t get_local_id_1_39;\n    \n    get_local_id_1_39 = get_local_id(1);\n    \n    int32_t get_group_id_0_40;\n    \n    get_group_id_0_40 = get_group_id(0);\n    \n    int32_t get_group_id_1_41;\n    \n    get_group_id_1_41 = get_group_id(1);\n    \n    int32_t get_group_id_2_42;\n    \n    get_group_id_2_42 = get_group_id(2);\n    \n    int32_t our_array_offset_30 = get_group_id_2_42 * x_elems_5 * y_elems_6;\n    int32_t odata_offset_33 = squot32(destoffset_1, 4) + our_array_offset_30;\n    int32_t idata_offset_34 = squot32(srcoffset_3, 4) + our_array_offset_30;\n    int32_t x_index_31 = get_group_id_0_40 * 16 * mulx_9 + get_local_id_0_38 +\n            srem32(get_local_id_1_39, mulx_9) * 16;\n    int32_t y_index_32 = get_group_id_1_41 * 16 + squot32(get_local_id_1_39,\n                                                          mulx_9);\n    int32_t index_in_35 = y_index_32 * x_elems_5 + x_index_31;\n    \n    if (slt32(x_index_31, x_elems_5) && (slt32(y_index_32, y_elems_6) &&\n                                         slt32(index_in_35, in_elems_7))) {\n        ((__local int32_t *) block_11)[get_local_id_1_39 * 17 +\n                                       get_local_id_0_38] = ((__global\n                                                              int32_t *) srcmem_2)[idata_offset_34 +\n                                                                                   index_in_35];\n    }\n    barrier(CLK_LOCAL_MEM_FENCE);\n    x_index_31 = get_group_id_1_41 * 16 + squot32(get_local_id_0_38, mulx_9);\n    y_index_32 = get_group_id_0_40 * 16 * mulx_9 + get_local_id_1_39 +\n        srem32(get_local_id_0_38, mulx_9) * 16;\n    \n    int32_t index_out_36 = y_index_32 * y_elems_6 + x_index_31;\n    \n    if (slt32(x_index_31, y_elems_6) && (slt32(y_index_32, x_elems_5) &&\n                                         slt32(index_out_36, out_elems_8))) {\n        ((__global int32_t *) destmem_0)[odata_offset_33 + index_out_36]",
            " =\n            ((__local int32_t *) block_11)[get_local_id_0_38 * 17 +\n                                           get_local_id_1_39];\n    }\n}\n__kernel void map_transpose_i32_low_width(const int block_dim0, const\n                                          int block_dim1, const int block_dim2,\n                                          uint block_11_backing_offset_0,\n                                          int32_t destoffset_1,\n                                          int32_t srcoffset_3,\n                                          int32_t num_arrays_4,\n                                          int32_t x_elems_5, int32_t y_elems_6,\n                                          int32_t in_elems_7,\n                                          int32_t out_elems_8, int32_t mulx_9,\n                                          int32_t muly_10, __global\n                                          unsigned char *destmem_0, __global\n                                          unsigned char *srcmem_2)\n{\n    volatile char *block_11_backing_0 = &shared_mem[block_11_backing_offset_0];\n    __local char *block_11;\n    \n    block_11 = (__local char *) block_11_backing_0;\n    \n    int32_t get_global_id_0_37;\n    \n    get_global_id_0_37 = get_global_id(0);\n    \n    int32_t get_local_id_0_38;\n    \n    get_local_id_0_38 = get_local_id(0);\n    \n    int32_t get_local_id_1_39;\n    \n    get_local_id_1_39 = get_local_id(1);\n    \n    int32_t get_group_id_0_40;\n    \n    get_group_id_0_40 = get_group_id(0);\n    \n    int32_t get_group_id_1_41;\n    \n    get_group_id_1_41 = get_group_id(1);\n    \n    int32_t get_group_id_2_42;\n    \n    get_group_id_2_42 = get_group_id(2);\n    \n    int32_t our_array_offset_30 = get_group_id_2_42 * x_elems_5 * y_elems_6;\n    int32_t odata_offset_33 = squot32(destoffset_1, 4) + our_array_offset_30;\n    int32_t idata_offset_34 = squot32(srcoffset_3, 4) + our_array_offset_30;\n    int32_t x_index_31 = get_group_id_0_40 * 16 + squot32(get_local_id_0_38,\n                                  ",
            "                        muly_10);\n    int32_t y_index_32 = get_group_id_1_41 * 16 * muly_10 + get_local_id_1_39 +\n            srem32(get_local_id_0_38, muly_10) * 16;\n    int32_t index_in_35 = y_index_32 * x_elems_5 + x_index_31;\n    \n    if (slt32(x_index_31, x_elems_5) && (slt32(y_index_32, y_elems_6) &&\n                                         slt32(index_in_35, in_elems_7))) {\n        ((__local int32_t *) block_11)[get_local_id_1_39 * 17 +\n                                       get_local_id_0_38] = ((__global\n                                                              int32_t *) srcmem_2)[idata_offset_34 +\n                                                                                   index_in_35];\n    }\n    barrier(CLK_LOCAL_MEM_FENCE);\n    x_index_31 = get_group_id_1_41 * 16 * muly_10 + get_local_id_0_38 +\n        srem32(get_local_id_1_39, muly_10) * 16;\n    y_index_32 = get_group_id_0_40 * 16 + squot32(get_local_id_1_39, muly_10);\n    \n    int32_t index_out_36 = y_index_32 * y_elems_6 + x_index_31;\n    \n    if (slt32(x_index_31, y_elems_6) && (slt32(y_index_32, x_elems_5) &&\n                                         slt32(index_out_36, out_elems_8))) {\n        ((__global int32_t *) destmem_0)[odata_offset_33 + index_out_36] =\n            ((__local int32_t *) block_11)[get_local_id_0_38 * 17 +\n                                           get_local_id_1_39];\n    }\n}\n__kernel void map_transpose_i32_small(uint block_11_backing_offset_0,\n                                      int32_t destoffset_1, int32_t srcoffset_3,\n                                      int32_t num_arrays_4, int32_t x_elems_5,\n                                      int32_t y_elems_6, int32_t in_elems_7,\n                                      int32_t out_elems_8, int32_t mulx_9,\n                                      int32_t muly_10, __global\n                                      unsigned char *destmem_0, __global\n                                      unsigned char *srcmem_2)\n{\n    const int block_",
            "dim0 = 0;\n    const int block_dim1 = 1;\n    const int block_dim2 = 2;\n    volatile char *block_11_backing_0 = &shared_mem[block_11_backing_offset_0];\n    __local char *block_11;\n    \n    block_11 = (__local char *) block_11_backing_0;\n    \n    int32_t get_global_id_0_37;\n    \n    get_global_id_0_37 = get_global_id(0);\n    \n    int32_t get_local_id_0_38;\n    \n    get_local_id_0_38 = get_local_id(0);\n    \n    int32_t get_local_id_1_39;\n    \n    get_local_id_1_39 = get_local_id(1);\n    \n    int32_t get_group_id_0_40;\n    \n    get_group_id_0_40 = get_group_id(0);\n    \n    int32_t get_group_id_1_41;\n    \n    get_group_id_1_41 = get_group_id(1);\n    \n    int32_t get_group_id_2_42;\n    \n    get_group_id_2_42 = get_group_id(2);\n    \n    int32_t our_array_offset_30 = squot32(get_global_id_0_37, y_elems_6 *\n                                          x_elems_5) * (y_elems_6 * x_elems_5);\n    int32_t x_index_31 = squot32(srem32(get_global_id_0_37, y_elems_6 *\n                                        x_elems_5), y_elems_6);\n    int32_t y_index_32 = srem32(get_global_id_0_37, y_elems_6);\n    int32_t odata_offset_33 = squot32(destoffset_1, 4) + our_array_offset_30;\n    int32_t idata_offset_34 = squot32(srcoffset_3, 4) + our_array_offset_30;\n    int32_t index_in_35 = y_index_32 * x_elems_5 + x_index_31;\n    int32_t index_out_36 = x_index_31 * y_elems_6 + y_index_32;\n    \n    if (slt32(get_global_id_0_37, in_elems_7)) {\n        ((__global int32_t *) destmem_0)[odata_offset_33 + index_out_36] =\n            ((__global int32_t *) srcmem_2)[idata_offset_34 + index_in_35];\n    }\n}\n__kernel void replicate_32962(__global unsigned char *mem_32958,\n                              int32_t num_elems_32959, unsigned char val_32960)\n{\n    const int block_dim0 = 0;\n    const int block_dim1 = 1;\n    const int block_dim2 = 2;\n    int32_t replicate_gtid_32962;\n    int32_t replicate_ltid_32963;\n    int32_t replicate_gid_32964;\n    \n    replicate_gtid_32962 = get_global_id(0);\n    replicate_ltid_32963 = g",
            "et_local_id(0);\n    replicate_gid_32964 = get_group_id(0);\n    if (slt32(replicate_gtid_32962, num_elems_32959)) {\n        ((__global bool *) mem_32958)[replicate_gtid_32962] = val_32960;\n    }\n}\n__kernel void replicate_33364(__global unsigned char *mem_33360,\n                              int32_t num_elems_33361, int32_t val_33362)\n{\n    const int block_dim0 = 0;\n    const int block_dim1 = 1;\n    const int block_dim2 = 2;\n    int32_t replicate_gtid_33364;\n    int32_t replicate_ltid_33365;\n    int32_t replicate_gid_33366;\n    \n    replicate_gtid_33364 = get_global_id(0);\n    replicate_ltid_33365 = get_local_id(0);\n    replicate_gid_33366 = get_group_id(0);\n    if (slt32(replicate_gtid_33364, num_elems_33361)) {\n        ((__global int32_t *) mem_33360)[replicate_gtid_33364] = val_33362;\n    }\n}\n__kernel void scan_stage1_32141(uint scan_arr_mem_33017_backing_offset_0,\n                                uint scan_arr_mem_33019_backing_offset_1,\n                                int32_t flat_dim_31615, __global\n                                unsigned char *x_mem_32790, __global\n                                unsigned char *x_mem_32791, __global\n                                unsigned char *mem_32795, __global\n                                unsigned char *mem_32797,\n                                int32_t num_threads_33005)\n{\n    const int32_t segscan_group_sizze_32136 =\n                  quicksort1zisegscan_group_sizze_32135;\n    const int block_dim0 = 0;\n    const int block_dim1 = 1;\n    const int block_dim2 = 2;\n    volatile char *scan_arr_mem_33017_backing_0 =\n                  &shared_mem[scan_arr_mem_33017_backing_offset_0];\n    volatile char *scan_arr_mem_33019_backing_1 =\n                  &shared_mem[scan_arr_mem_33019_backing_offset_1];\n    int32_t global_tid_33012;\n    int32_t local_tid_33013;\n    int32_t group_sizze_33016;\n    int32_t wave_sizze_33015;\n    int32_t group_tid_33014;\n    \n    global_tid_33012 = get_global_id(0);\n    local_tid_33013 = get_local_id",
            "(0);\n    group_sizze_33016 = get_local_size(0);\n    wave_sizze_33015 = LOCKSTEP_WIDTH;\n    group_tid_33014 = get_group_id(0);\n    \n    int32_t phys_tid_32141 = global_tid_33012;\n    __local char *scan_arr_mem_33017;\n    \n    scan_arr_mem_33017 = (__local char *) scan_arr_mem_33017_backing_0;\n    \n    __local char *scan_arr_mem_33019;\n    \n    scan_arr_mem_33019 = (__local char *) scan_arr_mem_33019_backing_1;\n    \n    int32_t x_31723;\n    bool x_31724;\n    int32_t x_31725;\n    bool x_31726;\n    \n    x_31723 = 0;\n    x_31724 = 0;\n    for (int32_t j_33021 = 0; j_33021 < squot32(flat_dim_31615 +\n                                                num_threads_33005 - 1,\n                                                num_threads_33005); j_33021++) {\n        int32_t chunk_offset_33022 = segscan_group_sizze_32136 * j_33021 +\n                group_tid_33014 * (segscan_group_sizze_32136 *\n                                   squot32(flat_dim_31615 + num_threads_33005 -\n                                           1, num_threads_33005));\n        int32_t flat_idx_33023 = chunk_offset_33022 + local_tid_33013;\n        int32_t gtid_32140 = flat_idx_33023;\n        \n        // threads in bounds read input; others get neutral element\n        {\n            if (slt32(gtid_32140, flat_dim_31615)) {\n                int32_t x_31729 = ((__global\n                                    int32_t *) x_mem_32790)[gtid_32140];\n                bool x_31730 = ((__global bool *) x_mem_32791)[gtid_32140];\n                \n                // write to-scan values to parameters\n                {\n                    x_31725 = x_31729;\n                    x_31726 = x_31730;\n                }\n                // write mapped values results to global memory\n                { }\n            } else {\n                x_31725 = 0;\n                x_31726 = 0;\n            }\n        }\n        // combine with carry and write to local memory\n        {\n            int32_t res_31727;\n            \n            if (x_31726) {\n    ",
            "            res_31727 = x_31725;\n            } else {\n                res_31727 = x_31723;\n            }\n            \n            bool res_31728 = x_31724 || x_31726;\n            \n            ((__local int32_t *) scan_arr_mem_33017)[local_tid_33013] =\n                res_31727;\n            ((__local bool *) scan_arr_mem_33019)[local_tid_33013] = res_31728;\n        }\n        \n        int32_t x_33006;\n        bool x_33007;\n        int32_t x_33008;\n        bool x_33009;\n        int32_t x_33024;\n        bool x_33025;\n        int32_t x_33026;\n        bool x_33027;\n        int32_t skip_threads_33030;\n        \n        if (slt32(local_tid_33013, segscan_group_sizze_32136)) {\n            x_33008 = ((volatile __local\n                        int32_t *) scan_arr_mem_33017)[local_tid_33013];\n            x_33009 = ((volatile __local\n                        bool *) scan_arr_mem_33019)[local_tid_33013];\n        }\n        // in-block scan (hopefully no barriers needed)\n        {\n            skip_threads_33030 = 1;\n            while (slt32(skip_threads_33030, 32)) {\n                if (sle32(skip_threads_33030, local_tid_33013 -\n                          squot32(local_tid_33013, 32) * 32) &&\n                    slt32(local_tid_33013, segscan_group_sizze_32136)) {\n                    // read operands\n                    {\n                        x_33006 = ((volatile __local\n                                    int32_t *) scan_arr_mem_33017)[local_tid_33013 -\n                                                                   skip_threads_33030];\n                        x_33007 = ((volatile __local\n                                    bool *) scan_arr_mem_33019)[local_tid_33013 -\n                                                                skip_threads_33030];\n                    }\n                    // perform operation\n                    {\n                        int32_t res_33010;\n                        \n                        if (x_33009) {\n                            res_33010",
            " = x_33008;\n                        } else {\n                            res_33010 = x_33006;\n                        }\n                        \n                        bool res_33011 = x_33007 || x_33009;\n                        \n                        x_33008 = res_33010;\n                        x_33009 = res_33011;\n                    }\n                }\n                if (sle32(wave_sizze_33015, skip_threads_33030)) {\n                    barrier(CLK_LOCAL_MEM_FENCE);\n                }\n                if (sle32(skip_threads_33030, local_tid_33013 -\n                          squot32(local_tid_33013, 32) * 32) &&\n                    slt32(local_tid_33013, segscan_group_sizze_32136)) {\n                    // write result\n                    {\n                        ((volatile __local\n                          int32_t *) scan_arr_mem_33017)[local_tid_33013] =\n                            x_33008;\n                        ((volatile __local\n                          bool *) scan_arr_mem_33019)[local_tid_33013] =\n                            x_33009;\n                    }\n                }\n                if (sle32(wave_sizze_33015, skip_threads_33030)) {\n                    barrier(CLK_LOCAL_MEM_FENCE);\n                }\n                skip_threads_33030 *= 2;\n            }\n        }\n        barrier(CLK_LOCAL_MEM_FENCE);\n        // last thread of block 'i' writes its result to offset 'i'\n        {\n            if ((local_tid_33013 - squot32(local_tid_33013, 32) * 32) == 31 &&\n                slt32(local_tid_33013, segscan_group_sizze_32136)) {\n                ((volatile __local\n                  int32_t *) scan_arr_mem_33017)[squot32(local_tid_33013, 32)] =\n                    x_33008;\n                ((volatile __local\n                  bool *) scan_arr_mem_33019)[squot32(local_tid_33013, 32)] =\n                    x_33009;\n            }\n        }\n        barrier(CLK_LOCAL_MEM_FENCE);\n        // scan the first block, after which offset 'i' contains carry-in for warp ",
            "'i+1'\n        {\n            int32_t skip_threads_33031;\n            \n            if (squot32(local_tid_33013, 32) == 0 && slt32(local_tid_33013,\n                                                           segscan_group_sizze_32136)) {\n                x_33026 = ((volatile __local\n                            int32_t *) scan_arr_mem_33017)[local_tid_33013];\n                x_33027 = ((volatile __local\n                            bool *) scan_arr_mem_33019)[local_tid_33013];\n            }\n            // in-block scan (hopefully no barriers needed)\n            {\n                skip_threads_33031 = 1;\n                while (slt32(skip_threads_33031, 32)) {\n                    if (sle32(skip_threads_33031, local_tid_33013 -\n                              squot32(local_tid_33013, 32) * 32) &&\n                        (squot32(local_tid_33013, 32) == 0 &&\n                         slt32(local_tid_33013, segscan_group_sizze_32136))) {\n                        // read operands\n                        {\n                            x_33024 = ((volatile __local\n                                        int32_t *) scan_arr_mem_33017)[local_tid_33013 -\n                                                                       skip_threads_33031];\n                            x_33025 = ((volatile __local\n                                        bool *) scan_arr_mem_33019)[local_tid_33013 -\n                                                                    skip_threads_33031];\n                        }\n                        // perform operation\n                        {\n                            int32_t res_33028;\n                            \n                            if (x_33027) {\n                                res_33028 = x_33026;\n                            } else {\n                                res_33028 = x_33024;\n                            }\n                            \n                            bool res_33029 = x_33025 || x_33027;\n                            \n               ",
            "             x_33026 = res_33028;\n                            x_33027 = res_33029;\n                        }\n                    }\n                    if (sle32(wave_sizze_33015, skip_threads_33031)) {\n                        barrier(CLK_LOCAL_MEM_FENCE);\n                    }\n                    if (sle32(skip_threads_33031, local_tid_33013 -\n                              squot32(local_tid_33013, 32) * 32) &&\n                        (squot32(local_tid_33013, 32) == 0 &&\n                         slt32(local_tid_33013, segscan_group_sizze_32136))) {\n                        // write result\n                        {\n                            ((volatile __local\n                              int32_t *) scan_arr_mem_33017)[local_tid_33013] =\n                                x_33026;\n                            ((volatile __local\n                              bool *) scan_arr_mem_33019)[local_tid_33013] =\n                                x_33027;\n                        }\n                    }\n                    if (sle32(wave_sizze_33015, skip_threads_33031)) {\n                        barrier(CLK_LOCAL_MEM_FENCE);\n                    }\n                    skip_threads_33031 *= 2;\n                }\n            }\n        }\n        barrier(CLK_LOCAL_MEM_FENCE);\n        // carry-in for every block except the first\n        {\n            if (!(squot32(local_tid_33013, 32) == 0 || !slt32(local_tid_33013,\n                                                              segscan_group_sizze_32136))) {\n                // read operands\n                {\n                    x_33006 = ((volatile __local\n                                int32_t *) scan_arr_mem_33017)[squot32(local_tid_33013,\n                                                                       32) - 1];\n                    x_33007 = ((volatile __local\n                                bool *) scan_arr_mem_33019)[squot32(local_tid_33013,\n                                                                    32) - 1];\n          ",
            "      }\n                // perform operation\n                {\n                    int32_t res_33010;\n                    \n                    if (x_33009) {\n                        res_33010 = x_33008;\n                    } else {\n                        res_33010 = x_33006;\n                    }\n                    \n                    bool res_33011 = x_33007 || x_33009;\n                    \n                    x_33008 = res_33010;\n                    x_33009 = res_33011;\n                }\n                // write final result\n                {\n                    ((volatile __local\n                      int32_t *) scan_arr_mem_33017)[local_tid_33013] = x_33008;\n                    ((volatile __local\n                      bool *) scan_arr_mem_33019)[local_tid_33013] = x_33009;\n                }\n            }\n        }\n        barrier(CLK_LOCAL_MEM_FENCE);\n        // restore correct values for first block\n        {\n            if (squot32(local_tid_33013, 32) == 0) {\n                ((volatile __local\n                  int32_t *) scan_arr_mem_33017)[local_tid_33013] = x_33008;\n                ((volatile __local\n                  bool *) scan_arr_mem_33019)[local_tid_33013] = x_33009;\n            }\n        }\n        barrier(CLK_LOCAL_MEM_FENCE);\n        // threads in bounds write partial scan result\n        {\n            if (slt32(gtid_32140, flat_dim_31615)) {\n                ((__global int32_t *) mem_32795)[gtid_32140] = ((__local\n                                                                 int32_t *) scan_arr_mem_33017)[local_tid_33013];\n                ((__global bool *) mem_32797)[gtid_32140] = ((__local\n                                                              bool *) scan_arr_mem_33019)[local_tid_33013];\n            }\n        }\n        barrier(CLK_LOCAL_MEM_FENCE);\n        // first thread reads last element as carry-in for next iteration\n        {\n            if (local_tid_33013 == 0) {\n                x_31723 = ((__local\n                            ",
            "int32_t *) scan_arr_mem_33017)[segscan_group_sizze_32136 -\n                                                           1];\n                x_31724 = ((__local\n                            bool *) scan_arr_mem_33019)[segscan_group_sizze_32136 -\n                                                        1];\n            }\n        }\n        barrier(CLK_LOCAL_MEM_FENCE);\n    }\n}\n__kernel void scan_stage1_32150(uint scan_arr_mem_33081_backing_offset_0,\n                                uint scan_arr_mem_33083_backing_offset_1,\n                                int32_t flat_dim_31615, __global\n                                unsigned char *x_mem_32790, __global\n                                unsigned char *x_mem_32791, __global\n                                unsigned char *mem_32795, __global\n                                unsigned char *mem_32797, __global\n                                unsigned char *mem_32801, __global\n                                unsigned char *mem_32803, __global\n                                unsigned char *mem_32805,\n                                int32_t num_threads_33069)\n{\n    const int32_t segscan_group_sizze_32145 =\n                  quicksort1zisegscan_group_sizze_32144;\n    const int block_dim0 = 0;\n    const int block_dim1 = 1;\n    const int block_dim2 = 2;\n    volatile char *scan_arr_mem_33081_backing_0 =\n                  &shared_mem[scan_arr_mem_33081_backing_offset_0];\n    volatile char *scan_arr_mem_33083_backing_1 =\n                  &shared_mem[scan_arr_mem_33083_backing_offset_1];\n    int32_t global_tid_33076;\n    int32_t local_tid_33077;\n    int32_t group_sizze_33080;\n    int32_t wave_sizze_33079;\n    int32_t group_tid_33078;\n    \n    global_tid_33076 = get_global_id(0);\n    local_tid_33077 = get_local_id(0);\n    group_sizze_33080 = get_local_size(0);\n    wave_sizze_33079 = LOCKSTEP_WIDTH;\n    group_tid_33078 = get_group_id(0);\n    \n    int32_t phys_tid_32150 = global_tid_33076;\n    __local char *scan_arr_mem_33081;\n    \n    scan_ar",
            "r_mem_33081 = (__local char *) scan_arr_mem_33081_backing_0;\n    \n    __local char *scan_arr_mem_33083;\n    \n    scan_arr_mem_33083 = (__local char *) scan_arr_mem_33083_backing_1;\n    \n    int32_t x_31742;\n    bool x_31743;\n    int32_t x_31744;\n    bool x_31745;\n    \n    x_31742 = 0;\n    x_31743 = 0;\n    for (int32_t j_33085 = 0; j_33085 < squot32(flat_dim_31615 +\n                                                num_threads_33069 - 1,\n                                                num_threads_33069); j_33085++) {\n        int32_t chunk_offset_33086 = segscan_group_sizze_32145 * j_33085 +\n                group_tid_33078 * (segscan_group_sizze_32145 *\n                                   squot32(flat_dim_31615 + num_threads_33069 -\n                                           1, num_threads_33069));\n        int32_t flat_idx_33087 = chunk_offset_33086 + local_tid_33077;\n        int32_t gtid_32149 = flat_idx_33087;\n        \n        // threads in bounds read input; others get neutral element\n        {\n            if (slt32(gtid_32149, flat_dim_31615)) {\n                int32_t x_31748 = ((__global int32_t *) mem_32795)[gtid_32149];\n                bool x_31749 = ((__global bool *) mem_32797)[gtid_32149];\n                int32_t x_31750 = ((__global\n                                    int32_t *) x_mem_32790)[gtid_32149];\n                bool x_31752 = ((__global bool *) x_mem_32791)[gtid_32149];\n                int32_t res_31753;\n                \n                if (x_31749) {\n                    res_31753 = x_31748;\n                } else {\n                    res_31753 = 0;\n                }\n                \n                int32_t res_31755;\n                \n                if (x_31749) {\n                    res_31755 = res_31753;\n                } else {\n                    res_31755 = 0;\n                }\n                \n                bool res_31756 = sle32(res_31755, x_31750);\n                \n                // write to-scan values to parameters\n                {\n  ",
            "                  x_31744 = gtid_32149;\n                    x_31745 = x_31752;\n                }\n                // write mapped values results to global memory\n                {\n                    ((__global bool *) mem_32805)[gtid_32149] = res_31756;\n                }\n            } else {\n                x_31744 = 0;\n                x_31745 = 0;\n            }\n        }\n        // combine with carry and write to local memory\n        {\n            int32_t res_31746;\n            \n            if (x_31745) {\n                res_31746 = x_31744;\n            } else {\n                res_31746 = x_31742;\n            }\n            \n            bool res_31747 = x_31743 || x_31745;\n            \n            ((__local int32_t *) scan_arr_mem_33081)[local_tid_33077] =\n                res_31746;\n            ((__local bool *) scan_arr_mem_33083)[local_tid_33077] = res_31747;\n        }\n        \n        int32_t x_33070;\n        bool x_33071;\n        int32_t x_33072;\n        bool x_33073;\n        int32_t x_33088;\n        bool x_33089;\n        int32_t x_33090;\n        bool x_33091;\n        int32_t skip_threads_33094;\n        \n        if (slt32(local_tid_33077, segscan_group_sizze_32145)) {\n            x_33072 = ((volatile __local\n                        int32_t *) scan_arr_mem_33081)[local_tid_33077];\n            x_33073 = ((volatile __local\n                        bool *) scan_arr_mem_33083)[local_tid_33077];\n        }\n        // in-block scan (hopefully no barriers needed)\n        {\n            skip_threads_33094 = 1;\n            while (slt32(skip_threads_33094, 32)) {\n                if (sle32(skip_threads_33094, local_tid_33077 -\n                          squot32(local_tid_33077, 32) * 32) &&\n                    slt32(local_tid_33077, segscan_group_sizze_32145)) {\n                    // read operands\n                    {\n                        x_33070 = ((volatile __local\n                                    int32_t *) scan_arr_mem_33081)[local_tid_33077 -\n                     ",
            "                                              skip_threads_33094];\n                        x_33071 = ((volatile __local\n                                    bool *) scan_arr_mem_33083)[local_tid_33077 -\n                                                                skip_threads_33094];\n                    }\n                    // perform operation\n                    {\n                        int32_t res_33074;\n                        \n                        if (x_33073) {\n                            res_33074 = x_33072;\n                        } else {\n                            res_33074 = x_33070;\n                        }\n                        \n                        bool res_33075 = x_33071 || x_33073;\n                        \n                        x_33072 = res_33074;\n                        x_33073 = res_33075;\n                    }\n                }\n                if (sle32(wave_sizze_33079, skip_threads_33094)) {\n                    barrier(CLK_LOCAL_MEM_FENCE);\n                }\n                if (sle32(skip_threads_33094, local_tid_33077 -\n                          squot32(local_tid_33077, 32) * 32) &&\n                    slt32(local_tid_33077, segscan_group_sizze_32145)) {\n                    // write result\n                    {\n                        ((volatile __local\n                          int32_t *) scan_arr_mem_33081)[local_tid_33077] =\n                            x_33072;\n                        ((volatile __local\n                          bool *) scan_arr_mem_33083)[local_tid_33077] =\n                            x_33073;\n                    }\n                }\n                if (sle32(wave_sizze_33079, skip_threads_33094)) {\n                    barrier(CLK_LOCAL_MEM_FENCE);\n                }\n                skip_threads_33094 *= 2;\n            }\n        }\n        barrier(CLK_LOCAL_MEM_FENCE);\n        // last thread of block 'i' writes its result to offset 'i'\n        {\n            if ((local_tid_33077 - squot32(local_tid_33077, 32) ",
            "* 32) == 31 &&\n                slt32(local_tid_33077, segscan_group_sizze_32145)) {\n                ((volatile __local\n                  int32_t *) scan_arr_mem_33081)[squot32(local_tid_33077, 32)] =\n                    x_33072;\n                ((volatile __local\n                  bool *) scan_arr_mem_33083)[squot32(local_tid_33077, 32)] =\n                    x_33073;\n            }\n        }\n        barrier(CLK_LOCAL_MEM_FENCE);\n        // scan the first block, after which offset 'i' contains carry-in for warp 'i+1'\n        {\n            int32_t skip_threads_33095;\n            \n            if (squot32(local_tid_33077, 32) == 0 && slt32(local_tid_33077,\n                                                           segscan_group_sizze_32145)) {\n                x_33090 = ((volatile __local\n                            int32_t *) scan_arr_mem_33081)[local_tid_33077];\n                x_33091 = ((volatile __local\n                            bool *) scan_arr_mem_33083)[local_tid_33077];\n            }\n            // in-block scan (hopefully no barriers needed)\n            {\n                skip_threads_33095 = 1;\n                while (slt32(skip_threads_33095, 32)) {\n                    if (sle32(skip_threads_33095, local_tid_33077 -\n                              squot32(local_tid_33077, 32) * 32) &&\n                        (squot32(local_tid_33077, 32) == 0 &&\n                         slt32(local_tid_33077, segscan_group_sizze_32145))) {\n                        // read operands\n                        {\n                            x_33088 = ((volatile __local\n                                        int32_t *) scan_arr_mem_33081)[local_tid_33077 -\n                                                                       skip_threads_33095];\n                            x_33089 = ((volatile __local\n                                        bool *) scan_arr_mem_33083)[local_tid_33077 -\n                                                                    skip_threads_33095];\n           ",
            "             }\n                        // perform operation\n                        {\n                            int32_t res_33092;\n                            \n                            if (x_33091) {\n                                res_33092 = x_33090;\n                            } else {\n                                res_33092 = x_33088;\n                            }\n                            \n                            bool res_33093 = x_33089 || x_33091;\n                            \n                            x_33090 = res_33092;\n                            x_33091 = res_33093;\n                        }\n                    }\n                    if (sle32(wave_sizze_33079, skip_threads_33095)) {\n                        barrier(CLK_LOCAL_MEM_FENCE);\n                    }\n                    if (sle32(skip_threads_33095, local_tid_33077 -\n                              squot32(local_tid_33077, 32) * 32) &&\n                        (squot32(local_tid_33077, 32) == 0 &&\n                         slt32(local_tid_33077, segscan_group_sizze_32145))) {\n                        // write result\n                        {\n                            ((volatile __local\n                              int32_t *) scan_arr_mem_33081)[local_tid_33077] =\n                                x_33090;\n                            ((volatile __local\n                              bool *) scan_arr_mem_33083)[local_tid_33077] =\n                                x_33091;\n                        }\n                    }\n                    if (sle32(wave_sizze_33079, skip_threads_33095)) {\n                        barrier(CLK_LOCAL_MEM_FENCE);\n                    }\n                    skip_threads_33095 *= 2;\n                }\n            }\n        }\n        barrier(CLK_LOCAL_MEM_FENCE);\n        // carry-in for every block except the first\n        {\n            if (!(squot32(local_tid_33077, 32) == 0 || !slt32(local_tid_33077,\n                                                              segsca",
            "n_group_sizze_32145))) {\n                // read operands\n                {\n                    x_33070 = ((volatile __local\n                                int32_t *) scan_arr_mem_33081)[squot32(local_tid_33077,\n                                                                       32) - 1];\n                    x_33071 = ((volatile __local\n                                bool *) scan_arr_mem_33083)[squot32(local_tid_33077,\n                                                                    32) - 1];\n                }\n                // perform operation\n                {\n                    int32_t res_33074;\n                    \n                    if (x_33073) {\n                        res_33074 = x_33072;\n                    } else {\n                        res_33074 = x_33070;\n                    }\n                    \n                    bool res_33075 = x_33071 || x_33073;\n                    \n                    x_33072 = res_33074;\n                    x_33073 = res_33075;\n                }\n                // write final result\n                {\n                    ((volatile __local\n                      int32_t *) scan_arr_mem_33081)[local_tid_33077] = x_33072;\n                    ((volatile __local\n                      bool *) scan_arr_mem_33083)[local_tid_33077] = x_33073;\n                }\n            }\n        }\n        barrier(CLK_LOCAL_MEM_FENCE);\n        // restore correct values for first block\n        {\n            if (squot32(local_tid_33077, 32) == 0) {\n                ((volatile __local\n                  int32_t *) scan_arr_mem_33081)[local_tid_33077] = x_33072;\n                ((volatile __local\n                  bool *) scan_arr_mem_33083)[local_tid_33077] = x_33073;\n            }\n        }\n        barrier(CLK_LOCAL_MEM_FENCE);\n        // threads in bounds write partial scan result\n        {\n            if (slt32(gtid_32149, flat_dim_31615)) {\n                ((__global int32_t *) mem_32801)[gtid_32149] = ((__local\n                          ",
            "                                       int32_t *) scan_arr_mem_33081)[local_tid_33077];\n                ((__global bool *) mem_32803)[gtid_32149] = ((__local\n                                                              bool *) scan_arr_mem_33083)[local_tid_33077];\n            }\n        }\n        barrier(CLK_LOCAL_MEM_FENCE);\n        // first thread reads last element as carry-in for next iteration\n        {\n            if (local_tid_33077 == 0) {\n                x_31742 = ((__local\n                            int32_t *) scan_arr_mem_33081)[segscan_group_sizze_32145 -\n                                                           1];\n                x_31743 = ((__local\n                            bool *) scan_arr_mem_33083)[segscan_group_sizze_32145 -\n                                                        1];\n            }\n        }\n        barrier(CLK_LOCAL_MEM_FENCE);\n    }\n}\n__kernel void scan_stage1_32190(uint scan_arr_mem_33156_backing_offset_0,\n                                uint scan_arr_mem_33158_backing_offset_1,\n                                int32_t flat_dim_31615, __global\n                                unsigned char *x_mem_32791, __global\n                                unsigned char *mem_32805, __global\n                                unsigned char *mem_32815, __global\n                                unsigned char *mem_32817,\n                                int32_t num_threads_33143)\n{\n    const int32_t segscan_group_sizze_32185 =\n                  quicksort1zisegscan_group_sizze_32184;\n    const int block_dim0 = 0;\n    const int block_dim1 = 1;\n    const int block_dim2 = 2;\n    volatile char *scan_arr_mem_33156_backing_0 =\n                  &shared_mem[scan_arr_mem_33156_backing_offset_0];\n    volatile char *scan_arr_mem_33158_backing_1 =\n                  &shared_mem[scan_arr_mem_33158_backing_offset_1];\n    int32_t global_tid_33151;\n    int32_t local_tid_33152;\n    int32_t group_sizze_33155;\n    int32_t wave_sizze_33154;\n    int32_t group_tid_33153;",
            "\n    \n    global_tid_33151 = get_global_id(0);\n    local_tid_33152 = get_local_id(0);\n    group_sizze_33155 = get_local_size(0);\n    wave_sizze_33154 = LOCKSTEP_WIDTH;\n    group_tid_33153 = get_group_id(0);\n    \n    int32_t phys_tid_32190 = global_tid_33151;\n    __local char *scan_arr_mem_33156;\n    \n    scan_arr_mem_33156 = (__local char *) scan_arr_mem_33156_backing_0;\n    \n    __local char *scan_arr_mem_33158;\n    \n    scan_arr_mem_33158 = (__local char *) scan_arr_mem_33158_backing_1;\n    \n    int32_t x_31775;\n    bool x_31776;\n    int32_t x_31777;\n    bool x_31778;\n    \n    x_31775 = 0;\n    x_31776 = 0;\n    for (int32_t j_33160 = 0; j_33160 < squot32(flat_dim_31615 +\n                                                num_threads_33143 - 1,\n                                                num_threads_33143); j_33160++) {\n        int32_t chunk_offset_33161 = segscan_group_sizze_32185 * j_33160 +\n                group_tid_33153 * (segscan_group_sizze_32185 *\n                                   squot32(flat_dim_31615 + num_threads_33143 -\n                                           1, num_threads_33143));\n        int32_t flat_idx_33162 = chunk_offset_33161 + local_tid_33152;\n        int32_t gtid_32189 = flat_idx_33162;\n        \n        // threads in bounds read input; others get neutral element\n        {\n            if (slt32(gtid_32189, flat_dim_31615)) {\n                bool x_31782 = ((__global bool *) mem_32805)[gtid_32189];\n                bool x_31783 = ((__global bool *) x_mem_32791)[gtid_32189];\n                int32_t res_31784;\n                \n                if (x_31782) {\n                    res_31784 = 1;\n                } else {\n                    res_31784 = 0;\n                }\n                // write to-scan values to parameters\n                {\n                    x_31777 = res_31784;\n                    x_31778 = x_31783;\n                }\n                // write mapped values results to global memory\n                { }\n            } else {\n     ",
            "           x_31777 = 0;\n                x_31778 = 0;\n            }\n        }\n        // combine with carry and write to local memory\n        {\n            int32_t res_31779;\n            \n            if (x_31778) {\n                res_31779 = x_31777;\n            } else {\n                int32_t res_31780 = x_31775 + x_31777;\n                \n                res_31779 = res_31780;\n            }\n            \n            bool res_31781 = x_31776 || x_31778;\n            \n            ((__local int32_t *) scan_arr_mem_33156)[local_tid_33152] =\n                res_31779;\n            ((__local bool *) scan_arr_mem_33158)[local_tid_33152] = res_31781;\n        }\n        \n        int32_t x_33144;\n        bool x_33145;\n        int32_t x_33146;\n        bool x_33147;\n        int32_t x_33163;\n        bool x_33164;\n        int32_t x_33165;\n        bool x_33166;\n        int32_t skip_threads_33170;\n        \n        if (slt32(local_tid_33152, segscan_group_sizze_32185)) {\n            x_33146 = ((volatile __local\n                        int32_t *) scan_arr_mem_33156)[local_tid_33152];\n            x_33147 = ((volatile __local\n                        bool *) scan_arr_mem_33158)[local_tid_33152];\n        }\n        // in-block scan (hopefully no barriers needed)\n        {\n            skip_threads_33170 = 1;\n            while (slt32(skip_threads_33170, 32)) {\n                if (sle32(skip_threads_33170, local_tid_33152 -\n                          squot32(local_tid_33152, 32) * 32) &&\n                    slt32(local_tid_33152, segscan_group_sizze_32185)) {\n                    // read operands\n                    {\n                        x_33144 = ((volatile __local\n                                    int32_t *) scan_arr_mem_33156)[local_tid_33152 -\n                                                                   skip_threads_33170];\n                        x_33145 = ((volatile __local\n                                    bool *) scan_arr_mem_33158)[local_tid_33152 -\n                      ",
            "                                          skip_threads_33170];\n                    }\n                    // perform operation\n                    {\n                        int32_t res_33148;\n                        \n                        if (x_33147) {\n                            res_33148 = x_33146;\n                        } else {\n                            int32_t res_33149 = x_33144 + x_33146;\n                            \n                            res_33148 = res_33149;\n                        }\n                        \n                        bool res_33150 = x_33145 || x_33147;\n                        \n                        x_33146 = res_33148;\n                        x_33147 = res_33150;\n                    }\n                }\n                if (sle32(wave_sizze_33154, skip_threads_33170)) {\n                    barrier(CLK_LOCAL_MEM_FENCE);\n                }\n                if (sle32(skip_threads_33170, local_tid_33152 -\n                          squot32(local_tid_33152, 32) * 32) &&\n                    slt32(local_tid_33152, segscan_group_sizze_32185)) {\n                    // write result\n                    {\n                        ((volatile __local\n                          int32_t *) scan_arr_mem_33156)[local_tid_33152] =\n                            x_33146;\n                        ((volatile __local\n                          bool *) scan_arr_mem_33158)[local_tid_33152] =\n                            x_33147;\n                    }\n                }\n                if (sle32(wave_sizze_33154, skip_threads_33170)) {\n                    barrier(CLK_LOCAL_MEM_FENCE);\n                }\n                skip_threads_33170 *= 2;\n            }\n        }\n        barrier(CLK_LOCAL_MEM_FENCE);\n        // last thread of block 'i' writes its result to offset 'i'\n        {\n            if ((local_tid_33152 - squot32(local_tid_33152, 32) * 32) == 31 &&\n                slt32(local_tid_33152, segscan_group_sizze_32185)) {\n                ((volatile __local\n       ",
            "           int32_t *) scan_arr_mem_33156)[squot32(local_tid_33152, 32)] =\n                    x_33146;\n                ((volatile __local\n                  bool *) scan_arr_mem_33158)[squot32(local_tid_33152, 32)] =\n                    x_33147;\n            }\n        }\n        barrier(CLK_LOCAL_MEM_FENCE);\n        // scan the first block, after which offset 'i' contains carry-in for warp 'i+1'\n        {\n            int32_t skip_threads_33171;\n            \n            if (squot32(local_tid_33152, 32) == 0 && slt32(local_tid_33152,\n                                                           segscan_group_sizze_32185)) {\n                x_33165 = ((volatile __local\n                            int32_t *) scan_arr_mem_33156)[local_tid_33152];\n                x_33166 = ((volatile __local\n                            bool *) scan_arr_mem_33158)[local_tid_33152];\n            }\n            // in-block scan (hopefully no barriers needed)\n            {\n                skip_threads_33171 = 1;\n                while (slt32(skip_threads_33171, 32)) {\n                    if (sle32(skip_threads_33171, local_tid_33152 -\n                              squot32(local_tid_33152, 32) * 32) &&\n                        (squot32(local_tid_33152, 32) == 0 &&\n                         slt32(local_tid_33152, segscan_group_sizze_32185))) {\n                        // read operands\n                        {\n                            x_33163 = ((volatile __local\n                                        int32_t *) scan_arr_mem_33156)[local_tid_33152 -\n                                                                       skip_threads_33171];\n                            x_33164 = ((volatile __local\n                                        bool *) scan_arr_mem_33158)[local_tid_33152 -\n                                                                    skip_threads_33171];\n                        }\n                        // perform operation\n                        {\n                            int32_t res_",
            "33167;\n                            \n                            if (x_33166) {\n                                res_33167 = x_33165;\n                            } else {\n                                int32_t res_33168 = x_33163 + x_33165;\n                                \n                                res_33167 = res_33168;\n                            }\n                            \n                            bool res_33169 = x_33164 || x_33166;\n                            \n                            x_33165 = res_33167;\n                            x_33166 = res_33169;\n                        }\n                    }\n                    if (sle32(wave_sizze_33154, skip_threads_33171)) {\n                        barrier(CLK_LOCAL_MEM_FENCE);\n                    }\n                    if (sle32(skip_threads_33171, local_tid_33152 -\n                              squot32(local_tid_33152, 32) * 32) &&\n                        (squot32(local_tid_33152, 32) == 0 &&\n                         slt32(local_tid_33152, segscan_group_sizze_32185))) {\n                        // write result\n                        {\n                            ((volatile __local\n                              int32_t *) scan_arr_mem_33156)[local_tid_33152] =\n                                x_33165;\n                            ((volatile __local\n                              bool *) scan_arr_mem_33158)[local_tid_33152] =\n                                x_33166;\n                        }\n                    }\n                    if (sle32(wave_sizze_33154, skip_threads_33171)) {\n                        barrier(CLK_LOCAL_MEM_FENCE);\n                    }\n                    skip_threads_33171 *= 2;\n                }\n            }\n        }\n        barrier(CLK_LOCAL_MEM_FENCE);\n        // carry-in for every block except the first\n        {\n            if (!(squot32(local_tid_33152, 32) == 0 || !slt32(local_tid_33152,\n                                                              segscan_group_sizze_32185)",
            ")) {\n                // read operands\n                {\n                    x_33144 = ((volatile __local\n                                int32_t *) scan_arr_mem_33156)[squot32(local_tid_33152,\n                                                                       32) - 1];\n                    x_33145 = ((volatile __local\n                                bool *) scan_arr_mem_33158)[squot32(local_tid_33152,\n                                                                    32) - 1];\n                }\n                // perform operation\n                {\n                    int32_t res_33148;\n                    \n                    if (x_33147) {\n                        res_33148 = x_33146;\n                    } else {\n                        int32_t res_33149 = x_33144 + x_33146;\n                        \n                        res_33148 = res_33149;\n                    }\n                    \n                    bool res_33150 = x_33145 || x_33147;\n                    \n                    x_33146 = res_33148;\n                    x_33147 = res_33150;\n                }\n                // write final result\n                {\n                    ((volatile __local\n                      int32_t *) scan_arr_mem_33156)[local_tid_33152] = x_33146;\n                    ((volatile __local\n                      bool *) scan_arr_mem_33158)[local_tid_33152] = x_33147;\n                }\n            }\n        }\n        barrier(CLK_LOCAL_MEM_FENCE);\n        // restore correct values for first block\n        {\n            if (squot32(local_tid_33152, 32) == 0) {\n                ((volatile __local\n                  int32_t *) scan_arr_mem_33156)[local_tid_33152] = x_33146;\n                ((volatile __local\n                  bool *) scan_arr_mem_33158)[local_tid_33152] = x_33147;\n            }\n        }\n        barrier(CLK_LOCAL_MEM_FENCE);\n        // threads in bounds write partial scan result\n        {\n            if (slt32(gtid_32189, flat_dim_31615)) {\n                ((__global in",
            "t32_t *) mem_32815)[gtid_32189] = ((__local\n                                                                 int32_t *) scan_arr_mem_33156)[local_tid_33152];\n                ((__global bool *) mem_32817)[gtid_32189] = ((__local\n                                                              bool *) scan_arr_mem_33158)[local_tid_33152];\n            }\n        }\n        barrier(CLK_LOCAL_MEM_FENCE);\n        // first thread reads last element as carry-in for next iteration\n        {\n            if (local_tid_33152 == 0) {\n                x_31775 = ((__local\n                            int32_t *) scan_arr_mem_33156)[segscan_group_sizze_32185 -\n                                                           1];\n                x_31776 = ((__local\n                            bool *) scan_arr_mem_33158)[segscan_group_sizze_32185 -\n                                                        1];\n            }\n        }\n        barrier(CLK_LOCAL_MEM_FENCE);\n    }\n}\n__kernel void scan_stage1_32199(uint scan_arr_mem_33225_backing_offset_0,\n                                uint scan_arr_mem_33227_backing_offset_1,\n                                int32_t flat_dim_31615, __global\n                                unsigned char *x_mem_32791, __global\n                                unsigned char *mem_32805, __global\n                                unsigned char *mem_32821, __global\n                                unsigned char *mem_32823,\n                                int32_t num_threads_33212)\n{\n    const int32_t segscan_group_sizze_32194 =\n                  quicksort1zisegscan_group_sizze_32193;\n    const int block_dim0 = 0;\n    const int block_dim1 = 1;\n    const int block_dim2 = 2;\n    volatile char *scan_arr_mem_33225_backing_0 =\n                  &shared_mem[scan_arr_mem_33225_backing_offset_0];\n    volatile char *scan_arr_mem_33227_backing_1 =\n                  &shared_mem[scan_arr_mem_33227_backing_offset_1];\n    int32_t global_tid_33220;\n    int32_t local_tid_33221;\n    int32_t group_s",
            "izze_33224;\n    int32_t wave_sizze_33223;\n    int32_t group_tid_33222;\n    \n    global_tid_33220 = get_global_id(0);\n    local_tid_33221 = get_local_id(0);\n    group_sizze_33224 = get_local_size(0);\n    wave_sizze_33223 = LOCKSTEP_WIDTH;\n    group_tid_33222 = get_group_id(0);\n    \n    int32_t phys_tid_32199 = global_tid_33220;\n    __local char *scan_arr_mem_33225;\n    \n    scan_arr_mem_33225 = (__local char *) scan_arr_mem_33225_backing_0;\n    \n    __local char *scan_arr_mem_33227;\n    \n    scan_arr_mem_33227 = (__local char *) scan_arr_mem_33227_backing_1;\n    \n    int32_t x_31795;\n    bool x_31796;\n    int32_t x_31797;\n    bool x_31798;\n    \n    x_31795 = 0;\n    x_31796 = 0;\n    for (int32_t j_33229 = 0; j_33229 < squot32(flat_dim_31615 +\n                                                num_threads_33212 - 1,\n                                                num_threads_33212); j_33229++) {\n        int32_t chunk_offset_33230 = segscan_group_sizze_32194 * j_33229 +\n                group_tid_33222 * (segscan_group_sizze_32194 *\n                                   squot32(flat_dim_31615 + num_threads_33212 -\n                                           1, num_threads_33212));\n        int32_t flat_idx_33231 = chunk_offset_33230 + local_tid_33221;\n        int32_t gtid_32198 = flat_idx_33231;\n        \n        // threads in bounds read input; others get neutral element\n        {\n            if (slt32(gtid_32198, flat_dim_31615)) {\n                bool x_31802 = ((__global bool *) mem_32805)[gtid_32198];\n                bool x_31803 = ((__global bool *) x_mem_32791)[gtid_32198];\n                int32_t res_31804;\n                \n                if (x_31802) {\n                    res_31804 = 0;\n                } else {\n                    res_31804 = 1;\n                }\n                // write to-scan values to parameters\n                {\n                    x_31797 = res_31804;\n                    x_31798 = x_31803;\n                }\n                // write mapped values r",
            "esults to global memory\n                { }\n            } else {\n                x_31797 = 0;\n                x_31798 = 0;\n            }\n        }\n        // combine with carry and write to local memory\n        {\n            int32_t res_31799;\n            \n            if (x_31798) {\n                res_31799 = x_31797;\n            } else {\n                int32_t res_31800 = x_31795 + x_31797;\n                \n                res_31799 = res_31800;\n            }\n            \n            bool res_31801 = x_31796 || x_31798;\n            \n            ((__local int32_t *) scan_arr_mem_33225)[local_tid_33221] =\n                res_31799;\n            ((__local bool *) scan_arr_mem_33227)[local_tid_33221] = res_31801;\n        }\n        \n        int32_t x_33213;\n        bool x_33214;\n        int32_t x_33215;\n        bool x_33216;\n        int32_t x_33232;\n        bool x_33233;\n        int32_t x_33234;\n        bool x_33235;\n        int32_t skip_threads_33239;\n        \n        if (slt32(local_tid_33221, segscan_group_sizze_32194)) {\n            x_33215 = ((volatile __local\n                        int32_t *) scan_arr_mem_33225)[local_tid_33221];\n            x_33216 = ((volatile __local\n                        bool *) scan_arr_mem_33227)[local_tid_33221];\n        }\n        // in-block scan (hopefully no barriers needed)\n        {\n            skip_threads_33239 = 1;\n            while (slt32(skip_threads_33239, 32)) {\n                if (sle32(skip_threads_33239, local_tid_33221 -\n                          squot32(local_tid_33221, 32) * 32) &&\n                    slt32(local_tid_33221, segscan_group_sizze_32194)) {\n                    // read operands\n                    {\n                        x_33213 = ((volatile __local\n                                    int32_t *) scan_arr_mem_33225)[local_tid_33221 -\n                                                                   skip_threads_33239];\n                        x_33214 = ((volatile __local\n                                  ",
            "  bool *) scan_arr_mem_33227)[local_tid_33221 -\n                                                                skip_threads_33239];\n                    }\n                    // perform operation\n                    {\n                        int32_t res_33217;\n                        \n                        if (x_33216) {\n                            res_33217 = x_33215;\n                        } else {\n                            int32_t res_33218 = x_33213 + x_33215;\n                            \n                            res_33217 = res_33218;\n                        }\n                        \n                        bool res_33219 = x_33214 || x_33216;\n                        \n                        x_33215 = res_33217;\n                        x_33216 = res_33219;\n                    }\n                }\n                if (sle32(wave_sizze_33223, skip_threads_33239)) {\n                    barrier(CLK_LOCAL_MEM_FENCE);\n                }\n                if (sle32(skip_threads_33239, local_tid_33221 -\n                          squot32(local_tid_33221, 32) * 32) &&\n                    slt32(local_tid_33221, segscan_group_sizze_32194)) {\n                    // write result\n                    {\n                        ((volatile __local\n                          int32_t *) scan_arr_mem_33225)[local_tid_33221] =\n                            x_33215;\n                        ((volatile __local\n                          bool *) scan_arr_mem_33227)[local_tid_33221] =\n                            x_33216;\n                    }\n                }\n                if (sle32(wave_sizze_33223, skip_threads_33239)) {\n                    barrier(CLK_LOCAL_MEM_FENCE);\n                }\n                skip_threads_33239 *= 2;\n            }\n        }\n        barrier(CLK_LOCAL_MEM_FENCE);\n        // last thread of block 'i' writes its result to offset 'i'\n        {\n            if ((local_tid_33221 - squot32(local_tid_33221, 32) * 32) == 31 &&\n                slt32(local_tid_33221, se",
            "gscan_group_sizze_32194)) {\n                ((volatile __local\n                  int32_t *) scan_arr_mem_33225)[squot32(local_tid_33221, 32)] =\n                    x_33215;\n                ((volatile __local\n                  bool *) scan_arr_mem_33227)[squot32(local_tid_33221, 32)] =\n                    x_33216;\n            }\n        }\n        barrier(CLK_LOCAL_MEM_FENCE);\n        // scan the first block, after which offset 'i' contains carry-in for warp 'i+1'\n        {\n            int32_t skip_threads_33240;\n            \n            if (squot32(local_tid_33221, 32) == 0 && slt32(local_tid_33221,\n                                                           segscan_group_sizze_32194)) {\n                x_33234 = ((volatile __local\n                            int32_t *) scan_arr_mem_33225)[local_tid_33221];\n                x_33235 = ((volatile __local\n                            bool *) scan_arr_mem_33227)[local_tid_33221];\n            }\n            // in-block scan (hopefully no barriers needed)\n            {\n                skip_threads_33240 = 1;\n                while (slt32(skip_threads_33240, 32)) {\n                    if (sle32(skip_threads_33240, local_tid_33221 -\n                              squot32(local_tid_33221, 32) * 32) &&\n                        (squot32(local_tid_33221, 32) == 0 &&\n                         slt32(local_tid_33221, segscan_group_sizze_32194))) {\n                        // read operands\n                        {\n                            x_33232 = ((volatile __local\n                                        int32_t *) scan_arr_mem_33225)[local_tid_33221 -\n                                                                       skip_threads_33240];\n                            x_33233 = ((volatile __local\n                                        bool *) scan_arr_mem_33227)[local_tid_33221 -\n                                                                    skip_threads_33240];\n                        }\n                        // perform operat",
            "ion\n                        {\n                            int32_t res_33236;\n                            \n                            if (x_33235) {\n                                res_33236 = x_33234;\n                            } else {\n                                int32_t res_33237 = x_33232 + x_33234;\n                                \n                                res_33236 = res_33237;\n                            }\n                            \n                            bool res_33238 = x_33233 || x_33235;\n                            \n                            x_33234 = res_33236;\n                            x_33235 = res_33238;\n                        }\n                    }\n                    if (sle32(wave_sizze_33223, skip_threads_33240)) {\n                        barrier(CLK_LOCAL_MEM_FENCE);\n                    }\n                    if (sle32(skip_threads_33240, local_tid_33221 -\n                              squot32(local_tid_33221, 32) * 32) &&\n                        (squot32(local_tid_33221, 32) == 0 &&\n                         slt32(local_tid_33221, segscan_group_sizze_32194))) {\n                        // write result\n                        {\n                            ((volatile __local\n                              int32_t *) scan_arr_mem_33225)[local_tid_33221] =\n                                x_33234;\n                            ((volatile __local\n                              bool *) scan_arr_mem_33227)[local_tid_33221] =\n                                x_33235;\n                        }\n                    }\n                    if (sle32(wave_sizze_33223, skip_threads_33240)) {\n                        barrier(CLK_LOCAL_MEM_FENCE);\n                    }\n                    skip_threads_33240 *= 2;\n                }\n            }\n        }\n        barrier(CLK_LOCAL_MEM_FENCE);\n        // carry-in for every block except the first\n        {\n            if (!(squot32(local_tid_33221, 32) == 0 || !slt32(local_tid_33221,\n                  ",
            "                                            segscan_group_sizze_32194))) {\n                // read operands\n                {\n                    x_33213 = ((volatile __local\n                                int32_t *) scan_arr_mem_33225)[squot32(local_tid_33221,\n                                                                       32) - 1];\n                    x_33214 = ((volatile __local\n                                bool *) scan_arr_mem_33227)[squot32(local_tid_33221,\n                                                                    32) - 1];\n                }\n                // perform operation\n                {\n                    int32_t res_33217;\n                    \n                    if (x_33216) {\n                        res_33217 = x_33215;\n                    } else {\n                        int32_t res_33218 = x_33213 + x_33215;\n                        \n                        res_33217 = res_33218;\n                    }\n                    \n                    bool res_33219 = x_33214 || x_33216;\n                    \n                    x_33215 = res_33217;\n                    x_33216 = res_33219;\n                }\n                // write final result\n                {\n                    ((volatile __local\n                      int32_t *) scan_arr_mem_33225)[local_tid_33221] = x_33215;\n                    ((volatile __local\n                      bool *) scan_arr_mem_33227)[local_tid_33221] = x_33216;\n                }\n            }\n        }\n        barrier(CLK_LOCAL_MEM_FENCE);\n        // restore correct values for first block\n        {\n            if (squot32(local_tid_33221, 32) == 0) {\n                ((volatile __local\n                  int32_t *) scan_arr_mem_33225)[local_tid_33221] = x_33215;\n                ((volatile __local\n                  bool *) scan_arr_mem_33227)[local_tid_33221] = x_33216;\n            }\n        }\n        barrier(CLK_LOCAL_MEM_FENCE);\n        // threads in bounds write partial scan result\n        {\n            ",
            "if (slt32(gtid_32198, flat_dim_31615)) {\n                ((__global int32_t *) mem_32821)[gtid_32198] = ((__local\n                                                                 int32_t *) scan_arr_mem_33225)[local_tid_33221];\n                ((__global bool *) mem_32823)[gtid_32198] = ((__local\n                                                              bool *) scan_arr_mem_33227)[local_tid_33221];\n            }\n        }\n        barrier(CLK_LOCAL_MEM_FENCE);\n        // first thread reads last element as carry-in for next iteration\n        {\n            if (local_tid_33221 == 0) {\n                x_31795 = ((__local\n                            int32_t *) scan_arr_mem_33225)[segscan_group_sizze_32194 -\n                                                           1];\n                x_31796 = ((__local\n                            bool *) scan_arr_mem_33227)[segscan_group_sizze_32194 -\n                                                        1];\n            }\n        }\n        barrier(CLK_LOCAL_MEM_FENCE);\n    }\n}\n__kernel void scan_stage1_32225(uint scan_arr_mem_33298_backing_offset_0,\n                                uint scan_arr_mem_33300_backing_offset_1,\n                                int32_t flat_dim_31615, int32_t m_31645,\n                                int32_t i_p_m_t_s_31665, __global\n                                unsigned char *x_mem_32791, __global\n                                unsigned char *mem_32826, __global\n                                unsigned char *mem_32830, __global\n                                unsigned char *mem_32832,\n                                int32_t num_threads_33286)\n{\n    const int32_t segscan_group_sizze_32220 =\n                  quicksort1zisegscan_group_sizze_32219;\n    const int block_dim0 = 0;\n    const int block_dim1 = 1;\n    const int block_dim2 = 2;\n    volatile char *scan_arr_mem_33298_backing_0 =\n                  &shared_mem[scan_arr_mem_33298_backing_offset_0];\n    volatile char *scan_arr_mem_33300_backing_1 =\n  ",
            "                &shared_mem[scan_arr_mem_33300_backing_offset_1];\n    int32_t global_tid_33293;\n    int32_t local_tid_33294;\n    int32_t group_sizze_33297;\n    int32_t wave_sizze_33296;\n    int32_t group_tid_33295;\n    \n    global_tid_33293 = get_global_id(0);\n    local_tid_33294 = get_local_id(0);\n    group_sizze_33297 = get_local_size(0);\n    wave_sizze_33296 = LOCKSTEP_WIDTH;\n    group_tid_33295 = get_group_id(0);\n    \n    int32_t phys_tid_32225 = global_tid_33293;\n    __local char *scan_arr_mem_33298;\n    \n    scan_arr_mem_33298 = (__local char *) scan_arr_mem_33298_backing_0;\n    \n    __local char *scan_arr_mem_33300;\n    \n    scan_arr_mem_33300 = (__local char *) scan_arr_mem_33300_backing_1;\n    \n    int32_t x_31826;\n    bool x_31827;\n    int32_t x_31828;\n    bool x_31829;\n    \n    x_31826 = 0;\n    x_31827 = 0;\n    for (int32_t j_33302 = 0; j_33302 < squot32(flat_dim_31615 +\n                                                num_threads_33286 - 1,\n                                                num_threads_33286); j_33302++) {\n        int32_t chunk_offset_33303 = segscan_group_sizze_32220 * j_33302 +\n                group_tid_33295 * (segscan_group_sizze_32220 *\n                                   squot32(flat_dim_31615 + num_threads_33286 -\n                                           1, num_threads_33286));\n        int32_t flat_idx_33304 = chunk_offset_33303 + local_tid_33294;\n        int32_t gtid_32224 = flat_idx_33304;\n        \n        // threads in bounds read input; others get neutral element\n        {\n            if (slt32(gtid_32224, flat_dim_31615)) {\n                int32_t j_t_s_32588 = -1 * gtid_32224;\n                int32_t j_p_i_t_s_32589 = m_31645 + j_t_s_32588;\n                int32_t x_31832 = ((__global\n                                    int32_t *) mem_32826)[j_p_i_t_s_32589];\n                int32_t j_p_i_t_s_32587 = i_p_m_t_s_31665 + j_t_s_32588;\n                bool x_31833 = ((__global bool *) x_mem_32791)[j_p_i_t_s_32587];\n                \n",
            "                // write to-scan values to parameters\n                {\n                    x_31828 = x_31832;\n                    x_31829 = x_31833;\n                }\n                // write mapped values results to global memory\n                { }\n            } else {\n                x_31828 = 0;\n                x_31829 = 0;\n            }\n        }\n        // combine with carry and write to local memory\n        {\n            int32_t res_31830;\n            \n            if (x_31829) {\n                res_31830 = x_31828;\n            } else {\n                res_31830 = x_31826;\n            }\n            \n            bool res_31831 = x_31827 || x_31829;\n            \n            ((__local int32_t *) scan_arr_mem_33298)[local_tid_33294] =\n                res_31830;\n            ((__local bool *) scan_arr_mem_33300)[local_tid_33294] = res_31831;\n        }\n        \n        int32_t x_33287;\n        bool x_33288;\n        int32_t x_33289;\n        bool x_33290;\n        int32_t x_33305;\n        bool x_33306;\n        int32_t x_33307;\n        bool x_33308;\n        int32_t skip_threads_33311;\n        \n        if (slt32(local_tid_33294, segscan_group_sizze_32220)) {\n            x_33289 = ((volatile __local\n                        int32_t *) scan_arr_mem_33298)[local_tid_33294];\n            x_33290 = ((volatile __local\n                        bool *) scan_arr_mem_33300)[local_tid_33294];\n        }\n        // in-block scan (hopefully no barriers needed)\n        {\n            skip_threads_33311 = 1;\n            while (slt32(skip_threads_33311, 32)) {\n                if (sle32(skip_threads_33311, local_tid_33294 -\n                          squot32(local_tid_33294, 32) * 32) &&\n                    slt32(local_tid_33294, segscan_group_sizze_32220)) {\n                    // read operands\n                    {\n                        x_33287 = ((volatile __local\n                                    int32_t *) scan_arr_mem_33298)[local_tid_33294 -\n                                         ",
            "                          skip_threads_33311];\n                        x_33288 = ((volatile __local\n                                    bool *) scan_arr_mem_33300)[local_tid_33294 -\n                                                                skip_threads_33311];\n                    }\n                    // perform operation\n                    {\n                        int32_t res_33291;\n                        \n                        if (x_33290) {\n                            res_33291 = x_33289;\n                        } else {\n                            res_33291 = x_33287;\n                        }\n                        \n                        bool res_33292 = x_33288 || x_33290;\n                        \n                        x_33289 = res_33291;\n                        x_33290 = res_33292;\n                    }\n                }\n                if (sle32(wave_sizze_33296, skip_threads_33311)) {\n                    barrier(CLK_LOCAL_MEM_FENCE);\n                }\n                if (sle32(skip_threads_33311, local_tid_33294 -\n                          squot32(local_tid_33294, 32) * 32) &&\n                    slt32(local_tid_33294, segscan_group_sizze_32220)) {\n                    // write result\n                    {\n                        ((volatile __local\n                          int32_t *) scan_arr_mem_33298)[local_tid_33294] =\n                            x_33289;\n                        ((volatile __local\n                          bool *) scan_arr_mem_33300)[local_tid_33294] =\n                            x_33290;\n                    }\n                }\n                if (sle32(wave_sizze_33296, skip_threads_33311)) {\n                    barrier(CLK_LOCAL_MEM_FENCE);\n                }\n                skip_threads_33311 *= 2;\n            }\n        }\n        barrier(CLK_LOCAL_MEM_FENCE);\n        // last thread of block 'i' writes its result to offset 'i'\n        {\n            if ((local_tid_33294 - squot32(local_tid_33294, 32) * 32) == 31 &&\n     ",
            "           slt32(local_tid_33294, segscan_group_sizze_32220)) {\n                ((volatile __local\n                  int32_t *) scan_arr_mem_33298)[squot32(local_tid_33294, 32)] =\n                    x_33289;\n                ((volatile __local\n                  bool *) scan_arr_mem_33300)[squot32(local_tid_33294, 32)] =\n                    x_33290;\n            }\n        }\n        barrier(CLK_LOCAL_MEM_FENCE);\n        // scan the first block, after which offset 'i' contains carry-in for warp 'i+1'\n        {\n            int32_t skip_threads_33312;\n            \n            if (squot32(local_tid_33294, 32) == 0 && slt32(local_tid_33294,\n                                                           segscan_group_sizze_32220)) {\n                x_33307 = ((volatile __local\n                            int32_t *) scan_arr_mem_33298)[local_tid_33294];\n                x_33308 = ((volatile __local\n                            bool *) scan_arr_mem_33300)[local_tid_33294];\n            }\n            // in-block scan (hopefully no barriers needed)\n            {\n                skip_threads_33312 = 1;\n                while (slt32(skip_threads_33312, 32)) {\n                    if (sle32(skip_threads_33312, local_tid_33294 -\n                              squot32(local_tid_33294, 32) * 32) &&\n                        (squot32(local_tid_33294, 32) == 0 &&\n                         slt32(local_tid_33294, segscan_group_sizze_32220))) {\n                        // read operands\n                        {\n                            x_33305 = ((volatile __local\n                                        int32_t *) scan_arr_mem_33298)[local_tid_33294 -\n                                                                       skip_threads_33312];\n                            x_33306 = ((volatile __local\n                                        bool *) scan_arr_mem_33300)[local_tid_33294 -\n                                                                    skip_threads_33312];\n                        }\n     ",
            "                   // perform operation\n                        {\n                            int32_t res_33309;\n                            \n                            if (x_33308) {\n                                res_33309 = x_33307;\n                            } else {\n                                res_33309 = x_33305;\n                            }\n                            \n                            bool res_33310 = x_33306 || x_33308;\n                            \n                            x_33307 = res_33309;\n                            x_33308 = res_33310;\n                        }\n                    }\n                    if (sle32(wave_sizze_33296, skip_threads_33312)) {\n                        barrier(CLK_LOCAL_MEM_FENCE);\n                    }\n                    if (sle32(skip_threads_33312, local_tid_33294 -\n                              squot32(local_tid_33294, 32) * 32) &&\n                        (squot32(local_tid_33294, 32) == 0 &&\n                         slt32(local_tid_33294, segscan_group_sizze_32220))) {\n                        // write result\n                        {\n                            ((volatile __local\n                              int32_t *) scan_arr_mem_33298)[local_tid_33294] =\n                                x_33307;\n                            ((volatile __local\n                              bool *) scan_arr_mem_33300)[local_tid_33294] =\n                                x_33308;\n                        }\n                    }\n                    if (sle32(wave_sizze_33296, skip_threads_33312)) {\n                        barrier(CLK_LOCAL_MEM_FENCE);\n                    }\n                    skip_threads_33312 *= 2;\n                }\n            }\n        }\n        barrier(CLK_LOCAL_MEM_FENCE);\n        // carry-in for every block except the first\n        {\n            if (!(squot32(local_tid_33294, 32) == 0 || !slt32(local_tid_33294,\n                                                              segscan_group_sizze_32220)",
            ")) {\n                // read operands\n                {\n                    x_33287 = ((volatile __local\n                                int32_t *) scan_arr_mem_33298)[squot32(local_tid_33294,\n                                                                       32) - 1];\n                    x_33288 = ((volatile __local\n                                bool *) scan_arr_mem_33300)[squot32(local_tid_33294,\n                                                                    32) - 1];\n                }\n                // perform operation\n                {\n                    int32_t res_33291;\n                    \n                    if (x_33290) {\n                        res_33291 = x_33289;\n                    } else {\n                        res_33291 = x_33287;\n                    }\n                    \n                    bool res_33292 = x_33288 || x_33290;\n                    \n                    x_33289 = res_33291;\n                    x_33290 = res_33292;\n                }\n                // write final result\n                {\n                    ((volatile __local\n                      int32_t *) scan_arr_mem_33298)[local_tid_33294] = x_33289;\n                    ((volatile __local\n                      bool *) scan_arr_mem_33300)[local_tid_33294] = x_33290;\n                }\n            }\n        }\n        barrier(CLK_LOCAL_MEM_FENCE);\n        // restore correct values for first block\n        {\n            if (squot32(local_tid_33294, 32) == 0) {\n                ((volatile __local\n                  int32_t *) scan_arr_mem_33298)[local_tid_33294] = x_33289;\n                ((volatile __local\n                  bool *) scan_arr_mem_33300)[local_tid_33294] = x_33290;\n            }\n        }\n        barrier(CLK_LOCAL_MEM_FENCE);\n        // threads in bounds write partial scan result\n        {\n            if (slt32(gtid_32224, flat_dim_31615)) {\n                ((__global int32_t *) mem_32830)[gtid_32224] = ((__local\n                                              ",
            "                   int32_t *) scan_arr_mem_33298)[local_tid_33294];\n                ((__global bool *) mem_32832)[gtid_32224] = ((__local\n                                                              bool *) scan_arr_mem_33300)[local_tid_33294];\n            }\n        }\n        barrier(CLK_LOCAL_MEM_FENCE);\n        // first thread reads last element as carry-in for next iteration\n        {\n            if (local_tid_33294 == 0) {\n                x_31826 = ((__local\n                            int32_t *) scan_arr_mem_33298)[segscan_group_sizze_32220 -\n                                                           1];\n                x_31827 = ((__local\n                            bool *) scan_arr_mem_33300)[segscan_group_sizze_32220 -\n                                                        1];\n            }\n        }\n        barrier(CLK_LOCAL_MEM_FENCE);\n    }\n}\n__kernel void scan_stage2_32141(uint scan_arr_mem_33049_backing_offset_0,\n                                uint scan_arr_mem_33051_backing_offset_1,\n                                int32_t flat_dim_31615,\n                                int32_t num_groups_32138, __global\n                                unsigned char *mem_32795, __global\n                                unsigned char *mem_32797,\n                                int32_t num_threads_33005)\n{\n    const int32_t segscan_group_sizze_32136 =\n                  quicksort1zisegscan_group_sizze_32135;\n    const int block_dim0 = 0;\n    const int block_dim1 = 1;\n    const int block_dim2 = 2;\n    volatile char *scan_arr_mem_33049_backing_0 =\n                  &shared_mem[scan_arr_mem_33049_backing_offset_0];\n    volatile char *scan_arr_mem_33051_backing_1 =\n                  &shared_mem[scan_arr_mem_33051_backing_offset_1];\n    int32_t global_tid_33044;\n    int32_t local_tid_33045;\n    int32_t group_sizze_33048;\n    int32_t wave_sizze_33047;\n    int32_t group_tid_33046;\n    \n    global_tid_33044 = get_global_id(0);\n    local_tid_33045 = get_local_id(0);\n    group_si",
            "zze_33048 = get_local_size(0);\n    wave_sizze_33047 = LOCKSTEP_WIDTH;\n    group_tid_33046 = get_group_id(0);\n    \n    int32_t phys_tid_32141 = global_tid_33044;\n    __local char *scan_arr_mem_33049;\n    \n    scan_arr_mem_33049 = (__local char *) scan_arr_mem_33049_backing_0;\n    \n    __local char *scan_arr_mem_33051;\n    \n    scan_arr_mem_33051 = (__local char *) scan_arr_mem_33051_backing_1;\n    \n    int32_t flat_idx_33053 = (local_tid_33045 + 1) *\n            (segscan_group_sizze_32136 * squot32(flat_dim_31615 +\n                                                 num_threads_33005 - 1,\n                                                 num_threads_33005)) - 1;\n    int32_t gtid_32140 = flat_idx_33053;\n    \n    // threads in bound read carries; others get neutral element\n    {\n        if (slt32(gtid_32140, flat_dim_31615)) {\n            ((__local int32_t *) scan_arr_mem_33049)[local_tid_33045] =\n                ((__global int32_t *) mem_32795)[gtid_32140];\n            ((__local bool *) scan_arr_mem_33051)[local_tid_33045] = ((__global\n                                                                       bool *) mem_32797)[gtid_32140];\n        } else {\n            ((__local int32_t *) scan_arr_mem_33049)[local_tid_33045] = 0;\n            ((__local bool *) scan_arr_mem_33051)[local_tid_33045] = 0;\n        }\n    }\n    \n    int32_t x_33032;\n    bool x_33033;\n    int32_t x_33034;\n    bool x_33035;\n    int32_t x_33054;\n    bool x_33055;\n    int32_t x_33056;\n    bool x_33057;\n    int32_t skip_threads_33060;\n    \n    if (slt32(local_tid_33045, num_groups_32138)) {\n        x_33034 = ((volatile __local\n                    int32_t *) scan_arr_mem_33049)[local_tid_33045];\n        x_33035 = ((volatile __local\n                    bool *) scan_arr_mem_33051)[local_tid_33045];\n    }\n    // in-block scan (hopefully no barriers needed)\n    {\n        skip_threads_33060 = 1;\n        while (slt32(skip_threads_33060, 32)) {\n            if (sle32(skip_threads_33060, local_tid_33045 -\n        ",
            "              squot32(local_tid_33045, 32) * 32) &&\n                slt32(local_tid_33045, num_groups_32138)) {\n                // read operands\n                {\n                    x_33032 = ((volatile __local\n                                int32_t *) scan_arr_mem_33049)[local_tid_33045 -\n                                                               skip_threads_33060];\n                    x_33033 = ((volatile __local\n                                bool *) scan_arr_mem_33051)[local_tid_33045 -\n                                                            skip_threads_33060];\n                }\n                // perform operation\n                {\n                    int32_t res_33036;\n                    \n                    if (x_33035) {\n                        res_33036 = x_33034;\n                    } else {\n                        res_33036 = x_33032;\n                    }\n                    \n                    bool res_33037 = x_33033 || x_33035;\n                    \n                    x_33034 = res_33036;\n                    x_33035 = res_33037;\n                }\n            }\n            if (sle32(wave_sizze_33047, skip_threads_33060)) {\n                barrier(CLK_LOCAL_MEM_FENCE);\n            }\n            if (sle32(skip_threads_33060, local_tid_33045 -\n                      squot32(local_tid_33045, 32) * 32) &&\n                slt32(local_tid_33045, num_groups_32138)) {\n                // write result\n                {\n                    ((volatile __local\n                      int32_t *) scan_arr_mem_33049)[local_tid_33045] = x_33034;\n                    ((volatile __local\n                      bool *) scan_arr_mem_33051)[local_tid_33045] = x_33035;\n                }\n            }\n            if (sle32(wave_sizze_33047, skip_threads_33060)) {\n                barrier(CLK_LOCAL_MEM_FENCE);\n            }\n            skip_threads_33060 *= 2;\n        }\n    }\n    barrier(CLK_LOCAL_MEM_FENCE);\n    // last thread of block 'i' writes its result to offset '",
            "i'\n    {\n        if ((local_tid_33045 - squot32(local_tid_33045, 32) * 32) == 31 &&\n            slt32(local_tid_33045, num_groups_32138)) {\n            ((volatile __local\n              int32_t *) scan_arr_mem_33049)[squot32(local_tid_33045, 32)] =\n                x_33034;\n            ((volatile __local\n              bool *) scan_arr_mem_33051)[squot32(local_tid_33045, 32)] =\n                x_33035;\n        }\n    }\n    barrier(CLK_LOCAL_MEM_FENCE);\n    // scan the first block, after which offset 'i' contains carry-in for warp 'i+1'\n    {\n        int32_t skip_threads_33061;\n        \n        if (squot32(local_tid_33045, 32) == 0 && slt32(local_tid_33045,\n                                                       num_groups_32138)) {\n            x_33056 = ((volatile __local\n                        int32_t *) scan_arr_mem_33049)[local_tid_33045];\n            x_33057 = ((volatile __local\n                        bool *) scan_arr_mem_33051)[local_tid_33045];\n        }\n        // in-block scan (hopefully no barriers needed)\n        {\n            skip_threads_33061 = 1;\n            while (slt32(skip_threads_33061, 32)) {\n                if (sle32(skip_threads_33061, local_tid_33045 -\n                          squot32(local_tid_33045, 32) * 32) &&\n                    (squot32(local_tid_33045, 32) == 0 && slt32(local_tid_33045,\n                                                                num_groups_32138))) {\n                    // read operands\n                    {\n                        x_33054 = ((volatile __local\n                                    int32_t *) scan_arr_mem_33049)[local_tid_33045 -\n                                                                   skip_threads_33061];\n                        x_33055 = ((volatile __local\n                                    bool *) scan_arr_mem_33051)[local_tid_33045 -\n                                                                skip_threads_33061];\n                    }\n                    // perform operation\n           ",
            "         {\n                        int32_t res_33058;\n                        \n                        if (x_33057) {\n                            res_33058 = x_33056;\n                        } else {\n                            res_33058 = x_33054;\n                        }\n                        \n                        bool res_33059 = x_33055 || x_33057;\n                        \n                        x_33056 = res_33058;\n                        x_33057 = res_33059;\n                    }\n                }\n                if (sle32(wave_sizze_33047, skip_threads_33061)) {\n                    barrier(CLK_LOCAL_MEM_FENCE);\n                }\n                if (sle32(skip_threads_33061, local_tid_33045 -\n                          squot32(local_tid_33045, 32) * 32) &&\n                    (squot32(local_tid_33045, 32) == 0 && slt32(local_tid_33045,\n                                                                num_groups_32138))) {\n                    // write result\n                    {\n                        ((volatile __local\n                          int32_t *) scan_arr_mem_33049)[local_tid_33045] =\n                            x_33056;\n                        ((volatile __local\n                          bool *) scan_arr_mem_33051)[local_tid_33045] =\n                            x_33057;\n                    }\n                }\n                if (sle32(wave_sizze_33047, skip_threads_33061)) {\n                    barrier(CLK_LOCAL_MEM_FENCE);\n                }\n                skip_threads_33061 *= 2;\n            }\n        }\n    }\n    barrier(CLK_LOCAL_MEM_FENCE);\n    // carry-in for every block except the first\n    {\n        if (!(squot32(local_tid_33045, 32) == 0 || !slt32(local_tid_33045,\n                                                          num_groups_32138))) {\n            // read operands\n            {\n                x_33032 = ((volatile __local\n                            int32_t *) scan_arr_mem_33049)[squot32(local_tid_33045,\n                         ",
            "                                          32) - 1];\n                x_33033 = ((volatile __local\n                            bool *) scan_arr_mem_33051)[squot32(local_tid_33045,\n                                                                32) - 1];\n            }\n            // perform operation\n            {\n                int32_t res_33036;\n                \n                if (x_33035) {\n                    res_33036 = x_33034;\n                } else {\n                    res_33036 = x_33032;\n                }\n                \n                bool res_33037 = x_33033 || x_33035;\n                \n                x_33034 = res_33036;\n                x_33035 = res_33037;\n            }\n            // write final result\n            {\n                ((volatile __local\n                  int32_t *) scan_arr_mem_33049)[local_tid_33045] = x_33034;\n                ((volatile __local\n                  bool *) scan_arr_mem_33051)[local_tid_33045] = x_33035;\n            }\n        }\n    }\n    barrier(CLK_LOCAL_MEM_FENCE);\n    // restore correct values for first block\n    {\n        if (squot32(local_tid_33045, 32) == 0) {\n            ((volatile __local int32_t *) scan_arr_mem_33049)[local_tid_33045] =\n                x_33034;\n            ((volatile __local bool *) scan_arr_mem_33051)[local_tid_33045] =\n                x_33035;\n        }\n    }\n    barrier(CLK_LOCAL_MEM_FENCE);\n    // threads in bounds write scanned carries\n    {\n        if (slt32(gtid_32140, flat_dim_31615)) {\n            ((__global int32_t *) mem_32795)[gtid_32140] = ((__local\n                                                             int32_t *) scan_arr_mem_33049)[local_tid_33045];\n            ((__global bool *) mem_32797)[gtid_32140] = ((__local\n                                                          bool *) scan_arr_mem_33051)[local_tid_33045];\n        }\n    }\n}\n__kernel void scan_stage2_32150(uint scan_arr_mem_33113_backing_offset_0,\n                                uint scan_arr_mem_33115_backing_offs",
            "et_1,\n                                int32_t flat_dim_31615,\n                                int32_t num_groups_32147, __global\n                                unsigned char *mem_32801, __global\n                                unsigned char *mem_32803,\n                                int32_t num_threads_33069)\n{\n    const int32_t segscan_group_sizze_32145 =\n                  quicksort1zisegscan_group_sizze_32144;\n    const int block_dim0 = 0;\n    const int block_dim1 = 1;\n    const int block_dim2 = 2;\n    volatile char *scan_arr_mem_33113_backing_0 =\n                  &shared_mem[scan_arr_mem_33113_backing_offset_0];\n    volatile char *scan_arr_mem_33115_backing_1 =\n                  &shared_mem[scan_arr_mem_33115_backing_offset_1];\n    int32_t global_tid_33108;\n    int32_t local_tid_33109;\n    int32_t group_sizze_33112;\n    int32_t wave_sizze_33111;\n    int32_t group_tid_33110;\n    \n    global_tid_33108 = get_global_id(0);\n    local_tid_33109 = get_local_id(0);\n    group_sizze_33112 = get_local_size(0);\n    wave_sizze_33111 = LOCKSTEP_WIDTH;\n    group_tid_33110 = get_group_id(0);\n    \n    int32_t phys_tid_32150 = global_tid_33108;\n    __local char *scan_arr_mem_33113;\n    \n    scan_arr_mem_33113 = (__local char *) scan_arr_mem_33113_backing_0;\n    \n    __local char *scan_arr_mem_33115;\n    \n    scan_arr_mem_33115 = (__local char *) scan_arr_mem_33115_backing_1;\n    \n    int32_t flat_idx_33117 = (local_tid_33109 + 1) *\n            (segscan_group_sizze_32145 * squot32(flat_dim_31615 +\n                                                 num_threads_33069 - 1,\n                                                 num_threads_33069)) - 1;\n    int32_t gtid_32149 = flat_idx_33117;\n    \n    // threads in bound read carries; others get neutral element\n    {\n        if (slt32(gtid_32149, flat_dim_31615)) {\n            ((__local int32_t *) scan_arr_mem_33113)[local_tid_33109] =\n                ((__global int32_t *) mem_32801)[gtid_32149];\n            ((__local bool *) scan_arr_mem_3",
            "3115)[local_tid_33109] = ((__global\n                                                                       bool *) mem_32803)[gtid_32149];\n        } else {\n            ((__local int32_t *) scan_arr_mem_33113)[local_tid_33109] = 0;\n            ((__local bool *) scan_arr_mem_33115)[local_tid_33109] = 0;\n        }\n    }\n    \n    int32_t x_33096;\n    bool x_33097;\n    int32_t x_33098;\n    bool x_33099;\n    int32_t x_33118;\n    bool x_33119;\n    int32_t x_33120;\n    bool x_33121;\n    int32_t skip_threads_33124;\n    \n    if (slt32(local_tid_33109, num_groups_32147)) {\n        x_33098 = ((volatile __local\n                    int32_t *) scan_arr_mem_33113)[local_tid_33109];\n        x_33099 = ((volatile __local\n                    bool *) scan_arr_mem_33115)[local_tid_33109];\n    }\n    // in-block scan (hopefully no barriers needed)\n    {\n        skip_threads_33124 = 1;\n        while (slt32(skip_threads_33124, 32)) {\n            if (sle32(skip_threads_33124, local_tid_33109 -\n                      squot32(local_tid_33109, 32) * 32) &&\n                slt32(local_tid_33109, num_groups_32147)) {\n                // read operands\n                {\n                    x_33096 = ((volatile __local\n                                int32_t *) scan_arr_mem_33113)[local_tid_33109 -\n                                                               skip_threads_33124];\n                    x_33097 = ((volatile __local\n                                bool *) scan_arr_mem_33115)[local_tid_33109 -\n                                                            skip_threads_33124];\n                }\n                // perform operation\n                {\n                    int32_t res_33100;\n                    \n                    if (x_33099) {\n                        res_33100 = x_33098;\n                    } else {\n                        res_33100 = x_33096;\n                    }\n                    \n                    bool res_33101 = x_33097 || x_33099;\n                    \n                 ",
            "   x_33098 = res_33100;\n                    x_33099 = res_33101;\n                }\n            }\n            if (sle32(wave_sizze_33111, skip_threads_33124)) {\n                barrier(CLK_LOCAL_MEM_FENCE);\n            }\n            if (sle32(skip_threads_33124, local_tid_33109 -\n                      squot32(local_tid_33109, 32) * 32) &&\n                slt32(local_tid_33109, num_groups_32147)) {\n                // write result\n                {\n                    ((volatile __local\n                      int32_t *) scan_arr_mem_33113)[local_tid_33109] = x_33098;\n                    ((volatile __local\n                      bool *) scan_arr_mem_33115)[local_tid_33109] = x_33099;\n                }\n            }\n            if (sle32(wave_sizze_33111, skip_threads_33124)) {\n                barrier(CLK_LOCAL_MEM_FENCE);\n            }\n            skip_threads_33124 *= 2;\n        }\n    }\n    barrier(CLK_LOCAL_MEM_FENCE);\n    // last thread of block 'i' writes its result to offset 'i'\n    {\n        if ((local_tid_33109 - squot32(local_tid_33109, 32) * 32) == 31 &&\n            slt32(local_tid_33109, num_groups_32147)) {\n            ((volatile __local\n              int32_t *) scan_arr_mem_33113)[squot32(local_tid_33109, 32)] =\n                x_33098;\n            ((volatile __local\n              bool *) scan_arr_mem_33115)[squot32(local_tid_33109, 32)] =\n                x_33099;\n        }\n    }\n    barrier(CLK_LOCAL_MEM_FENCE);\n    // scan the first block, after which offset 'i' contains carry-in for warp 'i+1'\n    {\n        int32_t skip_threads_33125;\n        \n        if (squot32(local_tid_33109, 32) == 0 && slt32(local_tid_33109,\n                                                       num_groups_32147)) {\n            x_33120 = ((volatile __local\n                        int32_t *) scan_arr_mem_33113)[local_tid_33109];\n            x_33121 = ((volatile __local\n                        bool *) scan_arr_mem_33115)[local_tid_33109];\n        }\n        // in-block scan (hopefully no",
            " barriers needed)\n        {\n            skip_threads_33125 = 1;\n            while (slt32(skip_threads_33125, 32)) {\n                if (sle32(skip_threads_33125, local_tid_33109 -\n                          squot32(local_tid_33109, 32) * 32) &&\n                    (squot32(local_tid_33109, 32) == 0 && slt32(local_tid_33109,\n                                                                num_groups_32147))) {\n                    // read operands\n                    {\n                        x_33118 = ((volatile __local\n                                    int32_t *) scan_arr_mem_33113)[local_tid_33109 -\n                                                                   skip_threads_33125];\n                        x_33119 = ((volatile __local\n                                    bool *) scan_arr_mem_33115)[local_tid_33109 -\n                                                                skip_threads_33125];\n                    }\n                    // perform operation\n                    {\n                        int32_t res_33122;\n                        \n                        if (x_33121) {\n                            res_33122 = x_33120;\n                        } else {\n                            res_33122 = x_33118;\n                        }\n                        \n                        bool res_33123 = x_33119 || x_33121;\n                        \n                        x_33120 = res_33122;\n                        x_33121 = res_33123;\n                    }\n                }\n                if (sle32(wave_sizze_33111, skip_threads_33125)) {\n                    barrier(CLK_LOCAL_MEM_FENCE);\n                }\n                if (sle32(skip_threads_33125, local_tid_33109 -\n                          squot32(local_tid_33109, 32) * 32) &&\n                    (squot32(local_tid_33109, 32) == 0 && slt32(local_tid_33109,\n                                                                num_groups_32147))) {\n                    // write result\n                    {\n      ",
            "                  ((volatile __local\n                          int32_t *) scan_arr_mem_33113)[local_tid_33109] =\n                            x_33120;\n                        ((volatile __local\n                          bool *) scan_arr_mem_33115)[local_tid_33109] =\n                            x_33121;\n                    }\n                }\n                if (sle32(wave_sizze_33111, skip_threads_33125)) {\n                    barrier(CLK_LOCAL_MEM_FENCE);\n                }\n                skip_threads_33125 *= 2;\n            }\n        }\n    }\n    barrier(CLK_LOCAL_MEM_FENCE);\n    // carry-in for every block except the first\n    {\n        if (!(squot32(local_tid_33109, 32) == 0 || !slt32(local_tid_33109,\n                                                          num_groups_32147))) {\n            // read operands\n            {\n                x_33096 = ((volatile __local\n                            int32_t *) scan_arr_mem_33113)[squot32(local_tid_33109,\n                                                                   32) - 1];\n                x_33097 = ((volatile __local\n                            bool *) scan_arr_mem_33115)[squot32(local_tid_33109,\n                                                                32) - 1];\n            }\n            // perform operation\n            {\n                int32_t res_33100;\n                \n                if (x_33099) {\n                    res_33100 = x_33098;\n                } else {\n                    res_33100 = x_33096;\n                }\n                \n                bool res_33101 = x_33097 || x_33099;\n                \n                x_33098 = res_33100;\n                x_33099 = res_33101;\n            }\n            // write final result\n            {\n                ((volatile __local\n                  int32_t *) scan_arr_mem_33113)[local_tid_33109] = x_33098;\n                ((volatile __local\n                  bool *) scan_arr_mem_33115)[local_tid_33109] = x_33099;\n            }\n        }\n    }\n    barrier(CLK",
            "_LOCAL_MEM_FENCE);\n    // restore correct values for first block\n    {\n        if (squot32(local_tid_33109, 32) == 0) {\n            ((volatile __local int32_t *) scan_arr_mem_33113)[local_tid_33109] =\n                x_33098;\n            ((volatile __local bool *) scan_arr_mem_33115)[local_tid_33109] =\n                x_33099;\n        }\n    }\n    barrier(CLK_LOCAL_MEM_FENCE);\n    // threads in bounds write scanned carries\n    {\n        if (slt32(gtid_32149, flat_dim_31615)) {\n            ((__global int32_t *) mem_32801)[gtid_32149] = ((__local\n                                                             int32_t *) scan_arr_mem_33113)[local_tid_33109];\n            ((__global bool *) mem_32803)[gtid_32149] = ((__local\n                                                          bool *) scan_arr_mem_33115)[local_tid_33109];\n        }\n    }\n}\n__kernel void scan_stage2_32190(uint scan_arr_mem_33191_backing_offset_0,\n                                uint scan_arr_mem_33193_backing_offset_1,\n                                int32_t flat_dim_31615,\n                                int32_t num_groups_32187, __global\n                                unsigned char *mem_32815, __global\n                                unsigned char *mem_32817,\n                                int32_t num_threads_33143)\n{\n    const int32_t segscan_group_sizze_32185 =\n                  quicksort1zisegscan_group_sizze_32184;\n    const int block_dim0 = 0;\n    const int block_dim1 = 1;\n    const int block_dim2 = 2;\n    volatile char *scan_arr_mem_33191_backing_0 =\n                  &shared_mem[scan_arr_mem_33191_backing_offset_0];\n    volatile char *scan_arr_mem_33193_backing_1 =\n                  &shared_mem[scan_arr_mem_33193_backing_offset_1];\n    int32_t global_tid_33186;\n    int32_t local_tid_33187;\n    int32_t group_sizze_33190;\n    int32_t wave_sizze_33189;\n    int32_t group_tid_33188;\n    \n    global_tid_33186 = get_global_id(0);\n    local_tid_33187 = get_local_id(0);\n    group_sizze_33190 = get_loca",
            "l_size(0);\n    wave_sizze_33189 = LOCKSTEP_WIDTH;\n    group_tid_33188 = get_group_id(0);\n    \n    int32_t phys_tid_32190 = global_tid_33186;\n    __local char *scan_arr_mem_33191;\n    \n    scan_arr_mem_33191 = (__local char *) scan_arr_mem_33191_backing_0;\n    \n    __local char *scan_arr_mem_33193;\n    \n    scan_arr_mem_33193 = (__local char *) scan_arr_mem_33193_backing_1;\n    \n    int32_t flat_idx_33195 = (local_tid_33187 + 1) *\n            (segscan_group_sizze_32185 * squot32(flat_dim_31615 +\n                                                 num_threads_33143 - 1,\n                                                 num_threads_33143)) - 1;\n    int32_t gtid_32189 = flat_idx_33195;\n    \n    // threads in bound read carries; others get neutral element\n    {\n        if (slt32(gtid_32189, flat_dim_31615)) {\n            ((__local int32_t *) scan_arr_mem_33191)[local_tid_33187] =\n                ((__global int32_t *) mem_32815)[gtid_32189];\n            ((__local bool *) scan_arr_mem_33193)[local_tid_33187] = ((__global\n                                                                       bool *) mem_32817)[gtid_32189];\n        } else {\n            ((__local int32_t *) scan_arr_mem_33191)[local_tid_33187] = 0;\n            ((__local bool *) scan_arr_mem_33193)[local_tid_33187] = 0;\n        }\n    }\n    \n    int32_t x_33172;\n    bool x_33173;\n    int32_t x_33174;\n    bool x_33175;\n    int32_t x_33196;\n    bool x_33197;\n    int32_t x_33198;\n    bool x_33199;\n    int32_t skip_threads_33203;\n    \n    if (slt32(local_tid_33187, num_groups_32187)) {\n        x_33174 = ((volatile __local\n                    int32_t *) scan_arr_mem_33191)[local_tid_33187];\n        x_33175 = ((volatile __local\n                    bool *) scan_arr_mem_33193)[local_tid_33187];\n    }\n    // in-block scan (hopefully no barriers needed)\n    {\n        skip_threads_33203 = 1;\n        while (slt32(skip_threads_33203, 32)) {\n            if (sle32(skip_threads_33203, local_tid_33187 -\n                      squot3",
            "2(local_tid_33187, 32) * 32) &&\n                slt32(local_tid_33187, num_groups_32187)) {\n                // read operands\n                {\n                    x_33172 = ((volatile __local\n                                int32_t *) scan_arr_mem_33191)[local_tid_33187 -\n                                                               skip_threads_33203];\n                    x_33173 = ((volatile __local\n                                bool *) scan_arr_mem_33193)[local_tid_33187 -\n                                                            skip_threads_33203];\n                }\n                // perform operation\n                {\n                    int32_t res_33176;\n                    \n                    if (x_33175) {\n                        res_33176 = x_33174;\n                    } else {\n                        int32_t res_33177 = x_33172 + x_33174;\n                        \n                        res_33176 = res_33177;\n                    }\n                    \n                    bool res_33178 = x_33173 || x_33175;\n                    \n                    x_33174 = res_33176;\n                    x_33175 = res_33178;\n                }\n            }\n            if (sle32(wave_sizze_33189, skip_threads_33203)) {\n                barrier(CLK_LOCAL_MEM_FENCE);\n            }\n            if (sle32(skip_threads_33203, local_tid_33187 -\n                      squot32(local_tid_33187, 32) * 32) &&\n                slt32(local_tid_33187, num_groups_32187)) {\n                // write result\n                {\n                    ((volatile __local\n                      int32_t *) scan_arr_mem_33191)[local_tid_33187] = x_33174;\n                    ((volatile __local\n                      bool *) scan_arr_mem_33193)[local_tid_33187] = x_33175;\n                }\n            }\n            if (sle32(wave_sizze_33189, skip_threads_33203)) {\n                barrier(CLK_LOCAL_MEM_FENCE);\n            }\n            skip_threads_33203 *= 2;\n        }\n    }\n    barrier(CLK_LOCAL_MEM",
            "_FENCE);\n    // last thread of block 'i' writes its result to offset 'i'\n    {\n        if ((local_tid_33187 - squot32(local_tid_33187, 32) * 32) == 31 &&\n            slt32(local_tid_33187, num_groups_32187)) {\n            ((volatile __local\n              int32_t *) scan_arr_mem_33191)[squot32(local_tid_33187, 32)] =\n                x_33174;\n            ((volatile __local\n              bool *) scan_arr_mem_33193)[squot32(local_tid_33187, 32)] =\n                x_33175;\n        }\n    }\n    barrier(CLK_LOCAL_MEM_FENCE);\n    // scan the first block, after which offset 'i' contains carry-in for warp 'i+1'\n    {\n        int32_t skip_threads_33204;\n        \n        if (squot32(local_tid_33187, 32) == 0 && slt32(local_tid_33187,\n                                                       num_groups_32187)) {\n            x_33198 = ((volatile __local\n                        int32_t *) scan_arr_mem_33191)[local_tid_33187];\n            x_33199 = ((volatile __local\n                        bool *) scan_arr_mem_33193)[local_tid_33187];\n        }\n        // in-block scan (hopefully no barriers needed)\n        {\n            skip_threads_33204 = 1;\n            while (slt32(skip_threads_33204, 32)) {\n                if (sle32(skip_threads_33204, local_tid_33187 -\n                          squot32(local_tid_33187, 32) * 32) &&\n                    (squot32(local_tid_33187, 32) == 0 && slt32(local_tid_33187,\n                                                                num_groups_32187))) {\n                    // read operands\n                    {\n                        x_33196 = ((volatile __local\n                                    int32_t *) scan_arr_mem_33191)[local_tid_33187 -\n                                                                   skip_threads_33204];\n                        x_33197 = ((volatile __local\n                                    bool *) scan_arr_mem_33193)[local_tid_33187 -\n                                                                skip_threads_33204];\n    ",
            "                }\n                    // perform operation\n                    {\n                        int32_t res_33200;\n                        \n                        if (x_33199) {\n                            res_33200 = x_33198;\n                        } else {\n                            int32_t res_33201 = x_33196 + x_33198;\n                            \n                            res_33200 = res_33201;\n                        }\n                        \n                        bool res_33202 = x_33197 || x_33199;\n                        \n                        x_33198 = res_33200;\n                        x_33199 = res_33202;\n                    }\n                }\n                if (sle32(wave_sizze_33189, skip_threads_33204)) {\n                    barrier(CLK_LOCAL_MEM_FENCE);\n                }\n                if (sle32(skip_threads_33204, local_tid_33187 -\n                          squot32(local_tid_33187, 32) * 32) &&\n                    (squot32(local_tid_33187, 32) == 0 && slt32(local_tid_33187,\n                                                                num_groups_32187))) {\n                    // write result\n                    {\n                        ((volatile __local\n                          int32_t *) scan_arr_mem_33191)[local_tid_33187] =\n                            x_33198;\n                        ((volatile __local\n                          bool *) scan_arr_mem_33193)[local_tid_33187] =\n                            x_33199;\n                    }\n                }\n                if (sle32(wave_sizze_33189, skip_threads_33204)) {\n                    barrier(CLK_LOCAL_MEM_FENCE);\n                }\n                skip_threads_33204 *= 2;\n            }\n        }\n    }\n    barrier(CLK_LOCAL_MEM_FENCE);\n    // carry-in for every block except the first\n    {\n        if (!(squot32(local_tid_33187, 32) == 0 || !slt32(local_tid_33187,\n                                                          num_groups_32187))) {\n            // read operands\n",
            "            {\n                x_33172 = ((volatile __local\n                            int32_t *) scan_arr_mem_33191)[squot32(local_tid_33187,\n                                                                   32) - 1];\n                x_33173 = ((volatile __local\n                            bool *) scan_arr_mem_33193)[squot32(local_tid_33187,\n                                                                32) - 1];\n            }\n            // perform operation\n            {\n                int32_t res_33176;\n                \n                if (x_33175) {\n                    res_33176 = x_33174;\n                } else {\n                    int32_t res_33177 = x_33172 + x_33174;\n                    \n                    res_33176 = res_33177;\n                }\n                \n                bool res_33178 = x_33173 || x_33175;\n                \n                x_33174 = res_33176;\n                x_33175 = res_33178;\n            }\n            // write final result\n            {\n                ((volatile __local\n                  int32_t *) scan_arr_mem_33191)[local_tid_33187] = x_33174;\n                ((volatile __local\n                  bool *) scan_arr_mem_33193)[local_tid_33187] = x_33175;\n            }\n        }\n    }\n    barrier(CLK_LOCAL_MEM_FENCE);\n    // restore correct values for first block\n    {\n        if (squot32(local_tid_33187, 32) == 0) {\n            ((volatile __local int32_t *) scan_arr_mem_33191)[local_tid_33187] =\n                x_33174;\n            ((volatile __local bool *) scan_arr_mem_33193)[local_tid_33187] =\n                x_33175;\n        }\n    }\n    barrier(CLK_LOCAL_MEM_FENCE);\n    // threads in bounds write scanned carries\n    {\n        if (slt32(gtid_32189, flat_dim_31615)) {\n            ((__global int32_t *) mem_32815)[gtid_32189] = ((__local\n                                                             int32_t *) scan_arr_mem_33191)[local_tid_33187];\n            ((__global bool *) mem_32817)[gtid_32189] = ((__local\n              ",
            "                                            bool *) scan_arr_mem_33193)[local_tid_33187];\n        }\n    }\n}\n__kernel void scan_stage2_32199(uint scan_arr_mem_33260_backing_offset_0,\n                                uint scan_arr_mem_33262_backing_offset_1,\n                                int32_t flat_dim_31615,\n                                int32_t num_groups_32196, __global\n                                unsigned char *mem_32821, __global\n                                unsigned char *mem_32823,\n                                int32_t num_threads_33212)\n{\n    const int32_t segscan_group_sizze_32194 =\n                  quicksort1zisegscan_group_sizze_32193;\n    const int block_dim0 = 0;\n    const int block_dim1 = 1;\n    const int block_dim2 = 2;\n    volatile char *scan_arr_mem_33260_backing_0 =\n                  &shared_mem[scan_arr_mem_33260_backing_offset_0];\n    volatile char *scan_arr_mem_33262_backing_1 =\n                  &shared_mem[scan_arr_mem_33262_backing_offset_1];\n    int32_t global_tid_33255;\n    int32_t local_tid_33256;\n    int32_t group_sizze_33259;\n    int32_t wave_sizze_33258;\n    int32_t group_tid_33257;\n    \n    global_tid_33255 = get_global_id(0);\n    local_tid_33256 = get_local_id(0);\n    group_sizze_33259 = get_local_size(0);\n    wave_sizze_33258 = LOCKSTEP_WIDTH;\n    group_tid_33257 = get_group_id(0);\n    \n    int32_t phys_tid_32199 = global_tid_33255;\n    __local char *scan_arr_mem_33260;\n    \n    scan_arr_mem_33260 = (__local char *) scan_arr_mem_33260_backing_0;\n    \n    __local char *scan_arr_mem_33262;\n    \n    scan_arr_mem_33262 = (__local char *) scan_arr_mem_33262_backing_1;\n    \n    int32_t flat_idx_33264 = (local_tid_33256 + 1) *\n            (segscan_group_sizze_32194 * squot32(flat_dim_31615 +\n                                                 num_threads_33212 - 1,\n                                                 num_threads_33212)) - 1;\n    int32_t gtid_32198 = flat_idx_33264;\n    \n    // threads in bound read carries; others get",
            " neutral element\n    {\n        if (slt32(gtid_32198, flat_dim_31615)) {\n            ((__local int32_t *) scan_arr_mem_33260)[local_tid_33256] =\n                ((__global int32_t *) mem_32821)[gtid_32198];\n            ((__local bool *) scan_arr_mem_33262)[local_tid_33256] = ((__global\n                                                                       bool *) mem_32823)[gtid_32198];\n        } else {\n            ((__local int32_t *) scan_arr_mem_33260)[local_tid_33256] = 0;\n            ((__local bool *) scan_arr_mem_33262)[local_tid_33256] = 0;\n        }\n    }\n    \n    int32_t x_33241;\n    bool x_33242;\n    int32_t x_33243;\n    bool x_33244;\n    int32_t x_33265;\n    bool x_33266;\n    int32_t x_33267;\n    bool x_33268;\n    int32_t skip_threads_33272;\n    \n    if (slt32(local_tid_33256, num_groups_32196)) {\n        x_33243 = ((volatile __local\n                    int32_t *) scan_arr_mem_33260)[local_tid_33256];\n        x_33244 = ((volatile __local\n                    bool *) scan_arr_mem_33262)[local_tid_33256];\n    }\n    // in-block scan (hopefully no barriers needed)\n    {\n        skip_threads_33272 = 1;\n        while (slt32(skip_threads_33272, 32)) {\n            if (sle32(skip_threads_33272, local_tid_33256 -\n                      squot32(local_tid_33256, 32) * 32) &&\n                slt32(local_tid_33256, num_groups_32196)) {\n                // read operands\n                {\n                    x_33241 = ((volatile __local\n                                int32_t *) scan_arr_mem_33260)[local_tid_33256 -\n                                                               skip_threads_33272];\n                    x_33242 = ((volatile __local\n                                bool *) scan_arr_mem_33262)[local_tid_33256 -\n                                                            skip_threads_33272];\n                }\n                // perform operation\n                {\n                    int32_t res_33245;\n                    \n                    if (x_33244) {\n       ",
            "                 res_33245 = x_33243;\n                    } else {\n                        int32_t res_33246 = x_33241 + x_33243;\n                        \n                        res_33245 = res_33246;\n                    }\n                    \n                    bool res_33247 = x_33242 || x_33244;\n                    \n                    x_33243 = res_33245;\n                    x_33244 = res_33247;\n                }\n            }\n            if (sle32(wave_sizze_33258, skip_threads_33272)) {\n                barrier(CLK_LOCAL_MEM_FENCE);\n            }\n            if (sle32(skip_threads_33272, local_tid_33256 -\n                      squot32(local_tid_33256, 32) * 32) &&\n                slt32(local_tid_33256, num_groups_32196)) {\n                // write result\n                {\n                    ((volatile __local\n                      int32_t *) scan_arr_mem_33260)[local_tid_33256] = x_33243;\n                    ((volatile __local\n                      bool *) scan_arr_mem_33262)[local_tid_33256] = x_33244;\n                }\n            }\n            if (sle32(wave_sizze_33258, skip_threads_33272)) {\n                barrier(CLK_LOCAL_MEM_FENCE);\n            }\n            skip_threads_33272 *= 2;\n        }\n    }\n    barrier(CLK_LOCAL_MEM_FENCE);\n    // last thread of block 'i' writes its result to offset 'i'\n    {\n        if ((local_tid_33256 - squot32(local_tid_33256, 32) * 32) == 31 &&\n            slt32(local_tid_33256, num_groups_32196)) {\n            ((volatile __local\n              int32_t *) scan_arr_mem_33260)[squot32(local_tid_33256, 32)] =\n                x_33243;\n            ((volatile __local\n              bool *) scan_arr_mem_33262)[squot32(local_tid_33256, 32)] =\n                x_33244;\n        }\n    }\n    barrier(CLK_LOCAL_MEM_FENCE);\n    // scan the first block, after which offset 'i' contains carry-in for warp 'i+1'\n    {\n        int32_t skip_threads_33273;\n        \n        if (squot32(local_tid_33256, 32) == 0 && slt32(local_tid_33256,\n         ",
            "                                              num_groups_32196)) {\n            x_33267 = ((volatile __local\n                        int32_t *) scan_arr_mem_33260)[local_tid_33256];\n            x_33268 = ((volatile __local\n                        bool *) scan_arr_mem_33262)[local_tid_33256];\n        }\n        // in-block scan (hopefully no barriers needed)\n        {\n            skip_threads_33273 = 1;\n            while (slt32(skip_threads_33273, 32)) {\n                if (sle32(skip_threads_33273, local_tid_33256 -\n                          squot32(local_tid_33256, 32) * 32) &&\n                    (squot32(local_tid_33256, 32) == 0 && slt32(local_tid_33256,\n                                                                num_groups_32196))) {\n                    // read operands\n                    {\n                        x_33265 = ((volatile __local\n                                    int32_t *) scan_arr_mem_33260)[local_tid_33256 -\n                                                                   skip_threads_33273];\n                        x_33266 = ((volatile __local\n                                    bool *) scan_arr_mem_33262)[local_tid_33256 -\n                                                                skip_threads_33273];\n                    }\n                    // perform operation\n                    {\n                        int32_t res_33269;\n                        \n                        if (x_33268) {\n                            res_33269 = x_33267;\n                        } else {\n                            int32_t res_33270 = x_33265 + x_33267;\n                            \n                            res_33269 = res_33270;\n                        }\n                        \n                        bool res_33271 = x_33266 || x_33268;\n                        \n                        x_33267 = res_33269;\n                        x_33268 = res_33271;\n                    }\n                }\n                if (sle32(wave_sizze_33258, skip_threads",
            "_33273)) {\n                    barrier(CLK_LOCAL_MEM_FENCE);\n                }\n                if (sle32(skip_threads_33273, local_tid_33256 -\n                          squot32(local_tid_33256, 32) * 32) &&\n                    (squot32(local_tid_33256, 32) == 0 && slt32(local_tid_33256,\n                                                                num_groups_32196))) {\n                    // write result\n                    {\n                        ((volatile __local\n                          int32_t *) scan_arr_mem_33260)[local_tid_33256] =\n                            x_33267;\n                        ((volatile __local\n                          bool *) scan_arr_mem_33262)[local_tid_33256] =\n                            x_33268;\n                    }\n                }\n                if (sle32(wave_sizze_33258, skip_threads_33273)) {\n                    barrier(CLK_LOCAL_MEM_FENCE);\n                }\n                skip_threads_33273 *= 2;\n            }\n        }\n    }\n    barrier(CLK_LOCAL_MEM_FENCE);\n    // carry-in for every block except the first\n    {\n        if (!(squot32(local_tid_33256, 32) == 0 || !slt32(local_tid_33256,\n                                                          num_groups_32196))) {\n            // read operands\n            {\n                x_33241 = ((volatile __local\n                            int32_t *) scan_arr_mem_33260)[squot32(local_tid_33256,\n                                                                   32) - 1];\n                x_33242 = ((volatile __local\n                            bool *) scan_arr_mem_33262)[squot32(local_tid_33256,\n                                                                32) - 1];\n            }\n            // perform operation\n            {\n                int32_t res_33245;\n                \n                if (x_33244) {\n                    res_33245 = x_33243;\n                } else {\n                    int32_t res_33246 = x_33241 + x_33243;\n                    \n                    res_33245 ",
            "= res_33246;\n                }\n                \n                bool res_33247 = x_33242 || x_33244;\n                \n                x_33243 = res_33245;\n                x_33244 = res_33247;\n            }\n            // write final result\n            {\n                ((volatile __local\n                  int32_t *) scan_arr_mem_33260)[local_tid_33256] = x_33243;\n                ((volatile __local\n                  bool *) scan_arr_mem_33262)[local_tid_33256] = x_33244;\n            }\n        }\n    }\n    barrier(CLK_LOCAL_MEM_FENCE);\n    // restore correct values for first block\n    {\n        if (squot32(local_tid_33256, 32) == 0) {\n            ((volatile __local int32_t *) scan_arr_mem_33260)[local_tid_33256] =\n                x_33243;\n            ((volatile __local bool *) scan_arr_mem_33262)[local_tid_33256] =\n                x_33244;\n        }\n    }\n    barrier(CLK_LOCAL_MEM_FENCE);\n    // threads in bounds write scanned carries\n    {\n        if (slt32(gtid_32198, flat_dim_31615)) {\n            ((__global int32_t *) mem_32821)[gtid_32198] = ((__local\n                                                             int32_t *) scan_arr_mem_33260)[local_tid_33256];\n            ((__global bool *) mem_32823)[gtid_32198] = ((__local\n                                                          bool *) scan_arr_mem_33262)[local_tid_33256];\n        }\n    }\n}\n__kernel void scan_stage2_32225(uint scan_arr_mem_33330_backing_offset_0,\n                                uint scan_arr_mem_33332_backing_offset_1,\n                                int32_t flat_dim_31615,\n                                int32_t num_groups_32222, __global\n                                unsigned char *mem_32830, __global\n                                unsigned char *mem_32832,\n                                int32_t num_threads_33286)\n{\n    const int32_t segscan_group_sizze_32220 =\n                  quicksort1zisegscan_group_sizze_32219;\n    const int block_dim0 = 0;\n    const int block_dim1 = 1;\n    const in",
            "t block_dim2 = 2;\n    volatile char *scan_arr_mem_33330_backing_0 =\n                  &shared_mem[scan_arr_mem_33330_backing_offset_0];\n    volatile char *scan_arr_mem_33332_backing_1 =\n                  &shared_mem[scan_arr_mem_33332_backing_offset_1];\n    int32_t global_tid_33325;\n    int32_t local_tid_33326;\n    int32_t group_sizze_33329;\n    int32_t wave_sizze_33328;\n    int32_t group_tid_33327;\n    \n    global_tid_33325 = get_global_id(0);\n    local_tid_33326 = get_local_id(0);\n    group_sizze_33329 = get_local_size(0);\n    wave_sizze_33328 = LOCKSTEP_WIDTH;\n    group_tid_33327 = get_group_id(0);\n    \n    int32_t phys_tid_32225 = global_tid_33325;\n    __local char *scan_arr_mem_33330;\n    \n    scan_arr_mem_33330 = (__local char *) scan_arr_mem_33330_backing_0;\n    \n    __local char *scan_arr_mem_33332;\n    \n    scan_arr_mem_33332 = (__local char *) scan_arr_mem_33332_backing_1;\n    \n    int32_t flat_idx_33334 = (local_tid_33326 + 1) *\n            (segscan_group_sizze_32220 * squot32(flat_dim_31615 +\n                                                 num_threads_33286 - 1,\n                                                 num_threads_33286)) - 1;\n    int32_t gtid_32224 = flat_idx_33334;\n    \n    // threads in bound read carries; others get neutral element\n    {\n        if (slt32(gtid_32224, flat_dim_31615)) {\n            ((__local int32_t *) scan_arr_mem_33330)[local_tid_33326] =\n                ((__global int32_t *) mem_32830)[gtid_32224];\n            ((__local bool *) scan_arr_mem_33332)[local_tid_33326] = ((__global\n                                                                       bool *) mem_32832)[gtid_32224];\n        } else {\n            ((__local int32_t *) scan_arr_mem_33330)[local_tid_33326] = 0;\n            ((__local bool *) scan_arr_mem_33332)[local_tid_33326] = 0;\n        }\n    }\n    \n    int32_t x_33313;\n    bool x_33314;\n    int32_t x_33315;\n    bool x_33316;\n    int32_t x_33335;\n    bool x_33336;\n    int32_t x_33337;\n    bool x_33338;\n    int32_",
            "t skip_threads_33341;\n    \n    if (slt32(local_tid_33326, num_groups_32222)) {\n        x_33315 = ((volatile __local\n                    int32_t *) scan_arr_mem_33330)[local_tid_33326];\n        x_33316 = ((volatile __local\n                    bool *) scan_arr_mem_33332)[local_tid_33326];\n    }\n    // in-block scan (hopefully no barriers needed)\n    {\n        skip_threads_33341 = 1;\n        while (slt32(skip_threads_33341, 32)) {\n            if (sle32(skip_threads_33341, local_tid_33326 -\n                      squot32(local_tid_33326, 32) * 32) &&\n                slt32(local_tid_33326, num_groups_32222)) {\n                // read operands\n                {\n                    x_33313 = ((volatile __local\n                                int32_t *) scan_arr_mem_33330)[local_tid_33326 -\n                                                               skip_threads_33341];\n                    x_33314 = ((volatile __local\n                                bool *) scan_arr_mem_33332)[local_tid_33326 -\n                                                            skip_threads_33341];\n                }\n                // perform operation\n                {\n                    int32_t res_33317;\n                    \n                    if (x_33316) {\n                        res_33317 = x_33315;\n                    } else {\n                        res_33317 = x_33313;\n                    }\n                    \n                    bool res_33318 = x_33314 || x_33316;\n                    \n                    x_33315 = res_33317;\n                    x_33316 = res_33318;\n                }\n            }\n            if (sle32(wave_sizze_33328, skip_threads_33341)) {\n                barrier(CLK_LOCAL_MEM_FENCE);\n            }\n            if (sle32(skip_threads_33341, local_tid_33326 -\n                      squot32(local_tid_33326, 32) * 32) &&\n                slt32(local_tid_33326, num_groups_32222)) {\n                // write result\n                {\n                    ((volatile __local\n ",
            "                     int32_t *) scan_arr_mem_33330)[local_tid_33326] = x_33315;\n                    ((volatile __local\n                      bool *) scan_arr_mem_33332)[local_tid_33326] = x_33316;\n                }\n            }\n            if (sle32(wave_sizze_33328, skip_threads_33341)) {\n                barrier(CLK_LOCAL_MEM_FENCE);\n            }\n            skip_threads_33341 *= 2;\n        }\n    }\n    barrier(CLK_LOCAL_MEM_FENCE);\n    // last thread of block 'i' writes its result to offset 'i'\n    {\n        if ((local_tid_33326 - squot32(local_tid_33326, 32) * 32) == 31 &&\n            slt32(local_tid_33326, num_groups_32222)) {\n            ((volatile __local\n              int32_t *) scan_arr_mem_33330)[squot32(local_tid_33326, 32)] =\n                x_33315;\n            ((volatile __local\n              bool *) scan_arr_mem_33332)[squot32(local_tid_33326, 32)] =\n                x_33316;\n        }\n    }\n    barrier(CLK_LOCAL_MEM_FENCE);\n    // scan the first block, after which offset 'i' contains carry-in for warp 'i+1'\n    {\n        int32_t skip_threads_33342;\n        \n        if (squot32(local_tid_33326, 32) == 0 && slt32(local_tid_33326,\n                                                       num_groups_32222)) {\n            x_33337 = ((volatile __local\n                        int32_t *) scan_arr_mem_33330)[local_tid_33326];\n            x_33338 = ((volatile __local\n                        bool *) scan_arr_mem_33332)[local_tid_33326];\n        }\n        // in-block scan (hopefully no barriers needed)\n        {\n            skip_threads_33342 = 1;\n            while (slt32(skip_threads_33342, 32)) {\n                if (sle32(skip_threads_33342, local_tid_33326 -\n                          squot32(local_tid_33326, 32) * 32) &&\n                    (squot32(local_tid_33326, 32) == 0 && slt32(local_tid_33326,\n                                                                num_groups_32222))) {\n                    // read operands\n                    {\n                    ",
            "    x_33335 = ((volatile __local\n                                    int32_t *) scan_arr_mem_33330)[local_tid_33326 -\n                                                                   skip_threads_33342];\n                        x_33336 = ((volatile __local\n                                    bool *) scan_arr_mem_33332)[local_tid_33326 -\n                                                                skip_threads_33342];\n                    }\n                    // perform operation\n                    {\n                        int32_t res_33339;\n                        \n                        if (x_33338) {\n                            res_33339 = x_33337;\n                        } else {\n                            res_33339 = x_33335;\n                        }\n                        \n                        bool res_33340 = x_33336 || x_33338;\n                        \n                        x_33337 = res_33339;\n                        x_33338 = res_33340;\n                    }\n                }\n                if (sle32(wave_sizze_33328, skip_threads_33342)) {\n                    barrier(CLK_LOCAL_MEM_FENCE);\n                }\n                if (sle32(skip_threads_33342, local_tid_33326 -\n                          squot32(local_tid_33326, 32) * 32) &&\n                    (squot32(local_tid_33326, 32) == 0 && slt32(local_tid_33326,\n                                                                num_groups_32222))) {\n                    // write result\n                    {\n                        ((volatile __local\n                          int32_t *) scan_arr_mem_33330)[local_tid_33326] =\n                            x_33337;\n                        ((volatile __local\n                          bool *) scan_arr_mem_33332)[local_tid_33326] =\n                            x_33338;\n                    }\n                }\n                if (sle32(wave_sizze_33328, skip_threads_33342)) {\n                    barrier(CLK_LOCAL_MEM_FENCE);\n                }\n            ",
            "    skip_threads_33342 *= 2;\n            }\n        }\n    }\n    barrier(CLK_LOCAL_MEM_FENCE);\n    // carry-in for every block except the first\n    {\n        if (!(squot32(local_tid_33326, 32) == 0 || !slt32(local_tid_33326,\n                                                          num_groups_32222))) {\n            // read operands\n            {\n                x_33313 = ((volatile __local\n                            int32_t *) scan_arr_mem_33330)[squot32(local_tid_33326,\n                                                                   32) - 1];\n                x_33314 = ((volatile __local\n                            bool *) scan_arr_mem_33332)[squot32(local_tid_33326,\n                                                                32) - 1];\n            }\n            // perform operation\n            {\n                int32_t res_33317;\n                \n                if (x_33316) {\n                    res_33317 = x_33315;\n                } else {\n                    res_33317 = x_33313;\n                }\n                \n                bool res_33318 = x_33314 || x_33316;\n                \n                x_33315 = res_33317;\n                x_33316 = res_33318;\n            }\n            // write final result\n            {\n                ((volatile __local\n                  int32_t *) scan_arr_mem_33330)[local_tid_33326] = x_33315;\n                ((volatile __local\n                  bool *) scan_arr_mem_33332)[local_tid_33326] = x_33316;\n            }\n        }\n    }\n    barrier(CLK_LOCAL_MEM_FENCE);\n    // restore correct values for first block\n    {\n        if (squot32(local_tid_33326, 32) == 0) {\n            ((volatile __local int32_t *) scan_arr_mem_33330)[local_tid_33326] =\n                x_33315;\n            ((volatile __local bool *) scan_arr_mem_33332)[local_tid_33326] =\n                x_33316;\n        }\n    }\n    barrier(CLK_LOCAL_MEM_FENCE);\n    // threads in bounds write scanned carries\n    {\n        if (slt32(gtid_32224, flat_dim_31615)) {\n         ",
            "   ((__global int32_t *) mem_32830)[gtid_32224] = ((__local\n                                                             int32_t *) scan_arr_mem_33330)[local_tid_33326];\n            ((__global bool *) mem_32832)[gtid_32224] = ((__local\n                                                          bool *) scan_arr_mem_33332)[local_tid_33326];\n        }\n    }\n}\n__kernel void scan_stage3_33062(int32_t flat_dim_31615, __global\n                                unsigned char *mem_32795, __global\n                                unsigned char *mem_32797,\n                                int32_t num_threads_33005)\n{\n    const int32_t segscan_group_sizze_32136 =\n                  quicksort1zisegscan_group_sizze_32135;\n    const int block_dim0 = 0;\n    const int block_dim1 = 1;\n    const int block_dim2 = 2;\n    int32_t scan_stage3_gtid_33062;\n    int32_t scan_stage3_ltid_33063;\n    int32_t scan_stage3_gid_33064;\n    \n    scan_stage3_gtid_33062 = get_global_id(0);\n    scan_stage3_ltid_33063 = get_local_id(0);\n    scan_stage3_gid_33064 = get_group_id(0);\n    \n    int32_t phys_tid_32141 = scan_stage3_gtid_33062;\n    int32_t gtid_32140 = scan_stage3_gtid_33062;\n    int32_t orig_group_33067 = squot32(scan_stage3_gtid_33062,\n                                       segscan_group_sizze_32136 *\n                                       squot32(flat_dim_31615 +\n                                               num_threads_33005 - 1,\n                                               num_threads_33005));\n    int32_t carry_in_flat_idx_33068 = orig_group_33067 *\n            (segscan_group_sizze_32136 * squot32(flat_dim_31615 +\n                                                 num_threads_33005 - 1,\n                                                 num_threads_33005)) - 1;\n    \n    if (slt32(scan_stage3_gtid_33062, flat_dim_31615)) {\n        if (!(orig_group_33067 == 0 || scan_stage3_gtid_33062 ==\n              (orig_group_33067 + 1) * (segscan_group_sizze_32136 *\n                                        squot",
            "32(flat_dim_31615 +\n                                                num_threads_33005 - 1,\n                                                num_threads_33005)) - 1)) {\n            int32_t x_33038;\n            bool x_33039;\n            int32_t x_33040;\n            bool x_33041;\n            \n            x_33038 = ((__global int32_t *) mem_32795)[carry_in_flat_idx_33068];\n            x_33039 = ((__global bool *) mem_32797)[carry_in_flat_idx_33068];\n            x_33040 = ((__global int32_t *) mem_32795)[gtid_32140];\n            x_33041 = ((__global bool *) mem_32797)[gtid_32140];\n            \n            int32_t res_33042;\n            \n            if (x_33041) {\n                res_33042 = x_33040;\n            } else {\n                res_33042 = x_33038;\n            }\n            \n            bool res_33043 = x_33039 || x_33041;\n            \n            x_33038 = res_33042;\n            x_33039 = res_33043;\n            ((__global int32_t *) mem_32795)[gtid_32140] = x_33038;\n            ((__global bool *) mem_32797)[gtid_32140] = x_33039;\n        }\n    }\n}\n__kernel void scan_stage3_33126(int32_t flat_dim_31615, __global\n                                unsigned char *mem_32801, __global\n                                unsigned char *mem_32803,\n                                int32_t num_threads_33069)\n{\n    const int32_t segscan_group_sizze_32145 =\n                  quicksort1zisegscan_group_sizze_32144;\n    const int block_dim0 = 0;\n    const int block_dim1 = 1;\n    const int block_dim2 = 2;\n    int32_t scan_stage3_gtid_33126;\n    int32_t scan_stage3_ltid_33127;\n    int32_t scan_stage3_gid_33128;\n    \n    scan_stage3_gtid_33126 = get_global_id(0);\n    scan_stage3_ltid_33127 = get_local_id(0);\n    scan_stage3_gid_33128 = get_group_id(0);\n    \n    int32_t phys_tid_32150 = scan_stage3_gtid_33126;\n    int32_t gtid_32149 = scan_stage3_gtid_33126;\n    int32_t orig_group_33131 = squot32(scan_stage3_gtid_33126,\n                                       segscan_group_sizze_32145 *\n  ",
            "                                     squot32(flat_dim_31615 +\n                                               num_threads_33069 - 1,\n                                               num_threads_33069));\n    int32_t carry_in_flat_idx_33132 = orig_group_33131 *\n            (segscan_group_sizze_32145 * squot32(flat_dim_31615 +\n                                                 num_threads_33069 - 1,\n                                                 num_threads_33069)) - 1;\n    \n    if (slt32(scan_stage3_gtid_33126, flat_dim_31615)) {\n        if (!(orig_group_33131 == 0 || scan_stage3_gtid_33126 ==\n              (orig_group_33131 + 1) * (segscan_group_sizze_32145 *\n                                        squot32(flat_dim_31615 +\n                                                num_threads_33069 - 1,\n                                                num_threads_33069)) - 1)) {\n            int32_t x_33102;\n            bool x_33103;\n            int32_t x_33104;\n            bool x_33105;\n            \n            x_33102 = ((__global int32_t *) mem_32801)[carry_in_flat_idx_33132];\n            x_33103 = ((__global bool *) mem_32803)[carry_in_flat_idx_33132];\n            x_33104 = ((__global int32_t *) mem_32801)[gtid_32149];\n            x_33105 = ((__global bool *) mem_32803)[gtid_32149];\n            \n            int32_t res_33106;\n            \n            if (x_33105) {\n                res_33106 = x_33104;\n            } else {\n                res_33106 = x_33102;\n            }\n            \n            bool res_33107 = x_33103 || x_33105;\n            \n            x_33102 = res_33106;\n            x_33103 = res_33107;\n            ((__global int32_t *) mem_32801)[gtid_32149] = x_33102;\n            ((__global bool *) mem_32803)[gtid_32149] = x_33103;\n        }\n    }\n}\n__kernel void scan_stage3_33205(int32_t flat_dim_31615, __global\n                                unsigned char *mem_32815, __global\n                                unsigned char *mem_32817,\n                                in",
            "t32_t num_threads_33143)\n{\n    const int32_t segscan_group_sizze_32185 =\n                  quicksort1zisegscan_group_sizze_32184;\n    const int block_dim0 = 0;\n    const int block_dim1 = 1;\n    const int block_dim2 = 2;\n    int32_t scan_stage3_gtid_33205;\n    int32_t scan_stage3_ltid_33206;\n    int32_t scan_stage3_gid_33207;\n    \n    scan_stage3_gtid_33205 = get_global_id(0);\n    scan_stage3_ltid_33206 = get_local_id(0);\n    scan_stage3_gid_33207 = get_group_id(0);\n    \n    int32_t phys_tid_32190 = scan_stage3_gtid_33205;\n    int32_t gtid_32189 = scan_stage3_gtid_33205;\n    int32_t orig_group_33210 = squot32(scan_stage3_gtid_33205,\n                                       segscan_group_sizze_32185 *\n                                       squot32(flat_dim_31615 +\n                                               num_threads_33143 - 1,\n                                               num_threads_33143));\n    int32_t carry_in_flat_idx_33211 = orig_group_33210 *\n            (segscan_group_sizze_32185 * squot32(flat_dim_31615 +\n                                                 num_threads_33143 - 1,\n                                                 num_threads_33143)) - 1;\n    \n    if (slt32(scan_stage3_gtid_33205, flat_dim_31615)) {\n        if (!(orig_group_33210 == 0 || scan_stage3_gtid_33205 ==\n              (orig_group_33210 + 1) * (segscan_group_sizze_32185 *\n                                        squot32(flat_dim_31615 +\n                                                num_threads_33143 - 1,\n                                                num_threads_33143)) - 1)) {\n            int32_t x_33179;\n            bool x_33180;\n            int32_t x_33181;\n            bool x_33182;\n            \n            x_33179 = ((__global int32_t *) mem_32815)[carry_in_flat_idx_33211];\n            x_33180 = ((__global bool *) mem_32817)[carry_in_flat_idx_33211];\n            x_33181 = ((__global int32_t *) mem_32815)[gtid_32189];\n            x_33182 = ((__global bool *) mem_32817)[gtid_32189];\n",
            "            \n            int32_t res_33183;\n            \n            if (x_33182) {\n                res_33183 = x_33181;\n            } else {\n                int32_t res_33184 = x_33179 + x_33181;\n                \n                res_33183 = res_33184;\n            }\n            \n            bool res_33185 = x_33180 || x_33182;\n            \n            x_33179 = res_33183;\n            x_33180 = res_33185;\n            ((__global int32_t *) mem_32815)[gtid_32189] = x_33179;\n            ((__global bool *) mem_32817)[gtid_32189] = x_33180;\n        }\n    }\n}\n__kernel void scan_stage3_33274(int32_t flat_dim_31615, __global\n                                unsigned char *mem_32821, __global\n                                unsigned char *mem_32823,\n                                int32_t num_threads_33212)\n{\n    const int32_t segscan_group_sizze_32194 =\n                  quicksort1zisegscan_group_sizze_32193;\n    const int block_dim0 = 0;\n    const int block_dim1 = 1;\n    const int block_dim2 = 2;\n    int32_t scan_stage3_gtid_33274;\n    int32_t scan_stage3_ltid_33275;\n    int32_t scan_stage3_gid_33276;\n    \n    scan_stage3_gtid_33274 = get_global_id(0);\n    scan_stage3_ltid_33275 = get_local_id(0);\n    scan_stage3_gid_33276 = get_group_id(0);\n    \n    int32_t phys_tid_32199 = scan_stage3_gtid_33274;\n    int32_t gtid_32198 = scan_stage3_gtid_33274;\n    int32_t orig_group_33279 = squot32(scan_stage3_gtid_33274,\n                                       segscan_group_sizze_32194 *\n                                       squot32(flat_dim_31615 +\n                                               num_threads_33212 - 1,\n                                               num_threads_33212));\n    int32_t carry_in_flat_idx_33280 = orig_group_33279 *\n            (segscan_group_sizze_32194 * squot32(flat_dim_31615 +\n                                                 num_threads_33212 - 1,\n                                                 num_threads_33212)) - 1;\n    \n    if (slt32(scan_stage3_gtid_332",
            "74, flat_dim_31615)) {\n        if (!(orig_group_33279 == 0 || scan_stage3_gtid_33274 ==\n              (orig_group_33279 + 1) * (segscan_group_sizze_32194 *\n                                        squot32(flat_dim_31615 +\n                                                num_threads_33212 - 1,\n                                                num_threads_33212)) - 1)) {\n            int32_t x_33248;\n            bool x_33249;\n            int32_t x_33250;\n            bool x_33251;\n            \n            x_33248 = ((__global int32_t *) mem_32821)[carry_in_flat_idx_33280];\n            x_33249 = ((__global bool *) mem_32823)[carry_in_flat_idx_33280];\n            x_33250 = ((__global int32_t *) mem_32821)[gtid_32198];\n            x_33251 = ((__global bool *) mem_32823)[gtid_32198];\n            \n            int32_t res_33252;\n            \n            if (x_33251) {\n                res_33252 = x_33250;\n            } else {\n                int32_t res_33253 = x_33248 + x_33250;\n                \n                res_33252 = res_33253;\n            }\n            \n            bool res_33254 = x_33249 || x_33251;\n            \n            x_33248 = res_33252;\n            x_33249 = res_33254;\n            ((__global int32_t *) mem_32821)[gtid_32198] = x_33248;\n            ((__global bool *) mem_32823)[gtid_32198] = x_33249;\n        }\n    }\n}\n__kernel void scan_stage3_33343(int32_t flat_dim_31615, __global\n                                unsigned char *mem_32830, __global\n                                unsigned char *mem_32832,\n                                int32_t num_threads_33286)\n{\n    const int32_t segscan_group_sizze_32220 =\n                  quicksort1zisegscan_group_sizze_32219;\n    const int block_dim0 = 0;\n    const int block_dim1 = 1;\n    const int block_dim2 = 2;\n    int32_t scan_stage3_gtid_33343;\n    int32_t scan_stage3_ltid_33344;\n    int32_t scan_stage3_gid_33345;\n    \n    scan_stage3_gtid_33343 = get_global_id(0);\n    scan_stage3_ltid_33344 = get_local_id(0);\n    scan_",
            "stage3_gid_33345 = get_group_id(0);\n    \n    int32_t phys_tid_32225 = scan_stage3_gtid_33343;\n    int32_t gtid_32224 = scan_stage3_gtid_33343;\n    int32_t orig_group_33348 = squot32(scan_stage3_gtid_33343,\n                                       segscan_group_sizze_32220 *\n                                       squot32(flat_dim_31615 +\n                                               num_threads_33286 - 1,\n                                               num_threads_33286));\n    int32_t carry_in_flat_idx_33349 = orig_group_33348 *\n            (segscan_group_sizze_32220 * squot32(flat_dim_31615 +\n                                                 num_threads_33286 - 1,\n                                                 num_threads_33286)) - 1;\n    \n    if (slt32(scan_stage3_gtid_33343, flat_dim_31615)) {\n        if (!(orig_group_33348 == 0 || scan_stage3_gtid_33343 ==\n              (orig_group_33348 + 1) * (segscan_group_sizze_32220 *\n                                        squot32(flat_dim_31615 +\n                                                num_threads_33286 - 1,\n                                                num_threads_33286)) - 1)) {\n            int32_t x_33319;\n            bool x_33320;\n            int32_t x_33321;\n            bool x_33322;\n            \n            x_33319 = ((__global int32_t *) mem_32830)[carry_in_flat_idx_33349];\n            x_33320 = ((__global bool *) mem_32832)[carry_in_flat_idx_33349];\n            x_33321 = ((__global int32_t *) mem_32830)[gtid_32224];\n            x_33322 = ((__global bool *) mem_32832)[gtid_32224];\n            \n            int32_t res_33323;\n            \n            if (x_33322) {\n                res_33323 = x_33321;\n            } else {\n                res_33323 = x_33319;\n            }\n            \n            bool res_33324 = x_33320 || x_33322;\n            \n            x_33319 = res_33323;\n            x_33320 = res_33324;\n            ((__global int32_t *) mem_32830)[gtid_32224] = x_33319;\n            ((__global bool *) m",
            "em_32832)[gtid_32224] = x_33320;\n        }\n    }\n}\n__kernel void segmap_32152(int32_t flat_dim_31615, __global\n                           unsigned char *mem_32801, __global\n                           unsigned char *mem_32803, __global\n                           unsigned char *mem_32808)\n{\n    const int32_t segmap_group_sizze_32156 =\n                  quicksort1zisegmap_group_sizze_32155;\n    const int block_dim0 = 0;\n    const int block_dim1 = 1;\n    const int block_dim2 = 2;\n    int32_t global_tid_33133;\n    int32_t local_tid_33134;\n    int32_t group_sizze_33137;\n    int32_t wave_sizze_33136;\n    int32_t group_tid_33135;\n    \n    global_tid_33133 = get_global_id(0);\n    local_tid_33134 = get_local_id(0);\n    group_sizze_33137 = get_local_size(0);\n    wave_sizze_33136 = LOCKSTEP_WIDTH;\n    group_tid_33135 = get_group_id(0);\n    \n    int32_t phys_tid_32152 = global_tid_33133;\n    int32_t gtid_32151 = group_tid_33135 * segmap_group_sizze_32156 +\n            local_tid_33134;\n    \n    if (slt32(gtid_32151, flat_dim_31615)) {\n        int32_t x_32163 = ((__global int32_t *) mem_32801)[gtid_32151];\n        bool x_32164 = ((__global bool *) mem_32803)[gtid_32151];\n        int32_t res_32165;\n        \n        if (x_32164) {\n            res_32165 = x_32163;\n        } else {\n            res_32165 = 0;\n        }\n        ((__global int32_t *) mem_32808)[gtid_32151] = res_32165;\n    }\n}\n__kernel void segmap_32168(int32_t flat_dim_31615, __global\n                           unsigned char *mem_32803, __global\n                           unsigned char *mem_32808, __global\n                           unsigned char *mem_32811)\n{\n    const int32_t segmap_group_sizze_32172 =\n                  quicksort1zisegmap_group_sizze_32171;\n    const int block_dim0 = 0;\n    const int block_dim1 = 1;\n    const int block_dim2 = 2;\n    int32_t global_tid_33138;\n    int32_t local_tid_33139;\n    int32_t group_sizze_33142;\n    int32_t wave_sizze_33141;\n    int32_t group_tid_33140;\n    \n    global_tid_33138 ",
            "= get_global_id(0);\n    local_tid_33139 = get_local_id(0);\n    group_sizze_33142 = get_local_size(0);\n    wave_sizze_33141 = LOCKSTEP_WIDTH;\n    group_tid_33140 = get_group_id(0);\n    \n    int32_t phys_tid_32168 = global_tid_33138;\n    int32_t gtid_32167 = group_tid_33140 * segmap_group_sizze_32172 +\n            local_tid_33139;\n    \n    if (slt32(gtid_32167, flat_dim_31615)) {\n        int32_t x_32179 = ((__global int32_t *) mem_32808)[gtid_32167];\n        bool x_32180 = ((__global bool *) mem_32803)[gtid_32167];\n        int32_t res_32181;\n        \n        if (x_32180) {\n            res_32181 = x_32179;\n        } else {\n            res_32181 = 0;\n        }\n        ((__global int32_t *) mem_32811)[gtid_32167] = res_32181;\n    }\n}\n__kernel void segmap_32201(int32_t flat_dim_31615, __global\n                           unsigned char *mem_32821, __global\n                           unsigned char *mem_32826)\n{\n    const int32_t segmap_group_sizze_32205 =\n                  quicksort1zisegmap_group_sizze_32204;\n    const int block_dim0 = 0;\n    const int block_dim1 = 1;\n    const int block_dim2 = 2;\n    int32_t global_tid_33281;\n    int32_t local_tid_33282;\n    int32_t group_sizze_33285;\n    int32_t wave_sizze_33284;\n    int32_t group_tid_33283;\n    \n    global_tid_33281 = get_global_id(0);\n    local_tid_33282 = get_local_id(0);\n    group_sizze_33285 = get_local_size(0);\n    wave_sizze_33284 = LOCKSTEP_WIDTH;\n    group_tid_33283 = get_group_id(0);\n    \n    int32_t phys_tid_32201 = global_tid_33281;\n    int32_t gtid_32200 = group_tid_33283 * segmap_group_sizze_32205 +\n            local_tid_33282;\n    \n    if (slt32(gtid_32200, flat_dim_31615)) {\n        int32_t x_32212 = ((__global int32_t *) mem_32821)[gtid_32200];\n        int32_t res_32216 = x_32212 - 1;\n        \n        ((__global int32_t *) mem_32826)[gtid_32200] = res_32216;\n    }\n}\n__kernel void segmap_32227(int32_t m_31645, int32_t n_31675, __global\n                           unsigned char *mem_32830, __global\n         ",
            "                  unsigned char *mem_32832, __global\n                           unsigned char *mem_32835)\n{\n    const int32_t segmap_group_sizze_32231 =\n                  quicksort1zisegmap_group_sizze_32230;\n    const int block_dim0 = 0;\n    const int block_dim1 = 1;\n    const int block_dim2 = 2;\n    int32_t global_tid_33350;\n    int32_t local_tid_33351;\n    int32_t group_sizze_33354;\n    int32_t wave_sizze_33353;\n    int32_t group_tid_33352;\n    \n    global_tid_33350 = get_global_id(0);\n    local_tid_33351 = get_local_id(0);\n    group_sizze_33354 = get_local_size(0);\n    wave_sizze_33353 = LOCKSTEP_WIDTH;\n    group_tid_33352 = get_group_id(0);\n    \n    int32_t phys_tid_32227 = global_tid_33350;\n    int32_t gtid_32226 = group_tid_33352 * segmap_group_sizze_32231 +\n            local_tid_33351;\n    \n    if (slt32(gtid_32226, n_31675)) {\n        int32_t j_t_s_32592 = -1 * gtid_32226;\n        int32_t j_p_i_t_s_32593 = m_31645 + j_t_s_32592;\n        int32_t x_32238 = ((__global int32_t *) mem_32830)[j_p_i_t_s_32593];\n        bool x_32239 = ((__global bool *) mem_32832)[j_p_i_t_s_32593];\n        int32_t res_32240;\n        \n        if (x_32239) {\n            res_32240 = x_32238;\n        } else {\n            res_32240 = 0;\n        }\n        ((__global int32_t *) mem_32835)[gtid_32226] = res_32240;\n    }\n}\n__kernel void segmap_32242(int32_t flat_dim_31615, __global\n                           unsigned char *mem_32835, __global\n                           unsigned char *mem_32838)\n{\n    const int32_t segmap_group_sizze_32246 =\n                  quicksort1zisegmap_group_sizze_32245;\n    const int block_dim0 = 0;\n    const int block_dim1 = 1;\n    const int block_dim2 = 2;\n    int32_t global_tid_33355;\n    int32_t local_tid_33356;\n    int32_t group_sizze_33359;\n    int32_t wave_sizze_33358;\n    int32_t group_tid_33357;\n    \n    global_tid_33355 = get_global_id(0);\n    local_tid_33356 = get_local_id(0);\n    group_sizze_33359 = get_local_size(0);\n    wave_sizze_33358 = LOCKSTEP_WID",
            "TH;\n    group_tid_33357 = get_group_id(0);\n    \n    int32_t phys_tid_32242 = global_tid_33355;\n    int32_t gtid_32241 = group_tid_33357 * segmap_group_sizze_32246 +\n            local_tid_33356;\n    \n    if (slt32(gtid_32241, flat_dim_31615)) {\n        int32_t x_32253 = ((__global int32_t *) mem_32835)[gtid_32241];\n        int32_t res_32254 = 1 + x_32253;\n        \n        ((__global int32_t *) mem_32838)[gtid_32241] = res_32254;\n    }\n}\n__kernel void segmap_32256(int32_t flat_dim_31615, __global\n                           unsigned char *x_mem_32791, __global\n                           unsigned char *mem_32811, __global\n                           unsigned char *mem_32838, __global\n                           unsigned char *mem_32844)\n{\n    const int32_t segmap_group_sizze_32260 =\n                  quicksort1zisegmap_group_sizze_32259;\n    const int block_dim0 = 0;\n    const int block_dim1 = 1;\n    const int block_dim2 = 2;\n    int32_t global_tid_33369;\n    int32_t local_tid_33370;\n    int32_t group_sizze_33373;\n    int32_t wave_sizze_33372;\n    int32_t group_tid_33371;\n    \n    global_tid_33369 = get_global_id(0);\n    local_tid_33370 = get_local_id(0);\n    group_sizze_33373 = get_local_size(0);\n    wave_sizze_33372 = LOCKSTEP_WIDTH;\n    group_tid_33371 = get_group_id(0);\n    \n    int32_t phys_tid_32256 = global_tid_33369;\n    int32_t gtid_32255 = group_tid_33371 * segmap_group_sizze_32260 +\n            local_tid_33370;\n    \n    if (slt32(gtid_32255, flat_dim_31615)) {\n        bool x_32267 = ((__global bool *) x_mem_32791)[gtid_32255];\n        int32_t res_32270;\n        \n        if (x_32267) {\n            int32_t x_32268 = ((__global int32_t *) mem_32811)[gtid_32255];\n            int32_t x_32269 = ((__global int32_t *) mem_32838)[gtid_32255];\n            int32_t x_32271 = x_32268 + x_32269;\n            int32_t res_32272 = 1 + x_32271;\n            \n            res_32270 = res_32272;\n        } else {\n            res_32270 = -1;\n        }\n        ((__global int32_t *) mem_",
            "32844)[gtid_32255] = res_32270;\n    }\n}\n__kernel void segmap_32274(int32_t flat_dim_31615, int32_t fill_arg_31617,\n                           __global unsigned char *mem_32844, __global\n                           unsigned char *mem_32846)\n{\n    const int32_t segmap_group_sizze_32278 =\n                  quicksort1zisegmap_group_sizze_32277;\n    const int block_dim0 = 0;\n    const int block_dim1 = 1;\n    const int block_dim2 = 2;\n    int32_t global_tid_33374;\n    int32_t local_tid_33375;\n    int32_t group_sizze_33378;\n    int32_t wave_sizze_33377;\n    int32_t group_tid_33376;\n    \n    global_tid_33374 = get_global_id(0);\n    local_tid_33375 = get_local_id(0);\n    group_sizze_33378 = get_local_size(0);\n    wave_sizze_33377 = LOCKSTEP_WIDTH;\n    group_tid_33376 = get_group_id(0);\n    \n    int32_t phys_tid_32274 = global_tid_33374;\n    int32_t write_i_32273 = group_tid_33376 * segmap_group_sizze_32278 +\n            local_tid_33375;\n    \n    if (slt32(write_i_32273, flat_dim_31615)) {\n        int32_t write_index_31854 = ((__global\n                                      int32_t *) mem_32844)[write_i_32273];\n        \n        if (sle32(0, write_index_31854) && slt32(write_index_31854,\n                                                 fill_arg_31617)) {\n            ((__global bool *) mem_32846)[write_index_31854] = 1;\n        }\n    }\n}\n__kernel void segmap_32286(int32_t flat_dim_31615, int32_t fill_arg_31617,\n                           int32_t m_31645, __global unsigned char *x_mem_32790,\n                           __global unsigned char *x_mem_32791, __global\n                           unsigned char *mem_32803, __global\n                           unsigned char *mem_32805, __global\n                           unsigned char *mem_32808, __global\n                           unsigned char *mem_32815, __global\n                           unsigned char *mem_32826, __global\n                           unsigned char *mem_32830, __global\n                           unsigned char *mem_32832, ",
            "__global\n                           unsigned char *mem_32841, __global\n                           unsigned char *mem_32846)\n{\n    const int32_t segmap_group_sizze_32290 =\n                  quicksort1zisegmap_group_sizze_32289;\n    const int block_dim0 = 0;\n    const int block_dim1 = 1;\n    const int block_dim2 = 2;\n    int32_t global_tid_33379;\n    int32_t local_tid_33380;\n    int32_t group_sizze_33383;\n    int32_t wave_sizze_33382;\n    int32_t group_tid_33381;\n    \n    global_tid_33379 = get_global_id(0);\n    local_tid_33380 = get_local_id(0);\n    group_sizze_33383 = get_local_size(0);\n    wave_sizze_33382 = LOCKSTEP_WIDTH;\n    group_tid_33381 = get_group_id(0);\n    \n    int32_t phys_tid_32286 = global_tid_33379;\n    int32_t write_i_32285 = group_tid_33381 * segmap_group_sizze_32290 +\n            local_tid_33380;\n    \n    if (slt32(write_i_32285, flat_dim_31615)) {\n        int32_t x_31864 = ((__global int32_t *) mem_32808)[write_i_32285];\n        bool x_31865 = ((__global bool *) mem_32803)[write_i_32285];\n        bool x_31868 = ((__global bool *) x_mem_32791)[write_i_32285];\n        int32_t x_31869 = ((__global int32_t *) mem_32815)[write_i_32285];\n        int32_t j_t_s_32598 = -1 * write_i_32285;\n        int32_t j_p_i_t_s_32599 = m_31645 + j_t_s_32598;\n        int32_t x_31870 = ((__global int32_t *) mem_32830)[j_p_i_t_s_32599];\n        bool x_31871 = ((__global bool *) mem_32832)[j_p_i_t_s_32599];\n        bool x_31872 = ((__global bool *) mem_32805)[write_i_32285];\n        int32_t x_31873 = ((__global int32_t *) mem_32826)[write_i_32285];\n        int32_t write_value_31874 = ((__global\n                                      int32_t *) x_mem_32790)[write_i_32285];\n        int32_t res_31881;\n        \n        if (x_31865) {\n            res_31881 = x_31864;\n        } else {\n            res_31881 = 0;\n        }\n        \n        int32_t res_31882 = x_31869 - 1;\n        int32_t res_31883;\n        \n        if (x_31871) {\n            res_31883 = x_31870;\n        } else {\n  ",
            "          res_31883 = 0;\n        }\n        \n        int32_t res_31884 = 1 + res_31883;\n        int32_t y_31885;\n        \n        if (x_31872) {\n            int32_t res_31886 = res_31882 + res_31884;\n            \n            y_31885 = res_31886;\n        } else {\n            y_31885 = x_31873;\n        }\n        \n        int32_t res_31887 = res_31881 + y_31885;\n        int32_t res_31888;\n        \n        if (x_31868) {\n            int32_t x_31889 = res_31881 + res_31884;\n            \n            res_31888 = x_31889;\n        } else {\n            res_31888 = -1;\n        }\n        if (sle32(0, res_31888) && slt32(res_31888, fill_arg_31617)) {\n            ((__global bool *) mem_32846)[res_31888] = 1;\n        }\n        if (sle32(0, res_31887) && slt32(res_31887, flat_dim_31615)) {\n            ((__global int32_t *) mem_32841)[res_31887] = write_value_31874;\n        }\n    }\n}\n__kernel void segmap_32311(int32_t n_31902, int32_t m_31903,\n                           int32_t fill_arg_31905, int32_t w_minus_1_31906,\n                           int32_t n_31908, int32_t j_p_s0_t_i_31909,\n                           int32_t num_groups_32317, __global\n                           unsigned char *mem_32791, __global\n                           unsigned char *mem_32795, __global\n                           unsigned char *mem_32798, __global\n                           unsigned char *mem_32810, __global\n                           unsigned char *mem_32812, __global\n                           unsigned char *mem_32815, __global\n                           unsigned char *mem_32817, __global\n                           unsigned char *mem_32819, __global\n                           unsigned char *mem_32822, __global\n                           unsigned char *mem_32824, __global\n                           unsigned char *mem_32827, __global\n                           unsigned char *mem_32830, __global\n                           unsigned char *mem_32833, __global\n                           unsigned char *mem_",
            "32836, __global\n                           unsigned char *mem_32839, __global\n                           unsigned char *mem_32841, __global\n                           unsigned char *mem_32844, __global\n                           unsigned char *mem_32847, __global\n                           unsigned char *mem_32850, __global\n                           unsigned char *mem_32853, __global\n                           unsigned char *mem_32855, __global\n                           unsigned char *mem_32861, __global\n                           unsigned char *double_buffer_mem_32886, __global\n                           unsigned char *double_buffer_mem_32887)\n{\n    const int32_t segmap_group_sizze_32315 = mainzisegmap_group_sizze_32314;\n    const int block_dim0 = 0;\n    const int block_dim1 = 1;\n    const int block_dim2 = 2;\n    int32_t global_tid_33473;\n    int32_t local_tid_33474;\n    int32_t group_sizze_33477;\n    int32_t wave_sizze_33476;\n    int32_t group_tid_33475;\n    \n    global_tid_33473 = get_global_id(0);\n    local_tid_33474 = get_local_id(0);\n    group_sizze_33477 = get_local_size(0);\n    wave_sizze_33476 = LOCKSTEP_WIDTH;\n    group_tid_33475 = get_group_id(0);\n    \n    int32_t phys_tid_32311 = global_tid_33473;\n    int32_t phys_group_id_33478;\n    \n    phys_group_id_33478 = get_group_id(0);\n    for (int32_t i_33479 = 0; i_33479 < squot32(squot32(n_31902 +\n                                                        segmap_group_sizze_32315 -\n                                                        1,\n                                                        segmap_group_sizze_32315) -\n                                                phys_group_id_33478 +\n                                                num_groups_32317 - 1,\n                                                num_groups_32317); i_33479++) {\n        int32_t virt_group_id_33480 = phys_group_id_33478 + i_33479 *\n                num_groups_32317;\n        int32_t gtid_32310 = virt_group_id_33480 * segmap_group_sizze_32",
            "315 +\n                local_tid_33474;\n        \n        if (slt32(gtid_32310, n_31902)) {\n            bool res_32321 = ((__global bool *) mem_32791)[gtid_32310];\n            bool res_32322 = !res_32321;\n            \n            for (int32_t i_33481 = 0; i_33481 < m_31903; i_33481++) {\n                ((__global int32_t *) double_buffer_mem_32886)[phys_tid_32311 +\n                                                               i_33481 *\n                                                               (num_groups_32317 *\n                                                                segmap_group_sizze_32315)] =\n                    ((__global int32_t *) mem_32795)[gtid_32310 + i_33481 *\n                                                     n_31902];\n            }\n            for (int32_t i_33482 = 0; i_33482 < fill_arg_31905; i_33482++) {\n                ((__global bool *) double_buffer_mem_32887)[phys_tid_32311 +\n                                                            i_33482 *\n                                                            (num_groups_32317 *\n                                                             segmap_group_sizze_32315)] =\n                    ((__global bool *) mem_32798)[gtid_32310 + i_33482 *\n                                                  n_31902];\n            }\n            \n            bool res_32323;\n            bool loop_while_32326 = res_32322;\n            \n            while (loop_while_32326) {\n                int32_t discard_32642;\n                bool discard_32643;\n                int32_t scanacc_32635;\n                bool scanacc_32636;\n                \n                scanacc_32635 = 0;\n                scanacc_32636 = 0;\n                for (int32_t i_32639 = 0; i_32639 < m_31903; i_32639++) {\n                    int32_t x_32353 = ((__global\n                                        int32_t *) double_buffer_mem_32886)[phys_tid_32311 +\n                                                                            i_32639 *\n            ",
            "                                                                (num_groups_32317 *\n                                                                             segmap_group_sizze_32315)];\n                    bool x_32354 = ((__global\n                                     bool *) double_buffer_mem_32887)[phys_tid_32311 +\n                                                                      i_32639 *\n                                                                      (num_groups_32317 *\n                                                                       segmap_group_sizze_32315)];\n                    int32_t res_32351;\n                    \n                    if (x_32354) {\n                        res_32351 = x_32353;\n                    } else {\n                        res_32351 = scanacc_32635;\n                    }\n                    \n                    bool res_32352 = x_32354 || scanacc_32636;\n                    \n                    ((__global int32_t *) mem_32810)[phys_tid_32311 + i_32639 *\n                                                     (num_groups_32317 *\n                                                      segmap_group_sizze_32315)] =\n                        res_32351;\n                    ((__global bool *) mem_32812)[phys_tid_32311 + i_32639 *\n                                                  (num_groups_32317 *\n                                                   segmap_group_sizze_32315)] =\n                        res_32352;\n                    \n                    int32_t scanacc_tmp_33486 = res_32351;\n                    bool scanacc_tmp_33487;\n                    \n                    scanacc_tmp_33487 = res_32352;\n                    scanacc_32635 = scanacc_tmp_33486;\n                    scanacc_32636 = scanacc_tmp_33487;\n                }\n                discard_32642 = scanacc_32635;\n                discard_32643 = scanacc_32636;\n                \n                int32_t discard_32656;\n                bool discard_32657;\n                int",
            "32_t scanacc_32647;\n                bool scanacc_32648;\n                \n                scanacc_32647 = 0;\n                scanacc_32648 = 0;\n                for (int32_t i_32652 = 0; i_32652 < m_31903; i_32652++) {\n                    int32_t x_32372 = ((__global\n                                        int32_t *) mem_32810)[phys_tid_32311 +\n                                                              i_32652 *\n                                                              (num_groups_32317 *\n                                                               segmap_group_sizze_32315)];\n                    bool x_32373 = ((__global\n                                     bool *) mem_32812)[phys_tid_32311 +\n                                                        i_32652 *\n                                                        (num_groups_32317 *\n                                                         segmap_group_sizze_32315)];\n                    int32_t x_32374 = ((__global\n                                        int32_t *) double_buffer_mem_32886)[phys_tid_32311 +\n                                                                            i_32652 *\n                                                                            (num_groups_32317 *\n                                                                             segmap_group_sizze_32315)];\n                    bool x_32376 = ((__global\n                                     bool *) double_buffer_mem_32887)[phys_tid_32311 +\n                                                                      i_32652 *\n                                                                      (num_groups_32317 *\n                                                                       segmap_group_sizze_32315)];\n                    int32_t res_32377;\n                    \n                    if (x_32373) {\n                        res_32377 = x_32372;\n                    } else {\n                        res_32377 = 0;\n                    }\n   ",
            "                 \n                    int32_t res_32379;\n                    \n                    if (x_32373) {\n                        res_32379 = res_32377;\n                    } else {\n                        res_32379 = 0;\n                    }\n                    \n                    bool res_32380 = sle32(res_32379, x_32374);\n                    int32_t res_32370;\n                    \n                    if (x_32376) {\n                        res_32370 = i_32652;\n                    } else {\n                        res_32370 = scanacc_32647;\n                    }\n                    \n                    bool res_32371 = x_32376 || scanacc_32648;\n                    \n                    ((__global int32_t *) mem_32815)[phys_tid_32311 + i_32652 *\n                                                     (num_groups_32317 *\n                                                      segmap_group_sizze_32315)] =\n                        res_32370;\n                    ((__global bool *) mem_32817)[phys_tid_32311 + i_32652 *\n                                                  (num_groups_32317 *\n                                                   segmap_group_sizze_32315)] =\n                        res_32371;\n                    ((__global bool *) mem_32819)[phys_tid_32311 + i_32652 *\n                                                  (num_groups_32317 *\n                                                   segmap_group_sizze_32315)] =\n                        res_32380;\n                    \n                    int32_t scanacc_tmp_33490 = res_32370;\n                    bool scanacc_tmp_33491;\n                    \n                    scanacc_tmp_33491 = res_32371;\n                    scanacc_32647 = scanacc_tmp_33490;\n                    scanacc_32648 = scanacc_tmp_33491;\n                }\n                discard_32656 = scanacc_32647;\n                discard_32657 = scanacc_32648;\n                for (int32_t i_32662 = 0; i_32662 < m_31903; i_32662++) {\n                    int32_t x_3",
            "2387 = ((__global\n                                        int32_t *) mem_32815)[phys_tid_32311 +\n                                                              i_32662 *\n                                                              (num_groups_32317 *\n                                                               segmap_group_sizze_32315)];\n                    bool x_32388 = ((__global\n                                     bool *) mem_32817)[phys_tid_32311 +\n                                                        i_32662 *\n                                                        (num_groups_32317 *\n                                                         segmap_group_sizze_32315)];\n                    int32_t res_32389;\n                    \n                    if (x_32388) {\n                        res_32389 = x_32387;\n                    } else {\n                        res_32389 = 0;\n                    }\n                    ((__global int32_t *) mem_32822)[phys_tid_32311 + i_32662 *\n                                                     (num_groups_32317 *\n                                                      segmap_group_sizze_32315)] =\n                        res_32389;\n                    for (int32_t i_33497 = 0; i_33497 < 1; i_33497++) {\n                        ((__global bool *) mem_32824)[phys_tid_32311 +\n                                                      num_groups_32317 *\n                                                      segmap_group_sizze_32315 *\n                                                      i_32662 + i_33497 *\n                                                      (num_groups_32317 *\n                                                       segmap_group_sizze_32315)] =\n                            ((__global bool *) mem_32817)[phys_tid_32311 +\n                                                          num_groups_32317 *\n                                                          segmap_group_sizze_32315 *\n                                             ",
            "             i_32662 + i_33497 *\n                                                          (num_groups_32317 *\n                                                           segmap_group_sizze_32315)];\n                    }\n                }\n                for (int32_t i_32667 = 0; i_32667 < m_31903; i_32667++) {\n                    int32_t x_32394 = ((__global\n                                        int32_t *) mem_32822)[phys_tid_32311 +\n                                                              i_32667 *\n                                                              (num_groups_32317 *\n                                                               segmap_group_sizze_32315)];\n                    bool x_32395 = ((__global\n                                     bool *) mem_32824)[phys_tid_32311 +\n                                                        i_32667 *\n                                                        (num_groups_32317 *\n                                                         segmap_group_sizze_32315)];\n                    int32_t res_32396;\n                    \n                    if (x_32395) {\n                        res_32396 = x_32394;\n                    } else {\n                        res_32396 = 0;\n                    }\n                    ((__global int32_t *) mem_32827)[phys_tid_32311 + i_32667 *\n                                                     (num_groups_32317 *\n                                                      segmap_group_sizze_32315)] =\n                        res_32396;\n                }\n                \n                int32_t discard_32678;\n                int32_t scanacc_32671 = 0;\n                \n                for (int32_t i_32675 = 0; i_32675 < m_31903; i_32675++) {\n                    bool x_32406 = ((__global\n                                     bool *) mem_32819)[phys_tid_32311 +\n                                                        i_32675 *\n                                                        (num_groups_32317 *",
            "\n                                                         segmap_group_sizze_32315)];\n                    bool x_32407 = ((__global\n                                     bool *) double_buffer_mem_32887)[phys_tid_32311 +\n                                                                      i_32675 *\n                                                                      (num_groups_32317 *\n                                                                       segmap_group_sizze_32315)];\n                    int32_t res_32408;\n                    \n                    if (x_32406) {\n                        res_32408 = 1;\n                    } else {\n                        res_32408 = 0;\n                    }\n                    \n                    int32_t res_32403;\n                    \n                    if (x_32407) {\n                        res_32403 = res_32408;\n                    } else {\n                        int32_t res_32404 = res_32408 + scanacc_32671;\n                        \n                        res_32403 = res_32404;\n                    }\n                    ((__global int32_t *) mem_32830)[phys_tid_32311 + i_32675 *\n                                                     (num_groups_32317 *\n                                                      segmap_group_sizze_32315)] =\n                        res_32403;\n                    \n                    int32_t scanacc_tmp_33499 = res_32403;\n                    \n                    scanacc_32671 = scanacc_tmp_33499;\n                }\n                discard_32678 = scanacc_32671;\n                \n                int32_t discard_32689;\n                int32_t scanacc_32682 = 0;\n                \n                for (int32_t i_32686 = 0; i_32686 < m_31903; i_32686++) {\n                    bool x_32426 = ((__global\n                                     bool *) mem_32819)[phys_tid_32311 +\n                                                        i_32686 *\n                                                        (num_gro",
            "ups_32317 *\n                                                         segmap_group_sizze_32315)];\n                    bool x_32427 = ((__global\n                                     bool *) double_buffer_mem_32887)[phys_tid_32311 +\n                                                                      i_32686 *\n                                                                      (num_groups_32317 *\n                                                                       segmap_group_sizze_32315)];\n                    int32_t res_32428;\n                    \n                    if (x_32426) {\n                        res_32428 = 0;\n                    } else {\n                        res_32428 = 1;\n                    }\n                    \n                    int32_t res_32423;\n                    \n                    if (x_32427) {\n                        res_32423 = res_32428;\n                    } else {\n                        int32_t res_32424 = res_32428 + scanacc_32682;\n                        \n                        res_32423 = res_32424;\n                    }\n                    ((__global int32_t *) mem_32833)[phys_tid_32311 + i_32686 *\n                                                     (num_groups_32317 *\n                                                      segmap_group_sizze_32315)] =\n                        res_32423;\n                    \n                    int32_t scanacc_tmp_33501 = res_32423;\n                    \n                    scanacc_32682 = scanacc_tmp_33501;\n                }\n                discard_32689 = scanacc_32682;\n                for (int32_t i_32693 = 0; i_32693 < m_31903; i_32693++) {\n                    int32_t x_32439 = ((__global\n                                        int32_t *) mem_32833)[phys_tid_32311 +\n                                                              i_32693 *\n                                                              (num_groups_32317 *\n                                                               segmap_gr",
            "oup_sizze_32315)];\n                    int32_t res_32443 = x_32439 - 1;\n                    \n                    ((__global int32_t *) mem_32836)[phys_tid_32311 + i_32693 *\n                                                     (num_groups_32317 *\n                                                      segmap_group_sizze_32315)] =\n                        res_32443;\n                }\n                \n                int32_t discard_32704;\n                bool discard_32705;\n                int32_t scanacc_32697;\n                bool scanacc_32698;\n                \n                scanacc_32697 = 0;\n                scanacc_32698 = 0;\n                for (int32_t i_32701 = 0; i_32701 < m_31903; i_32701++) {\n                    int32_t j_t_s_32767 = -1 * i_32701;\n                    int32_t j_p_i_t_s_32768 = w_minus_1_31906 + j_t_s_32767;\n                    int32_t x_32456 = ((__global\n                                        int32_t *) mem_32836)[phys_tid_32311 +\n                                                              j_p_i_t_s_32768 *\n                                                              (num_groups_32317 *\n                                                               segmap_group_sizze_32315)];\n                    int32_t j_p_i_t_s_32766 = j_p_s0_t_i_31909 + j_t_s_32767;\n                    bool x_32457 = ((__global\n                                     bool *) double_buffer_mem_32887)[phys_tid_32311 +\n                                                                      j_p_i_t_s_32766 *\n                                                                      (num_groups_32317 *\n                                                                       segmap_group_sizze_32315)];\n                    int32_t res_32454;\n                    \n                    if (x_32457) {\n                        res_32454 = x_32456;\n                    } else {\n                        res_32454 = scanacc_32697;\n                    }\n                    \n                    bool ",
            "res_32455 = x_32457 || scanacc_32698;\n                    \n                    ((__global int32_t *) mem_32839)[phys_tid_32311 + i_32701 *\n                                                     (num_groups_32317 *\n                                                      segmap_group_sizze_32315)] =\n                        res_32454;\n                    ((__global bool *) mem_32841)[phys_tid_32311 + i_32701 *\n                                                  (num_groups_32317 *\n                                                   segmap_group_sizze_32315)] =\n                        res_32455;\n                    \n                    int32_t scanacc_tmp_33504 = res_32454;\n                    bool scanacc_tmp_33505;\n                    \n                    scanacc_tmp_33505 = res_32455;\n                    scanacc_32697 = scanacc_tmp_33504;\n                    scanacc_32698 = scanacc_tmp_33505;\n                }\n                discard_32704 = scanacc_32697;\n                discard_32705 = scanacc_32698;\n                for (int32_t i_32708 = 0; i_32708 < n_31908; i_32708++) {\n                    int32_t j_t_s_32771 = -1 * i_32708;\n                    int32_t j_p_i_t_s_32772 = w_minus_1_31906 + j_t_s_32771;\n                    int32_t x_32461 = ((__global\n                                        int32_t *) mem_32839)[phys_tid_32311 +\n                                                              j_p_i_t_s_32772 *\n                                                              (num_groups_32317 *\n                                                               segmap_group_sizze_32315)];\n                    bool x_32462 = ((__global\n                                     bool *) mem_32841)[phys_tid_32311 +\n                                                        j_p_i_t_s_32772 *\n                                                        (num_groups_32317 *\n                                                         segmap_group_sizze_32315)];\n                    int32_t res_32463;\n         ",
            "           \n                    if (x_32462) {\n                        res_32463 = x_32461;\n                    } else {\n                        res_32463 = 0;\n                    }\n                    ((__global int32_t *) mem_32844)[phys_tid_32311 + i_32708 *\n                                                     (num_groups_32317 *\n                                                      segmap_group_sizze_32315)] =\n                        res_32463;\n                }\n                for (int32_t i_32712 = 0; i_32712 < m_31903; i_32712++) {\n                    int32_t x_32466 = ((__global\n                                        int32_t *) mem_32844)[phys_tid_32311 +\n                                                              i_32712 *\n                                                              (num_groups_32317 *\n                                                               segmap_group_sizze_32315)];\n                    int32_t res_32467 = 1 + x_32466;\n                    \n                    ((__global int32_t *) mem_32847)[phys_tid_32311 + i_32712 *\n                                                     (num_groups_32317 *\n                                                      segmap_group_sizze_32315)] =\n                        res_32467;\n                }\n                for (int32_t i_33510 = 0; i_33510 < m_31903; i_33510++) {\n                    ((__global int32_t *) mem_32850)[phys_tid_32311 + i_33510 *\n                                                     (num_groups_32317 *\n                                                      segmap_group_sizze_32315)] =\n                        0;\n                }\n                for (int32_t i_32716 = 0; i_32716 < m_31903; i_32716++) {\n                    bool x_32470 = ((__global\n                                     bool *) double_buffer_mem_32887)[phys_tid_32311 +\n                                                                      i_32716 *\n                                                                      (num_gr",
            "oups_32317 *\n                                                                       segmap_group_sizze_32315)];\n                    int32_t res_32473;\n                    \n                    if (x_32470) {\n                        int32_t x_32471 = ((__global\n                                            int32_t *) mem_32827)[phys_tid_32311 +\n                                                                  i_32716 *\n                                                                  (num_groups_32317 *\n                                                                   segmap_group_sizze_32315)];\n                        int32_t x_32472 = ((__global\n                                            int32_t *) mem_32847)[phys_tid_32311 +\n                                                                  i_32716 *\n                                                                  (num_groups_32317 *\n                                                                   segmap_group_sizze_32315)];\n                        int32_t x_32474 = x_32471 + x_32472;\n                        int32_t res_32475 = 1 + x_32474;\n                        \n                        res_32473 = res_32475;\n                    } else {\n                        res_32473 = -1;\n                    }\n                    ((__global int32_t *) mem_32853)[phys_tid_32311 + i_32716 *\n                                                     (num_groups_32317 *\n                                                      segmap_group_sizze_32315)] =\n                        res_32473;\n                }\n                for (int32_t i_33512 = 0; i_33512 < fill_arg_31905; i_33512++) {\n                    ((__global bool *) mem_32855)[phys_tid_32311 + i_33512 *\n                                                  (num_groups_32317 *\n                                                   segmap_group_sizze_32315)] =\n                        ((__global\n                          bool *) double_buffer_mem_32887)[phys_tid_32311 +\n                  ",
            "                                         i_33512 *\n                                                           (num_groups_32317 *\n                                                            segmap_group_sizze_32315)];\n                }\n                for (int32_t write_iter_32718 = 0; write_iter_32718 < m_31903;\n                     write_iter_32718++) {\n                    int32_t write_iv_32720 = ((__global\n                                               int32_t *) mem_32853)[phys_tid_32311 +\n                                                                     write_iter_32718 *\n                                                                     (num_groups_32317 *\n                                                                      segmap_group_sizze_32315)];\n                    bool less_than_zzero_32721 = slt32(write_iv_32720, 0);\n                    bool greater_than_sizze_32722 = sle32(fill_arg_31905,\n                                                          write_iv_32720);\n                    bool outside_bounds_dim_32723 = less_than_zzero_32721 ||\n                         greater_than_sizze_32722;\n                    \n                    if (!outside_bounds_dim_32723) {\n                        ((__global bool *) mem_32855)[phys_tid_32311 +\n                                                      write_iv_32720 *\n                                                      (num_groups_32317 *\n                                                       segmap_group_sizze_32315)] =\n                            1;\n                    }\n                }\n                for (int32_t write_iter_32727 = 0; write_iter_32727 < m_31903;\n                     write_iter_32727++) {\n                    int32_t write_iv_32730 = ((__global\n                                               int32_t *) mem_32822)[phys_tid_32311 +\n                                                                     write_iter_32727 *\n                                                                     (num_gr",
            "oups_32317 *\n                                                                      segmap_group_sizze_32315)];\n                    bool write_iv_32731 = ((__global\n                                            bool *) mem_32824)[phys_tid_32311 +\n                                                               write_iter_32727 *\n                                                               (num_groups_32317 *\n                                                                segmap_group_sizze_32315)];\n                    int32_t j_t_s_32780 = -1 * write_iter_32727;\n                    int32_t j_p_i_t_s_32781 = w_minus_1_31906 + j_t_s_32780;\n                    int32_t write_iv_32732 = ((__global\n                                               int32_t *) mem_32839)[phys_tid_32311 +\n                                                                     j_p_i_t_s_32781 *\n                                                                     (num_groups_32317 *\n                                                                      segmap_group_sizze_32315)];\n                    bool write_iv_32733 = ((__global\n                                            bool *) mem_32841)[phys_tid_32311 +\n                                                               j_p_i_t_s_32781 *\n                                                               (num_groups_32317 *\n                                                                segmap_group_sizze_32315)];\n                    bool write_iv_32734 = ((__global\n                                            bool *) double_buffer_mem_32887)[phys_tid_32311 +\n                                                                             write_iter_32727 *\n                                                                             (num_groups_32317 *\n                                                                              segmap_group_sizze_32315)];\n                    int32_t write_iv_32735 = ((__global\n                                               int",
            "32_t *) mem_32830)[phys_tid_32311 +\n                                                                     write_iter_32727 *\n                                                                     (num_groups_32317 *\n                                                                      segmap_group_sizze_32315)];\n                    bool write_iv_32738 = ((__global\n                                            bool *) mem_32819)[phys_tid_32311 +\n                                                               write_iter_32727 *\n                                                               (num_groups_32317 *\n                                                                segmap_group_sizze_32315)];\n                    int32_t write_iv_32739 = ((__global\n                                               int32_t *) mem_32836)[phys_tid_32311 +\n                                                                     write_iter_32727 *\n                                                                     (num_groups_32317 *\n                                                                      segmap_group_sizze_32315)];\n                    int32_t res_32494;\n                    \n                    if (write_iv_32731) {\n                        res_32494 = write_iv_32730;\n                    } else {\n                        res_32494 = 0;\n                    }\n                    \n                    int32_t res_32495 = write_iv_32735 - 1;\n                    int32_t res_32496;\n                    \n                    if (write_iv_32733) {\n                        res_32496 = write_iv_32732;\n                    } else {\n                        res_32496 = 0;\n                    }\n                    \n                    int32_t res_32497 = 1 + res_32496;\n                    int32_t y_32498;\n                    \n                    if (write_iv_32738) {\n                        int32_t res_32499 = res_32495 + res_32497;\n                        \n                        y_32498 = res_32499;\n",
            "                    } else {\n                        y_32498 = write_iv_32739;\n                    }\n                    \n                    int32_t res_32500 = res_32494 + y_32498;\n                    int32_t res_32501;\n                    \n                    if (write_iv_32734) {\n                        int32_t x_32502 = res_32494 + res_32497;\n                        \n                        res_32501 = x_32502;\n                    } else {\n                        res_32501 = -1;\n                    }\n                    \n                    bool less_than_zzero_32741 = slt32(res_32501, 0);\n                    bool greater_than_sizze_32742 = sle32(fill_arg_31905,\n                                                          res_32501);\n                    bool outside_bounds_dim_32743 = less_than_zzero_32741 ||\n                         greater_than_sizze_32742;\n                    \n                    if (!outside_bounds_dim_32743) {\n                        ((__global bool *) mem_32855)[phys_tid_32311 +\n                                                      res_32501 *\n                                                      (num_groups_32317 *\n                                                       segmap_group_sizze_32315)] =\n                            1;\n                    }\n                    \n                    bool less_than_zzero_32747 = slt32(res_32500, 0);\n                    bool greater_than_sizze_32748 = sle32(m_31903, res_32500);\n                    bool outside_bounds_dim_32749 = less_than_zzero_32747 ||\n                         greater_than_sizze_32748;\n                    \n                    if (!outside_bounds_dim_32749) {\n                        for (int32_t i_33516 = 0; i_33516 < 1; i_33516++) {\n                            ((__global int32_t *) mem_32850)[phys_tid_32311 +\n                                                             num_groups_32317 *\n                                                             segmap_group_sizze_32315 *\n          ",
            "                                                   res_32500 +\n                                                             i_33516 *\n                                                             (num_groups_32317 *\n                                                              segmap_group_sizze_32315)] =\n                                ((__global\n                                  int32_t *) double_buffer_mem_32886)[phys_tid_32311 +\n                                                                      num_groups_32317 *\n                                                                      segmap_group_sizze_32315 *\n                                                                      write_iter_32727 +\n                                                                      i_33516 *\n                                                                      (num_groups_32317 *\n                                                                       segmap_group_sizze_32315)];\n                        }\n                    }\n                }\n                \n                bool res_32503;\n                bool redout_32753 = 1;\n                \n                for (int32_t i_32754 = 0; i_32754 < fill_arg_31905; i_32754++) {\n                    bool x_32507 = ((__global\n                                     bool *) mem_32855)[phys_tid_32311 +\n                                                        i_32754 *\n                                                        (num_groups_32317 *\n                                                         segmap_group_sizze_32315)];\n                    bool x_32506 = x_32507 && redout_32753;\n                    bool redout_tmp_33517 = x_32506;\n                    \n                    redout_32753 = redout_tmp_33517;\n                }\n                res_32503 = redout_32753;\n                \n                bool res_32508 = !res_32503;\n                \n                for (int32_t i_33518 = 0; i_33518 < m_31903; i_33518++) {\n                    ((",
            "__global\n                      int32_t *) double_buffer_mem_32886)[phys_tid_32311 +\n                                                          i_33518 *\n                                                          (num_groups_32317 *\n                                                           segmap_group_sizze_32315)] =\n                        ((__global int32_t *) mem_32850)[phys_tid_32311 +\n                                                         i_33518 *\n                                                         (num_groups_32317 *\n                                                          segmap_group_sizze_32315)];\n                }\n                for (int32_t i_33519 = 0; i_33519 < fill_arg_31905; i_33519++) {\n                    ((__global bool *) double_buffer_mem_32887)[phys_tid_32311 +\n                                                                i_33519 *\n                                                                (num_groups_32317 *\n                                                                 segmap_group_sizze_32315)] =\n                        ((__global bool *) mem_32855)[phys_tid_32311 + i_33519 *\n                                                      (num_groups_32317 *\n                                                       segmap_group_sizze_32315)];\n                }\n                \n                bool loop_while_tmp_33483 = res_32508;\n                \n                loop_while_32326 = loop_while_tmp_33483;\n            }\n            res_32323 = loop_while_32326;\n            for (int32_t i_33520 = 0; i_33520 < m_31903; i_33520++) {\n                ((__global int32_t *) mem_32861)[gtid_32310 + i_33520 *\n                                                 n_31902] = ((__global\n                                                              int32_t *) double_buffer_mem_32886)[phys_tid_32311 +\n                                                                                                  i_33520 *\n                                                    ",
            "                                              (num_groups_32317 *\n                                                                                                   segmap_group_sizze_32315)];\n            }\n        }\n    }\n}\n__kernel void segmap_32531(int32_t n_31902, int32_t m_31903,\n                           int32_t fill_arg_31905, __global\n                           unsigned char *mem_32788)\n{\n    const int32_t segmap_group_sizze_32537 = mainzisegmap_group_sizze_32536;\n    const int block_dim0 = 0;\n    const int block_dim1 = 1;\n    const int block_dim2 = 2;\n    int32_t global_tid_33418;\n    int32_t local_tid_33419;\n    int32_t group_sizze_33422;\n    int32_t wave_sizze_33421;\n    int32_t group_tid_33420;\n    \n    global_tid_33418 = get_global_id(0);\n    local_tid_33419 = get_local_id(0);\n    group_sizze_33422 = get_local_size(0);\n    wave_sizze_33421 = LOCKSTEP_WIDTH;\n    group_tid_33420 = get_group_id(0);\n    \n    int32_t phys_tid_32531 = global_tid_33418;\n    int32_t gtid_32529 = group_tid_33420 * segmap_group_sizze_32537 +\n            local_tid_33419;\n    int32_t gtid_32530;\n    \n    gtid_32530 = group_tid_33420 * segmap_group_sizze_32537 + local_tid_33419 -\n        (group_tid_33420 * segmap_group_sizze_32537 + local_tid_33419);\n    if (slt32(gtid_32529, n_31902) && slt32(gtid_32530, 1)) {\n        if ((sle32(0, gtid_32529) && slt32(gtid_32529, n_31902)) && (sle32(0,\n                                                                           m_31903) &&\n                                                                     slt32(m_31903,\n                                                                           fill_arg_31905))) {\n            ((__global bool *) mem_32788)[gtid_32529 * fill_arg_31905 +\n                                          m_31903] = 1;\n        }\n    }\n}\n__kernel void segmap_32547(int32_t n_31902, int32_t fill_arg_31905, __global\n                           unsigned char *mem_32788)\n{\n    const int32_t segmap_group_sizze_32553 = mainzisegmap_gro",
            "up_sizze_32552;\n    const int block_dim0 = 0;\n    const int block_dim1 = 1;\n    const int block_dim2 = 2;\n    int32_t global_tid_33413;\n    int32_t local_tid_33414;\n    int32_t group_sizze_33417;\n    int32_t wave_sizze_33416;\n    int32_t group_tid_33415;\n    \n    global_tid_33413 = get_global_id(0);\n    local_tid_33414 = get_local_id(0);\n    group_sizze_33417 = get_local_size(0);\n    wave_sizze_33416 = LOCKSTEP_WIDTH;\n    group_tid_33415 = get_group_id(0);\n    \n    int32_t phys_tid_32547 = global_tid_33413;\n    int32_t gtid_32545 = group_tid_33415 * segmap_group_sizze_32553 +\n            local_tid_33414;\n    int32_t gtid_32546;\n    \n    gtid_32546 = group_tid_33415 * segmap_group_sizze_32553 + local_tid_33414 -\n        (group_tid_33415 * segmap_group_sizze_32553 + local_tid_33414);\n    if (slt32(gtid_32545, n_31902) && slt32(gtid_32546, 1)) {\n        if ((sle32(0, gtid_32545) && slt32(gtid_32545, n_31902)) && (sle32(0,\n                                                                           0) &&\n                                                                     slt32(0,\n                                                                           fill_arg_31905))) {\n            ((__global bool *) mem_32788)[gtid_32545 * fill_arg_31905] = 1;\n        }\n    }\n}\n__kernel void segred_large_32522(uint red_arr_mem_33453_backing_offset_0,\n                                 uint sync_arr_mem_33455_backing_offset_1,\n                                 int32_t n_31902, int32_t fill_arg_31905,\n                                 int32_t num_groups_32519, __global\n                                 unsigned char *mem_32788, __global\n                                 unsigned char *mem_32791,\n                                 int32_t vit_num_groups_33441,\n                                 int32_t thread_per_segment_33443, __global\n                                 unsigned char *group_res_arr_mem_33444,\n                                 __global unsigned char *counter_mem_33446)\n{\n    const i",
            "nt32_t segred_group_sizze_32517 = mainzisegred_group_sizze_32516;\n    const int block_dim0 = 0;\n    const int block_dim1 = 1;\n    const int block_dim2 = 2;\n    volatile char *red_arr_mem_33453_backing_0 =\n                  &shared_mem[red_arr_mem_33453_backing_offset_0];\n    volatile char *sync_arr_mem_33455_backing_1 =\n                  &shared_mem[sync_arr_mem_33455_backing_offset_1];\n    int32_t global_tid_33448;\n    int32_t local_tid_33449;\n    int32_t group_sizze_33452;\n    int32_t wave_sizze_33451;\n    int32_t group_tid_33450;\n    \n    global_tid_33448 = get_global_id(0);\n    local_tid_33449 = get_local_id(0);\n    group_sizze_33452 = get_local_size(0);\n    wave_sizze_33451 = LOCKSTEP_WIDTH;\n    group_tid_33450 = get_group_id(0);\n    \n    int32_t phys_tid_32522 = global_tid_33448;\n    __local char *red_arr_mem_33453;\n    \n    red_arr_mem_33453 = (__local char *) red_arr_mem_33453_backing_0;\n    \n    __local char *sync_arr_mem_33455;\n    \n    sync_arr_mem_33455 = (__local char *) sync_arr_mem_33455_backing_1;\n    \n    int32_t phys_group_id_33457;\n    \n    phys_group_id_33457 = get_group_id(0);\n    for (int32_t i_33458 = 0; i_33458 < squot32(vit_num_groups_33441 -\n                                                phys_group_id_33457 +\n                                                num_groups_32519 - 1,\n                                                num_groups_32519); i_33458++) {\n        int32_t virt_group_id_33459 = phys_group_id_33457 + i_33458 *\n                num_groups_32519;\n        int32_t gtid_32509 = squot32(virt_group_id_33459,\n                                     squot32(num_groups_32519 + smax32(1,\n                                                                       n_31902) -\n                                             1, smax32(1, n_31902)));\n        int32_t gtid_32521;\n        bool x_acc_33460;\n        int32_t chunk_sizze_33461 = smin32(squot32(fill_arg_31905 +\n                                                   segred_group_sizze_32517 *\n      ",
            "                                             squot32(num_groups_32519 +\n                                                           smax32(1, n_31902) -\n                                                           1, smax32(1,\n                                                                     n_31902)) -\n                                                   1, segred_group_sizze_32517 *\n                                                   squot32(num_groups_32519 +\n                                                           smax32(1, n_31902) -\n                                                           1, smax32(1,\n                                                                     n_31902))),\n                                           squot32(fill_arg_31905 -\n                                                   srem32(virt_group_id_33459 *\n                                                          segred_group_sizze_32517 +\n                                                          local_tid_33449,\n                                                          segred_group_sizze_32517 *\n                                                          squot32(num_groups_32519 +\n                                                                  smax32(1,\n                                                                         n_31902) -\n                                                                  1, smax32(1,\n                                                                            n_31902))) +\n                                                   thread_per_segment_33443 - 1,\n                                                   thread_per_segment_33443));\n        bool x_32523;\n        bool x_32524;\n        \n        // neutral-initialise the accumulators\n        {\n            x_acc_33460 = 1;\n        }\n        for (int32_t i_33465 = 0; i_33465 < chunk_sizze_33461; i_33465++) {\n            gtid_32521 = srem32(virt_group_id_33459 * segred_group_sizze_32517 +\n                               ",
            " local_tid_33449, segred_group_sizze_32517 *\n                                squot32(num_groups_32519 + smax32(1, n_31902) -\n                                        1, smax32(1, n_31902))) +\n                thread_per_segment_33443 * i_33465;\n            // apply map function\n            {\n                bool x_32527 = ((__global bool *) mem_32788)[gtid_32509 *\n                                                             fill_arg_31905 +\n                                                             gtid_32521];\n                \n                // save map-out results\n                { }\n                // load accumulator\n                {\n                    x_32523 = x_acc_33460;\n                }\n                // load new values\n                {\n                    x_32524 = x_32527;\n                }\n                // apply reduction operator\n                {\n                    bool x_32525 = x_32523 && x_32524;\n                    \n                    // store in accumulator\n                    {\n                        x_acc_33460 = x_32525;\n                    }\n                }\n            }\n        }\n        // to reduce current chunk, first store our result in memory\n        {\n            x_32523 = x_acc_33460;\n            ((__local bool *) red_arr_mem_33453)[local_tid_33449] = x_32523;\n        }\n        barrier(CLK_LOCAL_MEM_FENCE);\n        \n        int32_t offset_33466;\n        int32_t skip_waves_33467;\n        bool x_33462;\n        bool x_33463;\n        \n        offset_33466 = 0;\n        // participating threads read initial accumulator\n        {\n            if (slt32(local_tid_33449, segred_group_sizze_32517)) {\n                x_33462 = ((__local bool *) red_arr_mem_33453)[local_tid_33449 +\n                                                               offset_33466];\n            }\n        }\n        offset_33466 = 1;\n        while (slt32(offset_33466, wave_sizze_33451)) {\n            if (slt32(local_tid_33449 + offset_33466,\n                    ",
            "  segred_group_sizze_32517) && ((local_tid_33449 -\n                                                     squot32(local_tid_33449,\n                                                             wave_sizze_33451) *\n                                                     wave_sizze_33451) & (2 *\n                                                                          offset_33466 -\n                                                                          1)) ==\n                0) {\n                // read array element\n                {\n                    x_33463 = ((volatile __local\n                                bool *) red_arr_mem_33453)[local_tid_33449 +\n                                                           offset_33466];\n                }\n                // apply reduction operation\n                {\n                    bool x_33464 = x_33462 && x_33463;\n                    \n                    x_33462 = x_33464;\n                }\n                // write result of operation\n                {\n                    ((volatile __local\n                      bool *) red_arr_mem_33453)[local_tid_33449] = x_33462;\n                }\n            }\n            offset_33466 *= 2;\n        }\n        skip_waves_33467 = 1;\n        while (slt32(skip_waves_33467, squot32(segred_group_sizze_32517 +\n                                               wave_sizze_33451 - 1,\n                                               wave_sizze_33451))) {\n            barrier(CLK_LOCAL_MEM_FENCE);\n            offset_33466 = skip_waves_33467 * wave_sizze_33451;\n            if (slt32(local_tid_33449 + offset_33466,\n                      segred_group_sizze_32517) && ((local_tid_33449 -\n                                                     squot32(local_tid_33449,\n                                                             wave_sizze_33451) *\n                                                     wave_sizze_33451) == 0 &&\n                                                    (squot32(local_tid_33449,\n          ",
            "                                                   wave_sizze_33451) &\n                                                     (2 * skip_waves_33467 -\n                                                      1)) == 0)) {\n                // read array element\n                {\n                    x_33463 = ((__local\n                                bool *) red_arr_mem_33453)[local_tid_33449 +\n                                                           offset_33466];\n                }\n                // apply reduction operation\n                {\n                    bool x_33464 = x_33462 && x_33463;\n                    \n                    x_33462 = x_33464;\n                }\n                // write result of operation\n                {\n                    ((__local bool *) red_arr_mem_33453)[local_tid_33449] =\n                        x_33462;\n                }\n            }\n            skip_waves_33467 *= 2;\n        }\n        barrier(CLK_LOCAL_MEM_FENCE);\n        // first thread saves the result in accumulator\n        {\n            if (local_tid_33449 == 0) {\n                x_acc_33460 = x_33462;\n            }\n        }\n        if (squot32(num_groups_32519 + smax32(1, n_31902) - 1, smax32(1,\n                                                                      n_31902)) ==\n            1) {\n            // first thread in group saves final result to memory\n            {\n                if (local_tid_33449 == 0) {\n                    ((__global bool *) mem_32791)[gtid_32509] = x_acc_33460;\n                }\n            }\n        } else {\n            int32_t old_counter_33468;\n            \n            // first thread in group saves group result to global memory\n            {\n                if (local_tid_33449 == 0) {\n                    ((__global\n                      bool *) group_res_arr_mem_33444)[virt_group_id_33459 *\n                                                       segred_group_sizze_32517] =\n                        x_acc_33460;\n                    mem_fence_glob",
            "al();\n                    old_counter_33468 = atomic_add(&((volatile __global\n                                                      int *) counter_mem_33446)[srem32(squot32(virt_group_id_33459,\n                                                                                               squot32(num_groups_32519 +\n                                                                                                       smax32(1,\n                                                                                                              n_31902) -\n                                                                                                       1,\n                                                                                                       smax32(1,\n                                                                                                              n_31902))),\n                                                                                       10240)],\n                                                   (int) 1);\n                    ((__local bool *) sync_arr_mem_33455)[0] =\n                        old_counter_33468 == squot32(num_groups_32519 +\n                                                     smax32(1, n_31902) - 1,\n                                                     smax32(1, n_31902)) - 1;\n                }\n            }\n            barrier(CLK_LOCAL_MEM_FENCE);\n            barrier(CLK_GLOBAL_MEM_FENCE);\n            \n            bool is_last_group_33469 = ((__local bool *) sync_arr_mem_33455)[0];\n            \n            if (is_last_group_33469) {\n                if (local_tid_33449 == 0) {\n                    old_counter_33468 = atomic_add(&((volatile __global\n                                                      int *) counter_mem_33446)[srem32(squot32(virt_group_id_33459,\n                                                                                               squot32(num_groups_32519 +\n                                             ",
            "                                                          smax32(1,\n                                                                                                              n_31902) -\n                                                                                                       1,\n                                                                                                       smax32(1,\n                                                                                                              n_31902))),\n                                                                                       10240)],\n                                                   (int) (0 -\n                                                          squot32(num_groups_32519 +\n                                                                  smax32(1,\n                                                                         n_31902) -\n                                                                  1, smax32(1,\n                                                                            n_31902))));\n                }\n                // read in the per-group-results\n                {\n                    if (slt32(local_tid_33449, squot32(num_groups_32519 +\n                                                       smax32(1, n_31902) - 1,\n                                                       smax32(1, n_31902)))) {\n                        x_32523 = ((__global\n                                    bool *) group_res_arr_mem_33444)[(squot32(virt_group_id_33459,\n                                                                              squot32(num_groups_32519 +\n                                                                                      smax32(1,\n                                                                                             n_31902) -\n                                                                                      1,\n                                            ",
            "                                          smax32(1,\n                                                                                             n_31902))) *\n                                                                      squot32(num_groups_32519 +\n                                                                              smax32(1,\n                                                                                     n_31902) -\n                                                                              1,\n                                                                              smax32(1,\n                                                                                     n_31902)) +\n                                                                      local_tid_33449) *\n                                                                     segred_group_sizze_32517];\n                    } else {\n                        x_32523 = 1;\n                    }\n                    ((__local bool *) red_arr_mem_33453)[local_tid_33449] =\n                        x_32523;\n                }\n                barrier(CLK_LOCAL_MEM_FENCE);\n                // reduce the per-group results\n                {\n                    int32_t offset_33470;\n                    int32_t skip_waves_33471;\n                    bool x_33462;\n                    bool x_33463;\n                    \n                    offset_33470 = 0;\n                    // participating threads read initial accumulator\n                    {\n                        if (slt32(local_tid_33449, segred_group_sizze_32517)) {\n                            x_33462 = ((__local\n                                        bool *) red_arr_mem_33453)[local_tid_33449 +\n                                                                   offset_33470];\n                        }\n                    }\n                    offset_33470 = 1;\n                    while (slt32(offset_33470, wave_sizze_33451)) {\n                        if ",
            "(slt32(local_tid_33449 + offset_33470,\n                                  segred_group_sizze_32517) &&\n                            ((local_tid_33449 - squot32(local_tid_33449,\n                                                        wave_sizze_33451) *\n                              wave_sizze_33451) & (2 * offset_33470 - 1)) ==\n                            0) {\n                            // read array element\n                            {\n                                x_33463 = ((volatile __local\n                                            bool *) red_arr_mem_33453)[local_tid_33449 +\n                                                                       offset_33470];\n                            }\n                            // apply reduction operation\n                            {\n                                bool x_33464 = x_33462 && x_33463;\n                                \n                                x_33462 = x_33464;\n                            }\n                            // write result of operation\n                            {\n                                ((volatile __local\n                                  bool *) red_arr_mem_33453)[local_tid_33449] =\n                                    x_33462;\n                            }\n                        }\n                        offset_33470 *= 2;\n                    }\n                    skip_waves_33471 = 1;\n                    while (slt32(skip_waves_33471,\n                                 squot32(segred_group_sizze_32517 +\n                                         wave_sizze_33451 - 1,\n                                         wave_sizze_33451))) {\n                        barrier(CLK_LOCAL_MEM_FENCE);\n                        offset_33470 = skip_waves_33471 * wave_sizze_33451;\n                        if (slt32(local_tid_33449 + offset_33470,\n                                  segred_group_sizze_32517) &&\n                            ((local_tid_33449 - squot32(local_tid_33449,\n                      ",
            "                                  wave_sizze_33451) *\n                              wave_sizze_33451) == 0 &&\n                             (squot32(local_tid_33449, wave_sizze_33451) & (2 *\n                                                                            skip_waves_33471 -\n                                                                            1)) ==\n                             0)) {\n                            // read array element\n                            {\n                                x_33463 = ((__local\n                                            bool *) red_arr_mem_33453)[local_tid_33449 +\n                                                                       offset_33470];\n                            }\n                            // apply reduction operation\n                            {\n                                bool x_33464 = x_33462 && x_33463;\n                                \n                                x_33462 = x_33464;\n                            }\n                            // write result of operation\n                            {\n                                ((__local\n                                  bool *) red_arr_mem_33453)[local_tid_33449] =\n                                    x_33462;\n                            }\n                        }\n                        skip_waves_33471 *= 2;\n                    }\n                    // and back to memory with the final result\n                    {\n                        if (local_tid_33449 == 0) {\n                            ((__global bool *) mem_32791)[gtid_32509] = x_33462;\n                        }\n                    }\n                }\n            }\n        }\n    }\n}\n__kernel void segred_nonseg_32132(uint sync_arr_mem_32978_backing_offset_0,\n                                  uint red_arr_mem_32980_backing_offset_1,\n                                  int32_t fill_arg_31617,\n                                  int32_t num_groups_32127, __global\n                  ",
            "                unsigned char *mem_32786, __global\n                                  unsigned char *mem_32789, __global\n                                  unsigned char *counter_mem_32968, __global\n                                  unsigned char *group_res_arr_mem_32970,\n                                  int32_t num_threads_32972)\n{\n    const int32_t segred_group_sizze_32125 =\n                  quicksort1zisegred_group_sizze_32124;\n    const int block_dim0 = 0;\n    const int block_dim1 = 1;\n    const int block_dim2 = 2;\n    volatile char *sync_arr_mem_32978_backing_0 =\n                  &shared_mem[sync_arr_mem_32978_backing_offset_0];\n    volatile char *red_arr_mem_32980_backing_1 =\n                  &shared_mem[red_arr_mem_32980_backing_offset_1];\n    int32_t global_tid_32973;\n    int32_t local_tid_32974;\n    int32_t group_sizze_32977;\n    int32_t wave_sizze_32976;\n    int32_t group_tid_32975;\n    \n    global_tid_32973 = get_global_id(0);\n    local_tid_32974 = get_local_id(0);\n    group_sizze_32977 = get_local_size(0);\n    wave_sizze_32976 = LOCKSTEP_WIDTH;\n    group_tid_32975 = get_group_id(0);\n    \n    int32_t phys_tid_32132 = global_tid_32973;\n    __local char *sync_arr_mem_32978;\n    \n    sync_arr_mem_32978 = (__local char *) sync_arr_mem_32978_backing_0;\n    \n    __local char *red_arr_mem_32980;\n    \n    red_arr_mem_32980 = (__local char *) red_arr_mem_32980_backing_1;\n    \n    int32_t dummy_32130 = 0;\n    int32_t gtid_32131;\n    \n    gtid_32131 = 0;\n    \n    bool x_acc_32982;\n    int32_t chunk_sizze_32983 = smin32(squot32(fill_arg_31617 +\n                                               segred_group_sizze_32125 *\n                                               num_groups_32127 - 1,\n                                               segred_group_sizze_32125 *\n                                               num_groups_32127),\n                                       squot32(fill_arg_31617 - phys_tid_32132 +\n                                               num_threads_32972",
            " - 1,\n                                               num_threads_32972));\n    bool x_31639;\n    bool x_31640;\n    \n    // neutral-initialise the accumulators\n    {\n        x_acc_32982 = 1;\n    }\n    for (int32_t i_32987 = 0; i_32987 < chunk_sizze_32983; i_32987++) {\n        gtid_32131 = phys_tid_32132 + num_threads_32972 * i_32987;\n        // apply map function\n        {\n            bool x_31642 = ((__global bool *) mem_32786)[gtid_32131];\n            \n            // save map-out results\n            { }\n            // load accumulator\n            {\n                x_31639 = x_acc_32982;\n            }\n            // load new values\n            {\n                x_31640 = x_31642;\n            }\n            // apply reduction operator\n            {\n                bool x_31641 = x_31639 && x_31640;\n                \n                // store in accumulator\n                {\n                    x_acc_32982 = x_31641;\n                }\n            }\n        }\n    }\n    // to reduce current chunk, first store our result in memory\n    {\n        x_31639 = x_acc_32982;\n        ((__local bool *) red_arr_mem_32980)[local_tid_32974] = x_31639;\n    }\n    barrier(CLK_LOCAL_MEM_FENCE);\n    \n    int32_t offset_32988;\n    int32_t skip_waves_32989;\n    bool x_32984;\n    bool x_32985;\n    \n    offset_32988 = 0;\n    // participating threads read initial accumulator\n    {\n        if (slt32(local_tid_32974, segred_group_sizze_32125)) {\n            x_32984 = ((__local bool *) red_arr_mem_32980)[local_tid_32974 +\n                                                           offset_32988];\n        }\n    }\n    offset_32988 = 1;\n    while (slt32(offset_32988, wave_sizze_32976)) {\n        if (slt32(local_tid_32974 + offset_32988, segred_group_sizze_32125) &&\n            ((local_tid_32974 - squot32(local_tid_32974, wave_sizze_32976) *\n              wave_sizze_32976) & (2 * offset_32988 - 1)) == 0) {\n            // read array element\n            {\n                x_32985 = ((volatile __local\n        ",
            "                    bool *) red_arr_mem_32980)[local_tid_32974 +\n                                                       offset_32988];\n            }\n            // apply reduction operation\n            {\n                bool x_32986 = x_32984 && x_32985;\n                \n                x_32984 = x_32986;\n            }\n            // write result of operation\n            {\n                ((volatile __local bool *) red_arr_mem_32980)[local_tid_32974] =\n                    x_32984;\n            }\n        }\n        offset_32988 *= 2;\n    }\n    skip_waves_32989 = 1;\n    while (slt32(skip_waves_32989, squot32(segred_group_sizze_32125 +\n                                           wave_sizze_32976 - 1,\n                                           wave_sizze_32976))) {\n        barrier(CLK_LOCAL_MEM_FENCE);\n        offset_32988 = skip_waves_32989 * wave_sizze_32976;\n        if (slt32(local_tid_32974 + offset_32988, segred_group_sizze_32125) &&\n            ((local_tid_32974 - squot32(local_tid_32974, wave_sizze_32976) *\n              wave_sizze_32976) == 0 && (squot32(local_tid_32974,\n                                                 wave_sizze_32976) & (2 *\n                                                                      skip_waves_32989 -\n                                                                      1)) ==\n             0)) {\n            // read array element\n            {\n                x_32985 = ((__local bool *) red_arr_mem_32980)[local_tid_32974 +\n                                                               offset_32988];\n            }\n            // apply reduction operation\n            {\n                bool x_32986 = x_32984 && x_32985;\n                \n                x_32984 = x_32986;\n            }\n            // write result of operation\n            {\n                ((__local bool *) red_arr_mem_32980)[local_tid_32974] = x_32984;\n            }\n        }\n        skip_waves_32989 *= 2;\n    }\n    barrier(CLK_LOCAL_MEM_FENCE);\n    // first thread saves th",
            "e result in accumulator\n    {\n        if (local_tid_32974 == 0) {\n            x_acc_32982 = x_32984;\n        }\n    }\n    \n    int32_t old_counter_32990;\n    \n    // first thread in group saves group result to global memory\n    {\n        if (local_tid_32974 == 0) {\n            ((__global bool *) group_res_arr_mem_32970)[group_tid_32975 *\n                                                        segred_group_sizze_32125] =\n                x_acc_32982;\n            mem_fence_global();\n            old_counter_32990 = atomic_add(&((volatile __global\n                                              int *) counter_mem_32968)[0],\n                                           (int) 1);\n            ((__local bool *) sync_arr_mem_32978)[0] = old_counter_32990 ==\n                num_groups_32127 - 1;\n        }\n    }\n    barrier(CLK_LOCAL_MEM_FENCE);\n    barrier(CLK_GLOBAL_MEM_FENCE);\n    \n    bool is_last_group_32991 = ((__local bool *) sync_arr_mem_32978)[0];\n    \n    if (is_last_group_32991) {\n        if (local_tid_32974 == 0) {\n            old_counter_32990 = atomic_add(&((volatile __global\n                                              int *) counter_mem_32968)[0],\n                                           (int) (0 - num_groups_32127));\n        }\n        // read in the per-group-results\n        {\n            if (slt32(local_tid_32974, num_groups_32127)) {\n                x_31639 = ((__global\n                            bool *) group_res_arr_mem_32970)[local_tid_32974 *\n                                                             segred_group_sizze_32125];\n            } else {\n                x_31639 = 1;\n            }\n            ((__local bool *) red_arr_mem_32980)[local_tid_32974] = x_31639;\n        }\n        barrier(CLK_LOCAL_MEM_FENCE);\n        // reduce the per-group results\n        {\n            int32_t offset_32992;\n            int32_t skip_waves_32993;\n            bool x_32984;\n            bool x_32985;\n            \n            offset_32992 = 0;\n            // participating ",
            "threads read initial accumulator\n            {\n                if (slt32(local_tid_32974, segred_group_sizze_32125)) {\n                    x_32984 = ((__local\n                                bool *) red_arr_mem_32980)[local_tid_32974 +\n                                                           offset_32992];\n                }\n            }\n            offset_32992 = 1;\n            while (slt32(offset_32992, wave_sizze_32976)) {\n                if (slt32(local_tid_32974 + offset_32992,\n                          segred_group_sizze_32125) && ((local_tid_32974 -\n                                                         squot32(local_tid_32974,\n                                                                 wave_sizze_32976) *\n                                                         wave_sizze_32976) &\n                                                        (2 * offset_32992 -\n                                                         1)) == 0) {\n                    // read array element\n                    {\n                        x_32985 = ((volatile __local\n                                    bool *) red_arr_mem_32980)[local_tid_32974 +\n                                                               offset_32992];\n                    }\n                    // apply reduction operation\n                    {\n                        bool x_32986 = x_32984 && x_32985;\n                        \n                        x_32984 = x_32986;\n                    }\n                    // write result of operation\n                    {\n                        ((volatile __local\n                          bool *) red_arr_mem_32980)[local_tid_32974] = x_32984;\n                    }\n                }\n                offset_32992 *= 2;\n            }\n            skip_waves_32993 = 1;\n            while (slt32(skip_waves_32993, squot32(segred_group_sizze_32125 +\n                                                   wave_sizze_32976 - 1,\n                                                   wave_siz",
            "ze_32976))) {\n                barrier(CLK_LOCAL_MEM_FENCE);\n                offset_32992 = skip_waves_32993 * wave_sizze_32976;\n                if (slt32(local_tid_32974 + offset_32992,\n                          segred_group_sizze_32125) && ((local_tid_32974 -\n                                                         squot32(local_tid_32974,\n                                                                 wave_sizze_32976) *\n                                                         wave_sizze_32976) ==\n                                                        0 &&\n                                                        (squot32(local_tid_32974,\n                                                                 wave_sizze_32976) &\n                                                         (2 * skip_waves_32993 -\n                                                          1)) == 0)) {\n                    // read array element\n                    {\n                        x_32985 = ((__local\n                                    bool *) red_arr_mem_32980)[local_tid_32974 +\n                                                               offset_32992];\n                    }\n                    // apply reduction operation\n                    {\n                        bool x_32986 = x_32984 && x_32985;\n                        \n                        x_32984 = x_32986;\n                    }\n                    // write result of operation\n                    {\n                        ((__local bool *) red_arr_mem_32980)[local_tid_32974] =\n                            x_32984;\n                    }\n                }\n                skip_waves_32993 *= 2;\n            }\n            // and back to memory with the final result\n            {\n                if (local_tid_32974 == 0) {\n                    ((__global bool *) mem_32789)[0] = x_32984;\n                }\n            }\n        }\n    }\n}\n__kernel void segred_nonseg_32307(uint sync_arr_mem_33394_backing_offset_0,\n                    ",
            "              uint red_arr_mem_33396_backing_offset_1,\n                                  int32_t fill_arg_31617,\n                                  int32_t num_groups_32302, __global\n                                  unsigned char *mem_32846, __global\n                                  unsigned char *mem_32849, __global\n                                  unsigned char *counter_mem_33384, __global\n                                  unsigned char *group_res_arr_mem_33386,\n                                  int32_t num_threads_33388)\n{\n    const int32_t segred_group_sizze_32300 =\n                  quicksort1zisegred_group_sizze_32299;\n    const int block_dim0 = 0;\n    const int block_dim1 = 1;\n    const int block_dim2 = 2;\n    volatile char *sync_arr_mem_33394_backing_0 =\n                  &shared_mem[sync_arr_mem_33394_backing_offset_0];\n    volatile char *red_arr_mem_33396_backing_1 =\n                  &shared_mem[red_arr_mem_33396_backing_offset_1];\n    int32_t global_tid_33389;\n    int32_t local_tid_33390;\n    int32_t group_sizze_33393;\n    int32_t wave_sizze_33392;\n    int32_t group_tid_33391;\n    \n    global_tid_33389 = get_global_id(0);\n    local_tid_33390 = get_local_id(0);\n    group_sizze_33393 = get_local_size(0);\n    wave_sizze_33392 = LOCKSTEP_WIDTH;\n    group_tid_33391 = get_group_id(0);\n    \n    int32_t phys_tid_32307 = global_tid_33389;\n    __local char *sync_arr_mem_33394;\n    \n    sync_arr_mem_33394 = (__local char *) sync_arr_mem_33394_backing_0;\n    \n    __local char *red_arr_mem_33396;\n    \n    red_arr_mem_33396 = (__local char *) red_arr_mem_33396_backing_1;\n    \n    int32_t dummy_32305 = 0;\n    int32_t gtid_32306;\n    \n    gtid_32306 = 0;\n    \n    bool x_acc_33398;\n    int32_t chunk_sizze_33399 = smin32(squot32(fill_arg_31617 +\n                                               segred_group_sizze_32300 *\n                                               num_groups_32302 - 1,\n                                               segred_group_sizze_32300 *\n           ",
            "                                    num_groups_32302),\n                                       squot32(fill_arg_31617 - phys_tid_32307 +\n                                               num_threads_33388 - 1,\n                                               num_threads_33388));\n    bool x_31894;\n    bool x_31895;\n    \n    // neutral-initialise the accumulators\n    {\n        x_acc_33398 = 1;\n    }\n    for (int32_t i_33403 = 0; i_33403 < chunk_sizze_33399; i_33403++) {\n        gtid_32306 = phys_tid_32307 + num_threads_33388 * i_33403;\n        // apply map function\n        {\n            bool x_31897 = ((__global bool *) mem_32846)[gtid_32306];\n            \n            // save map-out results\n            { }\n            // load accumulator\n            {\n                x_31894 = x_acc_33398;\n            }\n            // load new values\n            {\n                x_31895 = x_31897;\n            }\n            // apply reduction operator\n            {\n                bool x_31896 = x_31894 && x_31895;\n                \n                // store in accumulator\n                {\n                    x_acc_33398 = x_31896;\n                }\n            }\n        }\n    }\n    // to reduce current chunk, first store our result in memory\n    {\n        x_31894 = x_acc_33398;\n        ((__local bool *) red_arr_mem_33396)[local_tid_33390] = x_31894;\n    }\n    barrier(CLK_LOCAL_MEM_FENCE);\n    \n    int32_t offset_33404;\n    int32_t skip_waves_33405;\n    bool x_33400;\n    bool x_33401;\n    \n    offset_33404 = 0;\n    // participating threads read initial accumulator\n    {\n        if (slt32(local_tid_33390, segred_group_sizze_32300)) {\n            x_33400 = ((__local bool *) red_arr_mem_33396)[local_tid_33390 +\n                                                           offset_33404];\n        }\n    }\n    offset_33404 = 1;\n    while (slt32(offset_33404, wave_sizze_33392)) {\n        if (slt32(local_tid_33390 + offset_33404, segred_group_sizze_32300) &&\n            ((local_tid_33390 - squot32(loca",
            "l_tid_33390, wave_sizze_33392) *\n              wave_sizze_33392) & (2 * offset_33404 - 1)) == 0) {\n            // read array element\n            {\n                x_33401 = ((volatile __local\n                            bool *) red_arr_mem_33396)[local_tid_33390 +\n                                                       offset_33404];\n            }\n            // apply reduction operation\n            {\n                bool x_33402 = x_33400 && x_33401;\n                \n                x_33400 = x_33402;\n            }\n            // write result of operation\n            {\n                ((volatile __local bool *) red_arr_mem_33396)[local_tid_33390] =\n                    x_33400;\n            }\n        }\n        offset_33404 *= 2;\n    }\n    skip_waves_33405 = 1;\n    while (slt32(skip_waves_33405, squot32(segred_group_sizze_32300 +\n                                           wave_sizze_33392 - 1,\n                                           wave_sizze_33392))) {\n        barrier(CLK_LOCAL_MEM_FENCE);\n        offset_33404 = skip_waves_33405 * wave_sizze_33392;\n        if (slt32(local_tid_33390 + offset_33404, segred_group_sizze_32300) &&\n            ((local_tid_33390 - squot32(local_tid_33390, wave_sizze_33392) *\n              wave_sizze_33392) == 0 && (squot32(local_tid_33390,\n                                                 wave_sizze_33392) & (2 *\n                                                                      skip_waves_33405 -\n                                                                      1)) ==\n             0)) {\n            // read array element\n            {\n                x_33401 = ((__local bool *) red_arr_mem_33396)[local_tid_33390 +\n                                                               offset_33404];\n            }\n            // apply reduction operation\n            {\n                bool x_33402 = x_33400 && x_33401;\n                \n                x_33400 = x_33402;\n            }\n            // write result of operation\n            {\n    ",
            "            ((__local bool *) red_arr_mem_33396)[local_tid_33390] = x_33400;\n            }\n        }\n        skip_waves_33405 *= 2;\n    }\n    barrier(CLK_LOCAL_MEM_FENCE);\n    // first thread saves the result in accumulator\n    {\n        if (local_tid_33390 == 0) {\n            x_acc_33398 = x_33400;\n        }\n    }\n    \n    int32_t old_counter_33406;\n    \n    // first thread in group saves group result to global memory\n    {\n        if (local_tid_33390 == 0) {\n            ((__global bool *) group_res_arr_mem_33386)[group_tid_33391 *\n                                                        segred_group_sizze_32300] =\n                x_acc_33398;\n            mem_fence_global();\n            old_counter_33406 = atomic_add(&((volatile __global\n                                              int *) counter_mem_33384)[0],\n                                           (int) 1);\n            ((__local bool *) sync_arr_mem_33394)[0] = old_counter_33406 ==\n                num_groups_32302 - 1;\n        }\n    }\n    barrier(CLK_LOCAL_MEM_FENCE);\n    barrier(CLK_GLOBAL_MEM_FENCE);\n    \n    bool is_last_group_33407 = ((__local bool *) sync_arr_mem_33394)[0];\n    \n    if (is_last_group_33407) {\n        if (local_tid_33390 == 0) {\n            old_counter_33406 = atomic_add(&((volatile __global\n                                              int *) counter_mem_33384)[0],\n                                           (int) (0 - num_groups_32302));\n        }\n        // read in the per-group-results\n        {\n            if (slt32(local_tid_33390, num_groups_32302)) {\n                x_31894 = ((__global\n                            bool *) group_res_arr_mem_33386)[local_tid_33390 *\n                                                             segred_group_sizze_32300];\n            } else {\n                x_31894 = 1;\n            }\n            ((__local bool *) red_arr_mem_33396)[local_tid_33390] = x_31894;\n        }\n        barrier(CLK_LOCAL_MEM_FENCE);\n        // reduce the per-group results\n      ",
            "  {\n            int32_t offset_33408;\n            int32_t skip_waves_33409;\n            bool x_33400;\n            bool x_33401;\n            \n            offset_33408 = 0;\n            // participating threads read initial accumulator\n            {\n                if (slt32(local_tid_33390, segred_group_sizze_32300)) {\n                    x_33400 = ((__local\n                                bool *) red_arr_mem_33396)[local_tid_33390 +\n                                                           offset_33408];\n                }\n            }\n            offset_33408 = 1;\n            while (slt32(offset_33408, wave_sizze_33392)) {\n                if (slt32(local_tid_33390 + offset_33408,\n                          segred_group_sizze_32300) && ((local_tid_33390 -\n                                                         squot32(local_tid_33390,\n                                                                 wave_sizze_33392) *\n                                                         wave_sizze_33392) &\n                                                        (2 * offset_33408 -\n                                                         1)) == 0) {\n                    // read array element\n                    {\n                        x_33401 = ((volatile __local\n                                    bool *) red_arr_mem_33396)[local_tid_33390 +\n                                                               offset_33408];\n                    }\n                    // apply reduction operation\n                    {\n                        bool x_33402 = x_33400 && x_33401;\n                        \n                        x_33400 = x_33402;\n                    }\n                    // write result of operation\n                    {\n                        ((volatile __local\n                          bool *) red_arr_mem_33396)[local_tid_33390] = x_33400;\n                    }\n                }\n                offset_33408 *= 2;\n            }\n            skip_waves_33409 = 1;\n          ",
            "  while (slt32(skip_waves_33409, squot32(segred_group_sizze_32300 +\n                                                   wave_sizze_33392 - 1,\n                                                   wave_sizze_33392))) {\n                barrier(CLK_LOCAL_MEM_FENCE);\n                offset_33408 = skip_waves_33409 * wave_sizze_33392;\n                if (slt32(local_tid_33390 + offset_33408,\n                          segred_group_sizze_32300) && ((local_tid_33390 -\n                                                         squot32(local_tid_33390,\n                                                                 wave_sizze_33392) *\n                                                         wave_sizze_33392) ==\n                                                        0 &&\n                                                        (squot32(local_tid_33390,\n                                                                 wave_sizze_33392) &\n                                                         (2 * skip_waves_33409 -\n                                                          1)) == 0)) {\n                    // read array element\n                    {\n                        x_33401 = ((__local\n                                    bool *) red_arr_mem_33396)[local_tid_33390 +\n                                                               offset_33408];\n                    }\n                    // apply reduction operation\n                    {\n                        bool x_33402 = x_33400 && x_33401;\n                        \n                        x_33400 = x_33402;\n                    }\n                    // write result of operation\n                    {\n                        ((__local bool *) red_arr_mem_33396)[local_tid_33390] =\n                            x_33400;\n                    }\n                }\n                skip_waves_33409 *= 2;\n            }\n            // and back to memory with the final result\n            {\n                if (local_tid_33390 == 0) {\n          ",
            "          ((__global bool *) mem_32849)[0] = x_33400;\n                }\n            }\n        }\n    }\n}\n__kernel void segred_small_32522(uint red_arr_mem_33431_backing_offset_0,\n                                 int32_t n_31902, int32_t fill_arg_31905,\n                                 int32_t num_groups_32519, __global\n                                 unsigned char *mem_32788, __global\n                                 unsigned char *mem_32791,\n                                 int32_t segment_sizze_nonzzero_33424)\n{\n    const int32_t segred_group_sizze_32517 = mainzisegred_group_sizze_32516;\n    const int block_dim0 = 0;\n    const int block_dim1 = 1;\n    const int block_dim2 = 2;\n    volatile char *red_arr_mem_33431_backing_0 =\n                  &shared_mem[red_arr_mem_33431_backing_offset_0];\n    int32_t global_tid_33426;\n    int32_t local_tid_33427;\n    int32_t group_sizze_33430;\n    int32_t wave_sizze_33429;\n    int32_t group_tid_33428;\n    \n    global_tid_33426 = get_global_id(0);\n    local_tid_33427 = get_local_id(0);\n    group_sizze_33430 = get_local_size(0);\n    wave_sizze_33429 = LOCKSTEP_WIDTH;\n    group_tid_33428 = get_group_id(0);\n    \n    int32_t phys_tid_32522 = global_tid_33426;\n    __local char *red_arr_mem_33431;\n    \n    red_arr_mem_33431 = (__local char *) red_arr_mem_33431_backing_0;\n    \n    int32_t phys_group_id_33433;\n    \n    phys_group_id_33433 = get_group_id(0);\n    for (int32_t i_33434 = 0; i_33434 < squot32(squot32(n_31902 +\n                                                        squot32(segred_group_sizze_32517,\n                                                                segment_sizze_nonzzero_33424) -\n                                                        1,\n                                                        squot32(segred_group_sizze_32517,\n                                                                segment_sizze_nonzzero_33424)) -\n                                                phys_group_id_33433 +\n                        ",
            "                        num_groups_32519 - 1,\n                                                num_groups_32519); i_33434++) {\n        int32_t virt_group_id_33435 = phys_group_id_33433 + i_33434 *\n                num_groups_32519;\n        int32_t gtid_32509 = squot32(local_tid_33427,\n                                     segment_sizze_nonzzero_33424) +\n                virt_group_id_33435 * squot32(segred_group_sizze_32517,\n                                              segment_sizze_nonzzero_33424);\n        int32_t gtid_32521 = srem32(local_tid_33427, fill_arg_31905);\n        \n        // apply map function if in bounds\n        {\n            if (slt32(0, fill_arg_31905) && (slt32(gtid_32509, n_31902) &&\n                                             slt32(local_tid_33427,\n                                                   fill_arg_31905 *\n                                                   squot32(segred_group_sizze_32517,\n                                                           segment_sizze_nonzzero_33424)))) {\n                bool x_32527 = ((__global bool *) mem_32788)[gtid_32509 *\n                                                             fill_arg_31905 +\n                                                             gtid_32521];\n                \n                // save map-out results\n                { }\n                // save results to be reduced\n                {\n                    ((__local bool *) red_arr_mem_33431)[local_tid_33427] =\n                        x_32527;\n                }\n            } else {\n                ((__local bool *) red_arr_mem_33431)[local_tid_33427] = 1;\n            }\n        }\n        barrier(CLK_LOCAL_MEM_FENCE);\n        if (slt32(0, fill_arg_31905)) {\n            // perform segmented scan to imitate reduction\n            {\n                bool x_32523;\n                bool x_32524;\n                bool x_33436;\n                bool x_33437;\n                int32_t skip_threads_33439;\n                \n                if (slt32(loca",
            "l_tid_33427, fill_arg_31905 *\n                          squot32(segred_group_sizze_32517,\n                                  segment_sizze_nonzzero_33424))) {\n                    x_32524 = ((volatile __local\n                                bool *) red_arr_mem_33431)[local_tid_33427];\n                }\n                // in-block scan (hopefully no barriers needed)\n                {\n                    skip_threads_33439 = 1;\n                    while (slt32(skip_threads_33439, 32)) {\n                        if (sle32(skip_threads_33439, local_tid_33427 -\n                                  squot32(local_tid_33427, 32) * 32) &&\n                            slt32(local_tid_33427, fill_arg_31905 *\n                                  squot32(segred_group_sizze_32517,\n                                          segment_sizze_nonzzero_33424))) {\n                            // read operands\n                            {\n                                x_32523 = ((volatile __local\n                                            bool *) red_arr_mem_33431)[local_tid_33427 -\n                                                                       skip_threads_33439];\n                            }\n                            // perform operation\n                            {\n                                if (!slt32(srem32(local_tid_33427,\n                                                  fill_arg_31905),\n                                           local_tid_33427 - (local_tid_33427 -\n                                                              skip_threads_33439))) {\n                                    bool x_32525 = x_32523 && x_32524;\n                                    \n                                    x_32524 = x_32525;\n                                }\n                            }\n                        }\n                        if (sle32(wave_sizze_33429, skip_threads_33439)) {\n                            barrier(CLK_LOCAL_MEM_FENCE);\n                        }\n                  ",
            "      if (sle32(skip_threads_33439, local_tid_33427 -\n                                  squot32(local_tid_33427, 32) * 32) &&\n                            slt32(local_tid_33427, fill_arg_31905 *\n                                  squot32(segred_group_sizze_32517,\n                                          segment_sizze_nonzzero_33424))) {\n                            // write result\n                            {\n                                ((volatile __local\n                                  bool *) red_arr_mem_33431)[local_tid_33427] =\n                                    x_32524;\n                            }\n                        }\n                        if (sle32(wave_sizze_33429, skip_threads_33439)) {\n                            barrier(CLK_LOCAL_MEM_FENCE);\n                        }\n                        skip_threads_33439 *= 2;\n                    }\n                }\n                barrier(CLK_LOCAL_MEM_FENCE);\n                // last thread of block 'i' writes its result to offset 'i'\n                {\n                    if ((local_tid_33427 - squot32(local_tid_33427, 32) * 32) ==\n                        31 && slt32(local_tid_33427, fill_arg_31905 *\n                                    squot32(segred_group_sizze_32517,\n                                            segment_sizze_nonzzero_33424))) {\n                        ((volatile __local\n                          bool *) red_arr_mem_33431)[squot32(local_tid_33427,\n                                                             32)] = x_32524;\n                    }\n                }\n                barrier(CLK_LOCAL_MEM_FENCE);\n                // scan the first block, after which offset 'i' contains carry-in for warp 'i+1'\n                {\n                    int32_t skip_threads_33440;\n                    \n                    if (squot32(local_tid_33427, 32) == 0 &&\n                        slt32(local_tid_33427, fill_arg_31905 *\n                              squot32(segred_group_sizze_32517,\n            ",
            "                          segment_sizze_nonzzero_33424))) {\n                        x_33437 = ((volatile __local\n                                    bool *) red_arr_mem_33431)[local_tid_33427];\n                    }\n                    // in-block scan (hopefully no barriers needed)\n                    {\n                        skip_threads_33440 = 1;\n                        while (slt32(skip_threads_33440, 32)) {\n                            if (sle32(skip_threads_33440, local_tid_33427 -\n                                      squot32(local_tid_33427, 32) * 32) &&\n                                (squot32(local_tid_33427, 32) == 0 &&\n                                 slt32(local_tid_33427, fill_arg_31905 *\n                                       squot32(segred_group_sizze_32517,\n                                               segment_sizze_nonzzero_33424)))) {\n                                // read operands\n                                {\n                                    x_33436 = ((volatile __local\n                                                bool *) red_arr_mem_33431)[local_tid_33427 -\n                                                                           skip_threads_33440];\n                                }\n                                // perform operation\n                                {\n                                    if (!slt32(srem32(local_tid_33427 * 32 +\n                                                      32 - 1, fill_arg_31905),\n                                               local_tid_33427 * 32 + 32 - 1 -\n                                               ((local_tid_33427 -\n                                                 skip_threads_33440) * 32 + 32 -\n                                                1))) {\n                                        bool x_33438 = x_33436 && x_33437;\n                                        \n                                        x_33437 = x_33438;\n                                    }\n                       ",
            "         }\n                            }\n                            if (sle32(wave_sizze_33429, skip_threads_33440)) {\n                                barrier(CLK_LOCAL_MEM_FENCE);\n                            }\n                            if (sle32(skip_threads_33440, local_tid_33427 -\n                                      squot32(local_tid_33427, 32) * 32) &&\n                                (squot32(local_tid_33427, 32) == 0 &&\n                                 slt32(local_tid_33427, fill_arg_31905 *\n                                       squot32(segred_group_sizze_32517,\n                                               segment_sizze_nonzzero_33424)))) {\n                                // write result\n                                {\n                                    ((volatile __local\n                                      bool *) red_arr_mem_33431)[local_tid_33427] =\n                                        x_33437;\n                                }\n                            }\n                            if (sle32(wave_sizze_33429, skip_threads_33440)) {\n                                barrier(CLK_LOCAL_MEM_FENCE);\n                            }\n                            skip_threads_33440 *= 2;\n                        }\n                    }\n                }\n                barrier(CLK_LOCAL_MEM_FENCE);\n                // carry-in for every block except the first\n                {\n                    if (!(squot32(local_tid_33427, 32) == 0 ||\n                          !slt32(local_tid_33427, fill_arg_31905 *\n                                 squot32(segred_group_sizze_32517,\n                                         segment_sizze_nonzzero_33424)))) {\n                        // read operands\n                        {\n                            x_32523 = ((volatile __local\n                                        bool *) red_arr_mem_33431)[squot32(local_tid_33427,\n                                                                           32) -\n                   ",
            "                                                1];\n                        }\n                        // perform operation\n                        {\n                            if (!slt32(srem32(local_tid_33427, fill_arg_31905),\n                                       local_tid_33427 -\n                                       (squot32(local_tid_33427, 32) * 32 -\n                                        1))) {\n                                bool x_32525 = x_32523 && x_32524;\n                                \n                                x_32524 = x_32525;\n                            }\n                        }\n                        // write final result\n                        {\n                            ((volatile __local\n                              bool *) red_arr_mem_33431)[local_tid_33427] =\n                                x_32524;\n                        }\n                    }\n                }\n                barrier(CLK_LOCAL_MEM_FENCE);\n                // restore correct values for first block\n                {\n                    if (squot32(local_tid_33427, 32) == 0) {\n                        ((volatile __local\n                          bool *) red_arr_mem_33431)[local_tid_33427] = x_32524;\n                    }\n                }\n                barrier(CLK_LOCAL_MEM_FENCE);\n            }\n        }\n        barrier(CLK_LOCAL_MEM_FENCE);\n        // save final values of segments\n        {\n            if (slt32(virt_group_id_33435 * squot32(segred_group_sizze_32517,\n                                                    segment_sizze_nonzzero_33424) +\n                      local_tid_33427, n_31902) && slt32(local_tid_33427,\n                                                         squot32(segred_group_sizze_32517,\n                                                                 segment_sizze_nonzzero_33424))) {\n                ((__global bool *) mem_32791)[virt_group_id_33435 *\n                                              squot32(segred_group_sizze_32517,\n  ",
            "                                                    segment_sizze_nonzzero_33424) +\n                                              local_tid_33427] = ((__local\n                                                                   bool *) red_arr_mem_33431)[(local_tid_33427 +\n                                                                                               1) *\n                                                                                              segment_sizze_nonzzero_33424 -\n                                                                                              1];\n            }\n        }\n        barrier(CLK_LOCAL_MEM_FENCE);\n    }\n}\n",
            NULL};
static int32_t counter_mem_realtype_33540[10240];
static int32_t counter_mem_realtype_33644[10] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
static int32_t counter_mem_realtype_33840[10] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
struct memblock_device {
    int *references;
    CUdeviceptr mem;
    int64_t size;
    const char *desc;
} ;
struct memblock {
    int *references;
    char *mem;
    int64_t size;
    const char *desc;
} ;
static const char *size_names[] = {"main.segmap_group_size_32314",
                                   "main.segmap_group_size_32536",
                                   "main.segmap_group_size_32552",
                                   "main.segmap_num_groups_32316",
                                   "main.segred_group_size_32516",
                                   "main.segred_num_groups_32518",
                                   "quicksort1.group_size_32965",
                                   "quicksort1.group_size_33065",
                                   "quicksort1.group_size_33129",
                                   "quicksort1.group_size_33208",
                                   "quicksort1.group_size_33277",
                                   "quicksort1.group_size_33346",
                                   "quicksort1.group_size_33367",
                                   "quicksort1.segmap_group_size_32155",
                                   "quicksort1.segmap_group_size_32171",
                                   "quicksort1.segmap_group_size_32204",
                                   "quicksort1.segmap_group_size_32230",
                                   "quicksort1.segmap_group_size_32245",
                                   "quicksort1.segmap_group_size_32259",
                                   "quicksort1.segmap_group_size_32277",
                                   "quicksort1.segmap_group_size_32289",
                                   "quicksort1.segred_group_size_32124",
                                   "quicksort1.segred_group_size_32299",
                                   "quicksort1.segred_num_groups_32126",
                                   "quicksort1.segred_num_groups_32301",
                                   "quicksort1.segscan_group_size_32135",
                                   "quicksort1.segscan_group_size_32144",
                                   "quicksort1.segscan_group_size_32184",
                                   "quicksort1.segscan_group_size_32193",
                                   "quicksort1.segscan_group_size_32219",
                                   "quicksort1.segscan_num_groups_32137",
                                   "quicksort1.segscan_num_groups_32146",
                                   "quicksort1.segscan_num_groups_32186",
                                   "quicksort1.segscan_num_groups_32195",
                                   "quicksort1.segscan_num_groups_32221"};
static const char *size_vars[] = {"mainzisegmap_group_sizze_32314",
                                  "mainzisegmap_group_sizze_32536",
                                  "mainzisegmap_group_sizze_32552",
                                  "mainzisegmap_num_groups_32316",
                                  "mainzisegred_group_sizze_32516",
                                  "mainzisegred_num_groups_32518",
                                  "quicksort1zigroup_sizze_32965",
                                  "quicksort1zigroup_sizze_33065",
                                  "quicksort1zigroup_sizze_33129",
                                  "quicksort1zigroup_sizze_33208",
                                  "quicksort1zigroup_sizze_33277",
                                  "quicksort1zigroup_sizze_33346",
                                  "quicksort1zigroup_sizze_33367",
                                  "quicksort1zisegmap_group_sizze_32155",
                                  "quicksort1zisegmap_group_sizze_32171",
                                  "quicksort1zisegmap_group_sizze_32204",
                                  "quicksort1zisegmap_group_sizze_32230",
                                  "quicksort1zisegmap_group_sizze_32245",
                                  "quicksort1zisegmap_group_sizze_32259",
                                  "quicksort1zisegmap_group_sizze_32277",
                                  "quicksort1zisegmap_group_sizze_32289",
                                  "quicksort1zisegred_group_sizze_32124",
                                  "quicksort1zisegred_group_sizze_32299",
                                  "quicksort1zisegred_num_groups_32126",
                                  "quicksort1zisegred_num_groups_32301",
                                  "quicksort1zisegscan_group_sizze_32135",
                                  "quicksort1zisegscan_group_sizze_32144",
                                  "quicksort1zisegscan_group_sizze_32184",
                                  "quicksort1zisegscan_group_sizze_32193",
                                  "quicksort1zisegscan_group_sizze_32219",
                                  "quicksort1zisegscan_num_groups_32137",
                                  "quicksort1zisegscan_num_groups_32146",
                                  "quicksort1zisegscan_num_groups_32186",
                                  "quicksort1zisegscan_num_groups_32195",
                                  "quicksort1zisegscan_num_groups_32221"};
static const char *size_classes[] = {"group_size", "group_size", "group_size",
                                     "num_groups", "group_size", "num_groups",
                                     "group_size", "group_size", "group_size",
                                     "group_size", "group_size", "group_size",
                                     "group_size", "group_size", "group_size",
                                     "group_size", "group_size", "group_size",
                                     "group_size", "group_size", "group_size",
                                     "group_size", "group_size", "num_groups",
                                     "num_groups", "group_size", "group_size",
                                     "group_size", "group_size", "group_size",
                                     "num_groups", "num_groups", "num_groups",
                                     "num_groups", "num_groups"};
int futhark_get_num_sizes(void)
{
    return 35;
}
const char *futhark_get_size_name(int i)
{
    return size_names[i];
}
const char *futhark_get_size_class(int i)
{
    return size_classes[i];
}
struct sizes {
    size_t mainzisegmap_group_sizze_32314;
    size_t mainzisegmap_group_sizze_32536;
    size_t mainzisegmap_group_sizze_32552;
    size_t mainzisegmap_num_groups_32316;
    size_t mainzisegred_group_sizze_32516;
    size_t mainzisegred_num_groups_32518;
    size_t quicksort1zigroup_sizze_32965;
    size_t quicksort1zigroup_sizze_33065;
    size_t quicksort1zigroup_sizze_33129;
    size_t quicksort1zigroup_sizze_33208;
    size_t quicksort1zigroup_sizze_33277;
    size_t quicksort1zigroup_sizze_33346;
    size_t quicksort1zigroup_sizze_33367;
    size_t quicksort1zisegmap_group_sizze_32155;
    size_t quicksort1zisegmap_group_sizze_32171;
    size_t quicksort1zisegmap_group_sizze_32204;
    size_t quicksort1zisegmap_group_sizze_32230;
    size_t quicksort1zisegmap_group_sizze_32245;
    size_t quicksort1zisegmap_group_sizze_32259;
    size_t quicksort1zisegmap_group_sizze_32277;
    size_t quicksort1zisegmap_group_sizze_32289;
    size_t quicksort1zisegred_group_sizze_32124;
    size_t quicksort1zisegred_group_sizze_32299;
    size_t quicksort1zisegred_num_groups_32126;
    size_t quicksort1zisegred_num_groups_32301;
    size_t quicksort1zisegscan_group_sizze_32135;
    size_t quicksort1zisegscan_group_sizze_32144;
    size_t quicksort1zisegscan_group_sizze_32184;
    size_t quicksort1zisegscan_group_sizze_32193;
    size_t quicksort1zisegscan_group_sizze_32219;
    size_t quicksort1zisegscan_num_groups_32137;
    size_t quicksort1zisegscan_num_groups_32146;
    size_t quicksort1zisegscan_num_groups_32186;
    size_t quicksort1zisegscan_num_groups_32195;
    size_t quicksort1zisegscan_num_groups_32221;
} ;
struct futhark_context_config {
    struct cuda_config cu_cfg;
    size_t sizes[35];
    int num_nvrtc_opts;
    const char **nvrtc_opts;
} ;
struct futhark_context_config *futhark_context_config_new(void)
{
    struct futhark_context_config *cfg =
                                  (struct futhark_context_config *) malloc(sizeof(struct futhark_context_config));
    
    if (cfg == NULL)
        return NULL;
    cfg->num_nvrtc_opts = 0;
    cfg->nvrtc_opts = (const char **) malloc(sizeof(const char *));
    cfg->nvrtc_opts[0] = NULL;
    cfg->sizes[0] = 0;
    cfg->sizes[1] = 0;
    cfg->sizes[2] = 0;
    cfg->sizes[3] = 0;
    cfg->sizes[4] = 0;
    cfg->sizes[5] = 0;
    cfg->sizes[6] = 0;
    cfg->sizes[7] = 0;
    cfg->sizes[8] = 0;
    cfg->sizes[9] = 0;
    cfg->sizes[10] = 0;
    cfg->sizes[11] = 0;
    cfg->sizes[12] = 0;
    cfg->sizes[13] = 0;
    cfg->sizes[14] = 0;
    cfg->sizes[15] = 0;
    cfg->sizes[16] = 0;
    cfg->sizes[17] = 0;
    cfg->sizes[18] = 0;
    cfg->sizes[19] = 0;
    cfg->sizes[20] = 0;
    cfg->sizes[21] = 0;
    cfg->sizes[22] = 0;
    cfg->sizes[23] = 0;
    cfg->sizes[24] = 0;
    cfg->sizes[25] = 0;
    cfg->sizes[26] = 0;
    cfg->sizes[27] = 0;
    cfg->sizes[28] = 0;
    cfg->sizes[29] = 0;
    cfg->sizes[30] = 0;
    cfg->sizes[31] = 0;
    cfg->sizes[32] = 0;
    cfg->sizes[33] = 0;
    cfg->sizes[34] = 0;
    cuda_config_init(&cfg->cu_cfg, 35, size_names, size_vars, cfg->sizes,
                     size_classes);
    return cfg;
}
void futhark_context_config_free(struct futhark_context_config *cfg)
{
    free(cfg->nvrtc_opts);
    free(cfg);
}
void futhark_context_config_add_nvrtc_option(struct futhark_context_config *cfg,
                                             const char *opt)
{
    cfg->nvrtc_opts[cfg->num_nvrtc_opts] = opt;
    cfg->num_nvrtc_opts++;
    cfg->nvrtc_opts = (const char **) realloc(cfg->nvrtc_opts,
                                              (cfg->num_nvrtc_opts + 1) *
                                              sizeof(const char *));
    cfg->nvrtc_opts[cfg->num_nvrtc_opts] = NULL;
}
void futhark_context_config_set_debugging(struct futhark_context_config *cfg,
                                          int flag)
{
    cfg->cu_cfg.logging = cfg->cu_cfg.debugging = flag;
}
void futhark_context_config_set_logging(struct futhark_context_config *cfg,
                                        int flag)
{
    cfg->cu_cfg.logging = flag;
}
void futhark_context_config_set_device(struct futhark_context_config *cfg, const
                                       char *s)
{
    set_preferred_device(&cfg->cu_cfg, s);
}
void futhark_context_config_dump_program_to(struct futhark_context_config *cfg,
                                            const char *path)
{
    cfg->cu_cfg.dump_program_to = path;
}
void futhark_context_config_load_program_from(struct futhark_context_config *cfg,
                                              const char *path)
{
    cfg->cu_cfg.load_program_from = path;
}
void futhark_context_config_dump_ptx_to(struct futhark_context_config *cfg,
                                        const char *path)
{
    cfg->cu_cfg.dump_ptx_to = path;
}
void futhark_context_config_load_ptx_from(struct futhark_context_config *cfg,
                                          const char *path)
{
    cfg->cu_cfg.load_ptx_from = path;
}
void futhark_context_config_set_default_group_size(struct futhark_context_config *cfg,
                                                   int size)
{
    cfg->cu_cfg.default_block_size = size;
    cfg->cu_cfg.default_block_size_changed = 1;
}
void futhark_context_config_set_default_num_groups(struct futhark_context_config *cfg,
                                                   int num)
{
    cfg->cu_cfg.default_grid_size = num;
    cfg->cu_cfg.default_grid_size_changed = 1;
}
void futhark_context_config_set_default_tile_size(struct futhark_context_config *cfg,
                                                  int size)
{
    cfg->cu_cfg.default_tile_size = size;
    cfg->cu_cfg.default_tile_size_changed = 1;
}
void futhark_context_config_set_default_threshold(struct futhark_context_config *cfg,
                                                  int size)
{
    cfg->cu_cfg.default_threshold = size;
}
int futhark_context_config_set_size(struct futhark_context_config *cfg, const
                                    char *size_name, size_t size_value)
{
    for (int i = 0; i < 35; i++) {
        if (strcmp(size_name, size_names[i]) == 0) {
            cfg->sizes[i] = size_value;
            return 0;
        }
    }
    if (strcmp(size_name, "default_group_size") == 0) {
        cfg->cu_cfg.default_block_size = size_value;
        return 0;
    }
    if (strcmp(size_name, "default_num_groups") == 0) {
        cfg->cu_cfg.default_grid_size = size_value;
        return 0;
    }
    if (strcmp(size_name, "default_threshold") == 0) {
        cfg->cu_cfg.default_threshold = size_value;
        return 0;
    }
    if (strcmp(size_name, "default_tile_size") == 0) {
        cfg->cu_cfg.default_tile_size = size_value;
        return 0;
    }
    return 1;
}
struct futhark_context {
    int detail_memory;
    int debugging;
    int profiling;
    lock_t lock;
    char *error;
    int64_t peak_mem_usage_device;
    int64_t cur_mem_usage_device;
    int64_t peak_mem_usage_default;
    int64_t cur_mem_usage_default;
    struct memblock_device counter_mem_33446;
    struct memblock_device counter_mem_32968;
    struct memblock_device counter_mem_33384;
    CUfunction map_transpose_bool;
    CUfunction map_transpose_bool_low_height;
    CUfunction map_transpose_bool_low_width;
    CUfunction map_transpose_bool_small;
    CUfunction map_transpose_i32;
    CUfunction map_transpose_i32_low_height;
    CUfunction map_transpose_i32_low_width;
    CUfunction map_transpose_i32_small;
    CUfunction replicate_32962;
    CUfunction replicate_33364;
    CUfunction scan_stage1_32141;
    CUfunction scan_stage1_32150;
    CUfunction scan_stage1_32190;
    CUfunction scan_stage1_32199;
    CUfunction scan_stage1_32225;
    CUfunction scan_stage2_32141;
    CUfunction scan_stage2_32150;
    CUfunction scan_stage2_32190;
    CUfunction scan_stage2_32199;
    CUfunction scan_stage2_32225;
    CUfunction scan_stage3_33062;
    CUfunction scan_stage3_33126;
    CUfunction scan_stage3_33205;
    CUfunction scan_stage3_33274;
    CUfunction scan_stage3_33343;
    CUfunction segmap_32152;
    CUfunction segmap_32168;
    CUfunction segmap_32201;
    CUfunction segmap_32227;
    CUfunction segmap_32242;
    CUfunction segmap_32256;
    CUfunction segmap_32274;
    CUfunction segmap_32286;
    CUfunction segmap_32311;
    CUfunction segmap_32531;
    CUfunction segmap_32547;
    CUfunction segred_large_32522;
    CUfunction segred_nonseg_32132;
    CUfunction segred_nonseg_32307;
    CUfunction segred_small_32522;
    struct cuda_context cuda;
    struct sizes sizes;
} ;
struct futhark_context *futhark_context_new(struct futhark_context_config *cfg)
{
    struct futhark_context *ctx =
                           (struct futhark_context *) malloc(sizeof(struct futhark_context));
    
    if (ctx == NULL)
        return NULL;
    ctx->profiling = ctx->debugging = ctx->detail_memory =
        cfg->cu_cfg.debugging;
    ctx->cuda.cfg = cfg->cu_cfg;
    create_lock(&ctx->lock);
    ctx->peak_mem_usage_device = 0;
    ctx->cur_mem_usage_device = 0;
    ctx->peak_mem_usage_default = 0;
    ctx->cur_mem_usage_default = 0;
    cuda_setup(&ctx->cuda, cuda_program, cfg->nvrtc_opts);
    CUDA_SUCCEED(cuModuleGetFunction(&ctx->map_transpose_bool, ctx->cuda.module,
                                     "map_transpose_bool"));
    CUDA_SUCCEED(cuModuleGetFunction(&ctx->map_transpose_bool_low_height,
                                     ctx->cuda.module,
                                     "map_transpose_bool_low_height"));
    CUDA_SUCCEED(cuModuleGetFunction(&ctx->map_transpose_bool_low_width,
                                     ctx->cuda.module,
                                     "map_transpose_bool_low_width"));
    CUDA_SUCCEED(cuModuleGetFunction(&ctx->map_transpose_bool_small,
                                     ctx->cuda.module,
                                     "map_transpose_bool_small"));
    CUDA_SUCCEED(cuModuleGetFunction(&ctx->map_transpose_i32, ctx->cuda.module,
                                     "map_transpose_i32"));
    CUDA_SUCCEED(cuModuleGetFunction(&ctx->map_transpose_i32_low_height,
                                     ctx->cuda.module,
                                     "map_transpose_i32_low_height"));
    CUDA_SUCCEED(cuModuleGetFunction(&ctx->map_transpose_i32_low_width,
                                     ctx->cuda.module,
                                     "map_transpose_i32_low_width"));
    CUDA_SUCCEED(cuModuleGetFunction(&ctx->map_transpose_i32_small,
                                     ctx->cuda.module,
                                     "map_transpose_i32_small"));
    CUDA_SUCCEED(cuModuleGetFunction(&ctx->replicate_32962, ctx->cuda.module,
                                     "replicate_32962"));
    CUDA_SUCCEED(cuModuleGetFunction(&ctx->replicate_33364, ctx->cuda.module,
                                     "replicate_33364"));
    CUDA_SUCCEED(cuModuleGetFunction(&ctx->scan_stage1_32141, ctx->cuda.module,
                                     "scan_stage1_32141"));
    CUDA_SUCCEED(cuModuleGetFunction(&ctx->scan_stage1_32150, ctx->cuda.module,
                                     "scan_stage1_32150"));
    CUDA_SUCCEED(cuModuleGetFunction(&ctx->scan_stage1_32190, ctx->cuda.module,
                                     "scan_stage1_32190"));
    CUDA_SUCCEED(cuModuleGetFunction(&ctx->scan_stage1_32199, ctx->cuda.module,
                                     "scan_stage1_32199"));
    CUDA_SUCCEED(cuModuleGetFunction(&ctx->scan_stage1_32225, ctx->cuda.module,
                                     "scan_stage1_32225"));
    CUDA_SUCCEED(cuModuleGetFunction(&ctx->scan_stage2_32141, ctx->cuda.module,
                                     "scan_stage2_32141"));
    CUDA_SUCCEED(cuModuleGetFunction(&ctx->scan_stage2_32150, ctx->cuda.module,
                                     "scan_stage2_32150"));
    CUDA_SUCCEED(cuModuleGetFunction(&ctx->scan_stage2_32190, ctx->cuda.module,
                                     "scan_stage2_32190"));
    CUDA_SUCCEED(cuModuleGetFunction(&ctx->scan_stage2_32199, ctx->cuda.module,
                                     "scan_stage2_32199"));
    CUDA_SUCCEED(cuModuleGetFunction(&ctx->scan_stage2_32225, ctx->cuda.module,
                                     "scan_stage2_32225"));
    CUDA_SUCCEED(cuModuleGetFunction(&ctx->scan_stage3_33062, ctx->cuda.module,
                                     "scan_stage3_33062"));
    CUDA_SUCCEED(cuModuleGetFunction(&ctx->scan_stage3_33126, ctx->cuda.module,
                                     "scan_stage3_33126"));
    CUDA_SUCCEED(cuModuleGetFunction(&ctx->scan_stage3_33205, ctx->cuda.module,
                                     "scan_stage3_33205"));
    CUDA_SUCCEED(cuModuleGetFunction(&ctx->scan_stage3_33274, ctx->cuda.module,
                                     "scan_stage3_33274"));
    CUDA_SUCCEED(cuModuleGetFunction(&ctx->scan_stage3_33343, ctx->cuda.module,
                                     "scan_stage3_33343"));
    CUDA_SUCCEED(cuModuleGetFunction(&ctx->segmap_32152, ctx->cuda.module,
                                     "segmap_32152"));
    CUDA_SUCCEED(cuModuleGetFunction(&ctx->segmap_32168, ctx->cuda.module,
                                     "segmap_32168"));
    CUDA_SUCCEED(cuModuleGetFunction(&ctx->segmap_32201, ctx->cuda.module,
                                     "segmap_32201"));
    CUDA_SUCCEED(cuModuleGetFunction(&ctx->segmap_32227, ctx->cuda.module,
                                     "segmap_32227"));
    CUDA_SUCCEED(cuModuleGetFunction(&ctx->segmap_32242, ctx->cuda.module,
                                     "segmap_32242"));
    CUDA_SUCCEED(cuModuleGetFunction(&ctx->segmap_32256, ctx->cuda.module,
                                     "segmap_32256"));
    CUDA_SUCCEED(cuModuleGetFunction(&ctx->segmap_32274, ctx->cuda.module,
                                     "segmap_32274"));
    CUDA_SUCCEED(cuModuleGetFunction(&ctx->segmap_32286, ctx->cuda.module,
                                     "segmap_32286"));
    CUDA_SUCCEED(cuModuleGetFunction(&ctx->segmap_32311, ctx->cuda.module,
                                     "segmap_32311"));
    CUDA_SUCCEED(cuModuleGetFunction(&ctx->segmap_32531, ctx->cuda.module,
                                     "segmap_32531"));
    CUDA_SUCCEED(cuModuleGetFunction(&ctx->segmap_32547, ctx->cuda.module,
                                     "segmap_32547"));
    CUDA_SUCCEED(cuModuleGetFunction(&ctx->segred_large_32522, ctx->cuda.module,
                                     "segred_large_32522"));
    CUDA_SUCCEED(cuModuleGetFunction(&ctx->segred_nonseg_32132,
                                     ctx->cuda.module, "segred_nonseg_32132"));
    CUDA_SUCCEED(cuModuleGetFunction(&ctx->segred_nonseg_32307,
                                     ctx->cuda.module, "segred_nonseg_32307"));
    CUDA_SUCCEED(cuModuleGetFunction(&ctx->segred_small_32522, ctx->cuda.module,
                                     "segred_small_32522"));
    {
        ctx->counter_mem_33446.references = NULL;
        ctx->counter_mem_33446.size = 0;
        CUDA_SUCCEED(cuMemAlloc(&ctx->counter_mem_33446.mem, (10240 >
                                                              0 ? 10240 : 1) *
                                sizeof(int32_t)));
        if (10240 > 0)
            CUDA_SUCCEED(cuMemcpyHtoD(ctx->counter_mem_33446.mem,
                                      counter_mem_realtype_33540, 10240 *
                                      sizeof(int32_t)));
    }
    {
        ctx->counter_mem_32968.references = NULL;
        ctx->counter_mem_32968.size = 0;
        CUDA_SUCCEED(cuMemAlloc(&ctx->counter_mem_32968.mem, (10 > 0 ? 10 : 1) *
                                sizeof(int32_t)));
        if (10 > 0)
            CUDA_SUCCEED(cuMemcpyHtoD(ctx->counter_mem_32968.mem,
                                      counter_mem_realtype_33644, 10 *
                                      sizeof(int32_t)));
    }
    {
        ctx->counter_mem_33384.references = NULL;
        ctx->counter_mem_33384.size = 0;
        CUDA_SUCCEED(cuMemAlloc(&ctx->counter_mem_33384.mem, (10 > 0 ? 10 : 1) *
                                sizeof(int32_t)));
        if (10 > 0)
            CUDA_SUCCEED(cuMemcpyHtoD(ctx->counter_mem_33384.mem,
                                      counter_mem_realtype_33840, 10 *
                                      sizeof(int32_t)));
    }
    ctx->sizes.mainzisegmap_group_sizze_32314 = cfg->sizes[0];
    ctx->sizes.mainzisegmap_group_sizze_32536 = cfg->sizes[1];
    ctx->sizes.mainzisegmap_group_sizze_32552 = cfg->sizes[2];
    ctx->sizes.mainzisegmap_num_groups_32316 = cfg->sizes[3];
    ctx->sizes.mainzisegred_group_sizze_32516 = cfg->sizes[4];
    ctx->sizes.mainzisegred_num_groups_32518 = cfg->sizes[5];
    ctx->sizes.quicksort1zigroup_sizze_32965 = cfg->sizes[6];
    ctx->sizes.quicksort1zigroup_sizze_33065 = cfg->sizes[7];
    ctx->sizes.quicksort1zigroup_sizze_33129 = cfg->sizes[8];
    ctx->sizes.quicksort1zigroup_sizze_33208 = cfg->sizes[9];
    ctx->sizes.quicksort1zigroup_sizze_33277 = cfg->sizes[10];
    ctx->sizes.quicksort1zigroup_sizze_33346 = cfg->sizes[11];
    ctx->sizes.quicksort1zigroup_sizze_33367 = cfg->sizes[12];
    ctx->sizes.quicksort1zisegmap_group_sizze_32155 = cfg->sizes[13];
    ctx->sizes.quicksort1zisegmap_group_sizze_32171 = cfg->sizes[14];
    ctx->sizes.quicksort1zisegmap_group_sizze_32204 = cfg->sizes[15];
    ctx->sizes.quicksort1zisegmap_group_sizze_32230 = cfg->sizes[16];
    ctx->sizes.quicksort1zisegmap_group_sizze_32245 = cfg->sizes[17];
    ctx->sizes.quicksort1zisegmap_group_sizze_32259 = cfg->sizes[18];
    ctx->sizes.quicksort1zisegmap_group_sizze_32277 = cfg->sizes[19];
    ctx->sizes.quicksort1zisegmap_group_sizze_32289 = cfg->sizes[20];
    ctx->sizes.quicksort1zisegred_group_sizze_32124 = cfg->sizes[21];
    ctx->sizes.quicksort1zisegred_group_sizze_32299 = cfg->sizes[22];
    ctx->sizes.quicksort1zisegred_num_groups_32126 = cfg->sizes[23];
    ctx->sizes.quicksort1zisegred_num_groups_32301 = cfg->sizes[24];
    ctx->sizes.quicksort1zisegscan_group_sizze_32135 = cfg->sizes[25];
    ctx->sizes.quicksort1zisegscan_group_sizze_32144 = cfg->sizes[26];
    ctx->sizes.quicksort1zisegscan_group_sizze_32184 = cfg->sizes[27];
    ctx->sizes.quicksort1zisegscan_group_sizze_32193 = cfg->sizes[28];
    ctx->sizes.quicksort1zisegscan_group_sizze_32219 = cfg->sizes[29];
    ctx->sizes.quicksort1zisegscan_num_groups_32137 = cfg->sizes[30];
    ctx->sizes.quicksort1zisegscan_num_groups_32146 = cfg->sizes[31];
    ctx->sizes.quicksort1zisegscan_num_groups_32186 = cfg->sizes[32];
    ctx->sizes.quicksort1zisegscan_num_groups_32195 = cfg->sizes[33];
    ctx->sizes.quicksort1zisegscan_num_groups_32221 = cfg->sizes[34];
    return ctx;
}
void futhark_context_free(struct futhark_context *ctx)
{
    cuda_cleanup(&ctx->cuda);
    free_lock(&ctx->lock);
    free(ctx);
}
int futhark_context_sync(struct futhark_context *ctx)
{
    CUDA_SUCCEED(cuCtxSynchronize());
    return 0;
}
char *futhark_context_get_error(struct futhark_context *ctx)
{
    return ctx->error;
}
void futhark_context_pause_profiling(struct futhark_context *ctx)
{
    (void) ctx;
}
void futhark_context_unpause_profiling(struct futhark_context *ctx)
{
    (void) ctx;
}
static int memblock_unref_device(struct futhark_context *ctx,
                                 struct memblock_device *block, const
                                 char *desc)
{
    if (block->references != NULL) {
        *block->references -= 1;
        if (ctx->detail_memory)
            fprintf(stderr,
                    "Unreferencing block %s (allocated as %s) in %s: %d references remaining.\n",
                    desc, block->desc, "space 'device'", *block->references);
        if (*block->references == 0) {
            ctx->cur_mem_usage_device -= block->size;
            CUDA_SUCCEED(cuda_free(&ctx->cuda, block->mem, block->desc));
            free(block->references);
            if (ctx->detail_memory)
                fprintf(stderr,
                        "%lld bytes freed (now allocated: %lld bytes)\n",
                        (long long) block->size,
                        (long long) ctx->cur_mem_usage_device);
        }
        block->references = NULL;
    }
    return 0;
}
static int memblock_alloc_device(struct futhark_context *ctx,
                                 struct memblock_device *block, int64_t size,
                                 const char *desc)
{
    if (size < 0)
        panic(1, "Negative allocation of %lld bytes attempted for %s in %s.\n",
              (long long) size, desc, "space 'device'",
              ctx->cur_mem_usage_device);
    
    int ret = memblock_unref_device(ctx, block, desc);
    
    ctx->cur_mem_usage_device += size;
    if (ctx->detail_memory)
        fprintf(stderr,
                "Allocating %lld bytes for %s in %s (then allocated: %lld bytes)",
                (long long) size, desc, "space 'device'",
                (long long) ctx->cur_mem_usage_device);
    if (ctx->cur_mem_usage_device > ctx->peak_mem_usage_device) {
        ctx->peak_mem_usage_device = ctx->cur_mem_usage_device;
        if (ctx->detail_memory)
            fprintf(stderr, " (new peak).\n");
    } else if (ctx->detail_memory)
        fprintf(stderr, ".\n");
    CUDA_SUCCEED(cuda_alloc(&ctx->cuda, size, desc, &block->mem));
    block->references = (int *) malloc(sizeof(int));
    *block->references = 1;
    block->size = size;
    block->desc = desc;
    return ret;
}
static int memblock_set_device(struct futhark_context *ctx,
                               struct memblock_device *lhs,
                               struct memblock_device *rhs, const
                               char *lhs_desc)
{
    int ret = memblock_unref_device(ctx, lhs, lhs_desc);
    
    (*rhs->references)++;
    *lhs = *rhs;
    return ret;
}
static int memblock_unref(struct futhark_context *ctx, struct memblock *block,
                          const char *desc)
{
    if (block->references != NULL) {
        *block->references -= 1;
        if (ctx->detail_memory)
            fprintf(stderr,
                    "Unreferencing block %s (allocated as %s) in %s: %d references remaining.\n",
                    desc, block->desc, "default space", *block->references);
        if (*block->references == 0) {
            ctx->cur_mem_usage_default -= block->size;
            free(block->mem);
            free(block->references);
            if (ctx->detail_memory)
                fprintf(stderr,
                        "%lld bytes freed (now allocated: %lld bytes)\n",
                        (long long) block->size,
                        (long long) ctx->cur_mem_usage_default);
        }
        block->references = NULL;
    }
    return 0;
}
static int memblock_alloc(struct futhark_context *ctx, struct memblock *block,
                          int64_t size, const char *desc)
{
    if (size < 0)
        panic(1, "Negative allocation of %lld bytes attempted for %s in %s.\n",
              (long long) size, desc, "default space",
              ctx->cur_mem_usage_default);
    
    int ret = memblock_unref(ctx, block, desc);
    
    ctx->cur_mem_usage_default += size;
    if (ctx->detail_memory)
        fprintf(stderr,
                "Allocating %lld bytes for %s in %s (then allocated: %lld bytes)",
                (long long) size, desc, "default space",
                (long long) ctx->cur_mem_usage_default);
    if (ctx->cur_mem_usage_default > ctx->peak_mem_usage_default) {
        ctx->peak_mem_usage_default = ctx->cur_mem_usage_default;
        if (ctx->detail_memory)
            fprintf(stderr, " (new peak).\n");
    } else if (ctx->detail_memory)
        fprintf(stderr, ".\n");
    block->mem = (char *) malloc(size);
    block->references = (int *) malloc(sizeof(int));
    *block->references = 1;
    block->size = size;
    block->desc = desc;
    return ret;
}
static int memblock_set(struct futhark_context *ctx, struct memblock *lhs,
                        struct memblock *rhs, const char *lhs_desc)
{
    int ret = memblock_unref(ctx, lhs, lhs_desc);
    
    (*rhs->references)++;
    *lhs = *rhs;
    return ret;
}
void futhark_debugging_report(struct futhark_context *ctx)
{
    if (ctx->detail_memory || ctx->profiling) {
        fprintf(stderr, "Peak memory usage for space 'device': %lld bytes.\n",
                (long long) ctx->peak_mem_usage_device);
        fprintf(stderr, "Peak memory usage for default space: %lld bytes.\n",
                (long long) ctx->peak_mem_usage_default);
    }
    if (ctx->profiling) { }
}
static int futrts_main(struct futhark_context *ctx,
                       struct memblock_device *out_mem_p_33521,
                       int32_t *out_out_arrsizze_33522,
                       int32_t *out_out_arrsizze_33523,
                       struct memblock_device input_mem_32784, int32_t n_31902,
                       int32_t m_31903);
static int futrts__map_transpose_bool(struct futhark_context *ctx,
                                      struct memblock_device destmem_0,
                                      int32_t destoffset_1,
                                      struct memblock_device srcmem_2,
                                      int32_t srcoffset_3, int32_t num_arrays_4,
                                      int32_t x_elems_5, int32_t y_elems_6,
                                      int32_t in_elems_7, int32_t out_elems_8);
static int futrts__map_transpose_i32(struct futhark_context *ctx,
                                     struct memblock_device destmem_0,
                                     int32_t destoffset_1,
                                     struct memblock_device srcmem_2,
                                     int32_t srcoffset_3, int32_t num_arrays_4,
                                     int32_t x_elems_5, int32_t y_elems_6,
                                     int32_t in_elems_7, int32_t out_elems_8);
static int futrts_quicksort1(struct futhark_context *ctx,
                             struct memblock_device *out_mem_p_33638,
                             int32_t *out_out_arrsizze_33639,
                             int32_t *out_out_arrsizze_33640,
                             struct memblock_device input_mem_32784,
                             int32_t n_31612, int32_t sizze_31613);
static int futrts__replicate_i32(struct futhark_context *ctx,
                                 struct memblock_device mem_33360,
                                 int32_t num_elems_33361, int32_t val_33362);
static int futrts__replicate_bool(struct futhark_context *ctx,
                                  struct memblock_device mem_32958,
                                  int32_t num_elems_32959, bool val_32960);
static inline int8_t add8(int8_t x, int8_t y)
{
    return x + y;
}
static inline int16_t add16(int16_t x, int16_t y)
{
    return x + y;
}
static inline int32_t add32(int32_t x, int32_t y)
{
    return x + y;
}
static inline int64_t add64(int64_t x, int64_t y)
{
    return x + y;
}
static inline int8_t sub8(int8_t x, int8_t y)
{
    return x - y;
}
static inline int16_t sub16(int16_t x, int16_t y)
{
    return x - y;
}
static inline int32_t sub32(int32_t x, int32_t y)
{
    return x - y;
}
static inline int64_t sub64(int64_t x, int64_t y)
{
    return x - y;
}
static inline int8_t mul8(int8_t x, int8_t y)
{
    return x * y;
}
static inline int16_t mul16(int16_t x, int16_t y)
{
    return x * y;
}
static inline int32_t mul32(int32_t x, int32_t y)
{
    return x * y;
}
static inline int64_t mul64(int64_t x, int64_t y)
{
    return x * y;
}
static inline uint8_t udiv8(uint8_t x, uint8_t y)
{
    return x / y;
}
static inline uint16_t udiv16(uint16_t x, uint16_t y)
{
    return x / y;
}
static inline uint32_t udiv32(uint32_t x, uint32_t y)
{
    return x / y;
}
static inline uint64_t udiv64(uint64_t x, uint64_t y)
{
    return x / y;
}
static inline uint8_t umod8(uint8_t x, uint8_t y)
{
    return x % y;
}
static inline uint16_t umod16(uint16_t x, uint16_t y)
{
    return x % y;
}
static inline uint32_t umod32(uint32_t x, uint32_t y)
{
    return x % y;
}
static inline uint64_t umod64(uint64_t x, uint64_t y)
{
    return x % y;
}
static inline int8_t sdiv8(int8_t x, int8_t y)
{
    int8_t q = x / y;
    int8_t r = x % y;
    
    return q - ((r != 0 && r < 0 != y < 0) ? 1 : 0);
}
static inline int16_t sdiv16(int16_t x, int16_t y)
{
    int16_t q = x / y;
    int16_t r = x % y;
    
    return q - ((r != 0 && r < 0 != y < 0) ? 1 : 0);
}
static inline int32_t sdiv32(int32_t x, int32_t y)
{
    int32_t q = x / y;
    int32_t r = x % y;
    
    return q - ((r != 0 && r < 0 != y < 0) ? 1 : 0);
}
static inline int64_t sdiv64(int64_t x, int64_t y)
{
    int64_t q = x / y;
    int64_t r = x % y;
    
    return q - ((r != 0 && r < 0 != y < 0) ? 1 : 0);
}
static inline int8_t smod8(int8_t x, int8_t y)
{
    int8_t r = x % y;
    
    return r + (r == 0 || (x > 0 && y > 0) || (x < 0 && y < 0) ? 0 : y);
}
static inline int16_t smod16(int16_t x, int16_t y)
{
    int16_t r = x % y;
    
    return r + (r == 0 || (x > 0 && y > 0) || (x < 0 && y < 0) ? 0 : y);
}
static inline int32_t smod32(int32_t x, int32_t y)
{
    int32_t r = x % y;
    
    return r + (r == 0 || (x > 0 && y > 0) || (x < 0 && y < 0) ? 0 : y);
}
static inline int64_t smod64(int64_t x, int64_t y)
{
    int64_t r = x % y;
    
    return r + (r == 0 || (x > 0 && y > 0) || (x < 0 && y < 0) ? 0 : y);
}
static inline int8_t squot8(int8_t x, int8_t y)
{
    return x / y;
}
static inline int16_t squot16(int16_t x, int16_t y)
{
    return x / y;
}
static inline int32_t squot32(int32_t x, int32_t y)
{
    return x / y;
}
static inline int64_t squot64(int64_t x, int64_t y)
{
    return x / y;
}
static inline int8_t srem8(int8_t x, int8_t y)
{
    return x % y;
}
static inline int16_t srem16(int16_t x, int16_t y)
{
    return x % y;
}
static inline int32_t srem32(int32_t x, int32_t y)
{
    return x % y;
}
static inline int64_t srem64(int64_t x, int64_t y)
{
    return x % y;
}
static inline int8_t smin8(int8_t x, int8_t y)
{
    return x < y ? x : y;
}
static inline int16_t smin16(int16_t x, int16_t y)
{
    return x < y ? x : y;
}
static inline int32_t smin32(int32_t x, int32_t y)
{
    return x < y ? x : y;
}
static inline int64_t smin64(int64_t x, int64_t y)
{
    return x < y ? x : y;
}
static inline uint8_t umin8(uint8_t x, uint8_t y)
{
    return x < y ? x : y;
}
static inline uint16_t umin16(uint16_t x, uint16_t y)
{
    return x < y ? x : y;
}
static inline uint32_t umin32(uint32_t x, uint32_t y)
{
    return x < y ? x : y;
}
static inline uint64_t umin64(uint64_t x, uint64_t y)
{
    return x < y ? x : y;
}
static inline int8_t smax8(int8_t x, int8_t y)
{
    return x < y ? y : x;
}
static inline int16_t smax16(int16_t x, int16_t y)
{
    return x < y ? y : x;
}
static inline int32_t smax32(int32_t x, int32_t y)
{
    return x < y ? y : x;
}
static inline int64_t smax64(int64_t x, int64_t y)
{
    return x < y ? y : x;
}
static inline uint8_t umax8(uint8_t x, uint8_t y)
{
    return x < y ? y : x;
}
static inline uint16_t umax16(uint16_t x, uint16_t y)
{
    return x < y ? y : x;
}
static inline uint32_t umax32(uint32_t x, uint32_t y)
{
    return x < y ? y : x;
}
static inline uint64_t umax64(uint64_t x, uint64_t y)
{
    return x < y ? y : x;
}
static inline uint8_t shl8(uint8_t x, uint8_t y)
{
    return x << y;
}
static inline uint16_t shl16(uint16_t x, uint16_t y)
{
    return x << y;
}
static inline uint32_t shl32(uint32_t x, uint32_t y)
{
    return x << y;
}
static inline uint64_t shl64(uint64_t x, uint64_t y)
{
    return x << y;
}
static inline uint8_t lshr8(uint8_t x, uint8_t y)
{
    return x >> y;
}
static inline uint16_t lshr16(uint16_t x, uint16_t y)
{
    return x >> y;
}
static inline uint32_t lshr32(uint32_t x, uint32_t y)
{
    return x >> y;
}
static inline uint64_t lshr64(uint64_t x, uint64_t y)
{
    return x >> y;
}
static inline int8_t ashr8(int8_t x, int8_t y)
{
    return x >> y;
}
static inline int16_t ashr16(int16_t x, int16_t y)
{
    return x >> y;
}
static inline int32_t ashr32(int32_t x, int32_t y)
{
    return x >> y;
}
static inline int64_t ashr64(int64_t x, int64_t y)
{
    return x >> y;
}
static inline uint8_t and8(uint8_t x, uint8_t y)
{
    return x & y;
}
static inline uint16_t and16(uint16_t x, uint16_t y)
{
    return x & y;
}
static inline uint32_t and32(uint32_t x, uint32_t y)
{
    return x & y;
}
static inline uint64_t and64(uint64_t x, uint64_t y)
{
    return x & y;
}
static inline uint8_t or8(uint8_t x, uint8_t y)
{
    return x | y;
}
static inline uint16_t or16(uint16_t x, uint16_t y)
{
    return x | y;
}
static inline uint32_t or32(uint32_t x, uint32_t y)
{
    return x | y;
}
static inline uint64_t or64(uint64_t x, uint64_t y)
{
    return x | y;
}
static inline uint8_t xor8(uint8_t x, uint8_t y)
{
    return x ^ y;
}
static inline uint16_t xor16(uint16_t x, uint16_t y)
{
    return x ^ y;
}
static inline uint32_t xor32(uint32_t x, uint32_t y)
{
    return x ^ y;
}
static inline uint64_t xor64(uint64_t x, uint64_t y)
{
    return x ^ y;
}
static inline char ult8(uint8_t x, uint8_t y)
{
    return x < y;
}
static inline char ult16(uint16_t x, uint16_t y)
{
    return x < y;
}
static inline char ult32(uint32_t x, uint32_t y)
{
    return x < y;
}
static inline char ult64(uint64_t x, uint64_t y)
{
    return x < y;
}
static inline char ule8(uint8_t x, uint8_t y)
{
    return x <= y;
}
static inline char ule16(uint16_t x, uint16_t y)
{
    return x <= y;
}
static inline char ule32(uint32_t x, uint32_t y)
{
    return x <= y;
}
static inline char ule64(uint64_t x, uint64_t y)
{
    return x <= y;
}
static inline char slt8(int8_t x, int8_t y)
{
    return x < y;
}
static inline char slt16(int16_t x, int16_t y)
{
    return x < y;
}
static inline char slt32(int32_t x, int32_t y)
{
    return x < y;
}
static inline char slt64(int64_t x, int64_t y)
{
    return x < y;
}
static inline char sle8(int8_t x, int8_t y)
{
    return x <= y;
}
static inline char sle16(int16_t x, int16_t y)
{
    return x <= y;
}
static inline char sle32(int32_t x, int32_t y)
{
    return x <= y;
}
static inline char sle64(int64_t x, int64_t y)
{
    return x <= y;
}
static inline int8_t pow8(int8_t x, int8_t y)
{
    int8_t res = 1, rem = y;
    
    while (rem != 0) {
        if (rem & 1)
            res *= x;
        rem >>= 1;
        x *= x;
    }
    return res;
}
static inline int16_t pow16(int16_t x, int16_t y)
{
    int16_t res = 1, rem = y;
    
    while (rem != 0) {
        if (rem & 1)
            res *= x;
        rem >>= 1;
        x *= x;
    }
    return res;
}
static inline int32_t pow32(int32_t x, int32_t y)
{
    int32_t res = 1, rem = y;
    
    while (rem != 0) {
        if (rem & 1)
            res *= x;
        rem >>= 1;
        x *= x;
    }
    return res;
}
static inline int64_t pow64(int64_t x, int64_t y)
{
    int64_t res = 1, rem = y;
    
    while (rem != 0) {
        if (rem & 1)
            res *= x;
        rem >>= 1;
        x *= x;
    }
    return res;
}
static inline bool itob_i8_bool(int8_t x)
{
    return x;
}
static inline bool itob_i16_bool(int16_t x)
{
    return x;
}
static inline bool itob_i32_bool(int32_t x)
{
    return x;
}
static inline bool itob_i64_bool(int64_t x)
{
    return x;
}
static inline int8_t btoi_bool_i8(bool x)
{
    return x;
}
static inline int16_t btoi_bool_i16(bool x)
{
    return x;
}
static inline int32_t btoi_bool_i32(bool x)
{
    return x;
}
static inline int64_t btoi_bool_i64(bool x)
{
    return x;
}
#define sext_i8_i8(x) ((int8_t) (int8_t) x)
#define sext_i8_i16(x) ((int16_t) (int8_t) x)
#define sext_i8_i32(x) ((int32_t) (int8_t) x)
#define sext_i8_i64(x) ((int64_t) (int8_t) x)
#define sext_i16_i8(x) ((int8_t) (int16_t) x)
#define sext_i16_i16(x) ((int16_t) (int16_t) x)
#define sext_i16_i32(x) ((int32_t) (int16_t) x)
#define sext_i16_i64(x) ((int64_t) (int16_t) x)
#define sext_i32_i8(x) ((int8_t) (int32_t) x)
#define sext_i32_i16(x) ((int16_t) (int32_t) x)
#define sext_i32_i32(x) ((int32_t) (int32_t) x)
#define sext_i32_i64(x) ((int64_t) (int32_t) x)
#define sext_i64_i8(x) ((int8_t) (int64_t) x)
#define sext_i64_i16(x) ((int16_t) (int64_t) x)
#define sext_i64_i32(x) ((int32_t) (int64_t) x)
#define sext_i64_i64(x) ((int64_t) (int64_t) x)
#define zext_i8_i8(x) ((uint8_t) (uint8_t) x)
#define zext_i8_i16(x) ((uint16_t) (uint8_t) x)
#define zext_i8_i32(x) ((uint32_t) (uint8_t) x)
#define zext_i8_i64(x) ((uint64_t) (uint8_t) x)
#define zext_i16_i8(x) ((uint8_t) (uint16_t) x)
#define zext_i16_i16(x) ((uint16_t) (uint16_t) x)
#define zext_i16_i32(x) ((uint32_t) (uint16_t) x)
#define zext_i16_i64(x) ((uint64_t) (uint16_t) x)
#define zext_i32_i8(x) ((uint8_t) (uint32_t) x)
#define zext_i32_i16(x) ((uint16_t) (uint32_t) x)
#define zext_i32_i32(x) ((uint32_t) (uint32_t) x)
#define zext_i32_i64(x) ((uint64_t) (uint32_t) x)
#define zext_i64_i8(x) ((uint8_t) (uint64_t) x)
#define zext_i64_i16(x) ((uint16_t) (uint64_t) x)
#define zext_i64_i32(x) ((uint32_t) (uint64_t) x)
#define zext_i64_i64(x) ((uint64_t) (uint64_t) x)
#ifdef __OPENCL_VERSION__
int32_t futrts_popc8(int8_t x)
{
    return popcount(x);
}
int32_t futrts_popc16(int16_t x)
{
    return popcount(x);
}
int32_t futrts_popc32(int32_t x)
{
    return popcount(x);
}
int32_t futrts_popc64(int64_t x)
{
    return popcount(x);
}
#elif __CUDA_ARCH__
int32_t futrts_popc8(int8_t x)
{
    return __popc(zext_i8_i32(x));
}
int32_t futrts_popc16(int16_t x)
{
    return __popc(zext_i16_i32(x));
}
int32_t futrts_popc32(int32_t x)
{
    return __popc(x);
}
int32_t futrts_popc64(int64_t x)
{
    return __popcll(x);
}
#else
int32_t futrts_popc8(int8_t x)
{
    int c = 0;
    
    for (; x; ++c)
        x &= x - 1;
    return c;
}
int32_t futrts_popc16(int16_t x)
{
    int c = 0;
    
    for (; x; ++c)
        x &= x - 1;
    return c;
}
int32_t futrts_popc32(int32_t x)
{
    int c = 0;
    
    for (; x; ++c)
        x &= x - 1;
    return c;
}
int32_t futrts_popc64(int64_t x)
{
    int c = 0;
    
    for (; x; ++c)
        x &= x - 1;
    return c;
}
#endif
#ifdef __OPENCL_VERSION__
int32_t futrts_clzz8(int8_t x)
{
    return clz(x);
}
int32_t futrts_clzz16(int16_t x)
{
    return clz(x);
}
int32_t futrts_clzz32(int32_t x)
{
    return clz(x);
}
int32_t futrts_clzz64(int64_t x)
{
    return clz(x);
}
#elif __CUDA_ARCH__
int32_t futrts_clzz8(int8_t x)
{
    return __clz(zext_i8_i32(x)) - 24;
}
int32_t futrts_clzz16(int16_t x)
{
    return __clz(zext_i16_i32(x)) - 16;
}
int32_t futrts_clzz32(int32_t x)
{
    return __clz(x);
}
int32_t futrts_clzz64(int64_t x)
{
    return __clzll(x);
}
#else
int32_t futrts_clzz8(int8_t x)
{
    int n = 0;
    int bits = sizeof(x) * 8;
    
    for (int i = 0; i < bits; i++) {
        if (x < 0)
            break;
        n++;
        x <<= 1;
    }
    return n;
}
int32_t futrts_clzz16(int16_t x)
{
    int n = 0;
    int bits = sizeof(x) * 8;
    
    for (int i = 0; i < bits; i++) {
        if (x < 0)
            break;
        n++;
        x <<= 1;
    }
    return n;
}
int32_t futrts_clzz32(int32_t x)
{
    int n = 0;
    int bits = sizeof(x) * 8;
    
    for (int i = 0; i < bits; i++) {
        if (x < 0)
            break;
        n++;
        x <<= 1;
    }
    return n;
}
int32_t futrts_clzz64(int64_t x)
{
    int n = 0;
    int bits = sizeof(x) * 8;
    
    for (int i = 0; i < bits; i++) {
        if (x < 0)
            break;
        n++;
        x <<= 1;
    }
    return n;
}
#endif
static inline float fdiv32(float x, float y)
{
    return x / y;
}
static inline float fadd32(float x, float y)
{
    return x + y;
}
static inline float fsub32(float x, float y)
{
    return x - y;
}
static inline float fmul32(float x, float y)
{
    return x * y;
}
static inline float fmin32(float x, float y)
{
    return fmin(x, y);
}
static inline float fmax32(float x, float y)
{
    return fmax(x, y);
}
static inline float fpow32(float x, float y)
{
    return pow(x, y);
}
static inline char cmplt32(float x, float y)
{
    return x < y;
}
static inline char cmple32(float x, float y)
{
    return x <= y;
}
static inline float sitofp_i8_f32(int8_t x)
{
    return x;
}
static inline float sitofp_i16_f32(int16_t x)
{
    return x;
}
static inline float sitofp_i32_f32(int32_t x)
{
    return x;
}
static inline float sitofp_i64_f32(int64_t x)
{
    return x;
}
static inline float uitofp_i8_f32(uint8_t x)
{
    return x;
}
static inline float uitofp_i16_f32(uint16_t x)
{
    return x;
}
static inline float uitofp_i32_f32(uint32_t x)
{
    return x;
}
static inline float uitofp_i64_f32(uint64_t x)
{
    return x;
}
static inline int8_t fptosi_f32_i8(float x)
{
    return x;
}
static inline int16_t fptosi_f32_i16(float x)
{
    return x;
}
static inline int32_t fptosi_f32_i32(float x)
{
    return x;
}
static inline int64_t fptosi_f32_i64(float x)
{
    return x;
}
static inline uint8_t fptoui_f32_i8(float x)
{
    return x;
}
static inline uint16_t fptoui_f32_i16(float x)
{
    return x;
}
static inline uint32_t fptoui_f32_i32(float x)
{
    return x;
}
static inline uint64_t fptoui_f32_i64(float x)
{
    return x;
}
static inline double fdiv64(double x, double y)
{
    return x / y;
}
static inline double fadd64(double x, double y)
{
    return x + y;
}
static inline double fsub64(double x, double y)
{
    return x - y;
}
static inline double fmul64(double x, double y)
{
    return x * y;
}
static inline double fmin64(double x, double y)
{
    return fmin(x, y);
}
static inline double fmax64(double x, double y)
{
    return fmax(x, y);
}
static inline double fpow64(double x, double y)
{
    return pow(x, y);
}
static inline char cmplt64(double x, double y)
{
    return x < y;
}
static inline char cmple64(double x, double y)
{
    return x <= y;
}
static inline double sitofp_i8_f64(int8_t x)
{
    return x;
}
static inline double sitofp_i16_f64(int16_t x)
{
    return x;
}
static inline double sitofp_i32_f64(int32_t x)
{
    return x;
}
static inline double sitofp_i64_f64(int64_t x)
{
    return x;
}
static inline double uitofp_i8_f64(uint8_t x)
{
    return x;
}
static inline double uitofp_i16_f64(uint16_t x)
{
    return x;
}
static inline double uitofp_i32_f64(uint32_t x)
{
    return x;
}
static inline double uitofp_i64_f64(uint64_t x)
{
    return x;
}
static inline int8_t fptosi_f64_i8(double x)
{
    return x;
}
static inline int16_t fptosi_f64_i16(double x)
{
    return x;
}
static inline int32_t fptosi_f64_i32(double x)
{
    return x;
}
static inline int64_t fptosi_f64_i64(double x)
{
    return x;
}
static inline uint8_t fptoui_f64_i8(double x)
{
    return x;
}
static inline uint16_t fptoui_f64_i16(double x)
{
    return x;
}
static inline uint32_t fptoui_f64_i32(double x)
{
    return x;
}
static inline uint64_t fptoui_f64_i64(double x)
{
    return x;
}
static inline float fpconv_f32_f32(float x)
{
    return x;
}
static inline double fpconv_f32_f64(float x)
{
    return x;
}
static inline float fpconv_f64_f32(double x)
{
    return x;
}
static inline double fpconv_f64_f64(double x)
{
    return x;
}
static inline float futrts_log32(float x)
{
    return log(x);
}
static inline float futrts_log2_32(float x)
{
    return log2(x);
}
static inline float futrts_log10_32(float x)
{
    return log10(x);
}
static inline float futrts_sqrt32(float x)
{
    return sqrt(x);
}
static inline float futrts_exp32(float x)
{
    return exp(x);
}
static inline float futrts_cos32(float x)
{
    return cos(x);
}
static inline float futrts_sin32(float x)
{
    return sin(x);
}
static inline float futrts_tan32(float x)
{
    return tan(x);
}
static inline float futrts_acos32(float x)
{
    return acos(x);
}
static inline float futrts_asin32(float x)
{
    return asin(x);
}
static inline float futrts_atan32(float x)
{
    return atan(x);
}
static inline float futrts_atan2_32(float x, float y)
{
    return atan2(x, y);
}
static inline float futrts_gamma32(float x)
{
    return tgamma(x);
}
static inline float futrts_lgamma32(float x)
{
    return lgamma(x);
}
static inline char futrts_isnan32(float x)
{
    return isnan(x);
}
static inline char futrts_isinf32(float x)
{
    return isinf(x);
}
static inline int32_t futrts_to_bits32(float x)
{
    union {
        float f;
        int32_t t;
    } p;
    
    p.f = x;
    return p.t;
}
static inline float futrts_from_bits32(int32_t x)
{
    union {
        int32_t f;
        float t;
    } p;
    
    p.f = x;
    return p.t;
}
#ifdef __OPENCL_VERSION__
static inline float fmod32(float x, float y)
{
    return fmod(x, y);
}
static inline float futrts_round32(float x)
{
    return rint(x);
}
static inline float futrts_floor32(float x)
{
    return floor(x);
}
static inline float futrts_ceil32(float x)
{
    return ceil(x);
}
static inline float futrts_lerp32(float v0, float v1, float t)
{
    return mix(v0, v1, t);
}
#else
static inline float fmod32(float x, float y)
{
    return fmodf(x, y);
}
static inline float futrts_round32(float x)
{
    return rintf(x);
}
static inline float futrts_floor32(float x)
{
    return floorf(x);
}
static inline float futrts_ceil32(float x)
{
    return ceilf(x);
}
static inline float futrts_lerp32(float v0, float v1, float t)
{
    return v0 + (v1 - v0) * t;
}
#endif
static inline double futrts_log64(double x)
{
    return log(x);
}
static inline double futrts_log2_64(double x)
{
    return log2(x);
}
static inline double futrts_log10_64(double x)
{
    return log10(x);
}
static inline double futrts_sqrt64(double x)
{
    return sqrt(x);
}
static inline double futrts_exp64(double x)
{
    return exp(x);
}
static inline double futrts_cos64(double x)
{
    return cos(x);
}
static inline double futrts_sin64(double x)
{
    return sin(x);
}
static inline double futrts_tan64(double x)
{
    return tan(x);
}
static inline double futrts_acos64(double x)
{
    return acos(x);
}
static inline double futrts_asin64(double x)
{
    return asin(x);
}
static inline double futrts_atan64(double x)
{
    return atan(x);
}
static inline double futrts_atan2_64(double x, double y)
{
    return atan2(x, y);
}
static inline double futrts_gamma64(double x)
{
    return tgamma(x);
}
static inline double futrts_lgamma64(double x)
{
    return lgamma(x);
}
static inline double futrts_round64(double x)
{
    return rint(x);
}
static inline double futrts_ceil64(double x)
{
    return ceil(x);
}
static inline double futrts_floor64(double x)
{
    return floor(x);
}
static inline char futrts_isnan64(double x)
{
    return isnan(x);
}
static inline char futrts_isinf64(double x)
{
    return isinf(x);
}
static inline int64_t futrts_to_bits64(double x)
{
    union {
        double f;
        int64_t t;
    } p;
    
    p.f = x;
    return p.t;
}
static inline double futrts_from_bits64(int64_t x)
{
    union {
        int64_t f;
        double t;
    } p;
    
    p.f = x;
    return p.t;
}
static inline float fmod64(float x, float y)
{
    return fmod(x, y);
}
#ifdef __OPENCL_VERSION__
static inline double futrts_lerp64(double v0, double v1, double t)
{
    return mix(v0, v1, t);
}
#else
static inline double futrts_lerp64(double v0, double v1, double t)
{
    return v0 + (v1 - v0) * t;
}
#endif
static int futrts_main(struct futhark_context *ctx,
                       struct memblock_device *out_mem_p_33521,
                       int32_t *out_out_arrsizze_33522,
                       int32_t *out_out_arrsizze_33523,
                       struct memblock_device input_mem_32784, int32_t n_31902,
                       int32_t m_31903)
{
    struct memblock_device out_mem_33410;
    
    out_mem_33410.references = NULL;
    
    int32_t out_arrsizze_33411;
    int32_t out_arrsizze_33412;
    int32_t fill_arg_31905 = 1 + m_31903;
    int32_t w_minus_1_31906 = m_31903 - 1;
    int32_t j_m_i_31907 = -1 - w_minus_1_31906;
    int32_t n_31908 = squot32(j_m_i_31907, -1);
    int32_t j_p_s0_t_i_31909 = 1 + w_minus_1_31906;
    int64_t binop_x_32786 = sext_i32_i64(n_31902);
    int64_t binop_y_32787 = sext_i32_i64(fill_arg_31905);
    int64_t bytes_32785 = binop_x_32786 * binop_y_32787;
    struct memblock_device mem_32788;
    
    mem_32788.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32788, bytes_32785, "mem_32788"))
        return 1;
    
    int call_ret_33524 = futrts__replicate_bool(ctx, mem_32788, n_31902 *
                                                fill_arg_31905, 0);
    
    assert(call_ret_33524 == 0);
    
    int32_t segmap_group_sizze_32553;
    
    segmap_group_sizze_32553 = ctx->sizes.mainzisegmap_group_sizze_32552;
    
    int64_t segmap_group_sizze_32554 = sext_i32_i64(segmap_group_sizze_32553);
    int64_t y_32555 = segmap_group_sizze_32554 - 1;
    int64_t x_32556 = y_32555 + binop_x_32786;
    int64_t segmap_usable_groups_64_32558 = squot64(x_32556,
                                                    segmap_group_sizze_32554);
    int32_t segmap_usable_groups_32559 =
            sext_i64_i32(segmap_usable_groups_64_32558);
    CUdeviceptr kernel_arg_33528 = mem_32788.mem;
    
    if ((((((1 && segmap_usable_groups_32559 != 0) && 1 != 0) && 1 != 0) &&
          segmap_group_sizze_32553 != 0) && 1 != 0) && 1 != 0) {
        int perm[3] = {0, 1, 2};
        
        if (1 > 1 << 16) {
            perm[1] = perm[0];
            perm[0] = 1;
        }
        if (1 > 1 << 16) {
            perm[2] = perm[0];
            perm[0] = 2;
        }
        
        size_t grid[3];
        
        grid[perm[0]] = segmap_usable_groups_32559;
        grid[perm[1]] = 1;
        grid[perm[2]] = 1;
        
        void *kernel_args_33525[] = {&n_31902, &fill_arg_31905,
                                     &kernel_arg_33528};
        int64_t time_start_33526 = 0, time_end_33527 = 0;
        
        if (ctx->debugging) {
            fprintf(stderr, "Launching %s with grid size (", "segmap_32547");
            fprintf(stderr, "%d", segmap_usable_groups_32559);
            fprintf(stderr, ", ");
            fprintf(stderr, "%d", 1);
            fprintf(stderr, ", ");
            fprintf(stderr, "%d", 1);
            fprintf(stderr, ") and block size (");
            fprintf(stderr, "%d", segmap_group_sizze_32553);
            fprintf(stderr, ", ");
            fprintf(stderr, "%d", 1);
            fprintf(stderr, ", ");
            fprintf(stderr, "%d", 1);
            fprintf(stderr, ").\n");
            time_start_33526 = get_wall_time();
        }
        CUDA_SUCCEED(cuLaunchKernel(ctx->segmap_32547, grid[0], grid[1],
                                    grid[2], segmap_group_sizze_32553, 1, 1, 0,
                                    NULL, kernel_args_33525, NULL));
        if (ctx->debugging) {
            CUDA_SUCCEED(cuCtxSynchronize());
            time_end_33527 = get_wall_time();
            fprintf(stderr, "Kernel %s runtime: %ldus\n", "segmap_32547",
                    time_end_33527 - time_start_33526);
        }
    }
    
    int32_t segmap_group_sizze_32537;
    
    segmap_group_sizze_32537 = ctx->sizes.mainzisegmap_group_sizze_32536;
    
    int64_t segmap_group_sizze_32538 = sext_i32_i64(segmap_group_sizze_32537);
    int64_t y_32539 = segmap_group_sizze_32538 - 1;
    int64_t x_32540 = y_32539 + binop_x_32786;
    int64_t segmap_usable_groups_64_32542 = squot64(x_32540,
                                                    segmap_group_sizze_32538);
    int32_t segmap_usable_groups_32543 =
            sext_i64_i32(segmap_usable_groups_64_32542);
    CUdeviceptr kernel_arg_33532 = mem_32788.mem;
    
    if ((((((1 && segmap_usable_groups_32543 != 0) && 1 != 0) && 1 != 0) &&
          segmap_group_sizze_32537 != 0) && 1 != 0) && 1 != 0) {
        int perm[3] = {0, 1, 2};
        
        if (1 > 1 << 16) {
            perm[1] = perm[0];
            perm[0] = 1;
        }
        if (1 > 1 << 16) {
            perm[2] = perm[0];
            perm[0] = 2;
        }
        
        size_t grid[3];
        
        grid[perm[0]] = segmap_usable_groups_32543;
        grid[perm[1]] = 1;
        grid[perm[2]] = 1;
        
        void *kernel_args_33529[] = {&n_31902, &m_31903, &fill_arg_31905,
                                     &kernel_arg_33532};
        int64_t time_start_33530 = 0, time_end_33531 = 0;
        
        if (ctx->debugging) {
            fprintf(stderr, "Launching %s with grid size (", "segmap_32531");
            fprintf(stderr, "%d", segmap_usable_groups_32543);
            fprintf(stderr, ", ");
            fprintf(stderr, "%d", 1);
            fprintf(stderr, ", ");
            fprintf(stderr, "%d", 1);
            fprintf(stderr, ") and block size (");
            fprintf(stderr, "%d", segmap_group_sizze_32537);
            fprintf(stderr, ", ");
            fprintf(stderr, "%d", 1);
            fprintf(stderr, ", ");
            fprintf(stderr, "%d", 1);
            fprintf(stderr, ").\n");
            time_start_33530 = get_wall_time();
        }
        CUDA_SUCCEED(cuLaunchKernel(ctx->segmap_32531, grid[0], grid[1],
                                    grid[2], segmap_group_sizze_32537, 1, 1, 0,
                                    NULL, kernel_args_33529, NULL));
        if (ctx->debugging) {
            CUDA_SUCCEED(cuCtxSynchronize());
            time_end_33531 = get_wall_time();
            fprintf(stderr, "Kernel %s runtime: %ldus\n", "segmap_32531",
                    time_end_33531 - time_start_33530);
        }
    }
    
    int64_t nest_sizze_32515 = binop_x_32786 * binop_y_32787;
    int32_t segred_group_sizze_32517;
    
    segred_group_sizze_32517 = ctx->sizes.mainzisegred_group_sizze_32516;
    
    int32_t num_groups_32519;
    int32_t max_num_groups_33423;
    
    max_num_groups_33423 = ctx->sizes.mainzisegred_num_groups_32518;
    num_groups_32519 = sext_i64_i32(smax64(1, smin64(squot64(nest_sizze_32515 +
                                                             sext_i32_i64(segred_group_sizze_32517) -
                                                             1,
                                                             sext_i32_i64(segred_group_sizze_32517)),
                                                     sext_i32_i64(max_num_groups_33423))));
    
    struct memblock_device mem_32791;
    
    mem_32791.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32791, binop_x_32786, "mem_32791"))
        return 1;
    if (slt32(fill_arg_31905 * 2, segred_group_sizze_32517)) {
        int32_t segment_sizze_nonzzero_33424 = smax32(1, fill_arg_31905);
        int32_t num_threads_33425 = num_groups_32519 * segred_group_sizze_32517;
        
        if (ctx->debugging)
            fprintf(stderr, "%s\n", "\n# SegRed-small");
        if (ctx->debugging)
            fprintf(stderr, "%s: %llu%c", "num_segments", (long long) n_31902,
                    '\n');
        if (ctx->debugging)
            fprintf(stderr, "%s: %llu%c", "segment_size",
                    (long long) fill_arg_31905, '\n');
        if (ctx->debugging)
            fprintf(stderr, "%s: %llu%c", "segments_per_group",
                    (long long) squot32(segred_group_sizze_32517,
                                        segment_sizze_nonzzero_33424), '\n');
        if (ctx->debugging)
            fprintf(stderr, "%s: %llu%c", "required_groups",
                    (long long) squot32(n_31902 +
                                        squot32(segred_group_sizze_32517,
                                                segment_sizze_nonzzero_33424) -
                                        1, squot32(segred_group_sizze_32517,
                                                   segment_sizze_nonzzero_33424)),
                    '\n');
        
        unsigned int shared_sizze_33536 = sizeof(bool) *
                     segred_group_sizze_32517;
        CUdeviceptr kernel_arg_33538 = mem_32788.mem;
        CUdeviceptr kernel_arg_33539 = mem_32791.mem;
        unsigned int shared_offset_33537 = 0;
        
        if ((((((1 && num_groups_32519 != 0) && 1 != 0) && 1 != 0) &&
              segred_group_sizze_32517 != 0) && 1 != 0) && 1 != 0) {
            int perm[3] = {0, 1, 2};
            
            if (1 > 1 << 16) {
                perm[1] = perm[0];
                perm[0] = 1;
            }
            if (1 > 1 << 16) {
                perm[2] = perm[0];
                perm[0] = 2;
            }
            
            size_t grid[3];
            
            grid[perm[0]] = num_groups_32519;
            grid[perm[1]] = 1;
            grid[perm[2]] = 1;
            
            void *kernel_args_33533[] = {&shared_offset_33537, &n_31902,
                                         &fill_arg_31905, &num_groups_32519,
                                         &kernel_arg_33538, &kernel_arg_33539,
                                         &segment_sizze_nonzzero_33424};
            int64_t time_start_33534 = 0, time_end_33535 = 0;
            
            if (ctx->debugging) {
                fprintf(stderr, "Launching %s with grid size (",
                        "segred_small_32522");
                fprintf(stderr, "%d", num_groups_32519);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ") and block size (");
                fprintf(stderr, "%d", segred_group_sizze_32517);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ").\n");
                time_start_33534 = get_wall_time();
            }
            CUDA_SUCCEED(cuLaunchKernel(ctx->segred_small_32522, grid[0],
                                        grid[1], grid[2],
                                        segred_group_sizze_32517, 1, 1, 0 +
                                        (shared_sizze_33536 + (8 -
                                                               shared_sizze_33536 %
                                                               8) % 8), NULL,
                                        kernel_args_33533, NULL));
            if (ctx->debugging) {
                CUDA_SUCCEED(cuCtxSynchronize());
                time_end_33535 = get_wall_time();
                fprintf(stderr, "Kernel %s runtime: %ldus\n",
                        "segred_small_32522", time_end_33535 -
                        time_start_33534);
            }
        }
    } else {
        int32_t vit_num_groups_33441 = squot32(num_groups_32519 + smax32(1,
                                                                         n_31902) -
                                               1, smax32(1, n_31902)) * n_31902;
        int32_t num_threads_33442 = num_groups_32519 * segred_group_sizze_32517;
        int32_t thread_per_segment_33443 = squot32(num_groups_32519 + smax32(1,
                                                                             n_31902) -
                                                   1, smax32(1, n_31902)) *
                segred_group_sizze_32517;
        
        if (ctx->debugging)
            fprintf(stderr, "%s\n", "\n# SegRed-large");
        if (ctx->debugging)
            fprintf(stderr, "%s: %llu%c", "num_segments", (long long) n_31902,
                    '\n');
        if (ctx->debugging)
            fprintf(stderr, "%s: %llu%c", "segment_size",
                    (long long) fill_arg_31905, '\n');
        if (ctx->debugging)
            fprintf(stderr, "%s: %llu%c", "virt_num_groups",
                    (long long) vit_num_groups_33441, '\n');
        if (ctx->debugging)
            fprintf(stderr, "%s: %llu%c", "num_groups",
                    (long long) num_groups_32519, '\n');
        if (ctx->debugging)
            fprintf(stderr, "%s: %llu%c", "group_size",
                    (long long) segred_group_sizze_32517, '\n');
        if (ctx->debugging)
            fprintf(stderr, "%s: %llu%c", "elems_per_thread",
                    (long long) squot32(fill_arg_31905 +
                                        segred_group_sizze_32517 *
                                        squot32(num_groups_32519 + smax32(1,
                                                                          n_31902) -
                                                1, smax32(1, n_31902)) - 1,
                                        segred_group_sizze_32517 *
                                        squot32(num_groups_32519 + smax32(1,
                                                                          n_31902) -
                                                1, smax32(1, n_31902))), '\n');
        if (ctx->debugging)
            fprintf(stderr, "%s: %llu%c", "groups_per_segment",
                    (long long) squot32(num_groups_32519 + smax32(1, n_31902) -
                                        1, smax32(1, n_31902)), '\n');
        
        struct memblock_device group_res_arr_mem_33444;
        
        group_res_arr_mem_33444.references = NULL;
        if (memblock_alloc_device(ctx, &group_res_arr_mem_33444, sizeof(bool) *
                                  (segred_group_sizze_32517 *
                                   vit_num_groups_33441),
                                  "group_res_arr_mem_33444"))
            return 1;
        
        struct memblock_device counter_mem_33446 = ctx->counter_mem_33446;
        unsigned int shared_sizze_33544 = sizeof(bool) *
                     segred_group_sizze_32517;
        unsigned int shared_sizze_33546 = sizeof(bool);
        CUdeviceptr kernel_arg_33548 = mem_32788.mem;
        CUdeviceptr kernel_arg_33549 = mem_32791.mem;
        CUdeviceptr kernel_arg_33550 = group_res_arr_mem_33444.mem;
        CUdeviceptr kernel_arg_33551 = counter_mem_33446.mem;
        unsigned int shared_offset_33545 = 0;
        unsigned int shared_offset_33547 = 0 + (shared_sizze_33544 + (8 -
                                                                      shared_sizze_33544 %
                                                                      8) % 8);
        
        if ((((((1 && num_groups_32519 != 0) && 1 != 0) && 1 != 0) &&
              segred_group_sizze_32517 != 0) && 1 != 0) && 1 != 0) {
            int perm[3] = {0, 1, 2};
            
            if (1 > 1 << 16) {
                perm[1] = perm[0];
                perm[0] = 1;
            }
            if (1 > 1 << 16) {
                perm[2] = perm[0];
                perm[0] = 2;
            }
            
            size_t grid[3];
            
            grid[perm[0]] = num_groups_32519;
            grid[perm[1]] = 1;
            grid[perm[2]] = 1;
            
            void *kernel_args_33541[] = {&shared_offset_33545,
                                         &shared_offset_33547, &n_31902,
                                         &fill_arg_31905, &num_groups_32519,
                                         &kernel_arg_33548, &kernel_arg_33549,
                                         &vit_num_groups_33441,
                                         &thread_per_segment_33443,
                                         &kernel_arg_33550, &kernel_arg_33551};
            int64_t time_start_33542 = 0, time_end_33543 = 0;
            
            if (ctx->debugging) {
                fprintf(stderr, "Launching %s with grid size (",
                        "segred_large_32522");
                fprintf(stderr, "%d", num_groups_32519);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ") and block size (");
                fprintf(stderr, "%d", segred_group_sizze_32517);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ").\n");
                time_start_33542 = get_wall_time();
            }
            CUDA_SUCCEED(cuLaunchKernel(ctx->segred_large_32522, grid[0],
                                        grid[1], grid[2],
                                        segred_group_sizze_32517, 1, 1, 0 +
                                        (shared_sizze_33544 + (8 -
                                                               shared_sizze_33544 %
                                                               8) % 8) +
                                        (shared_sizze_33546 + (8 -
                                                               shared_sizze_33546 %
                                                               8) % 8), NULL,
                                        kernel_args_33541, NULL));
            if (ctx->debugging) {
                CUDA_SUCCEED(cuCtxSynchronize());
                time_end_33543 = get_wall_time();
                fprintf(stderr, "Kernel %s runtime: %ldus\n",
                        "segred_large_32522", time_end_33543 -
                        time_start_33542);
            }
        }
        if (memblock_unref_device(ctx, &group_res_arr_mem_33444,
                                  "group_res_arr_mem_33444") != 0)
            return 1;
    }
    
    int32_t segmap_group_sizze_32315;
    
    segmap_group_sizze_32315 = ctx->sizes.mainzisegmap_group_sizze_32314;
    
    int32_t num_groups_32317;
    int32_t max_num_groups_33472;
    
    max_num_groups_33472 = ctx->sizes.mainzisegmap_num_groups_32316;
    num_groups_32317 = sext_i64_i32(smax64(1, smin64(squot64(binop_x_32786 +
                                                             sext_i32_i64(segmap_group_sizze_32315) -
                                                             1,
                                                             sext_i32_i64(segmap_group_sizze_32315)),
                                                     sext_i32_i64(max_num_groups_33472))));
    
    int32_t convop_x_32793 = n_31902 * m_31903;
    int64_t binop_x_32794 = sext_i32_i64(convop_x_32793);
    int64_t bytes_32792 = 4 * binop_x_32794;
    struct memblock_device mem_32795;
    
    mem_32795.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32795, bytes_32792, "mem_32795"))
        return 1;
    
    int call_ret_33552 = futrts__map_transpose_i32(ctx, mem_32795, 0,
                                                   input_mem_32784, 0, 1,
                                                   m_31903, n_31902, n_31902 *
                                                   m_31903, n_31902 * m_31903);
    
    assert(call_ret_33552 == 0);
    
    int32_t convop_x_32797 = n_31902 * fill_arg_31905;
    int64_t bytes_32796 = sext_i32_i64(convop_x_32797);
    struct memblock_device mem_32798;
    
    mem_32798.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32798, bytes_32796, "mem_32798"))
        return 1;
    
    int call_ret_33553 = futrts__map_transpose_bool(ctx, mem_32798, 0,
                                                    mem_32788, 0, 1,
                                                    fill_arg_31905, n_31902,
                                                    n_31902 * fill_arg_31905,
                                                    n_31902 * fill_arg_31905);
    
    assert(call_ret_33553 == 0);
    if (memblock_unref_device(ctx, &mem_32788, "mem_32788") != 0)
        return 1;
    
    struct memblock_device mem_32861;
    
    mem_32861.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32861, bytes_32792, "mem_32861"))
        return 1;
    
    int64_t binop_x_32802 = sext_i32_i64(m_31903);
    int64_t bytes_32801 = 4 * binop_x_32802;
    int64_t binop_x_32843 = sext_i32_i64(n_31908);
    int64_t bytes_32842 = 4 * binop_x_32843;
    int64_t binop_x_32890 = sext_i32_i64(m_31903);
    int64_t double_buffer_sizze_32891 = 4 * binop_x_32890;
    int64_t double_buffer_sizze_32892 = sext_i32_i64(fill_arg_31905);
    int32_t num_threads_32933 = segmap_group_sizze_32315 * num_groups_32317;
    int64_t num_threads64_32934 = sext_i32_i64(num_threads_32933);
    int64_t total_sizze_32935 = bytes_32801 * num_threads64_32934;
    struct memblock_device mem_32810;
    
    mem_32810.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32810, total_sizze_32935, "mem_32810"))
        return 1;
    
    int64_t total_sizze_32936 = binop_x_32802 * num_threads64_32934;
    struct memblock_device mem_32812;
    
    mem_32812.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32812, total_sizze_32936, "mem_32812"))
        return 1;
    
    int64_t total_sizze_32937 = bytes_32801 * num_threads64_32934;
    struct memblock_device mem_32815;
    
    mem_32815.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32815, total_sizze_32937, "mem_32815"))
        return 1;
    
    int64_t total_sizze_32938 = binop_x_32802 * num_threads64_32934;
    struct memblock_device mem_32817;
    
    mem_32817.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32817, total_sizze_32938, "mem_32817"))
        return 1;
    
    int64_t total_sizze_32939 = binop_x_32802 * num_threads64_32934;
    struct memblock_device mem_32819;
    
    mem_32819.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32819, total_sizze_32939, "mem_32819"))
        return 1;
    
    int64_t total_sizze_32940 = bytes_32801 * num_threads64_32934;
    struct memblock_device mem_32822;
    
    mem_32822.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32822, total_sizze_32940, "mem_32822"))
        return 1;
    
    int64_t total_sizze_32941 = binop_x_32802 * num_threads64_32934;
    struct memblock_device mem_32824;
    
    mem_32824.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32824, total_sizze_32941, "mem_32824"))
        return 1;
    
    int64_t total_sizze_32942 = bytes_32801 * num_threads64_32934;
    struct memblock_device mem_32827;
    
    mem_32827.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32827, total_sizze_32942, "mem_32827"))
        return 1;
    
    int64_t total_sizze_32943 = bytes_32801 * num_threads64_32934;
    struct memblock_device mem_32830;
    
    mem_32830.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32830, total_sizze_32943, "mem_32830"))
        return 1;
    
    int64_t total_sizze_32944 = bytes_32801 * num_threads64_32934;
    struct memblock_device mem_32833;
    
    mem_32833.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32833, total_sizze_32944, "mem_32833"))
        return 1;
    
    int64_t total_sizze_32945 = bytes_32801 * num_threads64_32934;
    struct memblock_device mem_32836;
    
    mem_32836.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32836, total_sizze_32945, "mem_32836"))
        return 1;
    
    int64_t total_sizze_32946 = bytes_32801 * num_threads64_32934;
    struct memblock_device mem_32839;
    
    mem_32839.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32839, total_sizze_32946, "mem_32839"))
        return 1;
    
    int64_t total_sizze_32947 = binop_x_32802 * num_threads64_32934;
    struct memblock_device mem_32841;
    
    mem_32841.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32841, total_sizze_32947, "mem_32841"))
        return 1;
    
    int64_t total_sizze_32948 = bytes_32842 * num_threads64_32934;
    struct memblock_device mem_32844;
    
    mem_32844.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32844, total_sizze_32948, "mem_32844"))
        return 1;
    
    int64_t total_sizze_32949 = bytes_32801 * num_threads64_32934;
    struct memblock_device mem_32847;
    
    mem_32847.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32847, total_sizze_32949, "mem_32847"))
        return 1;
    
    int64_t total_sizze_32950 = bytes_32801 * num_threads64_32934;
    struct memblock_device mem_32850;
    
    mem_32850.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32850, total_sizze_32950, "mem_32850"))
        return 1;
    
    int64_t total_sizze_32951 = bytes_32801 * num_threads64_32934;
    struct memblock_device mem_32853;
    
    mem_32853.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32853, total_sizze_32951, "mem_32853"))
        return 1;
    
    int64_t total_sizze_32952 = binop_y_32787 * num_threads64_32934;
    struct memblock_device mem_32855;
    
    mem_32855.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32855, total_sizze_32952, "mem_32855"))
        return 1;
    
    int64_t total_sizze_32953 = double_buffer_sizze_32891 * num_threads64_32934;
    struct memblock_device double_buffer_mem_32886;
    
    double_buffer_mem_32886.references = NULL;
    if (memblock_alloc_device(ctx, &double_buffer_mem_32886, total_sizze_32953,
                              "double_buffer_mem_32886"))
        return 1;
    
    int64_t total_sizze_32954 = double_buffer_sizze_32892 * num_threads64_32934;
    struct memblock_device double_buffer_mem_32887;
    
    double_buffer_mem_32887.references = NULL;
    if (memblock_alloc_device(ctx, &double_buffer_mem_32887, total_sizze_32954,
                              "double_buffer_mem_32887"))
        return 1;
    
    CUdeviceptr kernel_arg_33557 = mem_32791.mem;
    CUdeviceptr kernel_arg_33558 = mem_32795.mem;
    CUdeviceptr kernel_arg_33559 = mem_32798.mem;
    CUdeviceptr kernel_arg_33560 = mem_32810.mem;
    CUdeviceptr kernel_arg_33561 = mem_32812.mem;
    CUdeviceptr kernel_arg_33562 = mem_32815.mem;
    CUdeviceptr kernel_arg_33563 = mem_32817.mem;
    CUdeviceptr kernel_arg_33564 = mem_32819.mem;
    CUdeviceptr kernel_arg_33565 = mem_32822.mem;
    CUdeviceptr kernel_arg_33566 = mem_32824.mem;
    CUdeviceptr kernel_arg_33567 = mem_32827.mem;
    CUdeviceptr kernel_arg_33568 = mem_32830.mem;
    CUdeviceptr kernel_arg_33569 = mem_32833.mem;
    CUdeviceptr kernel_arg_33570 = mem_32836.mem;
    CUdeviceptr kernel_arg_33571 = mem_32839.mem;
    CUdeviceptr kernel_arg_33572 = mem_32841.mem;
    CUdeviceptr kernel_arg_33573 = mem_32844.mem;
    CUdeviceptr kernel_arg_33574 = mem_32847.mem;
    CUdeviceptr kernel_arg_33575 = mem_32850.mem;
    CUdeviceptr kernel_arg_33576 = mem_32853.mem;
    CUdeviceptr kernel_arg_33577 = mem_32855.mem;
    CUdeviceptr kernel_arg_33578 = mem_32861.mem;
    CUdeviceptr kernel_arg_33579 = double_buffer_mem_32886.mem;
    CUdeviceptr kernel_arg_33580 = double_buffer_mem_32887.mem;
    
    if ((((((1 && num_groups_32317 != 0) && 1 != 0) && 1 != 0) &&
          segmap_group_sizze_32315 != 0) && 1 != 0) && 1 != 0) {
        int perm[3] = {0, 1, 2};
        
        if (1 > 1 << 16) {
            perm[1] = perm[0];
            perm[0] = 1;
        }
        if (1 > 1 << 16) {
            perm[2] = perm[0];
            perm[0] = 2;
        }
        
        size_t grid[3];
        
        grid[perm[0]] = num_groups_32317;
        grid[perm[1]] = 1;
        grid[perm[2]] = 1;
        
        void *kernel_args_33554[] = {&n_31902, &m_31903, &fill_arg_31905,
                                     &w_minus_1_31906, &n_31908,
                                     &j_p_s0_t_i_31909, &num_groups_32317,
                                     &kernel_arg_33557, &kernel_arg_33558,
                                     &kernel_arg_33559, &kernel_arg_33560,
                                     &kernel_arg_33561, &kernel_arg_33562,
                                     &kernel_arg_33563, &kernel_arg_33564,
                                     &kernel_arg_33565, &kernel_arg_33566,
                                     &kernel_arg_33567, &kernel_arg_33568,
                                     &kernel_arg_33569, &kernel_arg_33570,
                                     &kernel_arg_33571, &kernel_arg_33572,
                                     &kernel_arg_33573, &kernel_arg_33574,
                                     &kernel_arg_33575, &kernel_arg_33576,
                                     &kernel_arg_33577, &kernel_arg_33578,
                                     &kernel_arg_33579, &kernel_arg_33580};
        int64_t time_start_33555 = 0, time_end_33556 = 0;
        
        if (ctx->debugging) {
            fprintf(stderr, "Launching %s with grid size (", "segmap_32311");
            fprintf(stderr, "%d", num_groups_32317);
            fprintf(stderr, ", ");
            fprintf(stderr, "%d", 1);
            fprintf(stderr, ", ");
            fprintf(stderr, "%d", 1);
            fprintf(stderr, ") and block size (");
            fprintf(stderr, "%d", segmap_group_sizze_32315);
            fprintf(stderr, ", ");
            fprintf(stderr, "%d", 1);
            fprintf(stderr, ", ");
            fprintf(stderr, "%d", 1);
            fprintf(stderr, ").\n");
            time_start_33555 = get_wall_time();
        }
        CUDA_SUCCEED(cuLaunchKernel(ctx->segmap_32311, grid[0], grid[1],
                                    grid[2], segmap_group_sizze_32315, 1, 1, 0,
                                    NULL, kernel_args_33554, NULL));
        if (ctx->debugging) {
            CUDA_SUCCEED(cuCtxSynchronize());
            time_end_33556 = get_wall_time();
            fprintf(stderr, "Kernel %s runtime: %ldus\n", "segmap_32311",
                    time_end_33556 - time_start_33555);
        }
    }
    if (memblock_unref_device(ctx, &mem_32791, "mem_32791") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32795, "mem_32795") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32798, "mem_32798") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32810, "mem_32810") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32812, "mem_32812") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32815, "mem_32815") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32817, "mem_32817") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32819, "mem_32819") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32822, "mem_32822") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32824, "mem_32824") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32827, "mem_32827") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32830, "mem_32830") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32833, "mem_32833") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32836, "mem_32836") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32839, "mem_32839") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32841, "mem_32841") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32844, "mem_32844") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32847, "mem_32847") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32850, "mem_32850") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32853, "mem_32853") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32855, "mem_32855") != 0)
        return 1;
    if (memblock_unref_device(ctx, &double_buffer_mem_32886,
                              "double_buffer_mem_32886") != 0)
        return 1;
    if (memblock_unref_device(ctx, &double_buffer_mem_32887,
                              "double_buffer_mem_32887") != 0)
        return 1;
    
    int64_t binop_x_32865 = binop_x_32786 * binop_x_32802;
    int64_t bytes_32862 = 4 * binop_x_32865;
    struct memblock_device mem_32866;
    
    mem_32866.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32866, bytes_32862, "mem_32866"))
        return 1;
    
    int call_ret_33581 = futrts__map_transpose_i32(ctx, mem_32866, 0, mem_32861,
                                                   0, 1, n_31902, m_31903,
                                                   n_31902 * m_31903, n_31902 *
                                                   m_31903);
    
    assert(call_ret_33581 == 0);
    if (memblock_unref_device(ctx, &mem_32861, "mem_32861") != 0)
        return 1;
    out_arrsizze_33411 = n_31902;
    out_arrsizze_33412 = m_31903;
    if (memblock_set_device(ctx, &out_mem_33410, &mem_32866, "mem_32866") != 0)
        return 1;
    (*out_mem_p_33521).references = NULL;
    if (memblock_set_device(ctx, &*out_mem_p_33521, &out_mem_33410,
                            "out_mem_33410") != 0)
        return 1;
    *out_out_arrsizze_33522 = out_arrsizze_33411;
    *out_out_arrsizze_33523 = out_arrsizze_33412;
    if (memblock_unref_device(ctx, &mem_32866, "mem_32866") != 0)
        return 1;
    if (memblock_unref_device(ctx, &double_buffer_mem_32887,
                              "double_buffer_mem_32887") != 0)
        return 1;
    if (memblock_unref_device(ctx, &double_buffer_mem_32886,
                              "double_buffer_mem_32886") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32855, "mem_32855") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32853, "mem_32853") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32850, "mem_32850") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32847, "mem_32847") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32844, "mem_32844") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32841, "mem_32841") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32839, "mem_32839") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32836, "mem_32836") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32833, "mem_32833") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32830, "mem_32830") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32827, "mem_32827") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32824, "mem_32824") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32822, "mem_32822") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32819, "mem_32819") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32817, "mem_32817") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32815, "mem_32815") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32812, "mem_32812") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32810, "mem_32810") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32861, "mem_32861") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32798, "mem_32798") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32795, "mem_32795") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32791, "mem_32791") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32788, "mem_32788") != 0)
        return 1;
    if (memblock_unref_device(ctx, &out_mem_33410, "out_mem_33410") != 0)
        return 1;
    return 0;
}
static int futrts__map_transpose_bool(struct futhark_context *ctx,
                                      struct memblock_device destmem_0,
                                      int32_t destoffset_1,
                                      struct memblock_device srcmem_2,
                                      int32_t srcoffset_3, int32_t num_arrays_4,
                                      int32_t x_elems_5, int32_t y_elems_6,
                                      int32_t in_elems_7, int32_t out_elems_8)
{
    if (!(num_arrays_4 == 0 || (x_elems_5 == 0 || y_elems_6 == 0))) {
        int32_t muly_10 = squot32(16, x_elems_5);
        int32_t mulx_9 = squot32(16, y_elems_6);
        
        if (in_elems_7 == out_elems_8 && ((num_arrays_4 == 1 || x_elems_5 *
                                           y_elems_6 == in_elems_7) &&
                                          (x_elems_5 == 1 || y_elems_6 == 1))) {
            CUDA_SUCCEED(cuMemcpy(destmem_0.mem + destoffset_1, srcmem_2.mem +
                                  srcoffset_3, in_elems_7 * sizeof(bool)));
        } else {
            if (sle32(x_elems_5, 8) && slt32(16, y_elems_6)) {
                unsigned int shared_sizze_33585 = 272;
                CUdeviceptr kernel_arg_33587 = destmem_0.mem;
                CUdeviceptr kernel_arg_33588 = srcmem_2.mem;
                unsigned int shared_offset_33586 = 0;
                
                if ((((((1 && squot32(x_elems_5 + 16 - 1, 16) != 0) &&
                        squot32(squot32(y_elems_6 + muly_10 - 1, muly_10) + 16 -
                                1, 16) != 0) && num_arrays_4 != 0) && 16 !=
                      0) && 16 != 0) && 1 != 0) {
                    int perm[3] = {0, 1, 2};
                    
                    if (squot32(squot32(y_elems_6 + muly_10 - 1, muly_10) + 16 -
                                1, 16) > 1 << 16) {
                        perm[1] = perm[0];
                        perm[0] = 1;
                    }
                    if (num_arrays_4 > 1 << 16) {
                        perm[2] = perm[0];
                        perm[0] = 2;
                    }
                    
                    size_t grid[3];
                    
                    grid[perm[0]] = squot32(x_elems_5 + 16 - 1, 16);
                    grid[perm[1]] = squot32(squot32(y_elems_6 + muly_10 - 1,
                                                    muly_10) + 16 - 1, 16);
                    grid[perm[2]] = num_arrays_4;
                    
                    void *kernel_args_33582[] = {&perm[0], &perm[1], &perm[2],
                                                 &shared_offset_33586,
                                                 &destoffset_1, &srcoffset_3,
                                                 &num_arrays_4, &x_elems_5,
                                                 &y_elems_6, &in_elems_7,
                                                 &out_elems_8, &mulx_9,
                                                 &muly_10, &kernel_arg_33587,
                                                 &kernel_arg_33588};
                    int64_t time_start_33583 = 0, time_end_33584 = 0;
                    
                    if (ctx->debugging) {
                        fprintf(stderr, "Launching %s with grid size (",
                                "map_transpose_bool_low_width");
                        fprintf(stderr, "%d", squot32(x_elems_5 + 16 - 1, 16));
                        fprintf(stderr, ", ");
                        fprintf(stderr, "%d", squot32(squot32(y_elems_6 +
                                                              muly_10 - 1,
                                                              muly_10) + 16 - 1,
                                                      16));
                        fprintf(stderr, ", ");
                        fprintf(stderr, "%d", num_arrays_4);
                        fprintf(stderr, ") and block size (");
                        fprintf(stderr, "%d", 16);
                        fprintf(stderr, ", ");
                        fprintf(stderr, "%d", 16);
                        fprintf(stderr, ", ");
                        fprintf(stderr, "%d", 1);
                        fprintf(stderr, ").\n");
                        time_start_33583 = get_wall_time();
                    }
                    CUDA_SUCCEED(cuLaunchKernel(ctx->map_transpose_bool_low_width,
                                                grid[0], grid[1], grid[2], 16,
                                                16, 1, 0 + (shared_sizze_33585 +
                                                            (8 -
                                                             shared_sizze_33585 %
                                                             8) % 8), NULL,
                                                kernel_args_33582, NULL));
                    if (ctx->debugging) {
                        CUDA_SUCCEED(cuCtxSynchronize());
                        time_end_33584 = get_wall_time();
                        fprintf(stderr, "Kernel %s runtime: %ldus\n",
                                "map_transpose_bool_low_width", time_end_33584 -
                                time_start_33583);
                    }
                }
            } else {
                if (sle32(y_elems_6, 8) && slt32(16, x_elems_5)) {
                    unsigned int shared_sizze_33592 = 272;
                    CUdeviceptr kernel_arg_33594 = destmem_0.mem;
                    CUdeviceptr kernel_arg_33595 = srcmem_2.mem;
                    unsigned int shared_offset_33593 = 0;
                    
                    if ((((((1 && squot32(squot32(x_elems_5 + mulx_9 - 1,
                                                  mulx_9) + 16 - 1, 16) != 0) &&
                            squot32(y_elems_6 + 16 - 1, 16) != 0) &&
                           num_arrays_4 != 0) && 16 != 0) && 16 != 0) && 1 !=
                        0) {
                        int perm[3] = {0, 1, 2};
                        
                        if (squot32(y_elems_6 + 16 - 1, 16) > 1 << 16) {
                            perm[1] = perm[0];
                            perm[0] = 1;
                        }
                        if (num_arrays_4 > 1 << 16) {
                            perm[2] = perm[0];
                            perm[0] = 2;
                        }
                        
                        size_t grid[3];
                        
                        grid[perm[0]] = squot32(squot32(x_elems_5 + mulx_9 - 1,
                                                        mulx_9) + 16 - 1, 16);
                        grid[perm[1]] = squot32(y_elems_6 + 16 - 1, 16);
                        grid[perm[2]] = num_arrays_4;
                        
                        void *kernel_args_33589[] = {&perm[0], &perm[1],
                                                     &perm[2],
                                                     &shared_offset_33593,
                                                     &destoffset_1,
                                                     &srcoffset_3,
                                                     &num_arrays_4, &x_elems_5,
                                                     &y_elems_6, &in_elems_7,
                                                     &out_elems_8, &mulx_9,
                                                     &muly_10,
                                                     &kernel_arg_33594,
                                                     &kernel_arg_33595};
                        int64_t time_start_33590 = 0, time_end_33591 = 0;
                        
                        if (ctx->debugging) {
                            fprintf(stderr, "Launching %s with grid size (",
                                    "map_transpose_bool_low_height");
                            fprintf(stderr, "%d", squot32(squot32(x_elems_5 +
                                                                  mulx_9 - 1,
                                                                  mulx_9) + 16 -
                                                          1, 16));
                            fprintf(stderr, ", ");
                            fprintf(stderr, "%d", squot32(y_elems_6 + 16 - 1,
                                                          16));
                            fprintf(stderr, ", ");
                            fprintf(stderr, "%d", num_arrays_4);
                            fprintf(stderr, ") and block size (");
                            fprintf(stderr, "%d", 16);
                            fprintf(stderr, ", ");
                            fprintf(stderr, "%d", 16);
                            fprintf(stderr, ", ");
                            fprintf(stderr, "%d", 1);
                            fprintf(stderr, ").\n");
                            time_start_33590 = get_wall_time();
                        }
                        CUDA_SUCCEED(cuLaunchKernel(ctx->map_transpose_bool_low_height,
                                                    grid[0], grid[1], grid[2],
                                                    16, 16, 1, 0 +
                                                    (shared_sizze_33592 + (8 -
                                                                           shared_sizze_33592 %
                                                                           8) %
                                                     8), NULL,
                                                    kernel_args_33589, NULL));
                        if (ctx->debugging) {
                            CUDA_SUCCEED(cuCtxSynchronize());
                            time_end_33591 = get_wall_time();
                            fprintf(stderr, "Kernel %s runtime: %ldus\n",
                                    "map_transpose_bool_low_height",
                                    time_end_33591 - time_start_33590);
                        }
                    }
                } else {
                    if (sle32(x_elems_5, 8) && sle32(y_elems_6, 8)) {
                        unsigned int shared_sizze_33599 = 1;
                        CUdeviceptr kernel_arg_33601 = destmem_0.mem;
                        CUdeviceptr kernel_arg_33602 = srcmem_2.mem;
                        unsigned int shared_offset_33600 = 0;
                        
                        if ((((((1 && squot32(num_arrays_4 * x_elems_5 *
                                              y_elems_6 + 256 - 1, 256) != 0) &&
                                1 != 0) && 1 != 0) && 256 != 0) && 1 != 0) &&
                            1 != 0) {
                            int perm[3] = {0, 1, 2};
                            
                            if (1 > 1 << 16) {
                                perm[1] = perm[0];
                                perm[0] = 1;
                            }
                            if (1 > 1 << 16) {
                                perm[2] = perm[0];
                                perm[0] = 2;
                            }
                            
                            size_t grid[3];
                            
                            grid[perm[0]] = squot32(num_arrays_4 * x_elems_5 *
                                                    y_elems_6 + 256 - 1, 256);
                            grid[perm[1]] = 1;
                            grid[perm[2]] = 1;
                            
                            void *kernel_args_33596[] = {&shared_offset_33600,
                                                         &destoffset_1,
                                                         &srcoffset_3,
                                                         &num_arrays_4,
                                                         &x_elems_5, &y_elems_6,
                                                         &in_elems_7,
                                                         &out_elems_8, &mulx_9,
                                                         &muly_10,
                                                         &kernel_arg_33601,
                                                         &kernel_arg_33602};
                            int64_t time_start_33597 = 0, time_end_33598 = 0;
                            
                            if (ctx->debugging) {
                                fprintf(stderr, "Launching %s with grid size (",
                                        "map_transpose_bool_small");
                                fprintf(stderr, "%d", squot32(num_arrays_4 *
                                                              x_elems_5 *
                                                              y_elems_6 + 256 -
                                                              1, 256));
                                fprintf(stderr, ", ");
                                fprintf(stderr, "%d", 1);
                                fprintf(stderr, ", ");
                                fprintf(stderr, "%d", 1);
                                fprintf(stderr, ") and block size (");
                                fprintf(stderr, "%d", 256);
                                fprintf(stderr, ", ");
                                fprintf(stderr, "%d", 1);
                                fprintf(stderr, ", ");
                                fprintf(stderr, "%d", 1);
                                fprintf(stderr, ").\n");
                                time_start_33597 = get_wall_time();
                            }
                            CUDA_SUCCEED(cuLaunchKernel(ctx->map_transpose_bool_small,
                                                        grid[0], grid[1],
                                                        grid[2], 256, 1, 1, 0 +
                                                        (shared_sizze_33599 +
                                                         (8 -
                                                          shared_sizze_33599 %
                                                          8) % 8), NULL,
                                                        kernel_args_33596,
                                                        NULL));
                            if (ctx->debugging) {
                                CUDA_SUCCEED(cuCtxSynchronize());
                                time_end_33598 = get_wall_time();
                                fprintf(stderr, "Kernel %s runtime: %ldus\n",
                                        "map_transpose_bool_small",
                                        time_end_33598 - time_start_33597);
                            }
                        }
                    } else {
                        unsigned int shared_sizze_33606 = 1056;
                        CUdeviceptr kernel_arg_33608 = destmem_0.mem;
                        CUdeviceptr kernel_arg_33609 = srcmem_2.mem;
                        unsigned int shared_offset_33607 = 0;
                        
                        if ((((((1 && squot32(x_elems_5 + 32 - 1, 32) != 0) &&
                                squot32(y_elems_6 + 32 - 1, 32) != 0) &&
                               num_arrays_4 != 0) && 32 != 0) && 8 != 0) && 1 !=
                            0) {
                            int perm[3] = {0, 1, 2};
                            
                            if (squot32(y_elems_6 + 32 - 1, 32) > 1 << 16) {
                                perm[1] = perm[0];
                                perm[0] = 1;
                            }
                            if (num_arrays_4 > 1 << 16) {
                                perm[2] = perm[0];
                                perm[0] = 2;
                            }
                            
                            size_t grid[3];
                            
                            grid[perm[0]] = squot32(x_elems_5 + 32 - 1, 32);
                            grid[perm[1]] = squot32(y_elems_6 + 32 - 1, 32);
                            grid[perm[2]] = num_arrays_4;
                            
                            void *kernel_args_33603[] = {&perm[0], &perm[1],
                                                         &perm[2],
                                                         &shared_offset_33607,
                                                         &destoffset_1,
                                                         &srcoffset_3,
                                                         &num_arrays_4,
                                                         &x_elems_5, &y_elems_6,
                                                         &in_elems_7,
                                                         &out_elems_8, &mulx_9,
                                                         &muly_10,
                                                         &kernel_arg_33608,
                                                         &kernel_arg_33609};
                            int64_t time_start_33604 = 0, time_end_33605 = 0;
                            
                            if (ctx->debugging) {
                                fprintf(stderr, "Launching %s with grid size (",
                                        "map_transpose_bool");
                                fprintf(stderr, "%d", squot32(x_elems_5 + 32 -
                                                              1, 32));
                                fprintf(stderr, ", ");
                                fprintf(stderr, "%d", squot32(y_elems_6 + 32 -
                                                              1, 32));
                                fprintf(stderr, ", ");
                                fprintf(stderr, "%d", num_arrays_4);
                                fprintf(stderr, ") and block size (");
                                fprintf(stderr, "%d", 32);
                                fprintf(stderr, ", ");
                                fprintf(stderr, "%d", 8);
                                fprintf(stderr, ", ");
                                fprintf(stderr, "%d", 1);
                                fprintf(stderr, ").\n");
                                time_start_33604 = get_wall_time();
                            }
                            CUDA_SUCCEED(cuLaunchKernel(ctx->map_transpose_bool,
                                                        grid[0], grid[1],
                                                        grid[2], 32, 8, 1, 0 +
                                                        (shared_sizze_33606 +
                                                         (8 -
                                                          shared_sizze_33606 %
                                                          8) % 8), NULL,
                                                        kernel_args_33603,
                                                        NULL));
                            if (ctx->debugging) {
                                CUDA_SUCCEED(cuCtxSynchronize());
                                time_end_33605 = get_wall_time();
                                fprintf(stderr, "Kernel %s runtime: %ldus\n",
                                        "map_transpose_bool", time_end_33605 -
                                        time_start_33604);
                            }
                        }
                    }
                }
            }
        }
    }
    return 0;
}
static int futrts__map_transpose_i32(struct futhark_context *ctx,
                                     struct memblock_device destmem_0,
                                     int32_t destoffset_1,
                                     struct memblock_device srcmem_2,
                                     int32_t srcoffset_3, int32_t num_arrays_4,
                                     int32_t x_elems_5, int32_t y_elems_6,
                                     int32_t in_elems_7, int32_t out_elems_8)
{
    if (!(num_arrays_4 == 0 || (x_elems_5 == 0 || y_elems_6 == 0))) {
        int32_t muly_10 = squot32(16, x_elems_5);
        int32_t mulx_9 = squot32(16, y_elems_6);
        
        if (in_elems_7 == out_elems_8 && ((num_arrays_4 == 1 || x_elems_5 *
                                           y_elems_6 == in_elems_7) &&
                                          (x_elems_5 == 1 || y_elems_6 == 1))) {
            CUDA_SUCCEED(cuMemcpy(destmem_0.mem + destoffset_1, srcmem_2.mem +
                                  srcoffset_3, in_elems_7 * sizeof(int32_t)));
        } else {
            if (sle32(x_elems_5, 8) && slt32(16, y_elems_6)) {
                unsigned int shared_sizze_33613 = 1088;
                CUdeviceptr kernel_arg_33615 = destmem_0.mem;
                CUdeviceptr kernel_arg_33616 = srcmem_2.mem;
                unsigned int shared_offset_33614 = 0;
                
                if ((((((1 && squot32(x_elems_5 + 16 - 1, 16) != 0) &&
                        squot32(squot32(y_elems_6 + muly_10 - 1, muly_10) + 16 -
                                1, 16) != 0) && num_arrays_4 != 0) && 16 !=
                      0) && 16 != 0) && 1 != 0) {
                    int perm[3] = {0, 1, 2};
                    
                    if (squot32(squot32(y_elems_6 + muly_10 - 1, muly_10) + 16 -
                                1, 16) > 1 << 16) {
                        perm[1] = perm[0];
                        perm[0] = 1;
                    }
                    if (num_arrays_4 > 1 << 16) {
                        perm[2] = perm[0];
                        perm[0] = 2;
                    }
                    
                    size_t grid[3];
                    
                    grid[perm[0]] = squot32(x_elems_5 + 16 - 1, 16);
                    grid[perm[1]] = squot32(squot32(y_elems_6 + muly_10 - 1,
                                                    muly_10) + 16 - 1, 16);
                    grid[perm[2]] = num_arrays_4;
                    
                    void *kernel_args_33610[] = {&perm[0], &perm[1], &perm[2],
                                                 &shared_offset_33614,
                                                 &destoffset_1, &srcoffset_3,
                                                 &num_arrays_4, &x_elems_5,
                                                 &y_elems_6, &in_elems_7,
                                                 &out_elems_8, &mulx_9,
                                                 &muly_10, &kernel_arg_33615,
                                                 &kernel_arg_33616};
                    int64_t time_start_33611 = 0, time_end_33612 = 0;
                    
                    if (ctx->debugging) {
                        fprintf(stderr, "Launching %s with grid size (",
                                "map_transpose_i32_low_width");
                        fprintf(stderr, "%d", squot32(x_elems_5 + 16 - 1, 16));
                        fprintf(stderr, ", ");
                        fprintf(stderr, "%d", squot32(squot32(y_elems_6 +
                                                              muly_10 - 1,
                                                              muly_10) + 16 - 1,
                                                      16));
                        fprintf(stderr, ", ");
                        fprintf(stderr, "%d", num_arrays_4);
                        fprintf(stderr, ") and block size (");
                        fprintf(stderr, "%d", 16);
                        fprintf(stderr, ", ");
                        fprintf(stderr, "%d", 16);
                        fprintf(stderr, ", ");
                        fprintf(stderr, "%d", 1);
                        fprintf(stderr, ").\n");
                        time_start_33611 = get_wall_time();
                    }
                    CUDA_SUCCEED(cuLaunchKernel(ctx->map_transpose_i32_low_width,
                                                grid[0], grid[1], grid[2], 16,
                                                16, 1, 0 + (shared_sizze_33613 +
                                                            (8 -
                                                             shared_sizze_33613 %
                                                             8) % 8), NULL,
                                                kernel_args_33610, NULL));
                    if (ctx->debugging) {
                        CUDA_SUCCEED(cuCtxSynchronize());
                        time_end_33612 = get_wall_time();
                        fprintf(stderr, "Kernel %s runtime: %ldus\n",
                                "map_transpose_i32_low_width", time_end_33612 -
                                time_start_33611);
                    }
                }
            } else {
                if (sle32(y_elems_6, 8) && slt32(16, x_elems_5)) {
                    unsigned int shared_sizze_33620 = 1088;
                    CUdeviceptr kernel_arg_33622 = destmem_0.mem;
                    CUdeviceptr kernel_arg_33623 = srcmem_2.mem;
                    unsigned int shared_offset_33621 = 0;
                    
                    if ((((((1 && squot32(squot32(x_elems_5 + mulx_9 - 1,
                                                  mulx_9) + 16 - 1, 16) != 0) &&
                            squot32(y_elems_6 + 16 - 1, 16) != 0) &&
                           num_arrays_4 != 0) && 16 != 0) && 16 != 0) && 1 !=
                        0) {
                        int perm[3] = {0, 1, 2};
                        
                        if (squot32(y_elems_6 + 16 - 1, 16) > 1 << 16) {
                            perm[1] = perm[0];
                            perm[0] = 1;
                        }
                        if (num_arrays_4 > 1 << 16) {
                            perm[2] = perm[0];
                            perm[0] = 2;
                        }
                        
                        size_t grid[3];
                        
                        grid[perm[0]] = squot32(squot32(x_elems_5 + mulx_9 - 1,
                                                        mulx_9) + 16 - 1, 16);
                        grid[perm[1]] = squot32(y_elems_6 + 16 - 1, 16);
                        grid[perm[2]] = num_arrays_4;
                        
                        void *kernel_args_33617[] = {&perm[0], &perm[1],
                                                     &perm[2],
                                                     &shared_offset_33621,
                                                     &destoffset_1,
                                                     &srcoffset_3,
                                                     &num_arrays_4, &x_elems_5,
                                                     &y_elems_6, &in_elems_7,
                                                     &out_elems_8, &mulx_9,
                                                     &muly_10,
                                                     &kernel_arg_33622,
                                                     &kernel_arg_33623};
                        int64_t time_start_33618 = 0, time_end_33619 = 0;
                        
                        if (ctx->debugging) {
                            fprintf(stderr, "Launching %s with grid size (",
                                    "map_transpose_i32_low_height");
                            fprintf(stderr, "%d", squot32(squot32(x_elems_5 +
                                                                  mulx_9 - 1,
                                                                  mulx_9) + 16 -
                                                          1, 16));
                            fprintf(stderr, ", ");
                            fprintf(stderr, "%d", squot32(y_elems_6 + 16 - 1,
                                                          16));
                            fprintf(stderr, ", ");
                            fprintf(stderr, "%d", num_arrays_4);
                            fprintf(stderr, ") and block size (");
                            fprintf(stderr, "%d", 16);
                            fprintf(stderr, ", ");
                            fprintf(stderr, "%d", 16);
                            fprintf(stderr, ", ");
                            fprintf(stderr, "%d", 1);
                            fprintf(stderr, ").\n");
                            time_start_33618 = get_wall_time();
                        }
                        CUDA_SUCCEED(cuLaunchKernel(ctx->map_transpose_i32_low_height,
                                                    grid[0], grid[1], grid[2],
                                                    16, 16, 1, 0 +
                                                    (shared_sizze_33620 + (8 -
                                                                           shared_sizze_33620 %
                                                                           8) %
                                                     8), NULL,
                                                    kernel_args_33617, NULL));
                        if (ctx->debugging) {
                            CUDA_SUCCEED(cuCtxSynchronize());
                            time_end_33619 = get_wall_time();
                            fprintf(stderr, "Kernel %s runtime: %ldus\n",
                                    "map_transpose_i32_low_height",
                                    time_end_33619 - time_start_33618);
                        }
                    }
                } else {
                    if (sle32(x_elems_5, 8) && sle32(y_elems_6, 8)) {
                        unsigned int shared_sizze_33627 = 1;
                        CUdeviceptr kernel_arg_33629 = destmem_0.mem;
                        CUdeviceptr kernel_arg_33630 = srcmem_2.mem;
                        unsigned int shared_offset_33628 = 0;
                        
                        if ((((((1 && squot32(num_arrays_4 * x_elems_5 *
                                              y_elems_6 + 256 - 1, 256) != 0) &&
                                1 != 0) && 1 != 0) && 256 != 0) && 1 != 0) &&
                            1 != 0) {
                            int perm[3] = {0, 1, 2};
                            
                            if (1 > 1 << 16) {
                                perm[1] = perm[0];
                                perm[0] = 1;
                            }
                            if (1 > 1 << 16) {
                                perm[2] = perm[0];
                                perm[0] = 2;
                            }
                            
                            size_t grid[3];
                            
                            grid[perm[0]] = squot32(num_arrays_4 * x_elems_5 *
                                                    y_elems_6 + 256 - 1, 256);
                            grid[perm[1]] = 1;
                            grid[perm[2]] = 1;
                            
                            void *kernel_args_33624[] = {&shared_offset_33628,
                                                         &destoffset_1,
                                                         &srcoffset_3,
                                                         &num_arrays_4,
                                                         &x_elems_5, &y_elems_6,
                                                         &in_elems_7,
                                                         &out_elems_8, &mulx_9,
                                                         &muly_10,
                                                         &kernel_arg_33629,
                                                         &kernel_arg_33630};
                            int64_t time_start_33625 = 0, time_end_33626 = 0;
                            
                            if (ctx->debugging) {
                                fprintf(stderr, "Launching %s with grid size (",
                                        "map_transpose_i32_small");
                                fprintf(stderr, "%d", squot32(num_arrays_4 *
                                                              x_elems_5 *
                                                              y_elems_6 + 256 -
                                                              1, 256));
                                fprintf(stderr, ", ");
                                fprintf(stderr, "%d", 1);
                                fprintf(stderr, ", ");
                                fprintf(stderr, "%d", 1);
                                fprintf(stderr, ") and block size (");
                                fprintf(stderr, "%d", 256);
                                fprintf(stderr, ", ");
                                fprintf(stderr, "%d", 1);
                                fprintf(stderr, ", ");
                                fprintf(stderr, "%d", 1);
                                fprintf(stderr, ").\n");
                                time_start_33625 = get_wall_time();
                            }
                            CUDA_SUCCEED(cuLaunchKernel(ctx->map_transpose_i32_small,
                                                        grid[0], grid[1],
                                                        grid[2], 256, 1, 1, 0 +
                                                        (shared_sizze_33627 +
                                                         (8 -
                                                          shared_sizze_33627 %
                                                          8) % 8), NULL,
                                                        kernel_args_33624,
                                                        NULL));
                            if (ctx->debugging) {
                                CUDA_SUCCEED(cuCtxSynchronize());
                                time_end_33626 = get_wall_time();
                                fprintf(stderr, "Kernel %s runtime: %ldus\n",
                                        "map_transpose_i32_small",
                                        time_end_33626 - time_start_33625);
                            }
                        }
                    } else {
                        unsigned int shared_sizze_33634 = 4224;
                        CUdeviceptr kernel_arg_33636 = destmem_0.mem;
                        CUdeviceptr kernel_arg_33637 = srcmem_2.mem;
                        unsigned int shared_offset_33635 = 0;
                        
                        if ((((((1 && squot32(x_elems_5 + 32 - 1, 32) != 0) &&
                                squot32(y_elems_6 + 32 - 1, 32) != 0) &&
                               num_arrays_4 != 0) && 32 != 0) && 8 != 0) && 1 !=
                            0) {
                            int perm[3] = {0, 1, 2};
                            
                            if (squot32(y_elems_6 + 32 - 1, 32) > 1 << 16) {
                                perm[1] = perm[0];
                                perm[0] = 1;
                            }
                            if (num_arrays_4 > 1 << 16) {
                                perm[2] = perm[0];
                                perm[0] = 2;
                            }
                            
                            size_t grid[3];
                            
                            grid[perm[0]] = squot32(x_elems_5 + 32 - 1, 32);
                            grid[perm[1]] = squot32(y_elems_6 + 32 - 1, 32);
                            grid[perm[2]] = num_arrays_4;
                            
                            void *kernel_args_33631[] = {&perm[0], &perm[1],
                                                         &perm[2],
                                                         &shared_offset_33635,
                                                         &destoffset_1,
                                                         &srcoffset_3,
                                                         &num_arrays_4,
                                                         &x_elems_5, &y_elems_6,
                                                         &in_elems_7,
                                                         &out_elems_8, &mulx_9,
                                                         &muly_10,
                                                         &kernel_arg_33636,
                                                         &kernel_arg_33637};
                            int64_t time_start_33632 = 0, time_end_33633 = 0;
                            
                            if (ctx->debugging) {
                                fprintf(stderr, "Launching %s with grid size (",
                                        "map_transpose_i32");
                                fprintf(stderr, "%d", squot32(x_elems_5 + 32 -
                                                              1, 32));
                                fprintf(stderr, ", ");
                                fprintf(stderr, "%d", squot32(y_elems_6 + 32 -
                                                              1, 32));
                                fprintf(stderr, ", ");
                                fprintf(stderr, "%d", num_arrays_4);
                                fprintf(stderr, ") and block size (");
                                fprintf(stderr, "%d", 32);
                                fprintf(stderr, ", ");
                                fprintf(stderr, "%d", 8);
                                fprintf(stderr, ", ");
                                fprintf(stderr, "%d", 1);
                                fprintf(stderr, ").\n");
                                time_start_33632 = get_wall_time();
                            }
                            CUDA_SUCCEED(cuLaunchKernel(ctx->map_transpose_i32,
                                                        grid[0], grid[1],
                                                        grid[2], 32, 8, 1, 0 +
                                                        (shared_sizze_33634 +
                                                         (8 -
                                                          shared_sizze_33634 %
                                                          8) % 8), NULL,
                                                        kernel_args_33631,
                                                        NULL));
                            if (ctx->debugging) {
                                CUDA_SUCCEED(cuCtxSynchronize());
                                time_end_33633 = get_wall_time();
                                fprintf(stderr, "Kernel %s runtime: %ldus\n",
                                        "map_transpose_i32", time_end_33633 -
                                        time_start_33632);
                            }
                        }
                    }
                }
            }
        }
    }
    return 0;
}
static int futrts_quicksort1(struct futhark_context *ctx,
                             struct memblock_device *out_mem_p_33638,
                             int32_t *out_out_arrsizze_33639,
                             int32_t *out_out_arrsizze_33640,
                             struct memblock_device input_mem_32784,
                             int32_t n_31612, int32_t sizze_31613)
{
    struct memblock_device out_mem_32955;
    
    out_mem_32955.references = NULL;
    
    int32_t out_arrsizze_32956;
    int32_t out_arrsizze_32957;
    int32_t flat_dim_31615 = n_31612 * sizze_31613;
    int32_t fill_arg_31617 = 1 + flat_dim_31615;
    int64_t bytes_32785 = sext_i32_i64(fill_arg_31617);
    struct memblock_device mem_32786;
    
    mem_32786.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32786, bytes_32785, "mem_32786"))
        return 1;
    
    int call_ret_33641 = futrts__replicate_bool(ctx, mem_32786, fill_arg_31617,
                                                0);
    
    assert(call_ret_33641 == 0);
    
    bool write_tmp_33642 = 1;
    
    CUDA_SUCCEED(cuMemcpyHtoD(mem_32786.mem + 0 * sizeof(bool),
                              &write_tmp_33642, sizeof(bool)));
    
    bool write_tmp_33643 = 1;
    
    CUDA_SUCCEED(cuMemcpyHtoD(mem_32786.mem + flat_dim_31615 * sizeof(bool),
                              &write_tmp_33643, sizeof(bool)));
    
    int32_t segred_group_sizze_32125;
    
    segred_group_sizze_32125 = ctx->sizes.quicksort1zisegred_group_sizze_32124;
    
    int32_t num_groups_32127;
    int32_t max_num_groups_32967;
    
    max_num_groups_32967 = ctx->sizes.quicksort1zisegred_num_groups_32126;
    num_groups_32127 = sext_i64_i32(smax64(1, smin64(squot64(bytes_32785 +
                                                             sext_i32_i64(segred_group_sizze_32125) -
                                                             1,
                                                             sext_i32_i64(segred_group_sizze_32125)),
                                                     sext_i32_i64(max_num_groups_32967))));
    
    struct memblock_device mem_32789;
    
    mem_32789.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32789, 1, "mem_32789"))
        return 1;
    
    struct memblock_device counter_mem_32968 = ctx->counter_mem_32968;
    struct memblock_device group_res_arr_mem_32970;
    
    group_res_arr_mem_32970.references = NULL;
    if (memblock_alloc_device(ctx, &group_res_arr_mem_32970, sizeof(bool) *
                              (segred_group_sizze_32125 * num_groups_32127),
                              "group_res_arr_mem_32970"))
        return 1;
    
    int32_t num_threads_32972 = num_groups_32127 * segred_group_sizze_32125;
    unsigned int shared_sizze_33648 = sizeof(bool);
    unsigned int shared_sizze_33650 = sizeof(bool) * segred_group_sizze_32125;
    CUdeviceptr kernel_arg_33652 = mem_32786.mem;
    CUdeviceptr kernel_arg_33653 = mem_32789.mem;
    CUdeviceptr kernel_arg_33654 = counter_mem_32968.mem;
    CUdeviceptr kernel_arg_33655 = group_res_arr_mem_32970.mem;
    unsigned int shared_offset_33649 = 0;
    unsigned int shared_offset_33651 = 0 + (shared_sizze_33648 + (8 -
                                                                  shared_sizze_33648 %
                                                                  8) % 8);
    
    if ((((((1 && num_groups_32127 != 0) && 1 != 0) && 1 != 0) &&
          segred_group_sizze_32125 != 0) && 1 != 0) && 1 != 0) {
        int perm[3] = {0, 1, 2};
        
        if (1 > 1 << 16) {
            perm[1] = perm[0];
            perm[0] = 1;
        }
        if (1 > 1 << 16) {
            perm[2] = perm[0];
            perm[0] = 2;
        }
        
        size_t grid[3];
        
        grid[perm[0]] = num_groups_32127;
        grid[perm[1]] = 1;
        grid[perm[2]] = 1;
        
        void *kernel_args_33645[] = {&shared_offset_33649, &shared_offset_33651,
                                     &fill_arg_31617, &num_groups_32127,
                                     &kernel_arg_33652, &kernel_arg_33653,
                                     &kernel_arg_33654, &kernel_arg_33655,
                                     &num_threads_32972};
        int64_t time_start_33646 = 0, time_end_33647 = 0;
        
        if (ctx->debugging) {
            fprintf(stderr, "Launching %s with grid size (",
                    "segred_nonseg_32132");
            fprintf(stderr, "%d", num_groups_32127);
            fprintf(stderr, ", ");
            fprintf(stderr, "%d", 1);
            fprintf(stderr, ", ");
            fprintf(stderr, "%d", 1);
            fprintf(stderr, ") and block size (");
            fprintf(stderr, "%d", segred_group_sizze_32125);
            fprintf(stderr, ", ");
            fprintf(stderr, "%d", 1);
            fprintf(stderr, ", ");
            fprintf(stderr, "%d", 1);
            fprintf(stderr, ").\n");
            time_start_33646 = get_wall_time();
        }
        CUDA_SUCCEED(cuLaunchKernel(ctx->segred_nonseg_32132, grid[0], grid[1],
                                    grid[2], segred_group_sizze_32125, 1, 1, 0 +
                                    (shared_sizze_33648 + (8 -
                                                           shared_sizze_33648 %
                                                           8) % 8) +
                                    (shared_sizze_33650 + (8 -
                                                           shared_sizze_33650 %
                                                           8) % 8), NULL,
                                    kernel_args_33645, NULL));
        if (ctx->debugging) {
            CUDA_SUCCEED(cuCtxSynchronize());
            time_end_33647 = get_wall_time();
            fprintf(stderr, "Kernel %s runtime: %ldus\n", "segred_nonseg_32132",
                    time_end_33647 - time_start_33646);
        }
    }
    
    bool read_res_33656;
    
    CUDA_SUCCEED(cuMemcpyDtoH(&read_res_33656, mem_32789.mem + 0 * sizeof(bool),
                              sizeof(bool)));
    
    bool res_31638 = read_res_33656;
    
    if (memblock_unref_device(ctx, &mem_32789, "mem_32789") != 0)
        return 1;
    
    bool res_31643 = !res_31638;
    int32_t m_31645 = flat_dim_31615 - 1;
    bool eq_x_zz_31654 = 0 == flat_dim_31615;
    int32_t i_p_m_t_s_31665 = 1 + m_31645;
    int32_t j_m_i_31674 = -1 - m_31645;
    int32_t n_31675 = squot32(j_m_i_31674, -1);
    bool empty_slice_31676 = n_31675 == 0;
    int32_t m_31677 = n_31675 - 1;
    int32_t m_t_s_31678 = -1 * m_31677;
    int32_t i_p_m_t_s_31679 = m_31645 + m_t_s_31678;
    bool zzero_leq_i_p_m_t_s_31680 = sle32(0, i_p_m_t_s_31679);
    bool i_p_m_t_s_leq_w_31681 = sle32(i_p_m_t_s_31679, flat_dim_31615);
    bool y_31683 = zzero_leq_i_p_m_t_s_31680 && i_p_m_t_s_leq_w_31681;
    bool ok_or_empty_31685 = empty_slice_31676 || y_31683;
    bool protect_assert_disj_31686 = res_31638 || ok_or_empty_31685;
    bool index_certs_31687;
    
    if (!protect_assert_disj_31686) {
        ctx->error = msgprintf("Error at\n%s\n%s%d%s%d%s\n",
                               " |-> Futhark/quicksort.fut:211:1-212:45\n |-> Futhark/quicksort.fut:212:20-45\n |-> /futlib/functional.fut:12:39-41\n |-> Futhark/quicksort.fut:208:18-67\n |-> /futlib/functional.fut:57:23-25\n |-> Futhark/quicksort.fut:191:8-52\n |-> Futhark/quicksort.fut:157:23-159:21\n |-> /futlib/functional.fut:12:39-41\n |-> Futhark/quicksort.fut:127:37-46\n `-> /futlib/array.fut:41:38-44\n",
                               "Index [::", -1,
                               "] out of bounds for array of shape [",
                               flat_dim_31615, "].");
        if (memblock_unref_device(ctx, &group_res_arr_mem_32970,
                                  "group_res_arr_mem_32970") != 0)
            return 1;
        if (memblock_unref_device(ctx, &mem_32789, "mem_32789") != 0)
            return 1;
        if (memblock_unref_device(ctx, &mem_32786, "mem_32786") != 0)
            return 1;
        if (memblock_unref_device(ctx, &out_mem_32955, "out_mem_32955") != 0)
            return 1;
        return 1;
    }
    
    bool dim_zzero_31688 = 0 == n_31675;
    bool both_empty_31689 = eq_x_zz_31654 && dim_zzero_31688;
    bool dim_match_31690 = flat_dim_31615 == n_31675;
    bool empty_or_match_31691 = both_empty_31689 || dim_match_31690;
    bool protect_assert_disj_31692 = res_31638 || empty_or_match_31691;
    bool empty_or_match_cert_31693;
    
    if (!protect_assert_disj_31692) {
        ctx->error = msgprintf("Error at\n%s\n%s%d%s\n",
                               " |-> Futhark/quicksort.fut:211:1-212:45\n |-> Futhark/quicksort.fut:212:20-45\n |-> /futlib/functional.fut:12:39-41\n |-> Futhark/quicksort.fut:208:18-67\n |-> /futlib/functional.fut:57:23-25\n |-> Futhark/quicksort.fut:191:8-52\n |-> Futhark/quicksort.fut:157:23-159:21\n |-> /futlib/functional.fut:12:39-41\n |-> Futhark/quicksort.fut:127:37-46\n `-> /futlib/array.fut:41:1-44\n",
                               "Function return value does not match shape of type [",
                               flat_dim_31615, "]t");
        if (memblock_unref_device(ctx, &group_res_arr_mem_32970,
                                  "group_res_arr_mem_32970") != 0)
            return 1;
        if (memblock_unref_device(ctx, &mem_32789, "mem_32789") != 0)
            return 1;
        if (memblock_unref_device(ctx, &mem_32786, "mem_32786") != 0)
            return 1;
        if (memblock_unref_device(ctx, &out_mem_32955, "out_mem_32955") != 0)
            return 1;
        return 1;
    }
    
    bool index_certs_31694;
    
    if (!protect_assert_disj_31686) {
        ctx->error = msgprintf("Error at\n%s\n%s%d%s%d%s\n",
                               " |-> Futhark/quicksort.fut:211:1-212:45\n |-> Futhark/quicksort.fut:212:20-45\n |-> /futlib/functional.fut:12:39-41\n |-> Futhark/quicksort.fut:208:18-67\n |-> /futlib/functional.fut:57:23-25\n |-> Futhark/quicksort.fut:191:8-52\n |-> Futhark/quicksort.fut:157:23-159:21\n |-> /futlib/functional.fut:12:39-41\n |-> Futhark/quicksort.fut:127:3-47\n |-> /futlib/functional.fut:12:39-41\n `-> /futlib/array.fut:41:38-44\n",
                               "Index [::", -1,
                               "] out of bounds for array of shape [",
                               flat_dim_31615, "].");
        if (memblock_unref_device(ctx, &group_res_arr_mem_32970,
                                  "group_res_arr_mem_32970") != 0)
            return 1;
        if (memblock_unref_device(ctx, &mem_32789, "mem_32789") != 0)
            return 1;
        if (memblock_unref_device(ctx, &mem_32786, "mem_32786") != 0)
            return 1;
        if (memblock_unref_device(ctx, &out_mem_32955, "out_mem_32955") != 0)
            return 1;
        return 1;
    }
    
    bool empty_or_match_cert_31695;
    
    if (!protect_assert_disj_31692) {
        ctx->error = msgprintf("Error at\n%s\n%s%d%s\n",
                               " |-> Futhark/quicksort.fut:211:1-212:45\n |-> Futhark/quicksort.fut:212:20-45\n |-> /futlib/functional.fut:12:39-41\n |-> Futhark/quicksort.fut:208:18-67\n |-> /futlib/functional.fut:57:23-25\n |-> Futhark/quicksort.fut:191:8-52\n `-> Futhark/quicksort.fut:149:1-160:8\n",
                               "Function return value does not match shape of type [",
                               flat_dim_31615, "]i32");
        if (memblock_unref_device(ctx, &group_res_arr_mem_32970,
                                  "group_res_arr_mem_32970") != 0)
            return 1;
        if (memblock_unref_device(ctx, &mem_32789, "mem_32789") != 0)
            return 1;
        if (memblock_unref_device(ctx, &mem_32786, "mem_32786") != 0)
            return 1;
        if (memblock_unref_device(ctx, &out_mem_32955, "out_mem_32955") != 0)
            return 1;
        return 1;
    }
    
    int64_t chunk_32133 = sext_i32_i64(flat_dim_31615);
    int32_t segscan_group_sizze_32136;
    
    segscan_group_sizze_32136 =
        ctx->sizes.quicksort1zisegscan_group_sizze_32135;
    
    int32_t num_groups_32138;
    int32_t max_num_groups_32994;
    
    max_num_groups_32994 = ctx->sizes.quicksort1zisegscan_num_groups_32137;
    num_groups_32138 = sext_i64_i32(smax64(1, smin64(squot64(chunk_32133 +
                                                             sext_i32_i64(segscan_group_sizze_32136) -
                                                             1,
                                                             sext_i32_i64(segscan_group_sizze_32136)),
                                                     sext_i32_i64(max_num_groups_32994))));
    
    int32_t segscan_group_sizze_32145;
    
    segscan_group_sizze_32145 =
        ctx->sizes.quicksort1zisegscan_group_sizze_32144;
    
    int32_t num_groups_32147;
    int32_t max_num_groups_32995;
    
    max_num_groups_32995 = ctx->sizes.quicksort1zisegscan_num_groups_32146;
    num_groups_32147 = sext_i64_i32(smax64(1, smin64(squot64(chunk_32133 +
                                                             sext_i32_i64(segscan_group_sizze_32145) -
                                                             1,
                                                             sext_i32_i64(segscan_group_sizze_32145)),
                                                     sext_i32_i64(max_num_groups_32995))));
    
    int32_t segmap_group_sizze_32156;
    
    segmap_group_sizze_32156 = ctx->sizes.quicksort1zisegmap_group_sizze_32155;
    
    int64_t segmap_group_sizze_32157 = sext_i32_i64(segmap_group_sizze_32156);
    int64_t y_32158 = segmap_group_sizze_32157 - 1;
    int64_t x_32159 = chunk_32133 + y_32158;
    int64_t segmap_usable_groups_64_32161;
    
    if (res_31643) {
        int64_t x_32602 = squot64(x_32159, segmap_group_sizze_32157);
        
        segmap_usable_groups_64_32161 = x_32602;
    } else {
        segmap_usable_groups_64_32161 = 0;
    }
    
    int32_t segmap_usable_groups_32162 =
            sext_i64_i32(segmap_usable_groups_64_32161);
    int32_t segmap_group_sizze_32172;
    
    segmap_group_sizze_32172 = ctx->sizes.quicksort1zisegmap_group_sizze_32171;
    
    int64_t segmap_group_sizze_32173 = sext_i32_i64(segmap_group_sizze_32172);
    int64_t y_32174 = segmap_group_sizze_32173 - 1;
    int64_t x_32175 = chunk_32133 + y_32174;
    int64_t segmap_usable_groups_64_32177;
    
    if (res_31643) {
        int64_t x_32604 = squot64(x_32175, segmap_group_sizze_32173);
        
        segmap_usable_groups_64_32177 = x_32604;
    } else {
        segmap_usable_groups_64_32177 = 0;
    }
    
    int32_t segmap_usable_groups_32178 =
            sext_i64_i32(segmap_usable_groups_64_32177);
    int32_t segscan_group_sizze_32185;
    
    segscan_group_sizze_32185 =
        ctx->sizes.quicksort1zisegscan_group_sizze_32184;
    
    int32_t num_groups_32187;
    int32_t max_num_groups_32996;
    
    max_num_groups_32996 = ctx->sizes.quicksort1zisegscan_num_groups_32186;
    num_groups_32187 = sext_i64_i32(smax64(1, smin64(squot64(chunk_32133 +
                                                             sext_i32_i64(segscan_group_sizze_32185) -
                                                             1,
                                                             sext_i32_i64(segscan_group_sizze_32185)),
                                                     sext_i32_i64(max_num_groups_32996))));
    
    int32_t segscan_group_sizze_32194;
    
    segscan_group_sizze_32194 =
        ctx->sizes.quicksort1zisegscan_group_sizze_32193;
    
    int32_t num_groups_32196;
    int32_t max_num_groups_32997;
    
    max_num_groups_32997 = ctx->sizes.quicksort1zisegscan_num_groups_32195;
    num_groups_32196 = sext_i64_i32(smax64(1, smin64(squot64(chunk_32133 +
                                                             sext_i32_i64(segscan_group_sizze_32194) -
                                                             1,
                                                             sext_i32_i64(segscan_group_sizze_32194)),
                                                     sext_i32_i64(max_num_groups_32997))));
    
    int32_t segmap_group_sizze_32205;
    
    segmap_group_sizze_32205 = ctx->sizes.quicksort1zisegmap_group_sizze_32204;
    
    int64_t segmap_group_sizze_32206 = sext_i32_i64(segmap_group_sizze_32205);
    int64_t y_32207 = segmap_group_sizze_32206 - 1;
    int64_t x_32208 = chunk_32133 + y_32207;
    int64_t segmap_usable_groups_64_32210;
    
    if (res_31643) {
        int64_t x_32606 = squot64(x_32208, segmap_group_sizze_32206);
        
        segmap_usable_groups_64_32210 = x_32606;
    } else {
        segmap_usable_groups_64_32210 = 0;
    }
    
    int32_t segmap_usable_groups_32211 =
            sext_i64_i32(segmap_usable_groups_64_32210);
    int32_t segscan_group_sizze_32220;
    
    segscan_group_sizze_32220 =
        ctx->sizes.quicksort1zisegscan_group_sizze_32219;
    
    int32_t num_groups_32222;
    int32_t max_num_groups_32998;
    
    max_num_groups_32998 = ctx->sizes.quicksort1zisegscan_num_groups_32221;
    num_groups_32222 = sext_i64_i32(smax64(1, smin64(squot64(chunk_32133 +
                                                             sext_i32_i64(segscan_group_sizze_32220) -
                                                             1,
                                                             sext_i32_i64(segscan_group_sizze_32220)),
                                                     sext_i32_i64(max_num_groups_32998))));
    
    int64_t n_32228 = sext_i32_i64(n_31675);
    int32_t segmap_group_sizze_32231;
    
    segmap_group_sizze_32231 = ctx->sizes.quicksort1zisegmap_group_sizze_32230;
    
    int64_t segmap_group_sizze_32232 = sext_i32_i64(segmap_group_sizze_32231);
    int64_t y_32233 = segmap_group_sizze_32232 - 1;
    int64_t x_32234 = n_32228 + y_32233;
    int64_t segmap_usable_groups_64_32236;
    
    if (res_31643) {
        int64_t x_32608 = squot64(x_32234, segmap_group_sizze_32232);
        
        segmap_usable_groups_64_32236 = x_32608;
    } else {
        segmap_usable_groups_64_32236 = 0;
    }
    
    int32_t segmap_usable_groups_32237 =
            sext_i64_i32(segmap_usable_groups_64_32236);
    int32_t segmap_group_sizze_32246;
    
    segmap_group_sizze_32246 = ctx->sizes.quicksort1zisegmap_group_sizze_32245;
    
    int64_t segmap_group_sizze_32247 = sext_i32_i64(segmap_group_sizze_32246);
    int64_t y_32248 = segmap_group_sizze_32247 - 1;
    int64_t x_32249 = chunk_32133 + y_32248;
    int64_t segmap_usable_groups_64_32251;
    
    if (res_31643) {
        int64_t x_32610 = squot64(x_32249, segmap_group_sizze_32247);
        
        segmap_usable_groups_64_32251 = x_32610;
    } else {
        segmap_usable_groups_64_32251 = 0;
    }
    
    int32_t segmap_usable_groups_32252 =
            sext_i64_i32(segmap_usable_groups_64_32251);
    int32_t segmap_group_sizze_32260;
    
    segmap_group_sizze_32260 = ctx->sizes.quicksort1zisegmap_group_sizze_32259;
    
    int64_t segmap_group_sizze_32261 = sext_i32_i64(segmap_group_sizze_32260);
    int64_t y_32262 = segmap_group_sizze_32261 - 1;
    int64_t x_32263 = chunk_32133 + y_32262;
    int64_t segmap_usable_groups_64_32265;
    
    if (res_31643) {
        int64_t x_32612 = squot64(x_32263, segmap_group_sizze_32261);
        
        segmap_usable_groups_64_32265 = x_32612;
    } else {
        segmap_usable_groups_64_32265 = 0;
    }
    
    int32_t segmap_usable_groups_32266 =
            sext_i64_i32(segmap_usable_groups_64_32265);
    int32_t segmap_group_sizze_32278;
    
    segmap_group_sizze_32278 = ctx->sizes.quicksort1zisegmap_group_sizze_32277;
    
    int64_t segmap_group_sizze_32279 = sext_i32_i64(segmap_group_sizze_32278);
    int64_t y_32280 = segmap_group_sizze_32279 - 1;
    int64_t x_32281 = chunk_32133 + y_32280;
    int64_t segmap_usable_groups_64_32283;
    
    if (res_31643) {
        int64_t x_32614 = squot64(x_32281, segmap_group_sizze_32279);
        
        segmap_usable_groups_64_32283 = x_32614;
    } else {
        segmap_usable_groups_64_32283 = 0;
    }
    
    int32_t segmap_usable_groups_32284 =
            sext_i64_i32(segmap_usable_groups_64_32283);
    int32_t segmap_group_sizze_32290;
    
    segmap_group_sizze_32290 = ctx->sizes.quicksort1zisegmap_group_sizze_32289;
    
    int64_t segmap_group_sizze_32291 = sext_i32_i64(segmap_group_sizze_32290);
    int64_t y_32292 = segmap_group_sizze_32291 - 1;
    int64_t x_32293 = chunk_32133 + y_32292;
    int64_t segmap_usable_groups_64_32295;
    
    if (res_31643) {
        int64_t x_32616 = squot64(x_32293, segmap_group_sizze_32291);
        
        segmap_usable_groups_64_32295 = x_32616;
    } else {
        segmap_usable_groups_64_32295 = 0;
    }
    
    int32_t segmap_usable_groups_32296 =
            sext_i64_i32(segmap_usable_groups_64_32295);
    int32_t segred_group_sizze_32300;
    
    segred_group_sizze_32300 = ctx->sizes.quicksort1zisegred_group_sizze_32299;
    
    int32_t num_groups_32302;
    int32_t max_num_groups_32999;
    
    max_num_groups_32999 = ctx->sizes.quicksort1zisegred_num_groups_32301;
    num_groups_32302 = sext_i64_i32(smax64(1, smin64(squot64(bytes_32785 +
                                                             sext_i32_i64(segred_group_sizze_32300) -
                                                             1,
                                                             sext_i32_i64(segred_group_sizze_32300)),
                                                     sext_i32_i64(max_num_groups_32999))));
    
    int64_t bytes_32793 = 4 * chunk_32133;
    struct memblock_device mem_32795;
    
    mem_32795.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32795, bytes_32793, "mem_32795"))
        return 1;
    
    struct memblock_device mem_32797;
    
    mem_32797.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32797, chunk_32133, "mem_32797"))
        return 1;
    
    struct memblock_device mem_32801;
    
    mem_32801.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32801, bytes_32793, "mem_32801"))
        return 1;
    
    struct memblock_device mem_32803;
    
    mem_32803.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32803, chunk_32133, "mem_32803"))
        return 1;
    
    struct memblock_device mem_32805;
    
    mem_32805.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32805, chunk_32133, "mem_32805"))
        return 1;
    
    struct memblock_device mem_32808;
    
    mem_32808.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32808, bytes_32793, "mem_32808"))
        return 1;
    
    struct memblock_device mem_32811;
    
    mem_32811.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32811, bytes_32793, "mem_32811"))
        return 1;
    
    struct memblock_device mem_32815;
    
    mem_32815.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32815, bytes_32793, "mem_32815"))
        return 1;
    
    struct memblock_device mem_32817;
    
    mem_32817.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32817, chunk_32133, "mem_32817"))
        return 1;
    
    struct memblock_device mem_32821;
    
    mem_32821.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32821, bytes_32793, "mem_32821"))
        return 1;
    
    struct memblock_device mem_32823;
    
    mem_32823.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32823, chunk_32133, "mem_32823"))
        return 1;
    
    struct memblock_device mem_32826;
    
    mem_32826.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32826, bytes_32793, "mem_32826"))
        return 1;
    
    struct memblock_device mem_32830;
    
    mem_32830.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32830, bytes_32793, "mem_32830"))
        return 1;
    
    struct memblock_device mem_32832;
    
    mem_32832.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32832, chunk_32133, "mem_32832"))
        return 1;
    
    int64_t bytes_32833 = 4 * n_32228;
    struct memblock_device mem_32835;
    
    mem_32835.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32835, bytes_32833, "mem_32835"))
        return 1;
    
    struct memblock_device mem_32838;
    
    mem_32838.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32838, bytes_32793, "mem_32838"))
        return 1;
    
    struct memblock_device mem_32844;
    
    mem_32844.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32844, bytes_32793, "mem_32844"))
        return 1;
    
    struct memblock_device mem_32849;
    
    mem_32849.references = NULL;
    if (memblock_alloc_device(ctx, &mem_32849, 1, "mem_32849"))
        return 1;
    
    struct memblock_device res_mem_32850;
    
    res_mem_32850.references = NULL;
    
    struct memblock_device res_mem_32851;
    
    res_mem_32851.references = NULL;
    
    bool res_31699;
    bool loop_while_31702;
    struct memblock_device x_mem_32790;
    
    x_mem_32790.references = NULL;
    
    struct memblock_device x_mem_32791;
    
    x_mem_32791.references = NULL;
    if (memblock_set_device(ctx, &x_mem_32790, &input_mem_32784,
                            "input_mem_32784") != 0)
        return 1;
    if (memblock_set_device(ctx, &x_mem_32791, &mem_32786, "mem_32786") != 0)
        return 1;
    loop_while_31702 = res_31643;
    while (loop_while_31702) {
        int32_t num_threads_33005 = num_groups_32138 *
                segscan_group_sizze_32136;
        unsigned int shared_sizze_33660 = sizeof(int32_t) *
                     segscan_group_sizze_32136;
        unsigned int shared_sizze_33662 = sizeof(bool) *
                     segscan_group_sizze_32136;
        CUdeviceptr kernel_arg_33664 = x_mem_32790.mem;
        CUdeviceptr kernel_arg_33665 = x_mem_32791.mem;
        CUdeviceptr kernel_arg_33666 = mem_32795.mem;
        CUdeviceptr kernel_arg_33667 = mem_32797.mem;
        unsigned int shared_offset_33661 = 0;
        unsigned int shared_offset_33663 = 0 + (shared_sizze_33660 + (8 -
                                                                      shared_sizze_33660 %
                                                                      8) % 8);
        
        if ((((((1 && num_groups_32138 != 0) && 1 != 0) && 1 != 0) &&
              segscan_group_sizze_32136 != 0) && 1 != 0) && 1 != 0) {
            int perm[3] = {0, 1, 2};
            
            if (1 > 1 << 16) {
                perm[1] = perm[0];
                perm[0] = 1;
            }
            if (1 > 1 << 16) {
                perm[2] = perm[0];
                perm[0] = 2;
            }
            
            size_t grid[3];
            
            grid[perm[0]] = num_groups_32138;
            grid[perm[1]] = 1;
            grid[perm[2]] = 1;
            
            void *kernel_args_33657[] = {&shared_offset_33661,
                                         &shared_offset_33663, &flat_dim_31615,
                                         &kernel_arg_33664, &kernel_arg_33665,
                                         &kernel_arg_33666, &kernel_arg_33667,
                                         &num_threads_33005};
            int64_t time_start_33658 = 0, time_end_33659 = 0;
            
            if (ctx->debugging) {
                fprintf(stderr, "Launching %s with grid size (",
                        "scan_stage1_32141");
                fprintf(stderr, "%d", num_groups_32138);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ") and block size (");
                fprintf(stderr, "%d", segscan_group_sizze_32136);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ").\n");
                time_start_33658 = get_wall_time();
            }
            CUDA_SUCCEED(cuLaunchKernel(ctx->scan_stage1_32141, grid[0],
                                        grid[1], grid[2],
                                        segscan_group_sizze_32136, 1, 1, 0 +
                                        (shared_sizze_33660 + (8 -
                                                               shared_sizze_33660 %
                                                               8) % 8) +
                                        (shared_sizze_33662 + (8 -
                                                               shared_sizze_33662 %
                                                               8) % 8), NULL,
                                        kernel_args_33657, NULL));
            if (ctx->debugging) {
                CUDA_SUCCEED(cuCtxSynchronize());
                time_end_33659 = get_wall_time();
                fprintf(stderr, "Kernel %s runtime: %ldus\n",
                        "scan_stage1_32141", time_end_33659 - time_start_33658);
            }
        }
        if (ctx->debugging)
            fprintf(stderr, "%s\n", "\n# SegScan");
        if (ctx->debugging)
            fprintf(stderr, "%s: %llu%c", "elems_per_group",
                    (long long) (segscan_group_sizze_32136 *
                                 squot32(flat_dim_31615 + num_threads_33005 - 1,
                                         num_threads_33005)), '\n');
        
        unsigned int shared_sizze_33671 = sizeof(int32_t) * num_groups_32138;
        unsigned int shared_sizze_33673 = sizeof(bool) * num_groups_32138;
        CUdeviceptr kernel_arg_33675 = mem_32795.mem;
        CUdeviceptr kernel_arg_33676 = mem_32797.mem;
        unsigned int shared_offset_33672 = 0;
        unsigned int shared_offset_33674 = 0 + (shared_sizze_33671 + (8 -
                                                                      shared_sizze_33671 %
                                                                      8) % 8);
        
        if ((((((1 && 1 != 0) && 1 != 0) && 1 != 0) && num_groups_32138 != 0) &&
             1 != 0) && 1 != 0) {
            int perm[3] = {0, 1, 2};
            
            if (1 > 1 << 16) {
                perm[1] = perm[0];
                perm[0] = 1;
            }
            if (1 > 1 << 16) {
                perm[2] = perm[0];
                perm[0] = 2;
            }
            
            size_t grid[3];
            
            grid[perm[0]] = 1;
            grid[perm[1]] = 1;
            grid[perm[2]] = 1;
            
            void *kernel_args_33668[] = {&shared_offset_33672,
                                         &shared_offset_33674, &flat_dim_31615,
                                         &num_groups_32138, &kernel_arg_33675,
                                         &kernel_arg_33676, &num_threads_33005};
            int64_t time_start_33669 = 0, time_end_33670 = 0;
            
            if (ctx->debugging) {
                fprintf(stderr, "Launching %s with grid size (",
                        "scan_stage2_32141");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ") and block size (");
                fprintf(stderr, "%d", num_groups_32138);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ").\n");
                time_start_33669 = get_wall_time();
            }
            CUDA_SUCCEED(cuLaunchKernel(ctx->scan_stage2_32141, grid[0],
                                        grid[1], grid[2], num_groups_32138, 1,
                                        1, 0 + (shared_sizze_33671 + (8 -
                                                                      shared_sizze_33671 %
                                                                      8) % 8) +
                                        (shared_sizze_33673 + (8 -
                                                               shared_sizze_33673 %
                                                               8) % 8), NULL,
                                        kernel_args_33668, NULL));
            if (ctx->debugging) {
                CUDA_SUCCEED(cuCtxSynchronize());
                time_end_33670 = get_wall_time();
                fprintf(stderr, "Kernel %s runtime: %ldus\n",
                        "scan_stage2_32141", time_end_33670 - time_start_33669);
            }
        }
        
        int32_t group_sizze_33065;
        
        group_sizze_33065 = ctx->sizes.quicksort1zigroup_sizze_33065;
        
        int32_t num_groups_33066;
        
        num_groups_33066 = squot32(flat_dim_31615 +
                                   sext_i32_i32(group_sizze_33065) - 1,
                                   sext_i32_i32(group_sizze_33065));
        
        CUdeviceptr kernel_arg_33680 = mem_32795.mem;
        CUdeviceptr kernel_arg_33681 = mem_32797.mem;
        
        if ((((((1 && num_groups_33066 != 0) && 1 != 0) && 1 != 0) &&
              group_sizze_33065 != 0) && 1 != 0) && 1 != 0) {
            int perm[3] = {0, 1, 2};
            
            if (1 > 1 << 16) {
                perm[1] = perm[0];
                perm[0] = 1;
            }
            if (1 > 1 << 16) {
                perm[2] = perm[0];
                perm[0] = 2;
            }
            
            size_t grid[3];
            
            grid[perm[0]] = num_groups_33066;
            grid[perm[1]] = 1;
            grid[perm[2]] = 1;
            
            void *kernel_args_33677[] = {&flat_dim_31615, &kernel_arg_33680,
                                         &kernel_arg_33681, &num_threads_33005};
            int64_t time_start_33678 = 0, time_end_33679 = 0;
            
            if (ctx->debugging) {
                fprintf(stderr, "Launching %s with grid size (",
                        "scan_stage3_33062");
                fprintf(stderr, "%d", num_groups_33066);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ") and block size (");
                fprintf(stderr, "%d", group_sizze_33065);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ").\n");
                time_start_33678 = get_wall_time();
            }
            CUDA_SUCCEED(cuLaunchKernel(ctx->scan_stage3_33062, grid[0],
                                        grid[1], grid[2], group_sizze_33065, 1,
                                        1, 0, NULL, kernel_args_33677, NULL));
            if (ctx->debugging) {
                CUDA_SUCCEED(cuCtxSynchronize());
                time_end_33679 = get_wall_time();
                fprintf(stderr, "Kernel %s runtime: %ldus\n",
                        "scan_stage3_33062", time_end_33679 - time_start_33678);
            }
        }
        
        int32_t num_threads_33069 = num_groups_32147 *
                segscan_group_sizze_32145;
        unsigned int shared_sizze_33685 = sizeof(int32_t) *
                     segscan_group_sizze_32145;
        unsigned int shared_sizze_33687 = sizeof(bool) *
                     segscan_group_sizze_32145;
        CUdeviceptr kernel_arg_33689 = x_mem_32790.mem;
        CUdeviceptr kernel_arg_33690 = x_mem_32791.mem;
        CUdeviceptr kernel_arg_33691 = mem_32795.mem;
        CUdeviceptr kernel_arg_33692 = mem_32797.mem;
        CUdeviceptr kernel_arg_33693 = mem_32801.mem;
        CUdeviceptr kernel_arg_33694 = mem_32803.mem;
        CUdeviceptr kernel_arg_33695 = mem_32805.mem;
        unsigned int shared_offset_33686 = 0;
        unsigned int shared_offset_33688 = 0 + (shared_sizze_33685 + (8 -
                                                                      shared_sizze_33685 %
                                                                      8) % 8);
        
        if ((((((1 && num_groups_32147 != 0) && 1 != 0) && 1 != 0) &&
              segscan_group_sizze_32145 != 0) && 1 != 0) && 1 != 0) {
            int perm[3] = {0, 1, 2};
            
            if (1 > 1 << 16) {
                perm[1] = perm[0];
                perm[0] = 1;
            }
            if (1 > 1 << 16) {
                perm[2] = perm[0];
                perm[0] = 2;
            }
            
            size_t grid[3];
            
            grid[perm[0]] = num_groups_32147;
            grid[perm[1]] = 1;
            grid[perm[2]] = 1;
            
            void *kernel_args_33682[] = {&shared_offset_33686,
                                         &shared_offset_33688, &flat_dim_31615,
                                         &kernel_arg_33689, &kernel_arg_33690,
                                         &kernel_arg_33691, &kernel_arg_33692,
                                         &kernel_arg_33693, &kernel_arg_33694,
                                         &kernel_arg_33695, &num_threads_33069};
            int64_t time_start_33683 = 0, time_end_33684 = 0;
            
            if (ctx->debugging) {
                fprintf(stderr, "Launching %s with grid size (",
                        "scan_stage1_32150");
                fprintf(stderr, "%d", num_groups_32147);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ") and block size (");
                fprintf(stderr, "%d", segscan_group_sizze_32145);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ").\n");
                time_start_33683 = get_wall_time();
            }
            CUDA_SUCCEED(cuLaunchKernel(ctx->scan_stage1_32150, grid[0],
                                        grid[1], grid[2],
                                        segscan_group_sizze_32145, 1, 1, 0 +
                                        (shared_sizze_33685 + (8 -
                                                               shared_sizze_33685 %
                                                               8) % 8) +
                                        (shared_sizze_33687 + (8 -
                                                               shared_sizze_33687 %
                                                               8) % 8), NULL,
                                        kernel_args_33682, NULL));
            if (ctx->debugging) {
                CUDA_SUCCEED(cuCtxSynchronize());
                time_end_33684 = get_wall_time();
                fprintf(stderr, "Kernel %s runtime: %ldus\n",
                        "scan_stage1_32150", time_end_33684 - time_start_33683);
            }
        }
        if (ctx->debugging)
            fprintf(stderr, "%s\n", "\n# SegScan");
        if (ctx->debugging)
            fprintf(stderr, "%s: %llu%c", "elems_per_group",
                    (long long) (segscan_group_sizze_32145 *
                                 squot32(flat_dim_31615 + num_threads_33069 - 1,
                                         num_threads_33069)), '\n');
        
        unsigned int shared_sizze_33699 = sizeof(int32_t) * num_groups_32147;
        unsigned int shared_sizze_33701 = sizeof(bool) * num_groups_32147;
        CUdeviceptr kernel_arg_33703 = mem_32801.mem;
        CUdeviceptr kernel_arg_33704 = mem_32803.mem;
        unsigned int shared_offset_33700 = 0;
        unsigned int shared_offset_33702 = 0 + (shared_sizze_33699 + (8 -
                                                                      shared_sizze_33699 %
                                                                      8) % 8);
        
        if ((((((1 && 1 != 0) && 1 != 0) && 1 != 0) && num_groups_32147 != 0) &&
             1 != 0) && 1 != 0) {
            int perm[3] = {0, 1, 2};
            
            if (1 > 1 << 16) {
                perm[1] = perm[0];
                perm[0] = 1;
            }
            if (1 > 1 << 16) {
                perm[2] = perm[0];
                perm[0] = 2;
            }
            
            size_t grid[3];
            
            grid[perm[0]] = 1;
            grid[perm[1]] = 1;
            grid[perm[2]] = 1;
            
            void *kernel_args_33696[] = {&shared_offset_33700,
                                         &shared_offset_33702, &flat_dim_31615,
                                         &num_groups_32147, &kernel_arg_33703,
                                         &kernel_arg_33704, &num_threads_33069};
            int64_t time_start_33697 = 0, time_end_33698 = 0;
            
            if (ctx->debugging) {
                fprintf(stderr, "Launching %s with grid size (",
                        "scan_stage2_32150");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ") and block size (");
                fprintf(stderr, "%d", num_groups_32147);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ").\n");
                time_start_33697 = get_wall_time();
            }
            CUDA_SUCCEED(cuLaunchKernel(ctx->scan_stage2_32150, grid[0],
                                        grid[1], grid[2], num_groups_32147, 1,
                                        1, 0 + (shared_sizze_33699 + (8 -
                                                                      shared_sizze_33699 %
                                                                      8) % 8) +
                                        (shared_sizze_33701 + (8 -
                                                               shared_sizze_33701 %
                                                               8) % 8), NULL,
                                        kernel_args_33696, NULL));
            if (ctx->debugging) {
                CUDA_SUCCEED(cuCtxSynchronize());
                time_end_33698 = get_wall_time();
                fprintf(stderr, "Kernel %s runtime: %ldus\n",
                        "scan_stage2_32150", time_end_33698 - time_start_33697);
            }
        }
        
        int32_t group_sizze_33129;
        
        group_sizze_33129 = ctx->sizes.quicksort1zigroup_sizze_33129;
        
        int32_t num_groups_33130;
        
        num_groups_33130 = squot32(flat_dim_31615 +
                                   sext_i32_i32(group_sizze_33129) - 1,
                                   sext_i32_i32(group_sizze_33129));
        
        CUdeviceptr kernel_arg_33708 = mem_32801.mem;
        CUdeviceptr kernel_arg_33709 = mem_32803.mem;
        
        if ((((((1 && num_groups_33130 != 0) && 1 != 0) && 1 != 0) &&
              group_sizze_33129 != 0) && 1 != 0) && 1 != 0) {
            int perm[3] = {0, 1, 2};
            
            if (1 > 1 << 16) {
                perm[1] = perm[0];
                perm[0] = 1;
            }
            if (1 > 1 << 16) {
                perm[2] = perm[0];
                perm[0] = 2;
            }
            
            size_t grid[3];
            
            grid[perm[0]] = num_groups_33130;
            grid[perm[1]] = 1;
            grid[perm[2]] = 1;
            
            void *kernel_args_33705[] = {&flat_dim_31615, &kernel_arg_33708,
                                         &kernel_arg_33709, &num_threads_33069};
            int64_t time_start_33706 = 0, time_end_33707 = 0;
            
            if (ctx->debugging) {
                fprintf(stderr, "Launching %s with grid size (",
                        "scan_stage3_33126");
                fprintf(stderr, "%d", num_groups_33130);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ") and block size (");
                fprintf(stderr, "%d", group_sizze_33129);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ").\n");
                time_start_33706 = get_wall_time();
            }
            CUDA_SUCCEED(cuLaunchKernel(ctx->scan_stage3_33126, grid[0],
                                        grid[1], grid[2], group_sizze_33129, 1,
                                        1, 0, NULL, kernel_args_33705, NULL));
            if (ctx->debugging) {
                CUDA_SUCCEED(cuCtxSynchronize());
                time_end_33707 = get_wall_time();
                fprintf(stderr, "Kernel %s runtime: %ldus\n",
                        "scan_stage3_33126", time_end_33707 - time_start_33706);
            }
        }
        
        CUdeviceptr kernel_arg_33713 = mem_32801.mem;
        CUdeviceptr kernel_arg_33714 = mem_32803.mem;
        CUdeviceptr kernel_arg_33715 = mem_32808.mem;
        
        if ((((((1 && segmap_usable_groups_32162 != 0) && 1 != 0) && 1 != 0) &&
              segmap_group_sizze_32156 != 0) && 1 != 0) && 1 != 0) {
            int perm[3] = {0, 1, 2};
            
            if (1 > 1 << 16) {
                perm[1] = perm[0];
                perm[0] = 1;
            }
            if (1 > 1 << 16) {
                perm[2] = perm[0];
                perm[0] = 2;
            }
            
            size_t grid[3];
            
            grid[perm[0]] = segmap_usable_groups_32162;
            grid[perm[1]] = 1;
            grid[perm[2]] = 1;
            
            void *kernel_args_33710[] = {&flat_dim_31615, &kernel_arg_33713,
                                         &kernel_arg_33714, &kernel_arg_33715};
            int64_t time_start_33711 = 0, time_end_33712 = 0;
            
            if (ctx->debugging) {
                fprintf(stderr, "Launching %s with grid size (",
                        "segmap_32152");
                fprintf(stderr, "%d", segmap_usable_groups_32162);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ") and block size (");
                fprintf(stderr, "%d", segmap_group_sizze_32156);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ").\n");
                time_start_33711 = get_wall_time();
            }
            CUDA_SUCCEED(cuLaunchKernel(ctx->segmap_32152, grid[0], grid[1],
                                        grid[2], segmap_group_sizze_32156, 1, 1,
                                        0, NULL, kernel_args_33710, NULL));
            if (ctx->debugging) {
                CUDA_SUCCEED(cuCtxSynchronize());
                time_end_33712 = get_wall_time();
                fprintf(stderr, "Kernel %s runtime: %ldus\n", "segmap_32152",
                        time_end_33712 - time_start_33711);
            }
        }
        
        CUdeviceptr kernel_arg_33719 = mem_32803.mem;
        CUdeviceptr kernel_arg_33720 = mem_32808.mem;
        CUdeviceptr kernel_arg_33721 = mem_32811.mem;
        
        if ((((((1 && segmap_usable_groups_32178 != 0) && 1 != 0) && 1 != 0) &&
              segmap_group_sizze_32172 != 0) && 1 != 0) && 1 != 0) {
            int perm[3] = {0, 1, 2};
            
            if (1 > 1 << 16) {
                perm[1] = perm[0];
                perm[0] = 1;
            }
            if (1 > 1 << 16) {
                perm[2] = perm[0];
                perm[0] = 2;
            }
            
            size_t grid[3];
            
            grid[perm[0]] = segmap_usable_groups_32178;
            grid[perm[1]] = 1;
            grid[perm[2]] = 1;
            
            void *kernel_args_33716[] = {&flat_dim_31615, &kernel_arg_33719,
                                         &kernel_arg_33720, &kernel_arg_33721};
            int64_t time_start_33717 = 0, time_end_33718 = 0;
            
            if (ctx->debugging) {
                fprintf(stderr, "Launching %s with grid size (",
                        "segmap_32168");
                fprintf(stderr, "%d", segmap_usable_groups_32178);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ") and block size (");
                fprintf(stderr, "%d", segmap_group_sizze_32172);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ").\n");
                time_start_33717 = get_wall_time();
            }
            CUDA_SUCCEED(cuLaunchKernel(ctx->segmap_32168, grid[0], grid[1],
                                        grid[2], segmap_group_sizze_32172, 1, 1,
                                        0, NULL, kernel_args_33716, NULL));
            if (ctx->debugging) {
                CUDA_SUCCEED(cuCtxSynchronize());
                time_end_33718 = get_wall_time();
                fprintf(stderr, "Kernel %s runtime: %ldus\n", "segmap_32168",
                        time_end_33718 - time_start_33717);
            }
        }
        
        int32_t num_threads_33143 = num_groups_32187 *
                segscan_group_sizze_32185;
        unsigned int shared_sizze_33725 = sizeof(int32_t) *
                     segscan_group_sizze_32185;
        unsigned int shared_sizze_33727 = sizeof(bool) *
                     segscan_group_sizze_32185;
        CUdeviceptr kernel_arg_33729 = x_mem_32791.mem;
        CUdeviceptr kernel_arg_33730 = mem_32805.mem;
        CUdeviceptr kernel_arg_33731 = mem_32815.mem;
        CUdeviceptr kernel_arg_33732 = mem_32817.mem;
        unsigned int shared_offset_33726 = 0;
        unsigned int shared_offset_33728 = 0 + (shared_sizze_33725 + (8 -
                                                                      shared_sizze_33725 %
                                                                      8) % 8);
        
        if ((((((1 && num_groups_32187 != 0) && 1 != 0) && 1 != 0) &&
              segscan_group_sizze_32185 != 0) && 1 != 0) && 1 != 0) {
            int perm[3] = {0, 1, 2};
            
            if (1 > 1 << 16) {
                perm[1] = perm[0];
                perm[0] = 1;
            }
            if (1 > 1 << 16) {
                perm[2] = perm[0];
                perm[0] = 2;
            }
            
            size_t grid[3];
            
            grid[perm[0]] = num_groups_32187;
            grid[perm[1]] = 1;
            grid[perm[2]] = 1;
            
            void *kernel_args_33722[] = {&shared_offset_33726,
                                         &shared_offset_33728, &flat_dim_31615,
                                         &kernel_arg_33729, &kernel_arg_33730,
                                         &kernel_arg_33731, &kernel_arg_33732,
                                         &num_threads_33143};
            int64_t time_start_33723 = 0, time_end_33724 = 0;
            
            if (ctx->debugging) {
                fprintf(stderr, "Launching %s with grid size (",
                        "scan_stage1_32190");
                fprintf(stderr, "%d", num_groups_32187);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ") and block size (");
                fprintf(stderr, "%d", segscan_group_sizze_32185);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ").\n");
                time_start_33723 = get_wall_time();
            }
            CUDA_SUCCEED(cuLaunchKernel(ctx->scan_stage1_32190, grid[0],
                                        grid[1], grid[2],
                                        segscan_group_sizze_32185, 1, 1, 0 +
                                        (shared_sizze_33725 + (8 -
                                                               shared_sizze_33725 %
                                                               8) % 8) +
                                        (shared_sizze_33727 + (8 -
                                                               shared_sizze_33727 %
                                                               8) % 8), NULL,
                                        kernel_args_33722, NULL));
            if (ctx->debugging) {
                CUDA_SUCCEED(cuCtxSynchronize());
                time_end_33724 = get_wall_time();
                fprintf(stderr, "Kernel %s runtime: %ldus\n",
                        "scan_stage1_32190", time_end_33724 - time_start_33723);
            }
        }
        if (ctx->debugging)
            fprintf(stderr, "%s\n", "\n# SegScan");
        if (ctx->debugging)
            fprintf(stderr, "%s: %llu%c", "elems_per_group",
                    (long long) (segscan_group_sizze_32185 *
                                 squot32(flat_dim_31615 + num_threads_33143 - 1,
                                         num_threads_33143)), '\n');
        
        unsigned int shared_sizze_33736 = sizeof(int32_t) * num_groups_32187;
        unsigned int shared_sizze_33738 = sizeof(bool) * num_groups_32187;
        CUdeviceptr kernel_arg_33740 = mem_32815.mem;
        CUdeviceptr kernel_arg_33741 = mem_32817.mem;
        unsigned int shared_offset_33737 = 0;
        unsigned int shared_offset_33739 = 0 + (shared_sizze_33736 + (8 -
                                                                      shared_sizze_33736 %
                                                                      8) % 8);
        
        if ((((((1 && 1 != 0) && 1 != 0) && 1 != 0) && num_groups_32187 != 0) &&
             1 != 0) && 1 != 0) {
            int perm[3] = {0, 1, 2};
            
            if (1 > 1 << 16) {
                perm[1] = perm[0];
                perm[0] = 1;
            }
            if (1 > 1 << 16) {
                perm[2] = perm[0];
                perm[0] = 2;
            }
            
            size_t grid[3];
            
            grid[perm[0]] = 1;
            grid[perm[1]] = 1;
            grid[perm[2]] = 1;
            
            void *kernel_args_33733[] = {&shared_offset_33737,
                                         &shared_offset_33739, &flat_dim_31615,
                                         &num_groups_32187, &kernel_arg_33740,
                                         &kernel_arg_33741, &num_threads_33143};
            int64_t time_start_33734 = 0, time_end_33735 = 0;
            
            if (ctx->debugging) {
                fprintf(stderr, "Launching %s with grid size (",
                        "scan_stage2_32190");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ") and block size (");
                fprintf(stderr, "%d", num_groups_32187);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ").\n");
                time_start_33734 = get_wall_time();
            }
            CUDA_SUCCEED(cuLaunchKernel(ctx->scan_stage2_32190, grid[0],
                                        grid[1], grid[2], num_groups_32187, 1,
                                        1, 0 + (shared_sizze_33736 + (8 -
                                                                      shared_sizze_33736 %
                                                                      8) % 8) +
                                        (shared_sizze_33738 + (8 -
                                                               shared_sizze_33738 %
                                                               8) % 8), NULL,
                                        kernel_args_33733, NULL));
            if (ctx->debugging) {
                CUDA_SUCCEED(cuCtxSynchronize());
                time_end_33735 = get_wall_time();
                fprintf(stderr, "Kernel %s runtime: %ldus\n",
                        "scan_stage2_32190", time_end_33735 - time_start_33734);
            }
        }
        
        int32_t group_sizze_33208;
        
        group_sizze_33208 = ctx->sizes.quicksort1zigroup_sizze_33208;
        
        int32_t num_groups_33209;
        
        num_groups_33209 = squot32(flat_dim_31615 +
                                   sext_i32_i32(group_sizze_33208) - 1,
                                   sext_i32_i32(group_sizze_33208));
        
        CUdeviceptr kernel_arg_33745 = mem_32815.mem;
        CUdeviceptr kernel_arg_33746 = mem_32817.mem;
        
        if ((((((1 && num_groups_33209 != 0) && 1 != 0) && 1 != 0) &&
              group_sizze_33208 != 0) && 1 != 0) && 1 != 0) {
            int perm[3] = {0, 1, 2};
            
            if (1 > 1 << 16) {
                perm[1] = perm[0];
                perm[0] = 1;
            }
            if (1 > 1 << 16) {
                perm[2] = perm[0];
                perm[0] = 2;
            }
            
            size_t grid[3];
            
            grid[perm[0]] = num_groups_33209;
            grid[perm[1]] = 1;
            grid[perm[2]] = 1;
            
            void *kernel_args_33742[] = {&flat_dim_31615, &kernel_arg_33745,
                                         &kernel_arg_33746, &num_threads_33143};
            int64_t time_start_33743 = 0, time_end_33744 = 0;
            
            if (ctx->debugging) {
                fprintf(stderr, "Launching %s with grid size (",
                        "scan_stage3_33205");
                fprintf(stderr, "%d", num_groups_33209);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ") and block size (");
                fprintf(stderr, "%d", group_sizze_33208);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ").\n");
                time_start_33743 = get_wall_time();
            }
            CUDA_SUCCEED(cuLaunchKernel(ctx->scan_stage3_33205, grid[0],
                                        grid[1], grid[2], group_sizze_33208, 1,
                                        1, 0, NULL, kernel_args_33742, NULL));
            if (ctx->debugging) {
                CUDA_SUCCEED(cuCtxSynchronize());
                time_end_33744 = get_wall_time();
                fprintf(stderr, "Kernel %s runtime: %ldus\n",
                        "scan_stage3_33205", time_end_33744 - time_start_33743);
            }
        }
        
        int32_t num_threads_33212 = num_groups_32196 *
                segscan_group_sizze_32194;
        unsigned int shared_sizze_33750 = sizeof(int32_t) *
                     segscan_group_sizze_32194;
        unsigned int shared_sizze_33752 = sizeof(bool) *
                     segscan_group_sizze_32194;
        CUdeviceptr kernel_arg_33754 = x_mem_32791.mem;
        CUdeviceptr kernel_arg_33755 = mem_32805.mem;
        CUdeviceptr kernel_arg_33756 = mem_32821.mem;
        CUdeviceptr kernel_arg_33757 = mem_32823.mem;
        unsigned int shared_offset_33751 = 0;
        unsigned int shared_offset_33753 = 0 + (shared_sizze_33750 + (8 -
                                                                      shared_sizze_33750 %
                                                                      8) % 8);
        
        if ((((((1 && num_groups_32196 != 0) && 1 != 0) && 1 != 0) &&
              segscan_group_sizze_32194 != 0) && 1 != 0) && 1 != 0) {
            int perm[3] = {0, 1, 2};
            
            if (1 > 1 << 16) {
                perm[1] = perm[0];
                perm[0] = 1;
            }
            if (1 > 1 << 16) {
                perm[2] = perm[0];
                perm[0] = 2;
            }
            
            size_t grid[3];
            
            grid[perm[0]] = num_groups_32196;
            grid[perm[1]] = 1;
            grid[perm[2]] = 1;
            
            void *kernel_args_33747[] = {&shared_offset_33751,
                                         &shared_offset_33753, &flat_dim_31615,
                                         &kernel_arg_33754, &kernel_arg_33755,
                                         &kernel_arg_33756, &kernel_arg_33757,
                                         &num_threads_33212};
            int64_t time_start_33748 = 0, time_end_33749 = 0;
            
            if (ctx->debugging) {
                fprintf(stderr, "Launching %s with grid size (",
                        "scan_stage1_32199");
                fprintf(stderr, "%d", num_groups_32196);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ") and block size (");
                fprintf(stderr, "%d", segscan_group_sizze_32194);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ").\n");
                time_start_33748 = get_wall_time();
            }
            CUDA_SUCCEED(cuLaunchKernel(ctx->scan_stage1_32199, grid[0],
                                        grid[1], grid[2],
                                        segscan_group_sizze_32194, 1, 1, 0 +
                                        (shared_sizze_33750 + (8 -
                                                               shared_sizze_33750 %
                                                               8) % 8) +
                                        (shared_sizze_33752 + (8 -
                                                               shared_sizze_33752 %
                                                               8) % 8), NULL,
                                        kernel_args_33747, NULL));
            if (ctx->debugging) {
                CUDA_SUCCEED(cuCtxSynchronize());
                time_end_33749 = get_wall_time();
                fprintf(stderr, "Kernel %s runtime: %ldus\n",
                        "scan_stage1_32199", time_end_33749 - time_start_33748);
            }
        }
        if (ctx->debugging)
            fprintf(stderr, "%s\n", "\n# SegScan");
        if (ctx->debugging)
            fprintf(stderr, "%s: %llu%c", "elems_per_group",
                    (long long) (segscan_group_sizze_32194 *
                                 squot32(flat_dim_31615 + num_threads_33212 - 1,
                                         num_threads_33212)), '\n');
        
        unsigned int shared_sizze_33761 = sizeof(int32_t) * num_groups_32196;
        unsigned int shared_sizze_33763 = sizeof(bool) * num_groups_32196;
        CUdeviceptr kernel_arg_33765 = mem_32821.mem;
        CUdeviceptr kernel_arg_33766 = mem_32823.mem;
        unsigned int shared_offset_33762 = 0;
        unsigned int shared_offset_33764 = 0 + (shared_sizze_33761 + (8 -
                                                                      shared_sizze_33761 %
                                                                      8) % 8);
        
        if ((((((1 && 1 != 0) && 1 != 0) && 1 != 0) && num_groups_32196 != 0) &&
             1 != 0) && 1 != 0) {
            int perm[3] = {0, 1, 2};
            
            if (1 > 1 << 16) {
                perm[1] = perm[0];
                perm[0] = 1;
            }
            if (1 > 1 << 16) {
                perm[2] = perm[0];
                perm[0] = 2;
            }
            
            size_t grid[3];
            
            grid[perm[0]] = 1;
            grid[perm[1]] = 1;
            grid[perm[2]] = 1;
            
            void *kernel_args_33758[] = {&shared_offset_33762,
                                         &shared_offset_33764, &flat_dim_31615,
                                         &num_groups_32196, &kernel_arg_33765,
                                         &kernel_arg_33766, &num_threads_33212};
            int64_t time_start_33759 = 0, time_end_33760 = 0;
            
            if (ctx->debugging) {
                fprintf(stderr, "Launching %s with grid size (",
                        "scan_stage2_32199");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ") and block size (");
                fprintf(stderr, "%d", num_groups_32196);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ").\n");
                time_start_33759 = get_wall_time();
            }
            CUDA_SUCCEED(cuLaunchKernel(ctx->scan_stage2_32199, grid[0],
                                        grid[1], grid[2], num_groups_32196, 1,
                                        1, 0 + (shared_sizze_33761 + (8 -
                                                                      shared_sizze_33761 %
                                                                      8) % 8) +
                                        (shared_sizze_33763 + (8 -
                                                               shared_sizze_33763 %
                                                               8) % 8), NULL,
                                        kernel_args_33758, NULL));
            if (ctx->debugging) {
                CUDA_SUCCEED(cuCtxSynchronize());
                time_end_33760 = get_wall_time();
                fprintf(stderr, "Kernel %s runtime: %ldus\n",
                        "scan_stage2_32199", time_end_33760 - time_start_33759);
            }
        }
        
        int32_t group_sizze_33277;
        
        group_sizze_33277 = ctx->sizes.quicksort1zigroup_sizze_33277;
        
        int32_t num_groups_33278;
        
        num_groups_33278 = squot32(flat_dim_31615 +
                                   sext_i32_i32(group_sizze_33277) - 1,
                                   sext_i32_i32(group_sizze_33277));
        
        CUdeviceptr kernel_arg_33770 = mem_32821.mem;
        CUdeviceptr kernel_arg_33771 = mem_32823.mem;
        
        if ((((((1 && num_groups_33278 != 0) && 1 != 0) && 1 != 0) &&
              group_sizze_33277 != 0) && 1 != 0) && 1 != 0) {
            int perm[3] = {0, 1, 2};
            
            if (1 > 1 << 16) {
                perm[1] = perm[0];
                perm[0] = 1;
            }
            if (1 > 1 << 16) {
                perm[2] = perm[0];
                perm[0] = 2;
            }
            
            size_t grid[3];
            
            grid[perm[0]] = num_groups_33278;
            grid[perm[1]] = 1;
            grid[perm[2]] = 1;
            
            void *kernel_args_33767[] = {&flat_dim_31615, &kernel_arg_33770,
                                         &kernel_arg_33771, &num_threads_33212};
            int64_t time_start_33768 = 0, time_end_33769 = 0;
            
            if (ctx->debugging) {
                fprintf(stderr, "Launching %s with grid size (",
                        "scan_stage3_33274");
                fprintf(stderr, "%d", num_groups_33278);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ") and block size (");
                fprintf(stderr, "%d", group_sizze_33277);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ").\n");
                time_start_33768 = get_wall_time();
            }
            CUDA_SUCCEED(cuLaunchKernel(ctx->scan_stage3_33274, grid[0],
                                        grid[1], grid[2], group_sizze_33277, 1,
                                        1, 0, NULL, kernel_args_33767, NULL));
            if (ctx->debugging) {
                CUDA_SUCCEED(cuCtxSynchronize());
                time_end_33769 = get_wall_time();
                fprintf(stderr, "Kernel %s runtime: %ldus\n",
                        "scan_stage3_33274", time_end_33769 - time_start_33768);
            }
        }
        
        CUdeviceptr kernel_arg_33775 = mem_32821.mem;
        CUdeviceptr kernel_arg_33776 = mem_32826.mem;
        
        if ((((((1 && segmap_usable_groups_32211 != 0) && 1 != 0) && 1 != 0) &&
              segmap_group_sizze_32205 != 0) && 1 != 0) && 1 != 0) {
            int perm[3] = {0, 1, 2};
            
            if (1 > 1 << 16) {
                perm[1] = perm[0];
                perm[0] = 1;
            }
            if (1 > 1 << 16) {
                perm[2] = perm[0];
                perm[0] = 2;
            }
            
            size_t grid[3];
            
            grid[perm[0]] = segmap_usable_groups_32211;
            grid[perm[1]] = 1;
            grid[perm[2]] = 1;
            
            void *kernel_args_33772[] = {&flat_dim_31615, &kernel_arg_33775,
                                         &kernel_arg_33776};
            int64_t time_start_33773 = 0, time_end_33774 = 0;
            
            if (ctx->debugging) {
                fprintf(stderr, "Launching %s with grid size (",
                        "segmap_32201");
                fprintf(stderr, "%d", segmap_usable_groups_32211);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ") and block size (");
                fprintf(stderr, "%d", segmap_group_sizze_32205);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ").\n");
                time_start_33773 = get_wall_time();
            }
            CUDA_SUCCEED(cuLaunchKernel(ctx->segmap_32201, grid[0], grid[1],
                                        grid[2], segmap_group_sizze_32205, 1, 1,
                                        0, NULL, kernel_args_33772, NULL));
            if (ctx->debugging) {
                CUDA_SUCCEED(cuCtxSynchronize());
                time_end_33774 = get_wall_time();
                fprintf(stderr, "Kernel %s runtime: %ldus\n", "segmap_32201",
                        time_end_33774 - time_start_33773);
            }
        }
        
        int32_t num_threads_33286 = num_groups_32222 *
                segscan_group_sizze_32220;
        unsigned int shared_sizze_33780 = sizeof(int32_t) *
                     segscan_group_sizze_32220;
        unsigned int shared_sizze_33782 = sizeof(bool) *
                     segscan_group_sizze_32220;
        CUdeviceptr kernel_arg_33784 = x_mem_32791.mem;
        CUdeviceptr kernel_arg_33785 = mem_32826.mem;
        CUdeviceptr kernel_arg_33786 = mem_32830.mem;
        CUdeviceptr kernel_arg_33787 = mem_32832.mem;
        unsigned int shared_offset_33781 = 0;
        unsigned int shared_offset_33783 = 0 + (shared_sizze_33780 + (8 -
                                                                      shared_sizze_33780 %
                                                                      8) % 8);
        
        if ((((((1 && num_groups_32222 != 0) && 1 != 0) && 1 != 0) &&
              segscan_group_sizze_32220 != 0) && 1 != 0) && 1 != 0) {
            int perm[3] = {0, 1, 2};
            
            if (1 > 1 << 16) {
                perm[1] = perm[0];
                perm[0] = 1;
            }
            if (1 > 1 << 16) {
                perm[2] = perm[0];
                perm[0] = 2;
            }
            
            size_t grid[3];
            
            grid[perm[0]] = num_groups_32222;
            grid[perm[1]] = 1;
            grid[perm[2]] = 1;
            
            void *kernel_args_33777[] = {&shared_offset_33781,
                                         &shared_offset_33783, &flat_dim_31615,
                                         &m_31645, &i_p_m_t_s_31665,
                                         &kernel_arg_33784, &kernel_arg_33785,
                                         &kernel_arg_33786, &kernel_arg_33787,
                                         &num_threads_33286};
            int64_t time_start_33778 = 0, time_end_33779 = 0;
            
            if (ctx->debugging) {
                fprintf(stderr, "Launching %s with grid size (",
                        "scan_stage1_32225");
                fprintf(stderr, "%d", num_groups_32222);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ") and block size (");
                fprintf(stderr, "%d", segscan_group_sizze_32220);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ").\n");
                time_start_33778 = get_wall_time();
            }
            CUDA_SUCCEED(cuLaunchKernel(ctx->scan_stage1_32225, grid[0],
                                        grid[1], grid[2],
                                        segscan_group_sizze_32220, 1, 1, 0 +
                                        (shared_sizze_33780 + (8 -
                                                               shared_sizze_33780 %
                                                               8) % 8) +
                                        (shared_sizze_33782 + (8 -
                                                               shared_sizze_33782 %
                                                               8) % 8), NULL,
                                        kernel_args_33777, NULL));
            if (ctx->debugging) {
                CUDA_SUCCEED(cuCtxSynchronize());
                time_end_33779 = get_wall_time();
                fprintf(stderr, "Kernel %s runtime: %ldus\n",
                        "scan_stage1_32225", time_end_33779 - time_start_33778);
            }
        }
        if (ctx->debugging)
            fprintf(stderr, "%s\n", "\n# SegScan");
        if (ctx->debugging)
            fprintf(stderr, "%s: %llu%c", "elems_per_group",
                    (long long) (segscan_group_sizze_32220 *
                                 squot32(flat_dim_31615 + num_threads_33286 - 1,
                                         num_threads_33286)), '\n');
        
        unsigned int shared_sizze_33791 = sizeof(int32_t) * num_groups_32222;
        unsigned int shared_sizze_33793 = sizeof(bool) * num_groups_32222;
        CUdeviceptr kernel_arg_33795 = mem_32830.mem;
        CUdeviceptr kernel_arg_33796 = mem_32832.mem;
        unsigned int shared_offset_33792 = 0;
        unsigned int shared_offset_33794 = 0 + (shared_sizze_33791 + (8 -
                                                                      shared_sizze_33791 %
                                                                      8) % 8);
        
        if ((((((1 && 1 != 0) && 1 != 0) && 1 != 0) && num_groups_32222 != 0) &&
             1 != 0) && 1 != 0) {
            int perm[3] = {0, 1, 2};
            
            if (1 > 1 << 16) {
                perm[1] = perm[0];
                perm[0] = 1;
            }
            if (1 > 1 << 16) {
                perm[2] = perm[0];
                perm[0] = 2;
            }
            
            size_t grid[3];
            
            grid[perm[0]] = 1;
            grid[perm[1]] = 1;
            grid[perm[2]] = 1;
            
            void *kernel_args_33788[] = {&shared_offset_33792,
                                         &shared_offset_33794, &flat_dim_31615,
                                         &num_groups_32222, &kernel_arg_33795,
                                         &kernel_arg_33796, &num_threads_33286};
            int64_t time_start_33789 = 0, time_end_33790 = 0;
            
            if (ctx->debugging) {
                fprintf(stderr, "Launching %s with grid size (",
                        "scan_stage2_32225");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ") and block size (");
                fprintf(stderr, "%d", num_groups_32222);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ").\n");
                time_start_33789 = get_wall_time();
            }
            CUDA_SUCCEED(cuLaunchKernel(ctx->scan_stage2_32225, grid[0],
                                        grid[1], grid[2], num_groups_32222, 1,
                                        1, 0 + (shared_sizze_33791 + (8 -
                                                                      shared_sizze_33791 %
                                                                      8) % 8) +
                                        (shared_sizze_33793 + (8 -
                                                               shared_sizze_33793 %
                                                               8) % 8), NULL,
                                        kernel_args_33788, NULL));
            if (ctx->debugging) {
                CUDA_SUCCEED(cuCtxSynchronize());
                time_end_33790 = get_wall_time();
                fprintf(stderr, "Kernel %s runtime: %ldus\n",
                        "scan_stage2_32225", time_end_33790 - time_start_33789);
            }
        }
        
        int32_t group_sizze_33346;
        
        group_sizze_33346 = ctx->sizes.quicksort1zigroup_sizze_33346;
        
        int32_t num_groups_33347;
        
        num_groups_33347 = squot32(flat_dim_31615 +
                                   sext_i32_i32(group_sizze_33346) - 1,
                                   sext_i32_i32(group_sizze_33346));
        
        CUdeviceptr kernel_arg_33800 = mem_32830.mem;
        CUdeviceptr kernel_arg_33801 = mem_32832.mem;
        
        if ((((((1 && num_groups_33347 != 0) && 1 != 0) && 1 != 0) &&
              group_sizze_33346 != 0) && 1 != 0) && 1 != 0) {
            int perm[3] = {0, 1, 2};
            
            if (1 > 1 << 16) {
                perm[1] = perm[0];
                perm[0] = 1;
            }
            if (1 > 1 << 16) {
                perm[2] = perm[0];
                perm[0] = 2;
            }
            
            size_t grid[3];
            
            grid[perm[0]] = num_groups_33347;
            grid[perm[1]] = 1;
            grid[perm[2]] = 1;
            
            void *kernel_args_33797[] = {&flat_dim_31615, &kernel_arg_33800,
                                         &kernel_arg_33801, &num_threads_33286};
            int64_t time_start_33798 = 0, time_end_33799 = 0;
            
            if (ctx->debugging) {
                fprintf(stderr, "Launching %s with grid size (",
                        "scan_stage3_33343");
                fprintf(stderr, "%d", num_groups_33347);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ") and block size (");
                fprintf(stderr, "%d", group_sizze_33346);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ").\n");
                time_start_33798 = get_wall_time();
            }
            CUDA_SUCCEED(cuLaunchKernel(ctx->scan_stage3_33343, grid[0],
                                        grid[1], grid[2], group_sizze_33346, 1,
                                        1, 0, NULL, kernel_args_33797, NULL));
            if (ctx->debugging) {
                CUDA_SUCCEED(cuCtxSynchronize());
                time_end_33799 = get_wall_time();
                fprintf(stderr, "Kernel %s runtime: %ldus\n",
                        "scan_stage3_33343", time_end_33799 - time_start_33798);
            }
        }
        
        CUdeviceptr kernel_arg_33805 = mem_32830.mem;
        CUdeviceptr kernel_arg_33806 = mem_32832.mem;
        CUdeviceptr kernel_arg_33807 = mem_32835.mem;
        
        if ((((((1 && segmap_usable_groups_32237 != 0) && 1 != 0) && 1 != 0) &&
              segmap_group_sizze_32231 != 0) && 1 != 0) && 1 != 0) {
            int perm[3] = {0, 1, 2};
            
            if (1 > 1 << 16) {
                perm[1] = perm[0];
                perm[0] = 1;
            }
            if (1 > 1 << 16) {
                perm[2] = perm[0];
                perm[0] = 2;
            }
            
            size_t grid[3];
            
            grid[perm[0]] = segmap_usable_groups_32237;
            grid[perm[1]] = 1;
            grid[perm[2]] = 1;
            
            void *kernel_args_33802[] = {&m_31645, &n_31675, &kernel_arg_33805,
                                         &kernel_arg_33806, &kernel_arg_33807};
            int64_t time_start_33803 = 0, time_end_33804 = 0;
            
            if (ctx->debugging) {
                fprintf(stderr, "Launching %s with grid size (",
                        "segmap_32227");
                fprintf(stderr, "%d", segmap_usable_groups_32237);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ") and block size (");
                fprintf(stderr, "%d", segmap_group_sizze_32231);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ").\n");
                time_start_33803 = get_wall_time();
            }
            CUDA_SUCCEED(cuLaunchKernel(ctx->segmap_32227, grid[0], grid[1],
                                        grid[2], segmap_group_sizze_32231, 1, 1,
                                        0, NULL, kernel_args_33802, NULL));
            if (ctx->debugging) {
                CUDA_SUCCEED(cuCtxSynchronize());
                time_end_33804 = get_wall_time();
                fprintf(stderr, "Kernel %s runtime: %ldus\n", "segmap_32227",
                        time_end_33804 - time_start_33803);
            }
        }
        
        CUdeviceptr kernel_arg_33811 = mem_32835.mem;
        CUdeviceptr kernel_arg_33812 = mem_32838.mem;
        
        if ((((((1 && segmap_usable_groups_32252 != 0) && 1 != 0) && 1 != 0) &&
              segmap_group_sizze_32246 != 0) && 1 != 0) && 1 != 0) {
            int perm[3] = {0, 1, 2};
            
            if (1 > 1 << 16) {
                perm[1] = perm[0];
                perm[0] = 1;
            }
            if (1 > 1 << 16) {
                perm[2] = perm[0];
                perm[0] = 2;
            }
            
            size_t grid[3];
            
            grid[perm[0]] = segmap_usable_groups_32252;
            grid[perm[1]] = 1;
            grid[perm[2]] = 1;
            
            void *kernel_args_33808[] = {&flat_dim_31615, &kernel_arg_33811,
                                         &kernel_arg_33812};
            int64_t time_start_33809 = 0, time_end_33810 = 0;
            
            if (ctx->debugging) {
                fprintf(stderr, "Launching %s with grid size (",
                        "segmap_32242");
                fprintf(stderr, "%d", segmap_usable_groups_32252);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ") and block size (");
                fprintf(stderr, "%d", segmap_group_sizze_32246);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ").\n");
                time_start_33809 = get_wall_time();
            }
            CUDA_SUCCEED(cuLaunchKernel(ctx->segmap_32242, grid[0], grid[1],
                                        grid[2], segmap_group_sizze_32246, 1, 1,
                                        0, NULL, kernel_args_33808, NULL));
            if (ctx->debugging) {
                CUDA_SUCCEED(cuCtxSynchronize());
                time_end_33810 = get_wall_time();
                fprintf(stderr, "Kernel %s runtime: %ldus\n", "segmap_32242",
                        time_end_33810 - time_start_33809);
            }
        }
        
        struct memblock_device mem_32841;
        
        mem_32841.references = NULL;
        if (memblock_alloc_device(ctx, &mem_32841, bytes_32793, "mem_32841"))
            return 1;
        
        int call_ret_33813 = futrts__replicate_i32(ctx, mem_32841,
                                                   flat_dim_31615, 0);
        
        assert(call_ret_33813 == 0);
        
        CUdeviceptr kernel_arg_33817 = x_mem_32791.mem;
        CUdeviceptr kernel_arg_33818 = mem_32811.mem;
        CUdeviceptr kernel_arg_33819 = mem_32838.mem;
        CUdeviceptr kernel_arg_33820 = mem_32844.mem;
        
        if ((((((1 && segmap_usable_groups_32266 != 0) && 1 != 0) && 1 != 0) &&
              segmap_group_sizze_32260 != 0) && 1 != 0) && 1 != 0) {
            int perm[3] = {0, 1, 2};
            
            if (1 > 1 << 16) {
                perm[1] = perm[0];
                perm[0] = 1;
            }
            if (1 > 1 << 16) {
                perm[2] = perm[0];
                perm[0] = 2;
            }
            
            size_t grid[3];
            
            grid[perm[0]] = segmap_usable_groups_32266;
            grid[perm[1]] = 1;
            grid[perm[2]] = 1;
            
            void *kernel_args_33814[] = {&flat_dim_31615, &kernel_arg_33817,
                                         &kernel_arg_33818, &kernel_arg_33819,
                                         &kernel_arg_33820};
            int64_t time_start_33815 = 0, time_end_33816 = 0;
            
            if (ctx->debugging) {
                fprintf(stderr, "Launching %s with grid size (",
                        "segmap_32256");
                fprintf(stderr, "%d", segmap_usable_groups_32266);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ") and block size (");
                fprintf(stderr, "%d", segmap_group_sizze_32260);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ").\n");
                time_start_33815 = get_wall_time();
            }
            CUDA_SUCCEED(cuLaunchKernel(ctx->segmap_32256, grid[0], grid[1],
                                        grid[2], segmap_group_sizze_32260, 1, 1,
                                        0, NULL, kernel_args_33814, NULL));
            if (ctx->debugging) {
                CUDA_SUCCEED(cuCtxSynchronize());
                time_end_33816 = get_wall_time();
                fprintf(stderr, "Kernel %s runtime: %ldus\n", "segmap_32256",
                        time_end_33816 - time_start_33815);
            }
        }
        
        struct memblock_device mem_32846;
        
        mem_32846.references = NULL;
        if (memblock_alloc_device(ctx, &mem_32846, bytes_32785, "mem_32846"))
            return 1;
        CUDA_SUCCEED(cuMemcpy(mem_32846.mem + 0, x_mem_32791.mem + 0,
                              fill_arg_31617 * sizeof(bool)));
        
        CUdeviceptr kernel_arg_33824 = mem_32844.mem;
        CUdeviceptr kernel_arg_33825 = mem_32846.mem;
        
        if ((((((1 && segmap_usable_groups_32284 != 0) && 1 != 0) && 1 != 0) &&
              segmap_group_sizze_32278 != 0) && 1 != 0) && 1 != 0) {
            int perm[3] = {0, 1, 2};
            
            if (1 > 1 << 16) {
                perm[1] = perm[0];
                perm[0] = 1;
            }
            if (1 > 1 << 16) {
                perm[2] = perm[0];
                perm[0] = 2;
            }
            
            size_t grid[3];
            
            grid[perm[0]] = segmap_usable_groups_32284;
            grid[perm[1]] = 1;
            grid[perm[2]] = 1;
            
            void *kernel_args_33821[] = {&flat_dim_31615, &fill_arg_31617,
                                         &kernel_arg_33824, &kernel_arg_33825};
            int64_t time_start_33822 = 0, time_end_33823 = 0;
            
            if (ctx->debugging) {
                fprintf(stderr, "Launching %s with grid size (",
                        "segmap_32274");
                fprintf(stderr, "%d", segmap_usable_groups_32284);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ") and block size (");
                fprintf(stderr, "%d", segmap_group_sizze_32278);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ").\n");
                time_start_33822 = get_wall_time();
            }
            CUDA_SUCCEED(cuLaunchKernel(ctx->segmap_32274, grid[0], grid[1],
                                        grid[2], segmap_group_sizze_32278, 1, 1,
                                        0, NULL, kernel_args_33821, NULL));
            if (ctx->debugging) {
                CUDA_SUCCEED(cuCtxSynchronize());
                time_end_33823 = get_wall_time();
                fprintf(stderr, "Kernel %s runtime: %ldus\n", "segmap_32274",
                        time_end_33823 - time_start_33822);
            }
        }
        
        CUdeviceptr kernel_arg_33829 = x_mem_32790.mem;
        CUdeviceptr kernel_arg_33830 = x_mem_32791.mem;
        CUdeviceptr kernel_arg_33831 = mem_32803.mem;
        CUdeviceptr kernel_arg_33832 = mem_32805.mem;
        CUdeviceptr kernel_arg_33833 = mem_32808.mem;
        CUdeviceptr kernel_arg_33834 = mem_32815.mem;
        CUdeviceptr kernel_arg_33835 = mem_32826.mem;
        CUdeviceptr kernel_arg_33836 = mem_32830.mem;
        CUdeviceptr kernel_arg_33837 = mem_32832.mem;
        CUdeviceptr kernel_arg_33838 = mem_32841.mem;
        CUdeviceptr kernel_arg_33839 = mem_32846.mem;
        
        if ((((((1 && segmap_usable_groups_32296 != 0) && 1 != 0) && 1 != 0) &&
              segmap_group_sizze_32290 != 0) && 1 != 0) && 1 != 0) {
            int perm[3] = {0, 1, 2};
            
            if (1 > 1 << 16) {
                perm[1] = perm[0];
                perm[0] = 1;
            }
            if (1 > 1 << 16) {
                perm[2] = perm[0];
                perm[0] = 2;
            }
            
            size_t grid[3];
            
            grid[perm[0]] = segmap_usable_groups_32296;
            grid[perm[1]] = 1;
            grid[perm[2]] = 1;
            
            void *kernel_args_33826[] = {&flat_dim_31615, &fill_arg_31617,
                                         &m_31645, &kernel_arg_33829,
                                         &kernel_arg_33830, &kernel_arg_33831,
                                         &kernel_arg_33832, &kernel_arg_33833,
                                         &kernel_arg_33834, &kernel_arg_33835,
                                         &kernel_arg_33836, &kernel_arg_33837,
                                         &kernel_arg_33838, &kernel_arg_33839};
            int64_t time_start_33827 = 0, time_end_33828 = 0;
            
            if (ctx->debugging) {
                fprintf(stderr, "Launching %s with grid size (",
                        "segmap_32286");
                fprintf(stderr, "%d", segmap_usable_groups_32296);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ") and block size (");
                fprintf(stderr, "%d", segmap_group_sizze_32290);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ").\n");
                time_start_33827 = get_wall_time();
            }
            CUDA_SUCCEED(cuLaunchKernel(ctx->segmap_32286, grid[0], grid[1],
                                        grid[2], segmap_group_sizze_32290, 1, 1,
                                        0, NULL, kernel_args_33826, NULL));
            if (ctx->debugging) {
                CUDA_SUCCEED(cuCtxSynchronize());
                time_end_33828 = get_wall_time();
                fprintf(stderr, "Kernel %s runtime: %ldus\n", "segmap_32286",
                        time_end_33828 - time_start_33827);
            }
        }
        
        struct memblock_device counter_mem_33384 = ctx->counter_mem_33384;
        struct memblock_device group_res_arr_mem_33386;
        
        group_res_arr_mem_33386.references = NULL;
        if (memblock_alloc_device(ctx, &group_res_arr_mem_33386, sizeof(bool) *
                                  (segred_group_sizze_32300 * num_groups_32302),
                                  "group_res_arr_mem_33386"))
            return 1;
        
        int32_t num_threads_33388 = num_groups_32302 * segred_group_sizze_32300;
        unsigned int shared_sizze_33844 = sizeof(bool);
        unsigned int shared_sizze_33846 = sizeof(bool) *
                     segred_group_sizze_32300;
        CUdeviceptr kernel_arg_33848 = mem_32846.mem;
        CUdeviceptr kernel_arg_33849 = mem_32849.mem;
        CUdeviceptr kernel_arg_33850 = counter_mem_33384.mem;
        CUdeviceptr kernel_arg_33851 = group_res_arr_mem_33386.mem;
        unsigned int shared_offset_33845 = 0;
        unsigned int shared_offset_33847 = 0 + (shared_sizze_33844 + (8 -
                                                                      shared_sizze_33844 %
                                                                      8) % 8);
        
        if ((((((1 && num_groups_32302 != 0) && 1 != 0) && 1 != 0) &&
              segred_group_sizze_32300 != 0) && 1 != 0) && 1 != 0) {
            int perm[3] = {0, 1, 2};
            
            if (1 > 1 << 16) {
                perm[1] = perm[0];
                perm[0] = 1;
            }
            if (1 > 1 << 16) {
                perm[2] = perm[0];
                perm[0] = 2;
            }
            
            size_t grid[3];
            
            grid[perm[0]] = num_groups_32302;
            grid[perm[1]] = 1;
            grid[perm[2]] = 1;
            
            void *kernel_args_33841[] = {&shared_offset_33845,
                                         &shared_offset_33847, &fill_arg_31617,
                                         &num_groups_32302, &kernel_arg_33848,
                                         &kernel_arg_33849, &kernel_arg_33850,
                                         &kernel_arg_33851, &num_threads_33388};
            int64_t time_start_33842 = 0, time_end_33843 = 0;
            
            if (ctx->debugging) {
                fprintf(stderr, "Launching %s with grid size (",
                        "segred_nonseg_32307");
                fprintf(stderr, "%d", num_groups_32302);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ") and block size (");
                fprintf(stderr, "%d", segred_group_sizze_32300);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ", ");
                fprintf(stderr, "%d", 1);
                fprintf(stderr, ").\n");
                time_start_33842 = get_wall_time();
            }
            CUDA_SUCCEED(cuLaunchKernel(ctx->segred_nonseg_32307, grid[0],
                                        grid[1], grid[2],
                                        segred_group_sizze_32300, 1, 1, 0 +
                                        (shared_sizze_33844 + (8 -
                                                               shared_sizze_33844 %
                                                               8) % 8) +
                                        (shared_sizze_33846 + (8 -
                                                               shared_sizze_33846 %
                                                               8) % 8), NULL,
                                        kernel_args_33841, NULL));
            if (ctx->debugging) {
                CUDA_SUCCEED(cuCtxSynchronize());
                time_end_33843 = get_wall_time();
                fprintf(stderr, "Kernel %s runtime: %ldus\n",
                        "segred_nonseg_32307", time_end_33843 -
                        time_start_33842);
            }
        }
        
        bool read_res_33852;
        
        CUDA_SUCCEED(cuMemcpyDtoH(&read_res_33852, mem_32849.mem + 0 *
                                  sizeof(bool), sizeof(bool)));
        
        bool res_31893 = read_res_33852;
        bool res_31898 = !res_31893;
        struct memblock_device x_mem_tmp_33000;
        
        x_mem_tmp_33000.references = NULL;
        if (memblock_set_device(ctx, &x_mem_tmp_33000, &mem_32841,
                                "mem_32841") != 0)
            return 1;
        
        struct memblock_device x_mem_tmp_33001;
        
        x_mem_tmp_33001.references = NULL;
        if (memblock_set_device(ctx, &x_mem_tmp_33001, &mem_32846,
                                "mem_32846") != 0)
            return 1;
        
        bool loop_while_tmp_33002;
        
        loop_while_tmp_33002 = res_31898;
        if (memblock_set_device(ctx, &x_mem_32790, &x_mem_tmp_33000,
                                "x_mem_tmp_33000") != 0)
            return 1;
        if (memblock_set_device(ctx, &x_mem_32791, &x_mem_tmp_33001,
                                "x_mem_tmp_33001") != 0)
            return 1;
        loop_while_31702 = loop_while_tmp_33002;
        if (memblock_unref_device(ctx, &x_mem_tmp_33001, "x_mem_tmp_33001") !=
            0)
            return 1;
        if (memblock_unref_device(ctx, &x_mem_tmp_33000, "x_mem_tmp_33000") !=
            0)
            return 1;
        if (memblock_unref_device(ctx, &group_res_arr_mem_33386,
                                  "group_res_arr_mem_33386") != 0)
            return 1;
        if (memblock_unref_device(ctx, &mem_32846, "mem_32846") != 0)
            return 1;
        if (memblock_unref_device(ctx, &mem_32841, "mem_32841") != 0)
            return 1;
    }
    if (memblock_set_device(ctx, &res_mem_32850, &x_mem_32790, "x_mem_32790") !=
        0)
        return 1;
    if (memblock_set_device(ctx, &res_mem_32851, &x_mem_32791, "x_mem_32791") !=
        0)
        return 1;
    res_31699 = loop_while_31702;
    if (memblock_unref_device(ctx, &mem_32786, "mem_32786") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32795, "mem_32795") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32797, "mem_32797") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32801, "mem_32801") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32803, "mem_32803") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32805, "mem_32805") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32808, "mem_32808") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32811, "mem_32811") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32815, "mem_32815") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32817, "mem_32817") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32821, "mem_32821") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32823, "mem_32823") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32826, "mem_32826") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32830, "mem_32830") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32832, "mem_32832") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32835, "mem_32835") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32838, "mem_32838") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32844, "mem_32844") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32849, "mem_32849") != 0)
        return 1;
    
    bool assert_arg_31899 = n_31612 == flat_dim_31615;
    bool dim_ok_31900;
    
    if (!assert_arg_31899) {
        ctx->error = msgprintf("Error at\n%s\n%s\n",
                               " |-> Futhark/quicksort.fut:211:1-212:45\n |-> Futhark/quicksort.fut:212:3-45\n |-> /futlib/functional.fut:12:39-41\n `-> /futlib/array.fut:87:3-33\n",
                               "new shape has different number of elements than old shape");
        if (memblock_unref_device(ctx, &x_mem_32791, "x_mem_32791") != 0)
            return 1;
        if (memblock_unref_device(ctx, &x_mem_32790, "x_mem_32790") != 0)
            return 1;
        if (memblock_unref_device(ctx, &res_mem_32851, "res_mem_32851") != 0)
            return 1;
        if (memblock_unref_device(ctx, &res_mem_32850, "res_mem_32850") != 0)
            return 1;
        if (memblock_unref_device(ctx, &mem_32849, "mem_32849") != 0)
            return 1;
        if (memblock_unref_device(ctx, &mem_32844, "mem_32844") != 0)
            return 1;
        if (memblock_unref_device(ctx, &mem_32838, "mem_32838") != 0)
            return 1;
        if (memblock_unref_device(ctx, &mem_32835, "mem_32835") != 0)
            return 1;
        if (memblock_unref_device(ctx, &mem_32832, "mem_32832") != 0)
            return 1;
        if (memblock_unref_device(ctx, &mem_32830, "mem_32830") != 0)
            return 1;
        if (memblock_unref_device(ctx, &mem_32826, "mem_32826") != 0)
            return 1;
        if (memblock_unref_device(ctx, &mem_32823, "mem_32823") != 0)
            return 1;
        if (memblock_unref_device(ctx, &mem_32821, "mem_32821") != 0)
            return 1;
        if (memblock_unref_device(ctx, &mem_32817, "mem_32817") != 0)
            return 1;
        if (memblock_unref_device(ctx, &mem_32815, "mem_32815") != 0)
            return 1;
        if (memblock_unref_device(ctx, &mem_32811, "mem_32811") != 0)
            return 1;
        if (memblock_unref_device(ctx, &mem_32808, "mem_32808") != 0)
            return 1;
        if (memblock_unref_device(ctx, &mem_32805, "mem_32805") != 0)
            return 1;
        if (memblock_unref_device(ctx, &mem_32803, "mem_32803") != 0)
            return 1;
        if (memblock_unref_device(ctx, &mem_32801, "mem_32801") != 0)
            return 1;
        if (memblock_unref_device(ctx, &mem_32797, "mem_32797") != 0)
            return 1;
        if (memblock_unref_device(ctx, &mem_32795, "mem_32795") != 0)
            return 1;
        if (memblock_unref_device(ctx, &group_res_arr_mem_32970,
                                  "group_res_arr_mem_32970") != 0)
            return 1;
        if (memblock_unref_device(ctx, &mem_32789, "mem_32789") != 0)
            return 1;
        if (memblock_unref_device(ctx, &mem_32786, "mem_32786") != 0)
            return 1;
        if (memblock_unref_device(ctx, &out_mem_32955, "out_mem_32955") != 0)
            return 1;
        return 1;
    }
    out_arrsizze_32956 = 1;
    out_arrsizze_32957 = n_31612;
    if (memblock_set_device(ctx, &out_mem_32955, &res_mem_32850,
                            "res_mem_32850") != 0)
        return 1;
    (*out_mem_p_33638).references = NULL;
    if (memblock_set_device(ctx, &*out_mem_p_33638, &out_mem_32955,
                            "out_mem_32955") != 0)
        return 1;
    *out_out_arrsizze_33639 = out_arrsizze_32956;
    *out_out_arrsizze_33640 = out_arrsizze_32957;
    if (memblock_unref_device(ctx, &x_mem_32791, "x_mem_32791") != 0)
        return 1;
    if (memblock_unref_device(ctx, &x_mem_32790, "x_mem_32790") != 0)
        return 1;
    if (memblock_unref_device(ctx, &res_mem_32851, "res_mem_32851") != 0)
        return 1;
    if (memblock_unref_device(ctx, &res_mem_32850, "res_mem_32850") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32849, "mem_32849") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32844, "mem_32844") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32838, "mem_32838") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32835, "mem_32835") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32832, "mem_32832") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32830, "mem_32830") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32826, "mem_32826") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32823, "mem_32823") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32821, "mem_32821") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32817, "mem_32817") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32815, "mem_32815") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32811, "mem_32811") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32808, "mem_32808") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32805, "mem_32805") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32803, "mem_32803") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32801, "mem_32801") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32797, "mem_32797") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32795, "mem_32795") != 0)
        return 1;
    if (memblock_unref_device(ctx, &group_res_arr_mem_32970,
                              "group_res_arr_mem_32970") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32789, "mem_32789") != 0)
        return 1;
    if (memblock_unref_device(ctx, &mem_32786, "mem_32786") != 0)
        return 1;
    if (memblock_unref_device(ctx, &out_mem_32955, "out_mem_32955") != 0)
        return 1;
    return 0;
}
static int futrts__replicate_i32(struct futhark_context *ctx,
                                 struct memblock_device mem_33360,
                                 int32_t num_elems_33361, int32_t val_33362)
{
    int32_t group_sizze_33367;
    
    group_sizze_33367 = ctx->sizes.quicksort1zigroup_sizze_33367;
    
    int32_t num_groups_33368;
    
    num_groups_33368 = squot32(num_elems_33361 +
                               sext_i32_i32(group_sizze_33367) - 1,
                               sext_i32_i32(group_sizze_33367));
    
    CUdeviceptr kernel_arg_33856 = mem_33360.mem;
    
    if ((((((1 && num_groups_33368 != 0) && 1 != 0) && 1 != 0) &&
          group_sizze_33367 != 0) && 1 != 0) && 1 != 0) {
        int perm[3] = {0, 1, 2};
        
        if (1 > 1 << 16) {
            perm[1] = perm[0];
            perm[0] = 1;
        }
        if (1 > 1 << 16) {
            perm[2] = perm[0];
            perm[0] = 2;
        }
        
        size_t grid[3];
        
        grid[perm[0]] = num_groups_33368;
        grid[perm[1]] = 1;
        grid[perm[2]] = 1;
        
        void *kernel_args_33853[] = {&kernel_arg_33856, &num_elems_33361,
                                     &val_33362};
        int64_t time_start_33854 = 0, time_end_33855 = 0;
        
        if (ctx->debugging) {
            fprintf(stderr, "Launching %s with grid size (", "replicate_33364");
            fprintf(stderr, "%d", num_groups_33368);
            fprintf(stderr, ", ");
            fprintf(stderr, "%d", 1);
            fprintf(stderr, ", ");
            fprintf(stderr, "%d", 1);
            fprintf(stderr, ") and block size (");
            fprintf(stderr, "%d", group_sizze_33367);
            fprintf(stderr, ", ");
            fprintf(stderr, "%d", 1);
            fprintf(stderr, ", ");
            fprintf(stderr, "%d", 1);
            fprintf(stderr, ").\n");
            time_start_33854 = get_wall_time();
        }
        CUDA_SUCCEED(cuLaunchKernel(ctx->replicate_33364, grid[0], grid[1],
                                    grid[2], group_sizze_33367, 1, 1, 0, NULL,
                                    kernel_args_33853, NULL));
        if (ctx->debugging) {
            CUDA_SUCCEED(cuCtxSynchronize());
            time_end_33855 = get_wall_time();
            fprintf(stderr, "Kernel %s runtime: %ldus\n", "replicate_33364",
                    time_end_33855 - time_start_33854);
        }
    }
    return 0;
}
static int futrts__replicate_bool(struct futhark_context *ctx,
                                  struct memblock_device mem_32958,
                                  int32_t num_elems_32959, bool val_32960)
{
    int32_t group_sizze_32965;
    
    group_sizze_32965 = ctx->sizes.quicksort1zigroup_sizze_32965;
    
    int32_t num_groups_32966;
    
    num_groups_32966 = squot32(num_elems_32959 +
                               sext_i32_i32(group_sizze_32965) - 1,
                               sext_i32_i32(group_sizze_32965));
    
    CUdeviceptr kernel_arg_33860 = mem_32958.mem;
    
    if ((((((1 && num_groups_32966 != 0) && 1 != 0) && 1 != 0) &&
          group_sizze_32965 != 0) && 1 != 0) && 1 != 0) {
        int perm[3] = {0, 1, 2};
        
        if (1 > 1 << 16) {
            perm[1] = perm[0];
            perm[0] = 1;
        }
        if (1 > 1 << 16) {
            perm[2] = perm[0];
            perm[0] = 2;
        }
        
        size_t grid[3];
        
        grid[perm[0]] = num_groups_32966;
        grid[perm[1]] = 1;
        grid[perm[2]] = 1;
        
        void *kernel_args_33857[] = {&kernel_arg_33860, &num_elems_32959,
                                     &val_32960};
        int64_t time_start_33858 = 0, time_end_33859 = 0;
        
        if (ctx->debugging) {
            fprintf(stderr, "Launching %s with grid size (", "replicate_32962");
            fprintf(stderr, "%d", num_groups_32966);
            fprintf(stderr, ", ");
            fprintf(stderr, "%d", 1);
            fprintf(stderr, ", ");
            fprintf(stderr, "%d", 1);
            fprintf(stderr, ") and block size (");
            fprintf(stderr, "%d", group_sizze_32965);
            fprintf(stderr, ", ");
            fprintf(stderr, "%d", 1);
            fprintf(stderr, ", ");
            fprintf(stderr, "%d", 1);
            fprintf(stderr, ").\n");
            time_start_33858 = get_wall_time();
        }
        CUDA_SUCCEED(cuLaunchKernel(ctx->replicate_32962, grid[0], grid[1],
                                    grid[2], group_sizze_32965, 1, 1, 0, NULL,
                                    kernel_args_33857, NULL));
        if (ctx->debugging) {
            CUDA_SUCCEED(cuCtxSynchronize());
            time_end_33859 = get_wall_time();
            fprintf(stderr, "Kernel %s runtime: %ldus\n", "replicate_32962",
                    time_end_33859 - time_start_33858);
        }
    }
    return 0;
}
struct futhark_i32_2d {
    struct memblock_device mem;
    int64_t shape[2];
} ;
struct futhark_i32_2d *futhark_new_i32_2d(struct futhark_context *ctx,
                                          int32_t *data, int64_t dim0,
                                          int64_t dim1)
{
    struct futhark_i32_2d *bad = NULL;
    struct futhark_i32_2d *arr =
                          (struct futhark_i32_2d *) malloc(sizeof(struct futhark_i32_2d));
    
    if (arr == NULL)
        return bad;
    lock_lock(&ctx->lock);
    arr->mem.references = NULL;
    if (memblock_alloc_device(ctx, &arr->mem, dim0 * dim1 * sizeof(int32_t),
                              "arr->mem"))
        return NULL;
    arr->shape[0] = dim0;
    arr->shape[1] = dim1;
    CUDA_SUCCEED(cuMemcpyHtoD(arr->mem.mem + 0, data + 0, dim0 * dim1 *
                              sizeof(int32_t)));
    lock_unlock(&ctx->lock);
    return arr;
}
struct futhark_i32_2d *futhark_new_raw_i32_2d(struct futhark_context *ctx,
                                              CUdeviceptr data, int offset,
                                              int64_t dim0, int64_t dim1)
{
    struct futhark_i32_2d *bad = NULL;
    struct futhark_i32_2d *arr =
                          (struct futhark_i32_2d *) malloc(sizeof(struct futhark_i32_2d));
    
    if (arr == NULL)
        return bad;
    lock_lock(&ctx->lock);
    arr->mem.references = NULL;
    if (memblock_alloc_device(ctx, &arr->mem, dim0 * dim1 * sizeof(int32_t),
                              "arr->mem"))
        return NULL;
    arr->shape[0] = dim0;
    arr->shape[1] = dim1;
    CUDA_SUCCEED(cuMemcpy(arr->mem.mem + 0, data + offset, dim0 * dim1 *
                          sizeof(int32_t)));
    lock_unlock(&ctx->lock);
    return arr;
}
int futhark_free_i32_2d(struct futhark_context *ctx, struct futhark_i32_2d *arr)
{
    lock_lock(&ctx->lock);
    if (memblock_unref_device(ctx, &arr->mem, "arr->mem") != 0)
        return 1;
    lock_unlock(&ctx->lock);
    free(arr);
    return 0;
}
int futhark_values_i32_2d(struct futhark_context *ctx,
                          struct futhark_i32_2d *arr, int32_t *data)
{
    lock_lock(&ctx->lock);
    CUDA_SUCCEED(cuMemcpyDtoH(data + 0, arr->mem.mem + 0, arr->shape[0] *
                              arr->shape[1] * sizeof(int32_t)));
    lock_unlock(&ctx->lock);
    return 0;
}
CUdeviceptr futhark_values_raw_i32_2d(struct futhark_context *ctx,
                                      struct futhark_i32_2d *arr)
{
    (void) ctx;
    return arr->mem.mem;
}
int64_t *futhark_shape_i32_2d(struct futhark_context *ctx,
                              struct futhark_i32_2d *arr)
{
    (void) ctx;
    return arr->shape;
}
int futhark_entry_main(struct futhark_context *ctx,
                       struct futhark_i32_2d **out0, const
                       struct futhark_i32_2d *in0)
{
    struct memblock_device input_mem_32784;
    
    input_mem_32784.references = NULL;
    
    int32_t n_31902;
    int32_t m_31903;
    struct memblock_device out_mem_33410;
    
    out_mem_33410.references = NULL;
    
    int32_t out_arrsizze_33411;
    int32_t out_arrsizze_33412;
    
    lock_lock(&ctx->lock);
    input_mem_32784 = in0->mem;
    n_31902 = in0->shape[0];
    m_31903 = in0->shape[1];
    
    int ret = futrts_main(ctx, &out_mem_33410, &out_arrsizze_33411,
                          &out_arrsizze_33412, input_mem_32784, n_31902,
                          m_31903);
    
    if (ret == 0) {
        assert((*out0 =
                (struct futhark_i32_2d *) malloc(sizeof(struct futhark_i32_2d))) !=
            NULL);
        (*out0)->mem = out_mem_33410;
        (*out0)->shape[0] = out_arrsizze_33411;
        (*out0)->shape[1] = out_arrsizze_33412;
    }
    lock_unlock(&ctx->lock);
    return ret;
}
int futhark_entry_quicksort1(struct futhark_context *ctx,
                             struct futhark_i32_2d **out0, const
                             struct futhark_i32_2d *in0)
{
    struct memblock_device input_mem_32784;
    
    input_mem_32784.references = NULL;
    
    int32_t n_31612;
    int32_t sizze_31613;
    struct memblock_device out_mem_32955;
    
    out_mem_32955.references = NULL;
    
    int32_t out_arrsizze_32956;
    int32_t out_arrsizze_32957;
    
    lock_lock(&ctx->lock);
    input_mem_32784 = in0->mem;
    sizze_31613 = in0->shape[0];
    n_31612 = in0->shape[1];
    
    int ret = futrts_quicksort1(ctx, &out_mem_32955, &out_arrsizze_32956,
                                &out_arrsizze_32957, input_mem_32784, n_31612,
                                sizze_31613);
    
    if (ret == 0) {
        assert((*out0 =
                (struct futhark_i32_2d *) malloc(sizeof(struct futhark_i32_2d))) !=
            NULL);
        (*out0)->mem = out_mem_32955;
        (*out0)->shape[0] = out_arrsizze_32956;
        (*out0)->shape[1] = out_arrsizze_32957;
    }
    lock_unlock(&ctx->lock);
    return ret;
}
